/*hdr
**	FILE:			Ted.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			23/4/1998
**	DESCRIPTION:	TextEditor
**
**	Copyright (C) 1998-2000, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"
#include "Ted.h"
#include "GAbout.h"

// Application identification
char *AppName = "Ted";

///////////////////////////////////////////////////////////////////////////////////////////////
char *ConvertSpaceToTab(char *In, int TabSize)
{
	StringPipe Out;
	
	char *l = In;
	while (l)
	{
		char Spaces[] = "                                ";

		int Pos = 0;
		for (int i=0; l[i]; i++)
		{
			if (l[i] == '\t')
			{
				int n = (((Pos + TabSize) / TabSize) * TabSize) - Pos;
				Out.BytePipe::Push((uchar*)Spaces, n);
				Pos += n;
			}
			else
			{
				Out.BytePipe::Push((uchar*)l+i, 1);
				Pos ++;
			}

			if (l[i] == '\n') break;
		}

		l = strchr(l, '\n');
		if (l) l++;
	}

	return Out.NewStr();
}

///////////////////////////////////////////////////////////////////////////////////////////////
Ted::Ted() :
	GDocApp(AppName, "MAIN")
{
	Text = 0;
	UseCrLf = 0;
	GApp::ObjInstance()->AppWnd = this;
	_Create();

	if (_LoadMenu())
	{
		if (_FileMenu)
		{
			int p = 5;
			_FileMenu->AppendItem("Page Set&up", IDM_PRINTSETUP, false, p++);
			_FileMenu->AppendItem("&Print", IDM_PRINT, false, p++);
			GMenuItem *i = _FileMenu->AppendItem("", -1, false, p++);
			if (i)
			{
				i->Separator(true);
			}
		}

		GSubMenu *Edit = Menu->AppendSub("&Edit");
		if (Edit)
		{
			/*
			Edit->AppendItem("&Undo", IDM_UNDO, false);
			Edit->AppendItem("&Redo", IDM_REDO, false);
			Edit->AppendSeparator();
			*/
			Edit->AppendItem("Cu&t", IDM_CUT, true);
			Edit->AppendItem("&Copy", IDM_COPY, true);
			Edit->AppendItem("&Paste", IDM_PASTE, true);
			Edit->AppendSeparator();
			Edit->AppendItem("&Find", IDM_FIND, true);
			Edit->AppendItem("R&eplace", IDM_REPLACE, true);
			Edit->AppendSeparator();
			
			UseCrLf = Edit->AppendItem("U&se CRLF", IDM_USE_CRLF, true);
			Edit->AppendItem("Convert <SP> to <TAB>", IDM_SP_TO_TAB, true);
		}

		GSubMenu *View = Menu->AppendSub("&View");
		if (View)
		{
			View->AppendItem("&Font", IDM_FONT, true);
			WordWrap = View->AppendItem("&Word wrap", IDM_WORD_WRAP, true);
		}

		GSubMenu *Help = Menu->AppendSub("&Help");
		if (Help)
		{
			Help->AppendItem("&Help", IDM_HELP, true);
			Help->AppendSeparator();
			Help->AppendItem("&About", IDM_ABOUT, true);
		}

		#ifdef WIN32
		Menu->Attach(this);
		#endif
	}

	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);

		StatusInfo[0] = Status->AppendPane("", -1);
		StatusInfo[1] = Status->AppendPane("", 150);
		if (StatusInfo[1]) StatusInfo[1]->Sunken(true);
	}

	Text = NEW(GTextView(100, 0, 0, 100, 100, this, AppName));
	if (Text)
	{
		Text->Attach(this);

		Text->SetPourLargest(true);
		Text->UpdateScrollBars();
		Text->Focus(true);
		Text->SetWrapType(TEXTED_WRAP_NONE);
		Text->SetAutoIndent(true);
	}

	Pour();
	OnNotify(Text, GTVN_CURSOR_CHANGED);
}

Ted::~Ted()
{
	_Destroy();
}

void Ted::GetFileTypes(GFileSelect *Dlg, bool Write)
{
	Dlg->Type("Code", "*.cpp;*.h;*.c");
	Dlg->Type("Text files", "*.txt;*.html;*.htm");
	Dlg->Type("All files", "*.*");
}

void Ted::OnFocus(bool b)
{
	if (b AND Text)
	{
		Text->Focus(true);
	}
}

void Ted::OnKey(VKey &K)
{
	Text->OnKey(K);
}

int Ted::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		#ifdef WIN32
		/*
		case WM_WINDOWPOSCHANGED: //WM_SIZE
		{
			if (Text) Text->UpdateScrollBars();
			break;
		}
		*/
		case WM_MOUSEWHEEL:
		{
			Text->OnEvent(Msg);
			break;
		}
		case 8000:
		{
			SetStatus(Msg->a, (char*)Msg->b);
			break;
		}
		#endif
	}

	return GDocApp::OnEvent(Msg);
}

bool Ted::OpenFile(char *File)
{
	bool Status = false;

	if (File)
	{
		if (Text->Open(File))
		{
			Status = true;
		}

		if (Status)
		{
			if (UseCrLf)
			{
				UseCrLf->Checked(Text->UseCrLf());
			}

			OnNotify(Text, GTVN_CURSOR_CHANGED);
		}
	}

	return Status;
}

bool Ted::SaveFile(char *Name)
{
	bool Status = false;
	
	if (Name)
	{
		Status = Text->Save(Name);
	}

	return Status;
}

void Ted::OnReceiveFiles(List<char> &Files)
{
	char ExeName[256];
	int i=0;

	LgiGetExeFile(ExeName); 

	for (char *File = Files.First(); File; File = Files.Next(), i++)
	{
		if (NOT i)
		{
			// first file can be handled by
			// the current instance of Ted
			_OpenFile(File);
		}
		else
		{
			// spawn new session of Ted to handle 
			// extra documents
			LgiExecute(ExeName, File, ".");
		}
	}
}

int Ted::OnCommand(int Cmd, int Event, GWndRef Wnd)
{
	switch (Cmd)
	{
		case IDM_CUT:
		{
			Text->Cut();
			break;
		}
		case IDM_COPY:
		{
			Text->Copy();
			break;
		}
		case IDM_PASTE:
		{
			Text->Paste();
			break;
		}
		case IDM_ABOUT:
		{
			GAbout Dlg(	this,
						AppName,
						VER,
						"Simple cross-platform text editor.\n"
						"http://members.ozemail.com.au/~fret/ted.html",
						0); // "about.bmp");
			break;
		}
		case IDM_USE_CRLF:
		{
			if (UseCrLf)
			{
				bool Use = NOT UseCrLf->Checked();
				UseCrLf->Checked(Use);
				Text->UseCrLf(Use);
			}
			break;
		}
		case IDM_WORD_WRAP:
		{
			if (WordWrap)
			{
				bool Wrap = NOT WordWrap->Checked();
				WordWrap->Checked(Wrap);
				Text->SetWrapType((Wrap)?TEXTED_WRAP_NONREFLOW:TEXTED_WRAP_NONE);
			}
			break;
		}
		case IDM_SP_TO_TAB:
		{
			if (Text)
			{
				char *Doc = Text->Name();
				if (Doc)
				{
					char *NewDoc = ConvertSpaceToTab(Doc, 4);
					if (NewDoc)
					{
						Text->Name(NewDoc);
						DeleteObj(NewDoc);
					}
				}
			}
			break;
		}
		case IDM_FONT:
		{
			if (Text)
			{
				GFont *f = Text->GetFont();
				if (f)
				{
					GdcFontType Type;
					if (Type.GetFromRef(f->Handle()))
					{
						if (Type.DoUI(this))
						{
							GFont *Temp = Type.CreateFont();
							if (Temp)
							{
								Text->SetFont(Temp);
								DeleteObj(Temp);
							}
						}
					}
				}
			}
			break;
		}
		case IDM_FIND:
		{
			Text->OpenFind();
			break;
		}
		case IDM_REPLACE:
		{
			Text->OpenReplace();
			break;
		}
		case IDM_HELP:
		{
			char ExeName[256];
			LgiGetExePath(ExeName);
			
			strcat(ExeName, DIR_STR);
			strcat(ExeName, "index.html");

			if (NOT LgiExecute(ExeName, "", "."))
			{
				LgiMsg(this, "Couldn't open help file in program directory.", AppName, MB_OK);
			}
			break;
		}
	}
	
	return GDocApp::OnCommand(Cmd, Event, Wnd);
}

int Ted::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl == Text)
	{
		if (Flags & GTVN_DOC_CHANGED)
		{
			SetDirty(true);
		}

		if (Flags & GTVN_CURSOR_CHANGED)
		{
			GdcPt2 p = Text->GetCursor();
			char Str[256];
			sprintf(Str, "Col: %i Line: %i of %i", p.x, p.y + 1, Text->GetLines());
			SetStatus(1, Str);
		}
	}

	return false;
}

void Ted::SetStatus(int Pos, char *Text)
{
	if (Pos >= 0 AND Pos < 3 AND StatusInfo[Pos] AND Text)
	{
		StatusInfo[Pos]->Name(Text);
	}
}
