/*hdr
**      FILE:           Gui.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>

#include "SimpleGameSound.h"
#include "Roster.h"

#include "Lgi.h"
#include "DragAndDrop.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
BViewRedir::BViewRedir(GView *wnd) :
	BView(	BRect(0, 0, 100, 100),
			"BViewRedir",
			B_FOLLOW_LEFT | B_FOLLOW_TOP,
			LGI_DRAW_VIEW_FLAGS)
{
	Wnd = wnd;
	ScrollView = 0;
	SetViewColor(B_TRANSPARENT_COLOR);
}

void BViewRedir::AttachedToWindow()
{
	Wnd->OnCreate();
}

void BViewRedir::DetachedFromWindow()
{
	Wnd->OnDestroy();
}

void BViewRedir::Draw(BRect UpdateRect)
{
	GdcViewDC DC(this);
	Wnd->_Paint(&DC);
	Window()->Sync();
}

void BViewRedir::FrameMoved(BPoint Point)
{
	if (NOT ScrollView)
	{
		Wnd->Sys_FrameMoved(Point);
	}
}

void BViewRedir::FrameResized(float width, float height)
{
	if (NOT ScrollView)
	{
		// Wnd->Pos.Dimension(width+1, height+1);
		Wnd->OnPosChange();
	}
}

void BViewRedir::Pulse()
{
	Wnd->OnPulse();
}

void BViewRedir::MessageReceived(BMessage *message)
{
	Wnd->OnEvent(message);
}

void BViewRedir::MakeFocus(bool f = true)
{
	BView::MakeFocus(f);
	Wnd->OnFocus(f);
}

void BViewRedir::KeyDown(const char *bytes, int32 numBytes)
{
	Wnd->Sys_KeyDown(bytes, numBytes);
}

void BViewRedir::KeyUp(const char *bytes, int32 numBytes)
{
	Wnd->Sys_KeyUp(bytes, numBytes);
}

void BViewRedir::MouseDown(BPoint point)
{
	Wnd->Sys_MouseDown(point);
}

void BViewRedir::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	Wnd->Sys_MouseMoved(point, transit, message);
}

bool BViewRedir::QuitRequested()
{
	GWindow *App = dynamic_cast<GWindow*>(Wnd);
	if (App)
	{
		return App->QuitRequested();
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////
GScroll::GScroll(GScrollView *notify, orientation posture) :
	BScrollBar(	(posture==B_VERTICAL)?BRect(0, 0, B_V_SCROLL_BAR_WIDTH, 100):BRect(0, 0, 100, B_H_SCROLL_BAR_HEIGHT),
				"Scrollbar",
				NULL,
				0, 99,
				posture)
{
	Notify = notify;
	Hide();
}

void GScroll::ValueChanged(float newValue)
{
	if (Notify)
	{
		Notify->OnNotify(this, newValue);
	}
}

///////////////////////////////////////////////////////////////////////////////////////
GScrollView::GScrollView(	char *name,
							GView *target,
							uint32 resizingMode,
							uint32 flags,
							bool horizontal,
							bool vertical,
							int border) :
				BView(		BRect(0, 0, 100, 100),
							name,
							resizingMode,
							flags | B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE)
{
	// Border
	Border = border;

	// Setup scrollbars
	X = Y = 0;
	if (horizontal)
	{
		X = NEW(GScroll(this, B_HORIZONTAL));
		if (X)
		{
			AddChild(X);
		}
	}

	if (vertical)
	{
		Y = NEW(GScroll(this, B_VERTICAL));
		if (Y)
		{
			AddChild(Y);
		}
	}

	// Setup target view
	Target = target;
	if (Target AND Target->Sys_View)
	{
		bool Vis = NOT Target->Sys_View->IsHidden();
		BView *Parent = Target->Sys_View->Parent();
		BView *LockView = (Parent) ? Parent : Target->Sys_View;
		bool Lock = LockView->LockLooper();

		// make our position the target's position
		Target->Sys_View->SetResizingMode(B_FOLLOW_LEFT | B_FOLLOW_TOP);
		GdcRegion r(Target->Sys_View->Frame());

		if (Parent)
		{
			// Replace the target view with ourselves
			Parent->RemoveChild(Target->Sys_View);
			Parent->AddChild(this);
		}

		// attach the target to us
		AddChild(Target->Sys_View);

		// set our position
		Target->Sys_View->MoveTo(0, 0);
		SetPos(r);
		if (Vis)
		{
			if (IsHidden())
			{
				Show();
			}
		}
		else
		{
			if (NOT IsHidden())
			{
				Hide();
			}
		}

		// Other setup
		SetViewColor(B_TRANSPARENT_COLOR);

		if (Lock) LockView->UnlockLooper();
	}
}

GScrollView::~GScrollView()
{
	if (Parent())
	{
		bool Lock = LockLooper();
		if (Target AND Target->Sys_View)
		{
			bool HasTarget = false;
			for (int i=0; i<CountChildren(); i++)
			{
				if (ChildAt(i) == Target->Sys_View)
				{
					HasTarget = true;
					break;
				}
			}
		
			if (HasTarget)
			{
				GdcRegion r(Frame());
				Target->Sys_View->RemoveSelf();
				Parent()->AddChild(Target->Sys_View);
				Target->GView::SetPos(r);

				switch (Border)
				{
					case SUNKEN:
						Target->Sunken(true);
						break;
					case RAISED:
						Target->Raised(true);
						break;
				}
			}
			
			Target = 0;
		}
		
		if (Lock) UnlockLooper();
		RemoveSelf();
	}
}

BScrollBar *GScrollView::ScrollBar(orientation posture)
{
	if (posture == B_VERTICAL)
	{
		return Y;
	}
	else if (posture == B_HORIZONTAL)
	{
		return X;
	}

	return 0;
}

void GScrollView::Draw(BRect update)
{
	GdcViewDC DC(this);
	GdcRegion r(Frame());
	r.Offset(-r.x1, -r.y1);
	
	switch (Border)
	{
		case SUNKEN:
		case RAISED:
		{
			DrawMoveDeepEdge(&DC, r, Border);
			break;
		}
	}

	// printf("\tGScrollView::Draw(%p, %i)\n", this, Border==SUNKEN);
}

void GScrollView::FrameResized(float width, float height)
{
	BView::FrameResized(width, height);
	GdcRegion r(Frame());
	GdcRegion tr(Target->Sys_View->Frame());

	/*
	printf("\tGScrollView::FrameResized pos(%i, %i, %i, %i) targetpos(%i, %i, %i, %i) targetvis(%i)\n", 
			r.x1, r.y1, r.x2, r.y2,
			tr.x1, tr.y1, tr.x2, tr.y2,
			NOT Target->IsHidden());
	*/

	r.Offset(-r.x1, -r.y1);

	
	// Border
	switch (Border)
	{
		case SUNKEN:
		case RAISED:
		{
			r.Size(2, 2);
			break;
		}
	}
	
	// Scrollbars
	if (X)
	{
		bool Lock = X->LockLooper();
		if (r.Y() > B_H_SCROLL_BAR_HEIGHT)
		{
			X->MoveTo(r.x1 - 1, r.y2 - B_H_SCROLL_BAR_HEIGHT);
			X->ResizeTo(r.X() - ((Y) ? B_V_SCROLL_BAR_WIDTH : 0), B_H_SCROLL_BAR_HEIGHT);
			r.y2 -= B_H_SCROLL_BAR_HEIGHT + 1;
			X->Show();
		}
		else if (NOT X->IsHidden())
		{
			X->Hide();
		}
		if (Lock) X->UnlockLooper();
	}
	if (Y)
	{
		bool Lock = Y->LockLooper();
		if (r.X() > B_V_SCROLL_BAR_WIDTH)
		{
			Y->MoveTo(r.x2 - B_V_SCROLL_BAR_WIDTH, r.y1 - 1);
			Y->ResizeTo(B_V_SCROLL_BAR_WIDTH, r.Y());
			r.x2 -= B_V_SCROLL_BAR_WIDTH + 1;
			Y->Show();
		}
		else if (NOT Y->IsHidden())
		{
			Y->Hide();
		}

		if (Lock) Y->UnlockLooper();
	}
	
	// View
	if (Target AND
		r.Valid())
	{
		Target->GView::SetPos(r);
		Target->OnPosChange();
	}
}

void GScrollView::OnNotify(GScroll *Bar, float Value)
{
	if (Target)
	{
		bool Hor = (int)X == (int)Bar;
		BMessage *m = NEW(BMessage(Hor ? WM_HSCROLL : WM_VSCROLL));
		m->AddInt32("a", Value);
		Target->GetBWindow()->PostMessage(m, Target->Sys_View);
	}
}

GdcRegion &GScrollView::GetPos()
{
	bool Lock = LockLooper();
	GdcRegion n(Frame());
	if (Lock) UnlockLooper();
	return n;
}

void GScrollView::SetPos(GdcRegion &r)
{
	Pos = r;

	bool Lock = LockLooper();
	MoveTo(Pos.x1, Pos.y1);
	ResizeTo(Pos.X()-1, Pos.Y()-1);
	if (Lock) UnlockLooper();
}

void GScrollView::AttachedToWindow()
{
	/*
	bool Lock = LockLooper();
	MoveTo(Pos.x1, Pos.y1);
	ResizeTo(Pos.X()-1, Pos.Y()-1);
	FrameResized(Pos.X()-1, Pos.Y()-1);
	if (Lock) UnlockLooper();
	*/
}

////////////////////////////////////////////////////////////////////////////////////////////////////
void _lgi_yield()
{
	BLooper *Looper = BLooper::LooperForThread(find_thread(0));
	if (Looper)
	{
		BWindow *Wnd = dynamic_cast<BWindow*>(Looper);
		if (Wnd)
		{
			if (Wnd->Lock())
			{
				Wnd->UpdateIfNeeded();
				Wnd->Unlock();
			}
		}
	}
}

void _lgi_assert(bool b, char *test, char *file, int line)
{
	if (NOT b)
	{
		LgiMsg(	0,
				"Assert failed, file: %s, line: %i\n"
				"%s\n",
				"Debug",
				MB_OK,
				file, line, test);
		exit(-1);					
	}
}

BMessage NewMsg(int m, int a, int b)
{
	BMessage Msg(m);
	Msg.AddInt32("a", a);
	Msg.AddInt32("b", b);
	return Msg;
}

void DrawBox(GdcBasePrimitives *pDC, GdcRegion &r, bool Sunken, bool Fill)
{
	if (Fill)
	{
		pDC->Colour(CBit(pDC->GetBits(), LC_MED));
		pDC->Rectangle(r.x1+1, r.y1+1, r.x2-1, r.y2-1);
	}

	pDC->Colour(CBit(pDC->GetBits(), (Sunken) ? LC_LIGHT : LC_LOW));
	pDC->Line(r.x2, r.y2, r.x2, r.y1);
	pDC->Line(r.x2, r.y2, r.x1, r.y2);

	pDC->Colour(CBit(pDC->GetBits(), (Sunken) ? LC_LOW : LC_LIGHT));
	pDC->Line(r.x1, r.y1, r.x1, r.y2);
	pDC->Line(r.x1, r.y1, r.x2, r.y1);
}

void DrawMoveDeepEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	COLOUR VLow = LC_SHADOW;
	COLOUR Low = LC_LOW;
	COLOUR High = LC_HIGH;
	COLOUR VHigh = LC_LIGHT;

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(Low, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(VLow, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(High, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VHigh, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case RAISED:
		{
			pDC->Colour(VHigh, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(Low, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VLow, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case CHISEL:
		{
			pDC->Colour(Low, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(Low, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(High, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
	}

	r.Size(2, 2);
}

void DrawMoveShallowEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(LC_LIGHT, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(LC_LOW, 24);
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
		case RAISED:
		{
			pDC->Colour(LC_LOW, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(LC_LIGHT, 24);
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
	}
}

void DrawMoveFill(GdcBasePrimitives *pDC, GdcRegion &r, int Width)
{
	pDC->Colour(LC_MED, 24);
	if (Width < 1 OR r.X() < (2 * Width) OR r.Y() < (2 * Width))
	{
		pDC->Rectangle(&r);
		r.ZOff(-1, -1);
	}
	else
	{
		pDC->Rectangle(r.x1, r.y1, r.x2, r.y1+Width-1);
		pDC->Rectangle(r.x1, r.y2-Width+1, r.x2, r.y2);
		pDC->Rectangle(r.x1, r.y1+Width, r.x1+Width-1, r.y2-Width);
		pDC->Rectangle(r.x2-Width+1, r.y1+Width, r.x2, r.y2-Width);
		r.Size(Width, Width);
	}
}

////////////////////////////////////////////////////////////////////////////
GMessage CreateMsg(int m, int a, int b)
{
	GMessage Msg(m);
	Msg.AddInt32("a", a);
	Msg.AddInt32("b", b);
	return Msg;
}

int MessageA(GMessage *m)
{
	int32 i=0;
	if (m) m->FindInt32("a", &i);
	return i;
}

int MessageB(GMessage *m)
{
	int32 i=0;
	if (m) m->FindInt32("b", &i);
	return i;
}

////////////////////////////////////////////////////////////////////////////
void LgiTrace(char *Msg, ...)
{
	if (Msg)
	{
		char Buffer[1025];
		LgiGetExeFile(Buffer);
		char *Dot = strrchr(Buffer, '.');
		if (Dot AND NOT strchr(Dot, DIR_CHAR))
		{
			strcpy(Dot+1, "txt");
		}
		else
		{
			strcat(Buffer, ".txt");
		}

		VFile f;
		if (f.Open(Buffer, O_WRITE))
		{
			va_list Arg;
			va_start(Arg, Msg);
			vsprintf(Buffer, Msg, Arg);
			va_end(Arg);

			f.Seek(0, SEEK_END);
			f.Write(Buffer, strlen(Buffer));
			f.Write((char*)"\r\n", 2);
		}
	}
}

bool LgiIsKeyDown(int Key)
{
	key_info KeyInfo;
	if (get_key_info(&KeyInfo) == B_OK)
	{
		switch (Key)
		{
		}
	}
	return false;
}

bool LgiTrimDir(char *Path)
{
	char *Last = Path;
	char *p = strchr(Last, DIR_CHAR);
	if (p)
	{
		while (p = strchr(p+1, DIR_CHAR))
		{
			Last = p;
		}
		*Last = 0;
		return true;
	}
	return false;
}

bool LgiGetExeFile(char *Dst)
{
	if (Dst)
	{
		#ifdef WIN32
		char Exe[256];
		if (GetModuleFileName(NULL, Exe, sizeof(Exe)) > 0)
		{
			strcpy(Dst, Exe);
			return TRUE;
		}
		#else if BEOS
		app_info Info;
		if (GApp::ObjInstance()->GetAppInfo(&Info) == B_OK)
		{
			BEntry e(&Info.ref);
			BPath p;
			if (e.GetPath(&p) == B_OK)
			{
				strcpy(Dst, p.Path());
				return TRUE;
			}
		}
		#endif
	}
	return FALSE;
}

bool LgiGetExePath(char *Dst)
{
	if (Dst)
	{
		if (LgiGetExeFile(Dst))
		{
			LgiTrimDir(Dst);
			return TRUE;
		}
	}
	return FALSE;
}

char *LgiFindFile(char *Name)
{
	char *Result = 0;

	if (Name)
	{
		#ifdef WIN32
		char ExePath[256];
		char File[256];
		if (LgiGetExePath(ExePath))
		{
			sprintf(File, "%s\\%s", ExePath, Name);
			if (FileExists(File))
			{
				Result = NewStr(File);
			}
			else
			{
				// ..
				strcpy(File, ExePath);
				TrimDir(File);
				sprintf(File+strlen(File), "\\%s", Name);
				if (FileExists(File))
				{
					Result = NewStr(File);
				}
				else
				{
					// ..\\code
					strcpy(File, ExePath);
					TrimDir(File);
					sprintf(File+strlen(File), "\\code\\%s", Name);
					if (FileExists(File))
					{
						Result = NewStr(File);
					}
				}
			}
		}
		#else if BEOS
		char ExePath[256];
		char File[256];
		if (LgiGetExePath(ExePath))
		{
			sprintf(File, "%s/%s", ExePath, Name);
			if (FileExists(File))
			{
				Result = NewStr(File);
			}
			else
			{
				// ..
				strcpy(File, ExePath);
				LgiTrimDir(File);
				sprintf(File+strlen(File), "/%s", Name);
				if (FileExists(File))
				{
					Result = NewStr(File);
				}
				else
				{
					// code
					strcpy(File, ExePath);
					sprintf(File+strlen(File), "/Code/%s", Name);
					if (FileExists(File))
					{
						Result = NewStr(File);
					}
				}
			}
		}
		#endif
	}

	return Result;
}

bool LgiExecute(char *File, char *Args, char *Dir)
{
	if (File)
	{
		char f[256];
		
		if (Dir)
		{
			sprintf(f, "%s/%s", Dir, File);
		}
		else
		{
			strcpy(f, File);
		}
	
		BRoster Roster;
		BEntry Entry(f);
		entry_ref Ref;
		if (Entry.GetRef(&Ref) == B_OK)
		{
			return Roster.Launch(&Ref) == B_OK;
		}
		else
		{
			if (strnicmp(File, "http://", 7) == 0)
			{
				if (Roster.FindApp("text/html", &Ref) == B_OK)
				{
					/*
					char *Arg[2] = {Ref.name, File};
					return Roster.Launch(&Ref, 2, Arg) == B_OK;
					*/
					char *Arg[1] = {File};
					return Roster.Launch(&Ref, 1, Arg) == B_OK;
				}
			}
		}
	}
	return false;
}

int LgiMsg(GView *Parent, char *Str, char *Title, int Type, ...)
{
	int Res = 0;
	if (Str)
	{
		char Buffer[1024];
		va_list Arg;

		va_start(Arg, Type);
		vsprintf(Buffer, Str, Arg);
		va_end(Arg);

		BAlert *Dlg = 0;
		
		switch (Type)
		{
			case MB_OK:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Ok"));
				break;
			}
			case MB_YESNO:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Yes", "No", NULL, B_WIDTH_AS_USUAL, B_IDEA_ALERT));
				break;
			}
			case MB_YESNOCANCEL:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Yes", "No", "Cancel", B_WIDTH_AS_USUAL, B_IDEA_ALERT));
				break;
			}
			default:
			{
				ASSERT(0);
				break;
			}
		}
		
		if (Dlg)
		{
			Res = Dlg->Go();
		}
		
		if (Res >= 0 AND
			(Type == MB_YESNO OR Type == MB_YESNOCANCEL))
		{
			if (Res == 0)
				return IDYES;
			if (Res == 1)
				return IDNO;
			if (Res == 2)
				return IDCANCEL;
		}
	}
	return Res;
}

int LgiRand(int Limit)
{
	return rand() % Limit;
}

long _lgi_play_sound(void *Ptr)
{
	BEntry Entry((char*)Ptr);
	entry_ref Ref;
	if (Entry.GetRef(&Ref) == B_OK)
	{
		BSimpleGameSound Sound(&Ref);
		if (Sound.InitCheck() == B_OK)
		{
			Sound.StartPlaying();
			while (Sound.IsPlaying())
			{
			}
			return true;
		}
	}
	
	return false;
}

bool LgiPlaySound(char *FileName, int ASync)
{
	if (ASync)
	{
		thread_id Thread = spawn_thread(_lgi_play_sound, "_lgi_play_sound", B_LOW_PRIORITY, FileName);
		if (Thread > 0)
		{
			resume_thread(Thread);
			return true;
		}
		return false;
	}

	return _lgi_play_sound(FileName);
}

bool LgiGetFileMimeType(char *File, char *Mime)
{
	bool Status = false;
	if (File AND Mime)
	{
		BFile f(File, B_READ_ONLY);
		BNodeInfo i(&f);
		Status = i.GetType(Mime) == B_OK;		
	}
	return Status;
}

bool LgiGetAppForMimeType(char *Mime, char *AppPath)
{
	bool Status = false;
	if (AppPath AND Mime)
	{
		BMimeType m(Mime);
		char DefApp[B_MIME_TYPE_LENGTH];
		if (m.GetPreferredApp(DefApp) == B_OK)
		{
			entry_ref ap;
			BMimeType a(DefApp);
			if (a.GetAppHint(&ap) == B_OK)
			{
				BPath p;
				BEntry e(&ap);
				if (e.GetPath(&p) == B_OK)
				{
					strcpy(AppPath, p.Path());
					Status = true;
				}
			}
		}
	}
	return Status;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GFlags::GFlags()
{
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
}

bool GFlags::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GFlags::Enabled(bool i)
{
	if (i) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

bool GFlags::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GFlags::Visible(bool i)
{
	if (i) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);
}

bool GFlags::Focus()
{
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GFlags::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

bool GFlags::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

void GFlags::DropTarget(bool i)
{
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
}

#define GWF_ALL (GWF_SUNKEN | GWF_FLAT | GWF_RAISED)

bool GFlags::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GFlags::Sunken(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
}

bool GFlags::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GFlags::Flat(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_FLAT);
}

bool GFlags::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GFlags::Raised(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_RAISED);
}

/*
bool GFlags::Border()
{
	return TestFlag(WndFlags, GWF_BORDER);
}

void GFlags::Border(bool i)
{
	if (i) SetFlag(WndFlags, GWF_BORDER);
	else ClearFlag(WndFlags, GWF_BORDER);
}
*/


////////////////////////////////////////////////////////////////////////////////////////////////////
GView::GView(BView *view)
{
	Sys_View = view;
	// Sys_View->SetViewColor(B_TRANSPARENT_32_BIT);
	WndFlags = GWF_ENABLED | GWF_VISIBLE;
	Pos.ZOff(0, 0);
	Parent = 0;
	Notify = 0;

	_CaptureThread = 0;
	_PulseThread = 0;
	_PulseRate = -1;
	_BorderSize = 2;
}

GView::~GView()
{
	if (_CaptureThread)
	{
		kill_thread(_CaptureThread);
	}
	if (_PulseThread)
	{
		kill_thread(_PulseThread);
	}

	BViewRedir *Redir = dynamic_cast<BViewRedir*>(Sys_View);
	DeleteObj(Redir);
	Sys_View = 0;
	WndFlags = 0;
}

#define ADJ_LEFT			1
#define ADJ_RIGHT			2
#define ADJ_UP				3
#define ADJ_DOWN			4

extern int IsAdjacent(GdcRegion &a, GdcRegion &b);
int IsAdjacent(GdcRegion &a, GdcRegion &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

extern GdcRegion JoinAdjacent(GdcRegion &a, GdcRegion &b, int Adj);
GdcRegion JoinAdjacent(GdcRegion &a, GdcRegion &b, int Adj)
{
	GdcRegion t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

void GView::_Paint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
	bool HasEdge = (Sunken() OR Raised()) AND _BorderSize;
	if (HasEdge)
	{
		if (_BorderSize == 2)
		{
			DrawMoveDeepEdge(pDC, r, Sunken() ? SUNKEN : RAISED);
		}
		else if (_BorderSize == 1)
		{
			DrawMoveShallowEdge(pDC, r, Sunken() ? SUNKEN : RAISED);
		}

		pDC->SetOrigin(-r.x1, -r.y1);
		r.Offset(-r.x1, -r.y1);
		pDC->_SetClient(&r);
	}
	
	OnPaint(pDC);
	
	if (HasEdge)
	{
		pDC->ClipRgn(0);
	}
}

GdcRegion *GView::FindLargest(GdcComplexRegion &r)
{
	int Pixels = 0;
	GdcRegion *Best = 0;
	static GdcRegion Final;

	// do initial run through the list to find largest single region
	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GdcRegion LastRgn = Final;
		int ThisPixels = Pixels;
		GdcRegion ThisRgn = Final;

		GdcComplexRegion TempList;
		for (GdcRegion *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GdcRegion *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GdcRegion t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GdcRegion *GView::FindSmallestFit(GdcComplexRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GdcRegion *Best = 0;

	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		if ((i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GdcRegion *GView::FindLargestEdge(GdcComplexRegion &r, int Edge)
{
	GdcRegion *Best = 0;

	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

int GView::GetCtrlValue(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Value() : -1;
}

void GView::SetCtrlValue(int Id, int i)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

void GView::SetId(int i)
{
	CtrlId = i;
}

bool GView::Sunken()
{
	return GFlags::Sunken();
}

void GView::Sunken(bool i)
{
	GFlags::Sunken(i);
}

bool GView::Flat()
{
	return GFlags::Flat();
}

void GView::Flat(bool i)
{
	GFlags::Flat(i);
}

bool GView::Raised()
{
	return GFlags::Raised();
}

void GView::Raised(bool i)
{
	GFlags::Raised(i);
}

void GView::Quit(bool DontDelete)
{
	BWindow *Wnd = GetBWindow();
	if (Wnd)
	{
		Wnd->PostMessage(NEW(BMessage(B_QUIT_REQUESTED)));
	}
}

bool GView::AttachChildren()
{
	bool Status = true;

	for (GView *c=Children.First(); c; c=Children.Next())
	{
		Status |= c->Attach(this);
	}

	return Status;
}

GdcRegion &GView::GetClient()
{
	/*
	bool Lock = GetBView()->LockLooper();
	BRect rc = GetBView()->Frame();
	if (Lock) GetBView()->UnlockLooper();

	static GdcRegion Client;
	Client.Set(0, 0, rc.right-rc.left, rc.bottom-rc.top);
	if (Sunken() OR Raised()) Client.Size(_BorderSize, _BorderSize);
	*/

	static GdcRegion Client;
	Client.ZOff(Pos.X()-1, Pos.Y()-1);
	if (Sunken() OR Raised()) Client.Size(_BorderSize, _BorderSize);
	
	return Client;
}

GFontRef GView::GetFont()
{
	return 0;
}

void GView::SetFont(GFontRef Fnt)
{
}

bool GView::IsOver(VMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

GView *GView::WindowFromPoint(int x, int y)
{
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		if (c->Pos.Overlap(x, y))
		{
			return c;
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

void GView::OnPosChange()
{
}

bool GView::SetPos(GdcRegion &p, bool Repaint)
{
	if (Sys_View)
	{
		bool Lock = Sys_View->LockLooper();
		Sys_View->MoveTo(p.x1, p.y1);
		Sys_View->ResizeTo(p.X()-1, p.Y()-1);
		
		if (Repaint)
		{
			Sys_View->Invalidate();
		}
		
		if (Lock ) Sys_View->UnlockLooper();
	}
	Pos = p;
	
	return TRUE;
}

class GuiLocker
{
	BHandler *Handler;
	bool Lock;

public:
	GuiLocker(BHandler *handler)
	{
		Lock = 0;
		Handler = handler;
		if (Handler)
		{
			Lock = Handler->LockLooperWithTimeout(20000) == B_OK;
		}
	}
	
	~GuiLocker()
	{
		if (Handler AND Lock)
		{
			Handler->UnlockLooper();
			Lock = 0;
		}
	}
	
	bool Locked()
	{
		return Lock;
	}
};

bool GView::Invalidate(GdcRegion *r, bool Repaint)
{
	// GuiLocker Lock(Sys_View);
	if (Sys_View->LockLooperWithTimeout(1000) == B_OK)
	{
		if (r)
		{
			GdcRegion c = GetClient();
			BRect Rc = *r;
			Rc.OffsetBy(c.x1, c.y1);
			Sys_View->Invalidate(Rc);
		}
		else
		{
			Sys_View->Invalidate();
		}
		
		if (Repaint)
		{
			BWindow *Wnd = Sys_View->Window();
			if (Wnd)
			{
				thread_id MyThread = find_thread(0);
				thread_id WndThread = Wnd->Thread();
				
				if (MyThread == WndThread)
				{
					// in thread
					Wnd->UpdateIfNeeded();
				}
				else
				{
					// out of thread
					if (r)
					{
						BRect Rc = *r;
						BRegion *Rgn = NEW(BRegion(Rc));
						Sys_View->ConstrainClippingRegion(Rgn);
						Sys_View->Draw(Rc);
						Sys_View->ConstrainClippingRegion(NULL);
					}
					else
					{
						BRect Rc = Sys_View->Frame();
						Sys_View->Draw(Rc);
					}
				}
			}
		}
		
		Sys_View->UnlockLooper();
	}
	
	return FALSE;
}

bool GView::InvalidateFrame(GdcRegion *r, bool Repaint)
{
	return FALSE;
}

long _lgi_pulse_thread(void *ptr)
{
	GView *Wnd = (GView*) ptr;

	if (Wnd)
	{
		while (Wnd->_PulseRate > 0)
		{
			LgiSleep(Wnd->_PulseRate);
			
			if (Wnd->GetBView()->LockLooperWithTimeout(1000) == B_OK)
			{
				Wnd->GetBWindow()->PostMessage(WM_PULSE, Wnd->GetBView());
				Wnd->GetBView()->UnlockLooper();
			}
		}
		
		Wnd->_PulseRate = -1;
		Wnd->_PulseThread = 0;
	}
	
	return 0;
}

void GView::SetPulse(int Length)
{
	if (Sys_View->LockLooper())
	{
		if (_PulseThread)
		{
			kill_thread(_PulseThread);
			_PulseThread = 0;
		}
	
		_PulseRate = Length;
		
		if (Length > 0)
		{
			_PulseThread = spawn_thread(_lgi_pulse_thread, "_lgi_pulse_thread", B_LOW_PRIORITY, this);
			if (_PulseThread > 0)
			{
				resume_thread(_PulseThread);
			}
		}
		
		Sys_View->UnlockLooper();
	}
}

BView *SearchChildren(BView *Parent, BPoint &Cursor)
{
	BView *Ret = 0;
	int32 Children = Parent->CountChildren();
	for (int i=0; i<Children; i++)
	{
		BView *View = Parent->ChildAt(i);
		if (View)
		{
			BRect r = View->Frame();
			if (Cursor.x >= r.left AND
				Cursor.y >= r.top AND
				Cursor.x < r.right AND
				Cursor.y < r.bottom)
			{
				BPoint Pt(Cursor.x-r.left, Cursor.y-r.top);
				BView *Child = SearchChildren(View, Pt);

				Cursor = Pt;
				if (Child)
				{
					Ret = Child;
				}
				else
				{
					Ret = View;
				}
				break;
			}
		}
	}

	return Ret;
}

bool GView::InThread()
{
	return find_thread(0) == GetBWindow()->Thread();
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	bool Status = false;

	if (GetBWindow() AND
		GetBWindow()->Lock())
	{
		BMessage *Msg = NEW(BMessage(Cmd));
		if (Msg)
		{
			Msg->AddInt32("a", a);
			Msg->AddInt32("b", b);
			Status = GetBWindow()->PostMessage(Msg, GetBView());
			DeleteObj(Msg);
		}
		GetBWindow()->Unlock();
	}
	
	return Status;
}

int GView::OnEvent(BMessage *Msg)
{
	switch (Message(Msg))
	{
		case B_SIMPLE_DATA:
		{
			if (GetParent())
			{
				GetParent()->OnEvent(Msg);
			}
			break;
		}
		case B_MOUSE_WHEEL_CHANGED:
		{
			float Wheel = 0.0;
			if (Msg->FindFloat("be:wheel_delta_y", &Wheel) == B_OK)
			{
				OnMouseWheel(Wheel * 3.0);
			}
			break;
		}
		case WM_PULSE:
		{
			OnPulse();
			break;
		}
		case WM_DRAG_DROP:
		{
			GDragDropSource *Source = 0;
			if (Msg->FindPointer("GDragDropSource", (void**) &Source) == B_OK)
			{
				BPoint Cursor;
				uint32 Buttons;
				GetBView()->GetMouse(&Cursor, &Buttons);
			
				GDragDropTarget *Target = 0;
				BView *TargetView = SearchChildren(Sys_View, Cursor);

				GdcPt2 MousePt(Cursor.x, Cursor.y);
				Target = dynamic_cast<GDragDropTarget*>(TargetView);
				if (Target)
				{
					if (Target->WillAccept(Source, MousePt, 0) != DROPEFFECT_NONE)
					{
						uchar *Data = 0;
						int Size = 0;
						if (Source->GetData(&Data, &Size))
						{
							Target->OnDrop(Data, Size, MousePt, 0);
						}
					}
				}
			}
			break;
		}
		case WM_CHANGE:
		{
			GView *Ctrl = dynamic_cast<GView*>((GView*) MessageA(Msg));
			if (Ctrl)
			{
				OnNotify(Ctrl, MessageB(Msg));
			}
			break;
		}
		case WM_COMMAND:
		{
			// printf("change...\n");
			int32 a=MessageA(Msg);
			return OnCommand(a&0xFFFF, a>>16, (GWndRef) MessageB(Msg));
		}
		default:
		{
			BMessage *Msg = GetBWindow()->CurrentMessage();
			if (Msg AND Msg->WasDropped())
			{
				Msg->PrintToStream();
			}
			break;
		}
	}

	return 0;
}

int GView::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl AND Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);
	Sys_View->SetName(n);
	return TRUE;
}

char *GView::Name()
{
	return GObject::Name();
}

bool GView::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GView::Enabled(bool e)
{
	if (e) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
	
	BControl *Ctrl = dynamic_cast<BControl*>(Sys_View);
	if (Ctrl)
	{
		bool Lock = Ctrl->LockLooper();
		Ctrl->SetEnabled(e);
		if (Lock) Ctrl->UnlockLooper();
	}

	Invalidate();
}

bool GView::Visible()
{
	return GFlags::Visible();
}

void GView::Visible(bool v)
{
	if (Sys_View)
	{
		bool Lock = Sys_View->LockLooper();
		if (v)
		{
			if (NOT GFlags::Visible())
			{
				Sys_View->Show();
			}
		}
		else
		{
			if (GFlags::Visible())
			{
				Sys_View->Hide();
			}
		}
		if (Lock) Sys_View->UnlockLooper();
	}

	GFlags::Visible(v);
}

bool GView::Focus()
{
	bool f = false;
	BWindow *Wnd = GetBWindow();

	if (Wnd AND Sys_View AND Sys_View->LockLooper())
	{
		BView *Cur = Wnd->CurrentFocus();
		f = (Sys_View == Cur);
		if (f) SetFlag(WndFlags, GWF_FOCUS);
		else ClearFlag(WndFlags, GWF_FOCUS);

		Sys_View->UnlockLooper();
	}

	return TestFlag(WndFlags, GWF_FOCUS);
}

void GView::Focus(bool f)
{
	if (f) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);

	if (Sys_View AND Sys_View->LockLooper())
	{
		Sys_View->MakeFocus(f);
		Sys_View->UnlockLooper();
	}
}

bool GView::IsCapturing()
{
	return false;
}

bool GView::Capture(bool c)
{
	return false;
}

bool GView::GetMouse(VMouse &m, bool ScreenCoords)
{
	if (GetBView()->LockLooper())
	{
		// get mouse state
		BPoint Cursor;
		uint32 Buttons;
		
		GetBView()->GetMouse(&Cursor, &Buttons);
		if (ScreenCoords)
		{
			GetBView()->ConvertToScreen(&Cursor);

			// fudge???
			Cursor.x += 20;
			Cursor.y += 20;
		}
		
		// position
		m.x = Cursor.x;
		m.y = Cursor.y;

		// buttons
		m.Flags =	(TestFlag(Buttons, B_PRIMARY_MOUSE_BUTTON) ? MK_LEFT : 0) |
					(TestFlag(Buttons, B_TERTIARY_MOUSE_BUTTON) ? MK_MIDDLE : 0) |
					(TestFlag(Buttons, B_SECONDARY_MOUSE_BUTTON) ? MK_RIGHT : 0);

		// key states
		key_info KeyInfo;
		if (get_key_info(&KeyInfo) == B_OK)
		{
			m.Flags |=	(TestFlag(KeyInfo.modifiers, B_CONTROL_KEY) ? MK_CTRL : 0) |
						(TestFlag(KeyInfo.modifiers, B_MENU_KEY) ? MK_ALT : 0) |
						(TestFlag(KeyInfo.modifiers, B_SHIFT_KEY) ? MK_SHIFT : 0);
		}
		
		GetBView()->UnlockLooper();
		
		return true;
	};

	return false;
}

int LastFunctionKey = 0;
void GView::Sys_KeyDown(const char *bytes, int32 numBytes)
{
	// printf("GView::Sys_KeyDown(%p, %i)\n", bytes, numBytes);

	key_info Info;
	get_key_info(&Info);

	for (int i=0; i<numBytes; i++)
	{
		VKey k;
		
		if (bytes[i] == B_FUNCTION_KEY)
		{
			k.c = 0;
			BMessage *msg = GetBWindow()->CurrentMessage(); 
			if (msg)
			{ 
				int32 key; 
				msg->FindInt32("key", &key); 
				switch (key)
				{ 
					case B_F1_KEY: 
						k.c = VK_F1;
						break;
					case B_F2_KEY: 
						k.c = VK_F2;
						break;
					case B_F3_KEY: 
						k.c = VK_F3;
						break;
					case B_F4_KEY: 
						k.c = VK_F4;
						break;
					case B_F5_KEY: 
						k.c = VK_F5;
						break;
					case B_F6_KEY: 
						k.c = VK_F6;
						break;
					case B_F7_KEY: 
						k.c = VK_F7;
						break;
					case B_F8_KEY: 
						k.c = VK_F8;
						break;
					case B_F9_KEY: 
						k.c = VK_F9;
						break;
					case B_F10_KEY: 
						k.c = VK_F10;
						break;
					case B_F11_KEY: 
						k.c = VK_F11;
						break;
					case B_F12_KEY: 
						k.c = VK_F12;
						break;
					default:
						k.c = key;
						break;
				}
			}

			k.IsChar = false;
			LastFunctionKey = k.c;
		}
		else
		{
			k.c = bytes[i];
			k.IsChar = ((k.c < 127) AND (k.c >= ' ')) OR
						(k.c == 9) OR
						(k.c == 8) OR
						(k.c == 10);
		}
		
		k.Flags = modifiers();
		k.Data = 0;
		k.Down = TRUE;
		
		OnKey(k);
	}
}

void GView::Sys_KeyUp(const char *bytes, int32 numBytes)
{
	key_info Info;
	get_key_info(&Info);

	for (int i=0; i<numBytes; i++)
	{
		VKey k;
		
		k.c = bytes[i];
		k.Flags = modifiers();
		k.Data = 0;
		k.Down = FALSE;
		k.IsChar = 0;
		
		OnKey(k);
	}
}

struct MouseThreadInfo
{
	GView *Wnd;
	int32 Clicks;
};

long MouseThread(MouseThreadInfo *Info)
{
	static int LastX = -DOUBLE_CLICK_THRESHOLD, LastY = -DOUBLE_CLICK_THRESHOLD;

	if (Info AND Info->Wnd)
	{
		VMouse m;
		BPoint Point;
		uint32 Buttons;
	
		if (Info->Wnd->GetBView()->LockLooper())
		{
			Info->Wnd->GetBView()->GetMouse(&Point, &Buttons, FALSE);
	
			m.x = Point.x;
			m.y = Point.y;
			int Keys = modifiers();
			m.Flags =	Buttons |
						(TestFlag(Keys, B_SHIFT_KEY) ? LGI_VMOUSE_SHIFT : 0) |
						(TestFlag(Keys, B_CONTROL_KEY) ? LGI_VMOUSE_ALT : 0) |
						(TestFlag(Keys, B_COMMAND_KEY) ? LGI_VMOUSE_CTRL : 0);
			m.Down(true);
			m.Double(	(Info->Clicks > 1) AND
						(abs(m.x-LastX) < DOUBLE_CLICK_THRESHOLD) AND
						(abs(m.y-LastY) < DOUBLE_CLICK_THRESHOLD));
			LastX = m.x;
			LastY = m.y;
			
			Info->Wnd->OnMouseClick(m);
			Info->Wnd->GetBView()->UnlockLooper();
		}
	
		bool Over = true;
		BRect r;
		if (Info->Wnd->GetBView()->LockLooper())
		{
			r = Info->Wnd->GetBView()->Frame();
			Info->Wnd->GetBView()->UnlockLooper();
		}
		int Sx = r.right - r.left;
		int Sy = r.bottom - r.top;
		
		while (Buttons)
		{
			if (Info->Wnd->GetBView()->LockLooper())
			{
				Info->Wnd->GetBView()->GetMouse(&Point, &Buttons, FALSE);
				if (m.x != Point.x OR m.y != Point.y)
				{
					m.x = Point.x;
					m.y = Point.y;
					int Keys = modifiers();
					m.Flags =	Buttons |
								(TestFlag(Keys, B_SHIFT_KEY) ? LGI_VMOUSE_SHIFT : 0) |
								(TestFlag(Keys, B_CONTROL_KEY) ? LGI_VMOUSE_ALT : 0) |
								(TestFlag(Keys, B_COMMAND_KEY) ? LGI_VMOUSE_SHIFT : 0);
					m.Double(false);

					bool NowOver = Point.x >= 0 AND Point.y >= 0 AND Point.x < Sx AND Point.y < Sy;
					if (NowOver != Over)
					{
						if (NowOver)
						{
							Info->Wnd->OnMouseEnter(m);
						}
						else
						{
							Info->Wnd->OnMouseExit(m);
						}
						
						Over = NowOver;
					}
					
					Info->Wnd->OnMouseMove(m);
				}
				
				Info->Wnd->GetBView()->UnlockLooper();
			}
			
			snooze(20 * 1000);
		}
		
		if (Info->Wnd->GetBView()->LockLooper())
		{
			Info->Wnd->GetBView()->GetMouse(&Point, &Buttons, FALSE);
			m.x = Point.x;
			m.y = Point.y;
			// m.Flags = ??;
			m.Down(false);
			m.Double(false);

			Info->Wnd->OnMouseClick(m);
			Info->Wnd->GetBView()->UnlockLooper();
		}
	}
	
	DeleteObj(Info);

	return 0;
}

void GView::Sys_MouseDown(BPoint Point)
{
	MouseThreadInfo *Info = NEW(MouseThreadInfo);
	if (Info)
	{
		Info->Wnd = this;
		Info->Clicks = 1;
	
		BMessage *Msg = (GetBWindow())?GetBWindow()->CurrentMessage():0;
		if (Msg)
		{
			Msg->FindInt32("clicks", &Info->Clicks);
		}

		thread_id Thread = spawn_thread(MouseThread, "LGI_MouseWatcher", B_NORMAL_PRIORITY, Info);
		if (Thread > 0)
		{
			resume_thread(Thread);
		}
	}
}

void GView::Sys_MouseMoved(BPoint Point, uint32 transit, const BMessage *message)
{
	VMouse m;
	ulong Buttons;

	Sys_View->GetMouse(&Point, &Buttons, FALSE);
	
	m.x = Point.x;
	m.y = Point.y;
	m.Flags = Buttons;
	m.Down(Buttons != 0);
	
	OnMouseMove(m);
}

void GView::Sys_FrameMoved(BPoint p)
{
	/*
	for (BView *v = GetBView()->Parent(); v; v = v->Parent())
	{
		BRect r = v->Frame();
		p.x -= r.left;
		p.y -= r.top;
	}
	*/

	// Pos.Offset(p.x-Pos.x1, p.y-Pos.y1);
	OnPosChange();
}

bool GView::IsAttached()
{
	return Sys_View AND Sys_View->Parent() != 0;
}

bool GView::Attach(GView *Wnd)
{
	bool Status = FALSE;
	Parent = Wnd;
	if (Parent)
	{
		BView *ParentSysView = Sys_View->Parent();
		if (ParentSysView != Parent->Sys_View)
		{
			if (ParentSysView)
			{
				// detach from the previous BView
				ParentSysView->RemoveChild(Sys_View);
			}

			// bool Lock = Sys_View->LockLooper();
			Sys_View->MoveTo(Pos.x1, Pos.y1);
			Sys_View->ResizeTo(Pos.X()-1, Pos.Y()-1);
			if (NOT GFlags::Visible() AND
				NOT Sys_View->IsHidden())
			{
				Sys_View->Hide();
			}
			if (NOT Enabled())
			{
				Enabled(false);
			}
			// if (Lock) Sys_View->UnlockLooper();

			if (Parent->Sys_View)
			{
				// attach to the new BView
				Parent->Sys_View->AddChild(Sys_View);
			}
		}

		if (NOT Parent->Children.HasItem(this))
		{
			// we aren't already a member of the parents children
			// list so add ourselves in.
			Parent->Children.Insert(this);
			Parent->OnChildrenChanged(this, true);
		}

		Status = true;
	}
	return Status;
}

bool GView::Detach()
{
	bool Status = FALSE;
	if (Parent)
	{
		Visible(FALSE);
		
		bool Lock = Parent->Sys_View->LockLooper();
		Parent->Sys_View->RemoveChild(Sys_View);
		if (Lock) Parent->Sys_View->UnlockLooper();
		
		Parent->Children.Delete(this);
		Parent->OnChildrenChanged(this, FALSE);

		Parent = 0;
		Status = TRUE;
	}
	return Status;
}

GView *GView::FindControl(GWndRef hCtrl)
{
	if (GetBView() == hCtrl)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GView *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

#define MIN_WINDOW_POS	24

void GView::MoveOnScreen()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);

	if (Pos.x2 >= Screen.X())
		Pos.Offset(Screen.X() - Pos.x2, 0);
	if (Pos.y2 >= Screen.Y())
		Pos.Offset(0, Screen.Y() - Pos.y2);
	if (Pos.x1 < 0)
		Pos.Offset(-Pos.x1, 0);
	if (Pos.y1 < MIN_WINDOW_POS)
		Pos.Offset(0, MIN_WINDOW_POS-Pos.y1);
}

void GView::MoveToCenter()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	Pos.Offset(-Pos.x1, -Pos.y1);
	Pos.Offset((Screen.X() - Pos.X()) / 2, (Screen.Y() - Pos.Y()) / 2);
}

void GView::MoveToMouse()
{
	VMouse m;
	if (GetMouse(m, true))
	{
		Pos.Offset(-Pos.x1, -Pos.y1);
		Pos.Offset(m.x-(Pos.X()/2), m.y-(Pos.Y()/2));
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
GWnd::GWnd(GView *notify) :
	BWindow(BRect(10, 10, 100, 100), "TestWnd", B_TITLED_WINDOW, B_ASYNCHRONOUS_CONTROLS, B_CURRENT_WORKSPACE)
{
	Notify = notify;
}

GWnd::GWnd(	GView *notify,
			BRect frame,
			char *title,
			window_type type,
			uint32 flags,
			uint32 workspaces)
	: BWindow(frame, title, type, flags, workspaces)		
{
	Notify = notify;
}

bool GWnd::QuitRequested()
{
	if (Notify)
	{
		return Notify->QuitRequested();
	}
	
	return true;
}

void GWnd::MessageReceived(BMessage *Msg)
{
	// Msg->PrintToStream();
	
	printf("Msg: %i\n", Msg->what);
	
	switch (Msg->what)
	{
		case B_MOUSE_DOWN:
		{
			printf("B_MOUSE_DOWN\n");
			break;
		}
		case B_MOUSE_MOVED:
		{
			printf("B_MOUSE_MOVED\n");
			break;
		}
	}

	BWindow::MessageReceived(Msg);
}

void GWnd::FrameMoved(BPoint origin)
{
	if (Notify)
	{
		Notify->Pos.Offset(	-Notify->Pos.x1 + origin.x,
							-Notify->Pos.y1 + origin.y);
	}
}

void GWnd::FrameResized(float width, float height)
{
}

////////////////////////////////////////////////////////////////////////////////////////////////
GWindow::GWindow() :
	GView(NEW(BViewRedir(this)))
{
	Menu = 0;
	GetBView()->SetResizingMode(B_FOLLOW_ALL_SIDES);
	GetBView()->SetViewColor(B_TRANSPARENT_COLOR);	// Rgb);

	Wnd = NEW(GWnd(this));
	if (Wnd)
	{
		BRect r = Wnd->Bounds();
		GetBView()->ResizeTo(r.Width(), r.Height());
		Wnd->AddChild(GetBView());
	}
}

GWindow::~GWindow()
{
	DeleteObj(Menu);

	if (GetBView()->LockLooper())
	{
		if (GetBView()->Window())
		{
			GetBView()->RemoveSelf();
		}

		Wnd->Hide();
		Wnd->Quit();
	}
}

bool GWindow::QuitRequested()
{
	bool Status = OnRequestClose();
	if (Status)
	{
		if (GetBView()->LockLooper())
		{
			GetBView()->RemoveSelf();
		}
		delete this;
	}
	return Status;
}

void GWindow::OnPaint(GdcBasePrimitives *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

void GWindow::OnChildrenChanged(GView *Wnd, bool Attaching)
{
}

void GWindow::OnPosChange()
{
	GView::OnPosChange();
	Pour();
}

void GWindow::OnCreate()
{
	/*
	if (Wnd)
	{
		BRect f = Wnd->Frame();
		ResizeTo(f.Width(), f.Height());
	}
	*/
}

bool GWindow::Visible()
{
	bool Status = FALSE;
	if (Wnd)
	{
		Status = NOT Wnd->IsHidden();
	}
	return Status;
}

void GWindow::Visible(bool v)
{
	if (Wnd)
	{
		if (v)
		{
			if (Wnd->IsHidden())
			{
				Wnd->Show();
				Invalidate(0, true);
			}
		}
		else
		{
			if (NOT Wnd->IsHidden())
			{
				Wnd->Hide();
			}
		}
	}

	if (v)
	{
		Pour();
	}
}

bool GWindow::Name(char *n)
{
	bool Status = GObject::Name(n);
	if (Wnd AND
		Sys_View->LockLooper())
	{
		Wnd->SetTitle(n);
		Sys_View->UnlockLooper();
	}
	return Status;
}

char *GWindow::Name()
{
	return GObject::Name();
}

GdcRegion &GWindow::GetPos()
{
	GdcRegion r(0, 0, 1, 1);
	if (Wnd)
	{
		BRect frame = Wnd->Frame();
		r.x1 = frame.left;
		r.y1 = frame.top;
		r.x2 = frame.right;
		r.y2 = frame.bottom;
	}
	return r;
}

bool GWindow::SetPos(GdcRegion &p, bool Repaint)
{
	if (Wnd)
	{
		Wnd->MoveTo(p.x1, p.y1);
		Wnd->ResizeTo(p.X(), p.Y());
	}
	return TRUE;
}

void GWindow::Pour()
{
	bool SafeToLock = false;

	if (GetBWindow()->IsLocked())
	{
		// I'm already locked... this could be bad if it's not me
		thread_id Thread = GetBWindow()->LockingThread();
		if (Thread != -1)
		{
			if (Thread == find_thread(NULL))
			{
				// it's all ok, I'm locking me
				SafeToLock = true;
			}
			else
			{
				// someone else is locking us
				// ok who is locking me?
				thread_info Info;
				if (get_thread_info(Thread, &Info) == B_OK)
				{
					printf("Evil locking thread: %i (%s)\n", Thread, Info.name);
				}
				else
				{
					printf("Couldn't get evil thread info\n");
				}
			}
		}
	}
	else
	{
		SafeToLock = true;
	}
	
	if (NOT SafeToLock)
	{
		return;
	}
	
	bool Lock = GetBView()->LockLooper();

	GetBWindow()->BeginViewTransaction();
	GdcRegion r(GetBView()->Frame());
	GdcComplexRegion Client(r);
	GdcComplexRegion Update;

	if (Menu AND Menu->Bar)
	{
		BRect Rc = Menu->Bar->Frame();
		GdcRegion Mp = Rc;
		Mp.x2 = 10000;
		Client.Subtract(&Mp);
	}

	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GdcRegion OldPos = w->Pos;
		Update.Union(&OldPos);

		if (w->Pour(Client))
		{
			if (NOT w->Visible())
			{
				w->Visible(true);
			}

			w->Invalidate();

			Client.Subtract(&w->Pos);
			Update.Subtract(&w->Pos);
		}
		else
		{
			// non-pourable
		}
	}

	GetBWindow()->EndViewTransaction();
	GetBWindow()->Flush();

	GetBView()->Invalidate();
	if (Lock) GetBView()->UnlockLooper();
}

int GWindow::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		case B_SIMPLE_DATA:
		{
			List<char> Files;
			int32 Count = 0;
			type_code Type = 0;
			
			if (Msg->GetInfo("refs", &Type, &Count) == B_OK)
			{
				for (int i=0; i<Count; i++)
				{
					entry_ref Ref;
					if (Msg->FindRef("refs", i, &Ref) == B_OK)
					{
						BPath Path;
						BEntry Entry(&Ref, true);
						Entry.GetPath(&Path);
						char *p = (char*) Path.Path();
						if (p)
						{
							Files.Insert(NewStr(p));
						}
					}
				}
			}
			
			if (Files.GetItems() > 0)
			{
				OnReceiveFiles(Files);
			}
			
			for (char *f=Files.First(); f; f=Files.Next())
			{
				DeleteArray(f);
			}
			break;
		}
		case WM_COMMAND:
		{
			int32 Cmd = 0;
			int32 Event = 0;
			Msg->FindInt32("Cmd", &Cmd);
			Msg->FindInt32("Event", &Event);
			OnCommand(Cmd, Event, 0);
			break;
		}
	}

	return GView::OnEvent(Msg);
}

/////////////////////////////////////////////////////////////////////////////
GApp *GApp::TheApp = 0;

GApp::GApp(char *Mime, char *lpCmdLine) : BApplication(Mime)
{
	LgiAssert(0);
	TheApp = this;

	srand(system_time());
	CmdLine = NewStr(lpCmdLine);
	LgiInitColours();
	AppWnd = 0;

	// System font
	SystemFont = 0;
	GdcFontType SysFontType;
	if (SysFontType.GetSystemFont("System"))
	{
		SystemFont = SysFontType.CreateFont();
		if (SystemFont)
		{
			SystemFont->Transparent(true);
		}
	}

	if (NOT SystemFont)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}
}

GApp::~GApp()
{
	// DeleteObj(AppWnd);	
	// DeleteObj(SystemFont);
	TheApp = 0;
}

void GApp::RefsReceived(BMessage *Msg)
{
	List<char> Files;
	int32 Count = 0;
	type_code Type = 0;
	
	if (Msg->GetInfo("refs", &Type, &Count) == B_OK)
	{
		for (int i=0; i<Count; i++)
		{
			entry_ref Ref;
			if (Msg->FindRef("refs", i, &Ref) == B_OK)
			{
				BPath Path;
				BEntry Entry(&Ref, true);
				Entry.GetPath(&Path);
				char *p = (char*) Path.Path();
				if (p)
				{
					Files.Insert(NewStr(p));
				}
			}
		}
	}
	
	if (AppWnd AND Files.GetItems() > 0)
	{
		AppWnd->OnReceiveFiles(Files);
	}
	
	for (char *f=Files.First(); f; f=Files.Next())
	{
		DeleteArray(f);
	}
}

void GApp::OnReceiveFiles(List<char> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

XmlTag *GApp::GetConfig(char *Tag)
{
	if (NOT Config)
	{
		char Path[256];
		if (LgiGetExePath(Path))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, "lgi.conf");
			if (FileExists(Path))
			{
				Config = NEW(Xml);
				if (Config)
				{
					if (Config->ParseXmlFile(Path) <= 0)
					{
						DeleteObj(Config);
					}
				}
			}
		}
	}

	if (Tag AND Config)
	{
		for (XmlTag *t=Config->Values.First(); t; t=Config->Values.Next())
		{
			if (t->Name AND
				stricmp(t->Name, Tag) == 0)
			{
				return t;
			}
		}
	}

	return 0;
}

bool GApp::GetOption(char *Dest, char Option)
{
	/*
	printf("GApp::GetOption(%c)\n", Option);
	printf("CmdLine: '%s'\n", CmdLine);
	*/

	char *c = CmdLine;
	while (c AND *c)
	{
		if (*c == '/' OR *c == '\\' OR *c == '-')
		{
			c++;
			if (*c == Option)
			{
				c++;
				if (*c)
				{
					while (*c AND strchr(" \t\r\n", *c))
					{
						c++;
					}

					for (char *i = c; *i AND *i != ' ';)
					{
						*Dest++ = *i++;
					}
					*Dest++ = 0;
					return TRUE;
				}
			}
		}

		c = strchr(c, ' ');
		if (c) c++;
	}

	return FALSE;
}

void GApp::OnCmdLine(char *Cmd)
{
	char WhiteSpace[] = " \r\n\t";

	DeleteArray(CmdLine);
	CmdLine = NewStr(Cmd);

	if (CmdLine)
	{
		List<char> Files;
		
		char *Delim = "\'\"";
		char *s;
		for (s = CmdLine; *s; )
		{
			// skip ws
			while (strchr(WhiteSpace, *s)) s++;

			// read to end of token
			char *e = s;
			if (strchr(Delim, *s))
			{
				for (e = ++s; *e AND NOT strchr(Delim, *e); e++);
			}
			else
			{
				for (; *e AND NOT strchr(WhiteSpace, *e); e++);
			}

			char *Arg = NewStr(s, (int)e-(int)s);
			if (Arg)
			{
				if (FileExists(Arg))
				{
					Files.Insert(Arg);
				}
				else
				{
					DeleteArray(Arg);
				}
			}

			// next
			s = (*e) ? e + 1 : e;
		}

		// call app
		if (Files.GetItems() > 0)
		{
			OnReceiveFiles(Files);
		}

		// clear up
		DeleteArrayList(&Files, char);
	}
}

void GApp::Exit(int Code)
{
	// DeleteObj(AppWnd);
	be_app->PostMessage(B_QUIT_REQUESTED);
}

////////////////////////////////////////////////////////////////////////////
int32 ThreadEntryPoint(void *i)
{
	if (i)
	{
		GThread *Thread = (GThread*) i;
		Thread->OnBeforeMain();
		Thread->ReturnValue = Thread->Main();
		Thread->OnAfterMain();
		Thread->State = LGITHREAD_EXITED;
		return Thread->ReturnValue;
	}
	return 0;
}

GThread::GThread()
{
	State = LGITHREAD_ASLEEP;
	ReturnValue = -1;
	hThread = spawn_thread(	ThreadEntryPoint,
							"Lgi.GThread", 
							B_NORMAL_PRIORITY, 
							this);
}

GThread::~GThread()
{
	if (NOT IsExited())
	{
		Terminate();
	}
}

int GThread::Priority()
{
	thread_info Info;
	if (get_thread_info(hThread, &Info) == B_OK)
	{
		return Info.priority;
	}
	return 0;
}

void GThread::Priority(int p)
{
	set_thread_priority(hThread, p);
}

int GThread::ExitCode()
{
	return ReturnValue;
}

bool GThread::IsExited()
{
	return State == LGITHREAD_EXITED;
}

bool GThread::IsRunning()
{
	return State == LGITHREAD_RUNNING;
}

void GThread::Run(bool b)
{
	if (b)
	{
		resume_thread(hThread);
		State = LGITHREAD_RUNNING;
		OnRun(b);
	}
	else
	{
		OnRun(b);
		suspend_thread(hThread);
		State = LGITHREAD_ASLEEP;
	}
}

void GThread::Terminate()
{
	if (kill_thread(hThread) == B_OK)
	{
		State = LGITHREAD_EXITED;
	}
}

int GThread::Main()
{
	return 0;
}

////////////////////////////////////////////////////////////////////////////
int GTrayIcon::Id = 1;

GTrayIcon::GTrayIcon(int icons)
{
	Val = 0;
	Msg = 0;
	Notify = 0;
	MyId = 0;
	Icons = max(icons, 1);
}

GTrayIcon::~GTrayIcon()
{
	Visible(FALSE);
}

int GTrayIcon::OnEvent(GMessage *Msg)
{
	return 0;
}

void GTrayIcon::OnMouseClick(VMouse &m)
{
}

bool GTrayIcon::LoadIcon(char *Str, int Index)
{
	return FALSE;
}

bool GTrayIcon::Visible()
{
	return 0;
}

void GTrayIcon::Visible(bool v)
{
	if (Visible() != v)
	{
		if (v)
		{
		}
		else
		{
		}
	}
}

int GTrayIcon::Value()
{
	return Val;
}

void GTrayIcon::Value(int v)
{
	Val = v;
}

///////////////////////////////////////////////////////////////////
bool LgiIsMounted(char *Name)
{
	return false;
}

bool LgiMountVolume(char *Name)
{
	return false;
}

////////////////////////////////////////////////////////////////////
GPrinter::GPrinter()
{
}

GPrinter::~GPrinter()
{
}

void *GPrinter::StartDC(GView *Parent)
{
	return NULL;
}

void GPrinter::EndDC()
{
}

bool GPrinter::Browse(GView *Parent)
{
	return false;
}

#define MAGIC_PRINTDLG					0xAAFF0100
#define MAGIC_DEVMODE					0xAAFF0101
#define MAGIC_DEVNAMES					0xAAFF0102

bool GPrinter::Serialize(pchar &Str, bool Write)
{
	if (Write)
	{
	}
	else
	{
	}

	return false;
}

