/*hdr
**      FILE:           GuiCommon.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Common windows and controls
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include "Lgi.h"

#include "Clipboard.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
GFileSelect::GFileSelect() : BHandler("GFileSelect")
{
	Size = 4096;
	FileStr = NEW(char[Size]);
	if (FileStr)
	{
		*FileStr = 0;
	}
	else
	{
		Size = 0;
	}
	
	MultiSel = FALSE;
	InitDir = 0;
	DefExt = 0;
	TitleStr = 0;
	Panel = 0;
}

GFileSelect::~GFileSelect()
{
	DeleteObj(Panel);
	DeleteArray(FileStr);
	DeleteArray(InitDir);
	DeleteArray(DefExt);
	DeleteArray(TitleStr);

	for (GFileType *Type = TypeList.First(); Type; Type = TypeList.Next())
	{
		DeleteObj(Type);
	}
}

char *GFileSelect::Name()
{
	return FileStr;
}

bool GFileSelect::Name(char *n)
{
	bool Status = FALSE;

	if (n AND FileStr)
	{
		strncpy(FileStr, n, Size);
	}

	return Status;
}

char *GFileSelect::operator [](int i)
{
	return NULL;
}

int GFileSelect::Names()
{
	return 1;
}

char *GFileSelect::File()
{
	char *Ret = 0;
	char *Last = FileStr;
	char *Dir = strchr(Last, '/');
	while (Dir)
	{
		Last = Dir + 1;
		Dir = strchr(Last, '/');
	}
	return Last;
}

char *GFileSelect::Extension()
{
	char *Ret = File();
	char *Dot = strchr(Ret, '.');
	if (Dot)
	{
		return Dot + 1;
	}
	return NULL;
}

int GFileSelect::Types()
{
	return TypeList.GetItems();
}

void GFileSelect::ClearTypes()
{
	DeleteObjList(&TypeList, GFileType);
}

GFileType *GFileSelect::TypeAt(int n)
{
	return TypeList[n];
}

bool GFileSelect::Type(char *Description, char *Extension, int Data)
{
	GFileType *Type = NEW(GFileType);
	if (Type)
	{
		Type->Description(Description);
		Type->Extension(Extension);
		TypeList.Insert(Type);
	}

	return Type != 0;
}

int GFileSelect::SelectedType()
{
	return -1;
}

GView *GFileSelect::Parent()
{
	return ParentWnd;
}

void GFileSelect::Parent(GView *Window)
{
	ParentWnd = Window;
}

bool GFileSelect::MultiSelect()
{
	return MultiSel;
}

void GFileSelect::MultiSelect(bool Multi)
{
	MultiSel = Multi;
}

char *GFileSelect::InitialDir()
{
	return InitDir;
}

void GFileSelect::InitialDir(char *Dir)
{
	DeleteArray(InitDir);
	InitDir = NewStr(Dir);
}

char *GFileSelect::Title()
{
	return TitleStr;
}

void GFileSelect::Title(char *title)
{
	DeleteArray(TitleStr);
	TitleStr = NewStr(title);
}

char *GFileSelect::DefaultExtension()
{
	return DefExt;
}

void GFileSelect::DefaultExtension(char *DefExt)
{
	DeleteArray(DefExt);
	DefExt = NewStr(DefExt);
}

void GFileSelect::Wait()
{
	thread_id	this_tid = find_thread(NULL);
	BLooper		*pLoop;
	BWindow		*pWin = 0;

	WaitForMessage = true;
	pLoop = BLooper::LooperForThread(this_tid);
	if (pLoop)
	{
		pWin = dynamic_cast<BWindow*>(pLoop);
	}

	if (pWin)
	{
		do
		{
			// update the window periodically
			pWin->UpdateIfNeeded();
			snooze(50);
		} while (Panel->IsShowing() AND WaitForMessage);
	}
	else
	{
		do
		{
			// just wait for exit
			snooze(50);
		} while (Panel->IsShowing() AND WaitForMessage);
	}
}

void GFileSelect::MessageReceived(BMessage *message)
{
	WaitForMessage = false;

	entry_ref ref;
	char *name = 0;
	if (message->FindRef("refs", &ref) == B_OK)
	{
		// open
		BEntry Entry(&ref, true);
		BPath Path;
		Entry.GetPath(&Path);
		strcpy(FileStr, Path.Path());
	}
	else if (message->FindRef("directory", &ref) == B_OK AND
			message->FindString("name", (const char**) &name) == B_OK)
	{
		// save
		BEntry Entry(&ref, true);
		BPath Path;
		Entry.GetPath(&Path);

		strcpy(FileStr, Path.Path());
		strcat(FileStr, DIR_STR);
		strcat(FileStr, name);
	}
}

bool GFileSelect::DoDialog(file_panel_mode Mode, int32 NodeType)
{
	BEntry Dir(InitDir);
	entry_ref DirRef;
	Dir.GetRef(&DirRef);

	Messenger = 0;
	Panel = NEW(BFilePanel(	Mode,
							NULL,
							&DirRef,
							NodeType,
							MultiSel,
							NEW(BMessage(WM_CHANGE)),
							NULL,
							TRUE, // modal
							FALSE)); // hide when done
	if (Panel)
	{
		if (ValidStr(FileStr))
		{
			Panel->SetSaveText(FileStr);
		}
	
		BWindow *Wnd = Panel->Window();
		if (Wnd->Lock())
		{
			Wnd->AddHandler(this);
			Wnd->Unlock();
		}
		
		Messenger = NEW(BMessenger(this));
		Panel->SetTarget(*Messenger);
		Panel->Show();
		Wait();
		Panel->Hide();
	}

	return strlen(FileStr) > 0;
}

bool GFileSelect::Open()
{
	return DoDialog(B_OPEN_PANEL);
}

bool GFileSelect::OpenFolder()
{
	return DoDialog(B_OPEN_PANEL, B_DIRECTORY_NODE);
}

bool GFileSelect::Save()
{
	return DoDialog(B_SAVE_PANEL);
}

///////////////////////////////////////////////////////////////////////////////////////////////
GFindReplace::GFindReplace(GView *Parent)
{
}

GFindReplace::~GFindReplace()
{
}

void GFindReplace::SetParent(GView *Parent)
{
}

bool GFindReplace::OpenFind()
{
	return FALSE;
}

bool GFindReplace::OpenReplace()
{
	return FALSE;
}

int GFindReplace::OnEvent(GMessage *Msg)
{
	int Status = FALSE;
	return Status;
}

bool GFindReplace::OnFind()
{
	return TRUE;
}

bool GFindReplace::OnReplace()
{
	return TRUE;
}

bool GFindReplace::CaseSensitive()
{
	return FALSE;
}

void GFindReplace::CaseSensitive(bool i)
{
}

bool GFindReplace::WholeWord()
{
	return 0;
}

void GFindReplace::WholeWord(bool i)
{
}

bool GFindReplace::BtnFindNext()
{
	return 0;
}

bool GFindReplace::BtnReplace()
{
	return 0;
}

bool GFindReplace::BtnReplaceAll()
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
GInput::GInput(GView *parent, char *InitStr, char *Msg, char *Title, bool Password)
{
	SetParent(parent);
	Str = 0;
	Pos.ZOff(300, 75);
	GView::Name(Title);
	MoveToCenter();
	Children.Insert(NEW(GText(-1, 5, 5, 200, 20, Msg)));
	Children.Insert(Edit = NEW(GEdit(200, 5, 30, 200, 20, InitStr)));
	GButton *Ok = 0;
	Children.Insert(Ok = NEW(GButton(IDOK, 215, 5, 60, 20, "Ok")));
	Children.Insert(NEW(GButton(IDCANCEL, 215, 30, 60, 20, "Cancel")));

	if (Ok)
	{
		Ok->Default(true);
	}
}

GInput::~GInput()
{
	DeleteArray(Str);
}

int GInput::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDOK:
		{
			Str = NewStr(Edit->Name());
			// fall thru
		}
		case IDCANCEL:
		{
			EndModal(Ctrl->GetId() == IDOK);
			break;
		}
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
int FontSizes[] = { 8, 9, 10, 11, 12, 14, 16, 18, 20, 24, 28, 32, 48, 72, 100, 0 };

GSelectFont::GSelectFont(GView *parent, char *Face)
{
	Font[0] = 0;
	Size = 0;

	if (Face)
	{
		char *Comma = strchr(Face, ',');
		if (Comma)
		{
			char *FontName = NewStr(Face, (int)Comma-(int)Face);
			if (FontName)
			{
				strcpy(Font, FontName);
				DeleteArray(FontName);
			}

			Size = atoi(Comma+2);
		}
	}
	
	Name("Select Font");
	Pos.ZOff(180, 100);
	MoveToCenter();

	int DefFont = 0;
	int DefSize = 0;
	
	Children.Insert(NEW(GText(-1, 10, 10, -1, -1, "Font:")));
	Children.Insert(CFont = NEW(GCombo(100, 50, 10, 100, 20, "Font")));
	if (CFont)
	{
		int32 numFamilies = count_font_families();
		for (int32 i = 0; i < numFamilies; i++)
		{
			font_family Temp;
			uint32 flags;
	
			if (get_font_family(i, &Temp, &flags) == B_OK)
			{
				CFont->Insert(Temp);
				if (stricmp(Temp, Font) == 0)
				{
					DefFont = i;
				}
	
				/*
				int32 numStyles = count_font_styles(family);
				for (int32 j = 0; j < numStyles; j++)
				{
					font_style style;
					if (get_font_style(family, j, &style, &flags) == B_OK)
					{
					}
				}
				*/
			}
		}
		
		CFont->Index(DefFont);
	}
	
	Children.Insert(NEW(GText(-1, 10, 40, -1, -1, "Size:")));
	Children.Insert(CSize = NEW(GCombo(100, 50, 40, 100, 20, "Size")));
	if (CSize)
	{
		int n=0;
		for (int *i = FontSizes; *i; i++, n++)
		{
			char Str[20];
			sprintf(Str, "%i", *i);
			CSize->Insert(Str);
			if (*i == Size)
			{
				DefSize = n;
			}
		}

		CSize->Index(DefSize);
	}
	
	Children.Insert(NEW(GButton(IDOK, 20, 70, 60, 20, "Ok")));
	Children.Insert(NEW(GButton(IDCANCEL, 90, 70, 60, 20, "Cancel")));
}

GSelectFont::~GSelectFont()
{
}

int GSelectFont::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDOK:
		{
			if (CFont AND CSize)
			{
				strcpy(Font, CFont->GCombo::Name());
				Size = FontSizes[CSize->Index()];
			}
			
			EndModal(1);
			break;
		}
		case IDCANCEL:
		{
			EndModal(0);
			break;
		}
	}
	
	return 0;
}



