/*hdr
**      FILE:           GuiDlg.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998 Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "Lgi.h"

#define TICKS_PER_SECOND					1000000
#define DOUBLE_CLICK_TIMEOUT				(TICKS_PER_SECOND/3)

#define DefaultViewColour()					SetViewColor(B_TRANSPARENT_COLOR); SetFlags(Flags() & ~B_FULL_UPDATE_ON_RESIZE)
// #define DefaultViewColour()					SetViewColor(216, 216, 216)

///////////////////////////////////////////////////////////////////////////////////////////
// Standard sub-class of any BControl
template <class b>
class BSub : public b
{
	GView *View;

public:
	BSub(GView *v) : b(BRect(0, 0, 1, 1), "BSub")
	{
		View = v;
	}

	void AttachedToWindow()
	{
		if (View) View->OnCreate();
	}

	/*
	void DetachedFromWindow()
	{
		if (View) View->OnDestroy();
	}
	
	void Draw(BRect UpdateRect) {}
	void FrameMoved(BPoint Point) {}
	void FrameResized(float width, float height) {}
	void Pulse() {}
	void MessageReceived(BMessage *message) {}
	void MakeFocus(bool f = true) {}
	void KeyDown(const char *bytes, int32 numBytes) {}
	void KeyUp(const char *bytes, int32 numBytes) {}
	void MouseDown(BPoint point) {}
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) {}
	bool QuitRequested() {}
	*/
};


///////////////////////////////////////////////////////////////////////////////////////////
extern rgb_color BeRGB(uchar r, uchar g, uchar b);

rgb_color BeRGB(uchar r, uchar g, uchar b)
{
	rgb_color c;
	c.red = r;
	c.green = g;
	c.blue = b;
	return c;
}

/////////////////////////////////////////////////////////////////////////////////////////
#define LGI_WATCHER_MSG			0x80000000

class GMouseWatcher : public GThread
{
	int TimeOut;
	GView *Notify;

public:
	GMouseWatcher(int timeout, GView *notify)
	{
		TimeOut = timeout;
		Notify = notify;
		if (Notify)
		{
			Run(true);
		}
	}
	
	void Convert(GdcPt2 &Pt)
	{
		BPoint BPt(Pt.x, Pt.y);
		Notify->GetBView()->ConvertFromScreen(&BPt);
		Pt.x = BPt.x;
		Pt.y = BPt.y;
	}
	
	int Main()
	{
		VMouse m;
		
		Notify->GetMouse(m);
		m.Flags |= LGI_WATCHER_MSG;
	
		while (m.Left() OR m.Right())
		{
			LgiSleep(TimeOut);

			VMouse Ms;
			Notify->GetMouse(Ms);
		
			if (Ms.x != m.x OR
				Ms.y != m.y)
			{
				if (Notify->GetBView()->LockLooper())
				{
					m.x = Ms.x;
					m.y = Ms.y;
					m.Flags |= LGI_WATCHER_MSG;
					Notify->OnMouseMove(m);

					Notify->GetBView()->UnlockLooper();
				}
			}
		}
	
		return 0;
	}
};



///////////////////////////////////////////////////////////////////////////////////////////
GDialog::GDialog() :
#ifdef LGI_RES
	ResObject(Res_Dialog),
#endif
	GView(NEW(BViewRedir(this)))
{
	ModalSem = 0;
	Name("Dialog");
	GetBView()->SetViewColor(216, 216, 216);
	GetBView()->SetResizingMode(B_FOLLOW_ALL);

	Wnd = NEW(GWnd(this, BRect(200, 200, 500, 500), "", B_MODAL_WINDOW, B_NOT_RESIZABLE));
	if (Wnd)
	{
		Wnd->SetLook(B_TITLED_WINDOW_LOOK);
		
		BRect r = Wnd->Bounds();
		GetBView()->ResizeTo(r.Width(), r.Height());
		Wnd->AddChild(GetBView());
	}
}

GDialog::~GDialog()
{
	EndModeless(0);
}

bool GDialog::QuitRequested()
{
	if (ModalSem)
	{
		EndModal();
	}
	
	return TRUE;
}

/*
GView *GDialog::FindControl(int Id)
{
	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}

	return FALSE;
}
*/

bool GDialog::LoadFromResource(char *Resource)
{
	char n[256];
	bool Status = GLgiRes::LoadFromResource(Resource, Children, &Pos, n);
	if (Status)
	{
		Name(n);
	}
	return Status;
}

status_t WaitForDelete(sem_id blocker, BWindow *Wnd)
{
	status_t	result;
	thread_id	this_tid = find_thread(NULL);

	// block until semaphore is deleted (modal is finished)
	if (Wnd)
	{
		do
		{
			// update the window periodically
			Wnd->UpdateIfNeeded();
			result = acquire_sem_etc(blocker, 1, B_TIMEOUT, 10000);
		} while (result != B_BAD_SEM_ID);
	}
	else
	{
		do
		{
			// just wait for exit
			// result = acquire_sem(blocker);
			result = acquire_sem_etc(blocker, 1, B_TIMEOUT, 10000);
		} while (result != B_BAD_SEM_ID);
	}
	
	return result;
}

int GDialog::DoModal()
{
	ModalRet = 0;
	ModalSem = create_sem(0, "ModalSem");

	if (Wnd->Lock())
	{
		Wnd->ResizeTo(Pos.X(), Pos.Y());
		Wnd->MoveTo(Pos.x1, Pos.y1);
		Wnd->SetTitle(GObject::Name());
		AttachChildren();
		Wnd->Show();
		Wnd->Unlock();

		OnCreate();
	}
	
	// MFC NOTE: WaitForDelete is different from CWnd::RunModalLoop
	// because it doesn't process messages.
	WaitForDelete(ModalSem, GetParent() ? GetParent()->GetBWindow() : 0);
	if (Wnd->Lock())
	{
		Wnd->Quit();
	}

	OnDestroy();
	return ModalRet;
}

int GDialog::DoModeless()
{
	if (Wnd->Lock())
	{
		Wnd->ResizeTo(Pos.X(), Pos.Y());
		Wnd->MoveTo(Pos.x1, Pos.y1);
		Wnd->SetTitle(GObject::Name());
		AttachChildren();
		Wnd->Show();
		Wnd->Unlock();
		
		OnCreate();
	}
	return FALSE;
}

int GDialog::OnEvent(BMessage *Msg)
{
	int Status = 0;

	// printf("GDialog::OnEvent(%i)\n", Msg->what);

	return Status;
}

void GDialog::EndModal(int Code)
{
	ModalRet = Code;
	GetBView()->RemoveSelf();
	if (Wnd->Lock())
	{
		Wnd->Hide();
		Wnd->Unlock();
	}
	delete_sem(ModalSem); // causes DoModal to unblock
}

void GDialog::EndModeless(int Code)
{
	if (Wnd->Lock())
	{
		GetBView()->RemoveSelf();
		Wnd->Hide();
		Wnd->Quit();
	}
	
	OnDestroy();
}

///////////////////////////////////////////////////////////////////////////////////////////
GControl::GControl(BView *view) : GView(view)
{
	Pos.ZOff(10, 10);
	SetId(0);
	Sunken(TRUE);
	Sys_LastClick = 0;
}

GControl::~GControl()
{
}

bool GControl::Attach(GView *parent)
{
	return GView::Attach(parent);
	/*
	Parent = parent;
	if (Parent AND Sys_View)
	{
		bool Lock = Parent->Sys_View->LockLooper();
		Parent->Sys_View->AddChild(Sys_View);
		Sys_View->ResizeTo(Pos.X(), Pos.Y());
		Sys_View->MoveTo(Pos.x1, Pos.y1);
		if (Lock) Parent->Sys_View->UnlockLooper();
		return TRUE;
	}
	return 0;
	*/
}

int GControl::OnEvent(BMessage *Msg)
{
	return 0;
}

GdcPt2 GControl::SizeOfStr(char *Str)
{
	GdcPt2 Pt(0, 0);
	if (Str)
	{
		BFont Font;
		Sys_View->GetFont(&Font);
		
		font_height Ht;
		Font.GetHeight(&Ht);
		
		Pt.y = Ht.ascent + Ht.descent + Ht.leading + 0.9999;
		Pt.x = Font.StringWidth(Str);
	}

	return Pt;
}

void GControl::MouseClickEvent(bool Down)
{
	static int LastX = -1, LastY = -1;

	VMouse m;
	BPoint Pt;
	uint32 Buttons;
	if (GetBView()->LockLooper())
	{
		GetBView()->GetMouse(&Pt, &Buttons, FALSE);
		m.x = Pt.x;
		m.y = Pt.y;
		m.Flags = Buttons;

		if (Down)
		{
			int32 Clicks = 1;
			GetBWindow()->CurrentMessage()->FindInt32("clicks", &Clicks);
			m.Double(	(Clicks > 1) AND
						(abs(m.x-LastX) < DOUBLE_CLICK_THRESHOLD) AND
						(abs(m.y-LastY) < DOUBLE_CLICK_THRESHOLD));
			
			LastX = m.x;
			LastY = m.y;
		}
		else
		{
			m.Double(false);
		}
		
		m.Down(Down);
		OnMouseClick(m);
		GetBView()->UnlockLooper();
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
// Button
class GButtonRedir : public BButton
{
	GButton *Obj;
	
public:
	GButtonRedir(GButton *obj) :
		BButton(BRect(0, 0, 1, 1), "Button", "", NEW(BMessage(WM_CHANGE)))
	{
		Obj = obj;
	}
	
	status_t Invoke(BMessage *message = NULL)
	{
		status_t Ret = BButton::Invoke(message);
	
		if (Obj)
		{
			GView *n = (Obj->GetNotify()) ? Obj->GetNotify() : Obj->GetParent();
			if (n)
			{
				n->OnNotify(Obj, 0);
			}
		}
		
		return Ret;
	}
	
	void AttachedToWindow()
	{
		if (Obj) Obj->OnCreate();
	}
	
	void KeyDown(const char *bytes, int32 numBytes)
	{
		if (Obj) Obj->Sys_KeyDown(bytes, numBytes);
	}
	
	void KeyUp(const char *bytes, int32 numBytes)
	{
		if (Obj) Obj->Sys_KeyUp(bytes, numBytes);
	}
};

GButton::GButton(int id, int x, int y, int cx, int cy, char *name)
	: GControl(Button = NEW(GButtonRedir(this)) )
#ifdef LGI_RES
	, ResObject(Res_Button)
#endif
{
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	if (Button)
	{
		Button->SetViewColor(216, 216, 216);
	}
}

GButton::~GButton()
{
}

bool GButton::Default()
{
	return (Button) ? Button->IsDefault() : false;
}

void GButton::Default(bool b)
{
	if (Button) Button->MakeDefault(b);
}

bool GButton::Name(char *s)
{
	bool Status = GObject::Name(s);
	if (Button)
	{
		bool Lock = Button->LockLooper();
		Button->SetLabel(s);
		if (Lock) Button->UnlockLooper();
	}
	return Status;
}

int GButton::OnEvent(GMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Edit
/*
B_FOLLOW_LEFT | B_FOLLOW_TOP,
B_FRAME_EVENTS | B_WILL_DRAW | B_NAVIGABLE
*/

class GEditRedir : public BTextControl
{
	friend class GEdit;
	
	GEdit *Obj;
	bool Ignore;

public:
	GEditRedir(GEdit *obj) :
		BTextControl(BRect(0, 0, 100, 20),
					"Edit",
					"",
					"",
					NEW(BMessage(WM_CHANGE)) )					
	{
		Obj = obj;
		Ignore = false;
		SetDivider(0);
		DefaultViewColour();
	}

	status_t Invoke(BMessage *message)
	{
		if (NOT Ignore AND Obj AND Obj->GetParent())
		{
			Obj->GetParent()->OnNotify(Obj, 13);
		}
		
		return BTextControl::Invoke(message);
	}
};

GEdit::GEdit(int id, int x, int y, int cx, int cy, char *name) :
	GControl(Edit = NEW(GEditRedir(this)) )
#ifdef LGI_RES
	, ResObject(Res_EditBox)
#endif
{
	GdcPt2 Size = SizeOfStr((name)?name:(char*)"A");
	if (cx < 0) cx = Size.x + 6;
	if (cy < 0) cy = Size.y + 4;

	MultiLineEdit = false;
	PasswordEdit = false;
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	SetPos(Pos);
	Name(name);
}

GEdit::~GEdit()
{
}

void GEdit::Password(bool i)
{
	PasswordEdit = i;
	if (Edit)
	{
		Edit->TextView()->HideTyping(PasswordEdit);
	}
}

bool GEdit::SetPos(GdcRegion &p, bool Repaint)
{
	Pos = p;
	if (Edit AND Edit->TextView())
	{
		bool Lock = Edit->LockLooper();
		Edit->MoveTo(p.x1, p.y1);
		Edit->ResizeTo(p.X()-1, p.Y()-1);
		Edit->TextView()->ResizeTo(p.X()-10, p.Y()-7);
		if (Lock) Edit->UnlockLooper();
	}
	return true;
}

void GEdit::MultiLine(bool m)
{
	MultiLineEdit = m;
}

void GEdit::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GEdit::Value()
{
	char *n = Name();
	return (n) ? atoi(n) : 0;
}

char *GEdit::Name()
{
	return (char*) ((Edit) ? Edit->Text() : 0);
}

bool GEdit::Name(char *s)
{
	GObject::Name(s);
	if (Edit)
	{
		bool Lock = Edit->LockLooper();
		((GEditRedir*)Edit)->Ignore = true;
		Edit->SetText(s);
		((GEditRedir*)Edit)->Ignore = false;
		if (Lock) Edit->UnlockLooper();
	}
	return true;
}

int GEdit::OnEvent(BMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Check box
class GCheckBoxRedir : public BCheckBox
{
	GCheckBox *Obj;
	
public:
	GCheckBoxRedir(GCheckBox *obj, int Init) :
		BCheckBox(BRect(0,0,1,1), "Check", "", NEW(BMessage(WM_CHANGE)))
	{
		Obj = obj;
		SetValue(Init);
		DefaultViewColour();
	}
};

GCheckBox::GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState)
	: GControl(Box = NEW(GCheckBoxRedir(this, InitState)) )
#ifdef LGI_RES
	, ResObject(Res_CheckBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Val = InitState;
	Three = FALSE;
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	if (name) Name(name);
}

GCheckBox::~GCheckBox()
{
}

int GCheckBox::Value()
{
	if (Box)
	{
		Val = Box->Value();
	}
	return Val;
}

void GCheckBox::Value(int b)
{
	Val = b;
	if (Box)
	{
		Box->SetValue(Val);
	}
}

bool GCheckBox::Name(char *s)
{
	bool Status = GObject::Name(s);
	if (Box)
	{
		bool Lock = Box->LockLooper();
		Box->SetLabel(s);
		if (Lock) Box->UnlockLooper();
	}
	return Status;
}

int GCheckBox::OnEvent(BMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Text
GText::GText(int id, int x, int y, int cx, int cy, char *name) :
	GControl(this),
	BStringView(BRect(0, 0, 1, 1), "Text", "")
#ifdef LGI_RES
	, ResObject(Res_StaticText)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x;
	if (cy < 0) cy = Size.y;

	GText::Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	DefaultViewColour();
}

GText::~GText()
{
}

// I didn't want to override this but I have to.
// The BeOS control can't handle multi-line text...
// So I handle that myself.
void GText::Draw(BRect updateRect)
{
	Sys_View->SetHighColor(216, 216, 216);
	Sys_View->SetLowColor(216, 216, 216);
	Sys_View->FillRect(BRect(0, 0, X(), Y()));
	Sys_View->SetHighColor(0, 0, 0);

	int y = 1;
	char *c = (char*) Text();
	if (c)
	{
		char *Start = c;
		font_height FntH;
	
		Sys_View->GetFontHeight(&FntH);
		int Ht = FntH.ascent + FntH.descent + FntH.leading + 1;
	
		while (*c)
		{
			if (*c == '\n')
			{
				int32 Len = (int32) c - (int32) Start;
	
				Sys_View->MovePenTo(0, y * Ht);
				Sys_View->DrawString(Start, Len);
				Start = c + 1;
				y++;
			}
			
			c++;
		}
	
		int32 Len = (int32) c - (int32) Start;
		Sys_View->MovePenTo(0, y * Ht);
		Sys_View->DrawString(Start, Len);
	}
}

int GText::OnEvent(BMessage *Msg)
{
	return 0;
}

bool GText::Name(char *n)
{
	if (LockLooper())
	{
		SetText((n)?n:"");
		UnlockLooper();
		return true;
	}
	else
	{
		SetText((n)?n:"");
	}
	return false;
}

char *GText::Name()
{
	char *t = "";
	bool Lock = LockLooper();
	t = (char*)Text();
	if (Lock) UnlockLooper();
	return t;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
int GRadioGroup::NextId = 20000;
GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init)
	: GControl(Box = NEW(BSub<BBox>(this)))
#ifdef LGI_RES
	, ResObject(Res_Group)
#endif
{
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	InitialValue = Init;
	if (name) Name(name);
}

GRadioGroup::~GRadioGroup()
{
}

bool GRadioGroup::Name(char *s)
{
	bool Status = GObject::Name(s);
	if (Box)
	{
		bool Lock = Box->LockLooper();
		Box->SetLabel(s);
		if (Lock) Box->UnlockLooper();
	}
	return Status;
}

void GRadioGroup::OnCreate()
{
	AttachChildren();
	Value(InitialValue);
}

int GRadioGroup::Value()
{
	if (Parent)
	{
		int i=0;
		for (	GView *w = Children.First();
				w;
				w = Children.Next(), i++)
		{
			GRadioButton *But = dynamic_cast<GRadioButton*>(w);
			if (But AND But->Value()) return i;
		}
	
		return -1;
	}

	return InitialValue;
}

void GRadioGroup::Value(int Which)
{
	if (Parent)
	{
		int i=0;
		for (	GView *w = Children.First();
				w;
				w = Children.Next(), i++)
		{
			GRadioButton *But = dynamic_cast<GRadioButton*>(w);
			if (But)
			{
				But->Value(Which == i);
			}
		}
	}
	else
	{
		InitialValue = Which;
	}
}

int GRadioGroup::OnNotify(GView *Ctrl, int Flags)
{
	GView *n = GetNotify() ? GetNotify() : GetParent();
	if (n)
	{
		if (dynamic_cast<GRadioButton*>(Ctrl))
		{
			n->OnNotify(this, Value());
		}
		else
		{
			n->OnNotify(Ctrl, Flags);
		}
	}
	return 0;
}

int GRadioGroup::OnEvent(BMessage *Msg)
{
	return 0;
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	GRadioButton *But = NEW(GRadioButton(NextId++, x, y, -1, -1, name));
	if (But)
	{
		Children.Insert(But);
	}
	return But;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
class GRadioButtonRedir : public BRadioButton
{
	GRadioButton *Obj;
	
public:
	GRadioButtonRedir(GRadioButton *obj) :
		BRadioButton(BRect(0,0,1,1), "RadioBtn", "", NEW(BMessage(WM_CHANGE)))
	{
		Obj = obj;
		DefaultViewColour();
	}
	
	void SetValue(int32 value)
	{
		BRadioButton::SetValue(value);
		if (value AND Obj)
		{
			GView *w = (Obj->GetNotify()) ? Obj->GetNotify() : Obj->GetParent();
			if (w)
			{
				w->OnNotify(Obj, 0);
			}
		}
	}
};

GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first)
	: GControl(Button = NEW(GRadioButtonRedir(this)) )
#ifdef LGI_RES
	, ResObject(Res_RadioBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	if (name) Name(name);
}

GRadioButton::~GRadioButton()
{
}

int GRadioButton::Value()
{
	int Status = 0;
	if (Button)
	{
		bool Lock = Button->LockLooper();
		Status = Button->Value();
		if (Lock) Button->UnlockLooper();
	}
	return Status;
}

void GRadioButton::Value(int i)
{
	if (Button)
	{
		bool Lock = Button->LockLooper();
		Button->SetValue(i);
		if (Lock) Button->UnlockLooper();
	}
}

bool GRadioButton::Name(char *s)
{
	bool Status = GObject::Name(s);
	if (Button)
	{
		bool Lock = Button->LockLooper();
		Button->SetLabel(s);
		if (Lock) Button->UnlockLooper();
	}
	return Status;
}

int GRadioButton::OnEvent(BMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////
// Slider control
GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert) :
	GControl(NEW(BViewRedir(this)))
#ifdef LGI_RES
	, ResObject(Res_Slider)
#endif
{
	SetId(id);
	if (vert)
	{
		Pos.Set(x, y, x+13, y+cy);
	}
	else
	{
		Pos.Set(x, y, x+cx, y+13);
	}
	
	Name(name);
	Vertical = vert;
	Min = 0;
	Max = 1;
	Val = 0;
	
	Thumb.ZOff(-1, -1);
	Track = false;
}

GSlider::~GSlider()
{
}

void GSlider::Value(int i)
{
	Val = i;
	Invalidate();
}

int GSlider::Value()
{
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	min = Min;
	max = Max;
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
	if (Val < Min) Val = Min;
	if (Val > Max) Val = Max;
}

int GSlider::OnEvent(BMessage *Msg)
{
	return 0;
}

#define SLIDER_BAR_SIZE		2
#define SLIDER_THUMB_WIDTH	10
#define SLIDER_THUMB_HEIGHT	6

void GSlider::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, X()-1, Y()-1);

	// fill area
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);

	// draw tray
	int Mid = (Vertical) ? r.X()/2 : r.Y()/2;
	GdcRegion Bar;
	if (Vertical)
	{
		Bar.ZOff(SLIDER_BAR_SIZE, Y()-1);
		Bar.Offset(Mid-(Bar.X()/2), 0);
	}
	else
	{
		Bar.ZOff(X()-1, SLIDER_BAR_SIZE);
		Bar.Offset(0, Mid-(Bar.Y()/2));
	}
	
	DrawMoveShallowEdge(pDC, Bar, SUNKEN);
	pDC->Colour(0);
	pDC->Line(Bar.x1, Bar.y1, Bar.x2, Bar.y2);
	
	// draw thumb
	int Length = ((Vertical) ? Y() : X()) - SLIDER_THUMB_HEIGHT - 1;
	int Pos = (Max != Min) ? Val * Length / (Max - Min) : 0;
	if (Vertical)
	{
		Thumb.ZOff(SLIDER_THUMB_WIDTH, SLIDER_THUMB_HEIGHT);
		Thumb.Offset(Mid - (Thumb.X()/2), Pos);
	}
	else
	{
		Thumb.ZOff(SLIDER_THUMB_HEIGHT, SLIDER_THUMB_WIDTH);
		Thumb.Offset(Pos, Mid - (Thumb.Y()/2));
	}
	
	GdcRegion t = Thumb;
	DrawMoveDeepEdge(pDC, t, RAISED);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&t);
}

void GSlider::OnMouseClick(VMouse &m)
{
	if (m.Down())
	{
		if (Thumb.Overlap(m.x, m.y))
		{
			Tx = m.x - Thumb.x1;
			Ty = m.y - Thumb.y1;
			Track = true;
		}
	}
	else
	{
		Track = false;
	}
}

void GSlider::OnMouseMove(VMouse &m)
{
	if (Track)
	{
		int Length = ((Vertical) ? Y() : X()) - SLIDER_THUMB_HEIGHT - 1;
		int NewPos = (Vertical) ? m.y - Ty : m.x - Tx;
		if (NewPos < 0) NewPos = 0;
		if (NewPos > Length) NewPos = Length;
		int NewVal = (Min != Max) ? (NewPos*(Max-Min)) / Length : 0;
		if (NewVal != Val)
		{
			Val = NewVal;
			Invalidate();
			
			if (GetParent())
			{
				GetParent()->OnNotify(this, Val);
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////
GBitmap::GBitmap(int id, int x, int y, char *FileName, bool Async)
	: GControl(this),
	BControl(	BRect(0, 0, 1, 1),
				NULL, "",
				NULL,
				B_FOLLOW_NONE,
				B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE)
#ifdef LGI_RES
	, ResObject(Res_Bitmap)
#endif
{
	SetId(id);
	Pos.ZOff(1, 1);
	Pos.Offset(x, y);
	pDC = 0;
	DefaultViewColour();

	int Opt = GdcD->SetOption(GDC_IMPORT_8_AS_16, true);
	pDC = LoadDC(FileName);
	GdcD->SetOption(GDC_IMPORT_8_AS_16, Opt);

	if (pDC)
	{
		Pos.Dimension(pDC->X()+4, pDC->Y()+4);
	}
}

GBitmap::~GBitmap()
{
	DeleteObj(pDC);
}

int GBitmap::OnEvent(BMessage *Msg)
{
	return 0;
}

void GBitmap::OnPaint(GdcBasePrimitives *pScreen)
{
	GdcRegion a(0, 0, Pos.X()-1, Pos.Y()-1);
	DrawMoveDeepEdge(pScreen, a, SUNKEN);
	if (pDC)
	{
		pScreen->Blt(2, 2, pDC);
	}
	else
	{
		pScreen->Colour(CBit(pScreen->GetBits(), Rgb24(255, 255, 255)));
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(VMouse &m)
{
	if (NOT m.Down() AND GetParent())
	{
		GDialog *p = (GDialog*) GetParent();
		p->OnNotify(this, 0);
	}
}

void GBitmap::SetDC(GdcBasePrimitives *pNewDC)
{
	DeleteObj(pDC);
	pDC = pNewDC;
}

GdcBasePrimitives *GBitmap::GetDC()
{
	return pDC;
}

///////////////////////////////////////////////////////////////////////////////////////////
GItemContainer::GItemContainer()
{
	Flags = 0;
	ImageList = 0;
}

GItemContainer::~GItemContainer()
{
	if (OwnList())
	{
		DeleteObj(ImageList);
	}
	else
	{
		ImageList = 0;
	}
}

bool GItemContainer::SetImageList(GImageList *list, bool Own)
{
	ImageList = list;
	OwnList(Own);
	AskImage(ImageList != NULL);
	return ImageList != NULL;
}

/*
bool GItemContainer::Load(char *File)
{
	GdcBasePrimitives *pDC = LoadDC(File);
	if (pDC)
	{
		ImageList = NEW(GImageList(24, 24, pDC));
		if (ImageList)
		{
			#ifdef WIN32
			ImageList->Create(pDC->X(), pDC->Y(), pDC->GetBits());
			#endif
		}
	}
	return pDC != 0;
}
*/

///////////////////////////////////////////////////////////////////////////////////////////
GProgress::GProgress(int id, int x, int y, int cx, int cy, char *name) :
	GControl(this),
	BView(BRect(x, y, x+10, y+10), "Progress", B_FOLLOW_NONE, B_WILL_DRAW)
#ifdef LGI_RES
	, ResObject(Res_Progress)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	c = Rgb24(50, 150, 255);
	DefaultViewColour();
}

GProgress::~GProgress()
{
}

void GProgress::Colour(COLOUR Col)
{
	c = Col;
	GView::Invalidate();
}

COLOUR GProgress::Colour()
{
	return c;
}

void GProgress::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, X()-1, Y()-1);
	DrawMoveShallowEdge(pDC, r, SUNKEN);
	
	if (High != 0)
	{
		int Pos = ((double) r.X() * (((double) Val - Low) / High));
		if (Pos > 0)
		{
			COLOUR High = Rgb24( (R24(c)+255)/2, (G24(c)+255)/2, (B24(c)+255)/2 );
			COLOUR Low = Rgb24( (R24(c)+0)/2, (G24(c)+0)/2, (B24(c)+0)/2 );
			GdcRegion p = r;

			p.x2 = p.x1 + Pos;
			r.x1 = p.x2 + 1;
			
			pDC->Colour(Low, 24);
			pDC->Line(p.x2, p.y2, p.x2, p.y1);
			pDC->Line(p.x2, p.y2, p.x1, p.y2);

			pDC->Colour(High, 24);
			pDC->Line(p.x1, p.y1, p.x2, p.y1);
			pDC->Line(p.x1, p.y1, p.x1, p.y2);

			p.Size(1, 1);
			pDC->Colour(c, 24);
			pDC->Rectangle(&p);
		}
	}

	if (r.Valid())
	{
		pDC->Colour(Rgb24(255, 255, 255), 24);
		pDC->Rectangle(&r);
	}
}

void GProgress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
	GView::Invalidate();
}

void GProgress::GetLimits(int &l, int &h)
{
	l = Low;
	h = High;
}

void GProgress::Value(int v)
{
	Val = limit(v, Low, High);
	GView::Invalidate();
}

int GProgress::Value()
{
	return Val;
}

int GProgress::OnEvent(BMessage *Msg)
{
	return 0;
}


//////////////////////////////////////////////////////////////////////////////////////
class GScrollBarRedir : public BScrollBar
{
	GScrollBar *Notify;
	
public:
	GScrollBarRedir(GScrollBar *notify,
					BRect frame,
					const char *name,
					BView *target, 
					float min,
					float max,
					orientation posture) :
		BScrollBar(	frame,
					name,
					target, 
					min,
					max,
					posture)
	{
		Notify = notify;
	}
	
	void ValueChanged(float newValue)
	{
		if (Notify)
		{
			GView *n = (Notify->GetNotify()) ? Notify->GetNotify() : Notify->GetParent();
			if (n)
			{
				n->OnNotify(Notify, newValue);
			}
		}
	}
};

GScrollBar::GScrollBar(BScrollBar *bar) :
	GControl(bar)
#ifdef LGI_RES
	, ResObject(Res_ScrollBar)
#endif
{
	Bar = bar;
}

GScrollBar::GScrollBar(int id, int x, int y, int cx, int cy, char *name) :
	GControl(Bar = NEW(GScrollBarRedir(	this,
										BRect(x, y, x+cx, y+cx),
										(name)?name:"scroll",
										0,
										0.0,
										100.0,
										(cx > cy) ? B_HORIZONTAL : B_VERTICAL)))
#ifdef LGI_RES
	, ResObject(Res_ScrollBar)
#endif
{
	if (Bar)
	{
		Bar->SetResizingMode(B_FOLLOW_LEFT | B_FOLLOW_TOP);
	}
	
	SetId(id);
	GdcRegion n(x, y,
				x + ((cx > cy) ? cx : B_V_SCROLL_BAR_WIDTH),
				y + ((cx <= cy) ? cy : B_H_SCROLL_BAR_HEIGHT));
	GView::SetPos(n);
	if (name) GView::Name(name);
	
	High = Low = PageSize = 0;
}

GScrollBar::~GScrollBar()
{
	Bar = 0;
}

bool GScrollBar::Vertical()
{
	bool Status = false;
	if (Bar AND Bar->LockLooper())
	{
		Status = Bar->Orientation() == B_VERTICAL;
		Bar->UnlockLooper();
	}
	return Status;
}

int GScrollBar::Pos()
{
	int i = 0;
	if (Bar AND Bar->LockLooper())
	{
		i = Bar->Value();
		Bar->UnlockLooper();
	}
	return i;
}

void GScrollBar::SetPos(int p)
{
	if (Bar AND Bar->LockLooper())
	{
		Bar->SetValue(p);
		Bar->UnlockLooper();
	}
}

void GScrollBar::Limits(int &Low, int &High)
{
	float l = 0.0, h = 0.0;
	if (Bar AND Bar->LockLooper())
	{
		Bar->GetRange(&l, &h);
		Bar->UnlockLooper();
	}
	Low = l;
	High = h;
}

void GScrollBar::SetLimits(int low, int high)
{
	if (Bar AND Bar->LockLooper())
	{
		// printf("Page=%i High=%i Low=%i\n", PageSize, High, Low);
		Low = low;
		High = high;
		if (PageSize > (High-Low))
		{
			Bar->SetRange(0, 0);
		}
		else
		{
			Bar->SetRange(Low, High-PageSize+1);
		}
		Bar->UnlockLooper();
	}
}

int GScrollBar::Page()
{
	float small = 1.0, large = 1.0;
	if (Bar AND Bar->LockLooper())
	{
		Bar->GetSteps(&small, &large);
		Bar->UnlockLooper();
	}
	return large;
}

void GScrollBar::SetPage(int p)
{
	if (Bar AND Bar->LockLooper())
	{
		PageSize = p;
		Bar->SetSteps(1.0, PageSize);
		Bar->SetProportion((double)PageSize / ((double)High-Low+1));
		Bar->UnlockLooper();
	}
}

bool GScrollBar::Attach(GView *parent)
{
	// parent->GetBView()->AddChild(new BScrollBar(BRect(10, 50, 200, 64), "Scroll", 0, 0.0, 100.0, B_HORIZONTAL)); 
	// return true;

	return GControl::Attach(parent);
}

int GScrollBar::OnEvent(BMessage *Msg)
{
	return FALSE;
}

