/*
**	FILE:		GuiViews.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Standard Views
**
**	Copyright (C) 1998, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
GLayout::GLayout() :
	GView(Redir = NEW(BViewRedir(this)))
{
	_PourLargest = false;
	VScroll = 0;
	HScroll = 0;
	XScroll = YScroll = false;
	Redir->SetScrollView(0);

	// GetBView()->SetResizingMode(B_FOLLOW_NONE);
	// GetBView()->SetViewColor(216, 216, 216);
	// GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
}

GLayout::~GLayout()
{
	DeleteObj(HScroll);
	DeleteObj(VScroll);
}

bool GLayout::GetPourLargest()
{
	return _PourLargest;
}

void GLayout::SetPourLargest(bool i)
{
	_PourLargest = i;
}

bool GLayout::Pour(GdcComplexRegion &r)
{
	if (_PourLargest)
	{
		GdcRegion *Best = FindLargest(r);
		if (Best)
		{
			SetPos(*Best, true);
			return true;
		}
	}
	
	return false;
}

void GLayout::GetScrollPos(int &x, int &y)
{
	if (HScroll)
	{
		x = HScroll->Pos();
	}
	else
	{
		x = 0;
	}

	if (VScroll)
	{
		y = VScroll->Pos();
	}
	else
	{
		y = 0;
	}
}

void GLayout::SetScrollPos(int x, int y)
{
	if (HScroll)
	{
		HScroll->SetPos(x);
	}

	if (VScroll)
	{
		VScroll->SetPos(y);
	}
}

bool GLayout::Attach(GView *p)
{
	bool Status = GView::Attach(p);
	
	if (HScroll OR VScroll)
	{
		SetScrollBars(HScroll!=0, VScroll!=0);
		
		if (Redir->GetScrollView())
		{
			if (HScroll)
			{
				HScroll->Bar = Redir->GetScrollView()->ScrollBar(B_HORIZONTAL);
			}
			if (VScroll)
			{
				VScroll->Bar = Redir->GetScrollView()->ScrollBar(B_VERTICAL);
			}
		}
	}
	
	Visible(GFlags::Visible());

	return Status;
}

bool GLayout::SetScrollBars(bool x, bool y)
{
	bool Has = (HScroll != 0) OR (VScroll != 0);
	bool Changing = Has ^ (x OR y);
	bool Attached = (GetBView() != 0) AND (GetBView()->Parent() != 0);

	if (Changing)
	{
		if (Attached AND
			Redir->GetScrollView())
		{
			delete Redir->GetScrollView();
			Redir->SetScrollView(0);
		}
	
		if (x)
		{
			if (NOT HScroll)
			{
				HScroll = NEW(GScrollBar(0));
				if (HScroll)
				{
					HScroll->SetId(IDC_HSCROLL);
				}
			}
		}
		else
		{
			DeleteObj(HScroll);
		}
		if (y)
		{
			if (NOT VScroll)
			{
				VScroll = NEW(GScrollBar(0));
				if (VScroll)
				{
					VScroll->SetId(IDC_VSCROLL);
				}
			}
		}
		else
		{
			DeleteObj(VScroll);
		}
	
		if ((x OR y) AND
			Attached)
		{
			GdcRegion r = GetPos();
			
			Redir->SetScrollView(NEW(GScrollView("GViewScroller",
												this,
												B_FOLLOW_LEFT | B_FOLLOW_TOP,
												0,
												x,
												y,
												Sunken())));
			if (Redir->GetScrollView())
			{
				Sunken(false);
				
				SetPos(r);
				if (HScroll)
				{
					HScroll->Bar = Redir->GetScrollView()->ScrollBar(B_HORIZONTAL);
				}
				if (VScroll)
				{
					VScroll->Bar = Redir->GetScrollView()->ScrollBar(B_VERTICAL);
				}
			}
		}
	}
	
	return true;
}

GdcRegion &GLayout::GetPos()
{
	if (Redir->GetScrollView())
	{
		return Redir->GetScrollView()->GetPos();
	}

	return GView::GetPos();
}

bool GLayout::SetPos(GdcRegion &p, bool Repaint)
{
	if (Redir->GetScrollView())
	{
		Redir->GetScrollView()->SetPos(p);
		if (Repaint AND
			Redir->GetScrollView()->LockLooper())
		{
			Redir->GetScrollView()->Invalidate();
			Redir->GetScrollView()->UnlockLooper();
		}
	}
	else
	{
		GView::SetPos(p, Repaint);
	}
	
	return TRUE;
}

bool GLayout::Visible()
{
	return GView::Visible();
}

void GLayout::Visible(bool i)
{
	if (Redir->GetScrollView())
	{
		// Set view visible
		bool Lock = Redir->GetScrollView()->LockLooper();
		if (i)
		{
			if (NOT GFlags::Visible())
			{
				Redir->GetScrollView()->Show();
			}
		}
		else
		{
			if (GFlags::Visible())
			{
				Redir->GetScrollView()->Hide();
			}
		}
		if (Lock) Redir->GetScrollView()->UnlockLooper();

		// keep track of visibility (used when we're not attached)
		GFlags::Visible(i);
	}
	else
	{
		GView::Visible(i);
	}
}

int GLayout::OnEvent(BMessage *Msg)
{
	switch (Message(Msg))
	{
		case WM_CHANGE:
		{
			if (GetParent())
			{
				GetParent()->OnEvent(Msg);
			}
			break;
		}
		case WM_VSCROLL:
		{
			if (VScroll)
			{
				OnNotify(VScroll, MessageA(Msg));
			}
			break;
		}
		case WM_HSCROLL:
		{
			if (HScroll)
			{
				OnNotify(HScroll, MessageA(Msg));
			}
			break;
		}
	}
	
	if (VScroll) VScroll->OnEvent(Msg);
	if (HScroll) HScroll->OnEvent(Msg);
	return GView::OnEvent(Msg);
}

//////////////////////////////////////////////////////////////////////////////////////
GStatusBar::GStatusBar()
{
	GView::Name("LGI_StatusBar");

	Font = NEW(GFont);
	if (Font)
	{
		Font->Transparent(TRUE);
		Font->Back(LC_MED);
		Font->Create("Ms Sans Serif", 10);
	}
}

GStatusBar::~GStatusBar()
{
	for (int i=0; i<GetSize(); i++)
	{
		DeleteObj((*this)[0]);
	}

	DeleteObj(Font);
}

bool GStatusBar::Pour(GdcComplexRegion &r)
{
	GdcRegion *Best = FindLargestEdge(r, GV_EDGE_BOTTOM);
	if (Best)
	{
		GdcRegion Take = *Best;
		
		int y = 21;
		// int y = 50;
		
		if (Take.Y() > y)
		{
			Take.y1 = max((Take.y2 - y), Take.y1);
			SetPos(Take);
			RePour();
			return TRUE;
		}
	}
	return FALSE;
}

void GStatusBar::OnPaint(GdcBasePrimitives *pDC)
{
	int Xm1 = pDC->X()-1;
	int Ym1 = pDC->Y()-1;

	pDC->Colour(LC_LOW, 24);
	pDC->Line(Xm1, Ym1, 0, Ym1);
	pDC->Line(Xm1, Ym1, Xm1, 0);

	pDC->Colour(LC_HIGH, 24);
	pDC->Line(0, 0, Xm1, 0);
	pDC->Line(0, 0, 0, Ym1);

	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(1, 1, Xm1-1, Ym1-1);

	#ifdef WIN32
	for (	GStatusPane *Pane = (GStatusPane*) Children.First();
			Pane;
			Pane = (GStatusPane*) Children.Next())
	{
		pDC->SetOrigin(-Pane->Pos.x1, -Pane->Pos.y1);
		Pane->OnPaint(pDC);
	}
	#endif
}

void GStatusBar::RePour()
{
	int X = Pos.X() - 5;
	int i = Children.GetItems()-1;

	for (	GStatusPane *Pane = (GStatusPane*) Children.Last();
			Pane;
			Pane = (GStatusPane*) Children.Prev(), i--)
	{
		if (i)
		{
			// not first
			X -= Pane->Width; 
			Pane->Pos.ZOff(Pane->Width, Pos.Y()-7);
			Pane->Pos.Offset(X, 3);
			X -= STATUSBAR_SEPARATOR;
		}
		else
		{
			// first
			Pane->Pos.ZOff(X-2, Pos.Y()-4);
			Pane->Pos.Offset(2, 2);
			X = 0;
		}

		Pane->SetPos(Pane->Pos);
	}
}

GStatusPane *GStatusBar::AppendPane(char *Text, int Width)
{
	GStatusPane *Pane = 0;
	if (Text)
	{
		Pane = NEW(GStatusPane);
		if (Pane)
		{
			Pane->SetParent(this);
			Pane->Name(Text);
			Pane->Width = Width;
			Children.Insert(Pane);

			#ifdef BEOS
			Pane->Attach(this);
			#endif
		}
	}
	return Pane;
}

bool GStatusBar::AppendPane(GStatusPane *Pane)
{
	if (Pane)
	{
		Pane->SetParent(this);
		Children.Insert(Pane);

		#ifdef BEOS
		Pane->Attach(this);
		#endif
	}
	return TRUE;
}

// Pane
GStatusPane::GStatusPane() :
	/*
	BView(	BRect(0, 0, 1, 1),
			NULL,
			B_FOLLOW_NONE,
			B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE),
	*/
	GView(NEW(BViewRedir(this)))
{
	Parent = 0;
	Flags = 0;
	Width = 32;
	Pos.ZOff(Width-1, 20);
	pDC = 0;
	_BorderSize = 1;
	
	GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
}

GStatusPane::~GStatusPane()
{
	DeleteObj(pDC)
}

bool GStatusPane::Name(char *n)
{
	bool Status = GObject::Name(n);
	GdcRegion p(0, 0, Pos.X()-1, Pos.Y()-1);
	p.Size(1, 1);
	Invalidate();
	return Status;
}

void GStatusPane::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
	int Off = 1;

	if (Sunken())
	{
		Off = 0;
		DrawMoveShallowEdge(pDC, r, SUNKEN);
	}
	
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);

	GStatusBar *Par = (GStatusBar*) GView::Parent;
	if (Par AND Par->Font)
	{
		Par->Font->Text(pDC, 2+Off, 1+Off, Name(), -1, &r);
	}
}

int GStatusPane::GetWidth()
{
	return Width;
}

void GStatusPane::SetWidth(int x)
{
	Width = x;
	if (GView::Parent)
	{
		((GStatusBar*) GView::Parent)->RePour();
	}
}

bool GStatusPane::Sunken()
{
	return GView::Sunken();
}

void GStatusPane::Sunken(bool s)
{
	bool Update = s ^ GView::Sunken();
	GView::Sunken(s);

	if (GetParent() AND Update)
	{
		GetParent()->Invalidate(&Pos);
	}
}

GdcBasePrimitives *GStatusPane::Bitmap()
{
	return pDC;
}

void GStatusPane::Bitmap(GdcBasePrimitives *pNewDC)
{
	DeleteObj(pDC);
	if (pNewDC)
	{
		pDC = NEW(GdcMemDC);
		if (pDC)
		{
			if (pDC->Create(pNewDC->X(), pNewDC->Y(), pNewDC->GetBits()))
			{
				pDC->Blt(0, 0, pNewDC);
			}
			else
			{
				DeleteObj(pDC);
			}
		}
	}

	if (GView::Parent)
	{
		GView::Parent->Invalidate(&Pos);
	}
}

//////////////////////////////////////////////////////////////////////////////
#define SPLITER_BORDER			4
#define SPLITER_BAR				5
#define SPLITER_INSET			2
#define SPLITER_MIN_X			(2*SPLITER_INSET)
#define SPLITER_MIN_Y			(2*SPLITER_INSET)

#define SPLITER_FULL_DRAG

GSplitter::GSplitter()
{
	Name("LGI_Spliter");
	Vertical = TRUE;
	SplitSet = SplitPos = -1;
	Moving = FALSE;
	SetPourLargest(true);
	_BorderSize = 1;

	BorderA = BorderB = true;
	PosA.ZOff(0, 0);
	PosB.ZOff(0, 0);
	ViewA = 0;
	ViewB = 0;
	Split(0);
	SplitFollow = false;
	Border(true);
	Raised(true);
	IsVertical(true);

	GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
}

GSplitter::~GSplitter()
{
	DeleteObj(ViewA);
	DeleteObj(ViewB);
}

void GSplitter::IsVertical(bool v)
{
	Vertical = v;
}

void GSplitter::SetViewA(GView *a, bool Border)
{
	BorderA = Border;
	ViewA = a;
	if (a AND GetParent())
	{
		a->Attach(this);
		a->Visible(true);

		CalcRegions(Pos);
		Invalidate();
	}
}

void GSplitter::SetViewB(GView *b, bool Border)
{
	BorderB = Border;
	ViewB = b;
	if (b AND GetParent())
	{
		b->Attach(this);
		b->Visible(true);

		CalcRegions(Pos);
		Invalidate();
	}
}

void GSplitter::Split(int s)
{
	SplitSet = s;
	int Limit = ((Vertical) ? Pos.X() : Pos.Y()) - 18;
	int NewPos = (s > 0) ? max(s, 4) : max(4, Limit+s);
	if (NewPos != SplitPos)
	{
		SplitPos = NewPos;
		CalcRegions(Pos);
		if (Visible())
		{
			GView::Invalidate();
		}
	}
}

void GSplitter::CalcRegions(GdcRegion Rect, bool Follow)
{
	PosA = Rect;
	PosA.Offset(-PosA.x1, -PosA.y1);
	if (Border())
	{
		PosA.Size(SPLITER_BORDER, SPLITER_BORDER);
	}
	
	//printf("this=%p, gsplitter::calc(%i, %i, %i, %i)\n", this, PosA.x1, PosA.y1, PosA.x2, PosA.y2);

	PosB = PosA;

	if (SplitFollow AND Follow)
	{
		SplitPos = max(1, ((Vertical) ? Pos.X()-10 : Pos.Y()-10) + SplitSet);
	}

	if (Vertical)
	{
		PosA.x2 = PosA.x1 + SplitPos;
		PosB.x1 = PosA.x2 + SPLITER_BAR;
	}
	else
	{
		PosA.y2 = PosA.y1 + SplitPos;
		PosB.y1 = PosA.y2 + SPLITER_BAR;
	}

	if (ViewA)
	{
		GdcRegion r = PosA;
		if (BorderA)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		r.Offset(Rect.x1, Rect.y1);
		ViewA->SetPos(r, true);
	}
	if (ViewB)
	{
		GdcRegion r = PosB;
		if (BorderB)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		r.Offset(Rect.x1, Rect.y1);
		ViewB->SetPos(r, true);
	}

	//printf("this=%p, posA(%i, %i, %i, %i)\n", this, PosA.x1, PosA.y1, PosA.x2, PosA.y2);
	//printf("this=%p, posB(%i, %i, %i, %i)\n", this, PosB.x1, PosB.y1, PosB.x2, PosB.y2);
}

void GSplitter::OnPosChange()
{
	CalcRegions(GetClient(), true);
}

void GSplitter::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r = GetClient();
	r.Offset(-r.x1, -r.y1);

	// printf("gsplitter::onpaint(%i, %i, %i, %i)\n", r.x1, r.y1, r.x2, r.y2);

	if (Border())
	{
		// edge
		DrawMoveFill(pDC, r, SPLITER_BORDER);
	}

	// bar
	pDC->Colour(LC_MED, 24);
	if (Vertical)
	{
		pDC->Rectangle(PosA.x2+1, r.y1, PosB.x1-1, r.y2);
	}
	else
	{
		pDC->Rectangle(r.x1, PosA.y2+1, r.x2, PosB.y1-1);
	}

	// PosA region
	r = PosA;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		/*
		pDC->Colour(Rgb24(255, 0, 0), 24);
		pDC->Rectangle(&r);
		*/

		if (BorderA)
		{
			DrawMoveDeepEdge(pDC, r, SUNKEN);
		}

		if (ViewA)
		{
		}
		else
		{
			pDC->Colour(0xFFFFFF);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	r = PosB;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		/*
		pDC->Colour(Rgb24(0, 0, 255), 24);
		pDC->Rectangle(&r);
		*/

		if (BorderB)
		{
			DrawMoveDeepEdge(pDC, r, SUNKEN);
		}

		if (ViewB)
		{
		}
		else
		{
			pDC->Colour(0xFFFFFF);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}
}

void GSplitter::OnMouseClick(VMouse &m)
{
	if (m.Down())
	{
		if (Vertical)
		{
			if (m.x > PosA.x2 AND m.x < PosB.x1 AND m.y > PosA.y1 AND m.y < PosA.y2)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x2;
				Bar.x2 = PosB.x1;
				Bar.y1 = PosA.y1;
				Bar.y2 = PosA.y2;
				Offset = m.x - Bar.x1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
		else
		{
			if (m.x > PosA.x1 AND m.x < PosA.x2 AND m.y > PosA.y2 AND m.y < PosB.y1)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x1;
				Bar.x2 = PosA.x2;
				Bar.y1 = PosA.y2;
				Bar.y2 = PosB.y1;
				Offset = m.y - Bar.y1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
	}
	else
	{
		#ifdef SPLITER_FULL_DRAG
		Capture(FALSE);
		Moving = FALSE;
		#else
		if (Moving)
		{
			HashRectangle(this, &Bar);
			Capture(FALSE);
			Moving = FALSE;

			int NewPos;
			if (Vertical)
			{
				NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
			}
			else
			{
				NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
			}

			if (NewPos != SplitPos)
			{
				SplitPos = NewPos;
				CalcRegions(Pos);
				GView::Invalidate();
			}
		}
		#endif
	}
}

void GSplitter::OnMouseMove(VMouse &m)
{
	if (Moving)
	{
		#ifdef SPLITER_FULL_DRAG
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}

		int NewPos;
		if (Vertical)
		{
			NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
		}
		else
		{
			NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
		}

		if (NewPos != SplitPos)
		{
			SplitPos = NewPos;
			CalcRegions(GetClient());
			Invalidate();
		}
		#else
		HashRectangle(this, &Bar);
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}
		HashRectangle(this, &Bar);
		#endif
	}
}

bool GSplitter::Attach(GView *p)
{
	bool Status = GLayout::Attach(p);
	if (Status)
	{
		if (ViewA)
		{
			ViewA->Attach(this);
		}

		if (ViewB)
		{
			ViewB->Attach(this);
		}
	}

	return Status;
}

GView *GSplitter::FindControl(GWndRef hCtrl)
{
	GView *c = 0;
	if (ViewA) c = ViewA->FindControl(hCtrl);
	if (c) return c;

	if (ViewB) c = ViewB->FindControl(hCtrl);
	return c;
}

