/*hdr
**	FILE:		16bit.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		20/3/97
**	DESCRIPTION:	16 bit primitives
**
**	Copyright (C) 1997, Matthew Allen
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Gdc2.h"

/****************************** Classes *************************************************************************************/
#define sPtr			((ushort*)Ptr)

bool GdcApp16::SetSurface(GdcSurface *d, GdcPalette *p, GdcSurface *a)
{
	if (d AND d->Bits == 16)
	{
		Dest = d;
		Pal = p;
		Ptr = d->Base;
		Alpha = 0;
		return TRUE;
	}
	return FALSE;
}

void GdcApp16::SetPtr(int x, int y)
{
	if (Dest AND Dest->Base)
	{
		Ptr = Dest->Base + ((y * Dest->Line) + x + x);
	}
	else
	{
		Ptr = 0;
		exit(0);
	}
}

void GdcApp16::IncX()
{
	Ptr += 2;
}

void GdcApp16::IncY()
{
	Ptr += Dest->Line;
}

void GdcApp16::IncPtr(int X, int Y)
{
	Ptr += (Y * Dest->Line) + X + X;
}

COLOUR GdcApp16::Get()
{
	return *sPtr;
}

// 16 bit set sub functions
void GdcApp16Set::Set()
{
	*sPtr = c;
}

void GdcApp16Set::VLine(int height)
{
	while (height--)
	{
		*sPtr = c;
		Ptr += Dest->Line;
	}
}

void GdcApp16Set::Rectangle(int x, int y)
{
#ifdef GDC_USE_ASM

// this duplicates the colour twice in eax to allow us to fill
// two pixels at per write. this means we are using the whole
// 32-bit bandwidth to the video card :)

	if (y > 0)
	{
		if (x > 1)
		{
			uchar *p = Ptr;
			COLOUR fill = c | (c << 16);
			int Line = Dest->Line;

			_asm {
				mov esi, p
				mov eax, fill
				mov edx, Line
				mov bx, ax
				shl eax, 16
				mov ax, bx
			LoopY:	mov edi, esi
				add esi, edx
				mov ecx, x
				shr ecx, 1
			LoopX:	mov [edi], eax
				add edi, 4
				dec ecx
				jnz LoopX
				test x, 1
				jz Next
				mov [edi], ax
			Next:	dec y
				jnz LoopY
			}
		}
		else if (x == 1)
		{
			VLine(y);
		}
	}
#else
	while (y--)
	{
		for (int n=0; n<x; n++) sPtr[n] = c;
		Ptr += Dest->Line;
	}
#endif
}

bool GdcApp16Set::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src)
	{
		switch (Src->Bits)
		{
			case 8:
			{
				ushort c[256];
				if (SPal)
				{
					GdcRGB *p = (*SPal)[0];
					for (int i=0; i<256; i++, p++)
					{
						c[i] = Rgb24To16(Rgb24(p->R, p->G, p->B));
					}
				}
				else
				{
					for (int i=0; i<256; i++)
					{
						c[i] = Rgb16(i, i, i);
					}
				}

				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						*d++ = c[*s++];
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 16:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					MemCpy(Ptr, s, Src->x * 2);
					s += Src->Line;
					Ptr += Dest->Line;
				}
				break;
			}
			case 24:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						*d++ = Rgb24To16(Rgb24(s[2], s[1], s[0]));
						s += 3;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 32:
			{
				ulong *s = (ulong*) Src->Base;
				ushort *d = (ushort*) Ptr;

				for (int y=0; y<Src->y; y++)
				{
					ulong *NextS = (ulong*) (((uchar*) s) + Src->Line);
					ushort *NextD = (ushort*) (((uchar*) d) + Dest->Line);

					for (int x=0; x<Src->x; x++)
					{
						*d = Rgb32To16(*s);
						d++;
						s++;
					}

					s = NextS;
					d = NextD;
				}

				Ptr = (uchar*) d;
				break;
			}
		}
	}
	return TRUE;
}

// 16 bit or sub functions
void GdcApp16Or::Set()
{
	*sPtr |= c;
}

void GdcApp16Or::VLine(int height)
{
	while (height--)
	{
		*sPtr |= c;
		Ptr += Dest->Line;
	}
}

void GdcApp16Or::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) sPtr[n] |= c;
		Ptr += Dest->Line;
	}
}

bool GdcApp16Or::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemOr(Ptr, s, Src->x * 2);
			s += Src->Line;
			Ptr += Dest->Line;
		}
	}
	return TRUE;
}

// 16 bit AND sub functions
void GdcApp16And::Set()
{
	*sPtr &= c;
}

void GdcApp16And::VLine(int height)
{
	while (height--)
	{
		*sPtr &= c;
		Ptr += Dest->Line;
	}
}

void GdcApp16And::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) sPtr[n] &= c;
		Ptr += Dest->Line;
	}
}

bool GdcApp16And::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src)
	{
		if (Src->Bits == Dest->Bits)
		{
			uchar *s = Src->Base;
			for (int y=0; y<Src->y; y++)
			{
				MemAnd(Ptr, s, Src->x * 2);
				s += Src->Line;
				Ptr += Dest->Line;
			}
		}
		else if (Src->Bits == 8)
		{
			uchar *s = Src->Base;
			for (int y=0; y<Src->y; y++)
			{
				uchar *Dp = Ptr;
				uchar *End = s + Src->x;

				for (uchar *Sp = s; Sp<End; Sp++)
				{
					*Dp++ &= *Sp;
					*Dp++ &= *Sp;
				}

				s += Src->Line;
				Ptr += Dest->Line;
			}
		}
	}
	return TRUE;
}

// 16 bit XOR sub functions
void GdcApp16Xor::Set()
{
	*sPtr ^= c;
}

void GdcApp16Xor::VLine(int height)
{
	while (height--)
	{
		*sPtr ^= c;
		Ptr += Dest->Line;
	}
}

void GdcApp16Xor::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) sPtr[n] ^= c;
		Ptr += Dest->Line;
	}
}

bool GdcApp16Xor::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemXor(Ptr, s, Src->x * 2);
			s += Src->Line;
			Ptr += Dest->Line;
		}
	}
	return TRUE;
}
