




#include <stdio.h>
#include <string.h>
#include <math.h>
#include "Gdc2.h"
#include "Lzw.h"

#define HasLZW

//////////////////////////////////////////////////////////////////
Lzw::Lzw()
{
	Meter = 0;
	BufSize = 0;
	Buffer = 0;
	CodeBuf = NEW(uchar[TIFF_LZW_BUFFER]);
}

Lzw::~Lzw()
{
	DeleteObj(Buffer);
	DeleteObj(CodeBuf);
}

void Lzw::SetBufSize(int i)
{
	BufSize = i;
}

void Lzw::InitializeTable()
{
	CodeSize = 9;
	ClearCode = 1 << (CodeSize - 1);
	EoiCode = ClearCode + 1;
	NextCode = EoiCode + 1;

	ZeroObj(CodeLength);
	ZeroObj(CodePtr);
	NextBuf = CodeBuf;
	for (int i=0; i<256; i++)
	{
		CodeLength[i] = 1;
		CodePtr[i] = NextBuf++;
		*CodePtr[i] = i;
	}
}

int Lzw::InputCode()
{
	int Code = 0;
	int b = CodeSize;
	uchar Mask[] = { 0, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF };

	while (CodeSize > 0)
	{
		if (BitsLeft < 1)
		{
			DeleteArray(Buffer);
			BitsLeft = 0;
			Buffer = NEW(uchar[BufSize]);
			if (Buffer)
			{
				if (Pipe)
				{
					int Size = Pipe->Sizeof();
					int Len = min(BufSize, Size);
					Pipe->Pop(Buffer, Len);
					BitsLeft += Len * 8;
				}
				else if (File)
				{
					BitsLeft = File->Read(Buffer, BufSize) * 8;
				}

				BufPos = Buffer;
				BitPos = 0;
				BufSize = 0;
			}

			if (BitsLeft <= 0)
			{
				return EoiCode;
			}
		}

		int ThisBits = min(8-BitPos, b);
		int Temp = Mask[ThisBits] & (*BufPos >> (8 - (ThisBits + BitPos)));
		Code |= Temp << (b - ThisBits);
		
		b -= ThisBits;
		BitsLeft -= ThisBits;
		BitPos += ThisBits;
		if (BitPos > 7)
		{
			BitPos -= 8;
			BufPos++;
		}
	}

	return Code;
}

bool Lzw::Decompress(BytePipe *Out, BytePipe *in, VFile *file)
{
	int Code = 0, OldCode = 0;
	uchar *OutString;

	Pipe = in;
	File = file;
	if (Out AND (Pipe OR File))
	{
		CodeSize = 9;
		ClearCode = 1 << (CodeSize - 1);
		EoiCode = ClearCode + 1;
		NextCode = EoiCode + 1;
		BitPos = 0;
		BitsLeft = 0;

		while ((Code = InputCode()) != EoiCode AND NextCode < 4096)
		{
			if (Code == ClearCode)
			{
				InitializeTable();
				Code = InputCode();
				if (Code == EoiCode)
				{
					break;
				}
				Out->Push(CodePtr[Code], CodeLength[Code]);
				OldCode = Code;
			}
			else
			{
				if (Code < NextCode)
				{
					Out->Push(CodePtr[Code], CodeLength[Code]);

					uchar *d = CodePtr[OldCode];
					int Length = CodeLength[OldCode];
					CodeLength[NextCode] = Length + 1;
					CodePtr[NextCode] = NextBuf;
					while (Length--)
					{
						*NextBuf++ = *d++;
					}
					*NextBuf++ = *CodePtr[Code];
					NextCode++;

					OldCode = Code;
				}
				else
				{
					if (Code > NextCode)
					{
						break;
					}
					else
					{
						int Length = CodeLength[OldCode];
						CodeLength[NextCode] = Length + 1;
						CodePtr[NextCode] = NextBuf;
						uchar *d = CodePtr[OldCode];
						while (Length--)
						{
							*NextBuf++ = *d++;
						}
						*NextBuf++ = *CodePtr[OldCode];
						Out->Push(CodePtr[NextCode], CodeLength[NextCode]);
						NextCode++;

						OldCode = Code;
					}
				}

				switch (NextCode)
				{
					case 511:
					case 1023:
					case 2047:
					{
						CodeSize++;
						break;
					}
					case 4095:
					{
						InitializeTable();
						break;
					}
				}
			}
		}

		return true;
	}

	return false;
}

#define MaxCode(CodeSize)  ((1 << (CodeSize))-1)
#define MaxHashTable  5003
#define MaxGIFBits  12
#if defined(HasLZW)
#define MaxGIFTable  (1 << MaxGIFBits)
#else
#define MaxGIFTable  MaxCode
#endif

void Lzw::Flush(bool Final)
{
	if (Final OR (BufPos >= Buffer + BufSize))
	{
		int Len = (int)BufPos - (int)Buffer;
		if (Pipe)
		{
			Pipe->Push(Buffer, Len);
		}
		else
		{
			File->Write(Buffer, Len);
		}
		BufPos = Buffer;
		BitsLeft = BufSize * sizeof(*BufPos);
	}
}

void Lzw::OutputCode(int code)
{
	// Emit a code.
	Datum |= ((long) code << BitPos);
	BitPos += CodeSize;
	
	while (BitPos >= 8)
	{
		// Add a character to current buffer
		Flush();
		if (BufPos < Buffer + BufSize)
		{
			*BufPos++ = Datum & 0xff;
		}
		
		Datum >>= 8;
		BitPos -= 8;
	}
	
	if (NextCode > MaxCode) 
	{
		CodeSize++;
		
		if (CodeSize == MaxGIFBits)
		{
			MaxCode=MaxGIFTable;
		}
		else
		{
			MaxCode=MaxCode(CodeSize);
		}
	}
}

bool Lzw::Compress(BytePipe *Out, BytePipe *In)
{
	bool Status = false;
	short index, WaitingCode;
	int InputLen = In->Sizeof();

	// Allocate encoder tables.
	uchar *packet = NEW(uchar[256]);
	short *hash_code = NEW(short[MaxHashTable]);
	short *hash_prefix = NEW(short[MaxHashTable]);
	uchar *hash_suffix = NEW(uchar[MaxHashTable]);
	uchar *Input = NEW(uchar[InputLen]);

	if (In AND
		Out AND
		Input AND
		packet AND
		hash_code AND
		hash_prefix AND
		hash_suffix)
	{
		// Initialize GIF encoder.
		CodeSize = 9;
		MaxCode = MaxCode(CodeSize);
		ClearCode = 1 << (CodeSize - 1);
		EoiCode = ClearCode + 1;
		NextCode = ClearCode + 2;
		BitPos = Datum = 0;
		BufSize = 16 << 10;
		BufPos = Buffer = NEW(uchar[BufSize]);
		Pipe = Out;
		memset(hash_code, 0, sizeof(*hash_code)*MaxHashTable);

		OutputCode(ClearCode);

		// Encode pixels.
		In->Pop(Input, InputLen);
		uchar *End = Input + InputLen;
		WaitingCode = *Input;

		DoEvery Do(200);
		if (Meter)
		{
			Meter->SetLimits(0, InputLen-1);
		}

		for (uchar *p = Input + 1; p < End; p++)
		{
			// Probe hash table.
			index = *p;
			int k = ((int) index << (MaxGIFBits - 8)) + WaitingCode;
			if (k >= MaxHashTable)
			{
				k -= MaxHashTable;
			}

			#if defined(HasLZW)
			if (hash_code[k] > 0)
			{
				if ((hash_prefix[k] == WaitingCode) AND
					(hash_suffix[k] == index))
				{
					WaitingCode = hash_code[k];
					continue;
				}
				
				bool next_pixel = false;
				int displacement = (k == 0) ? 1 : MaxHashTable - k;

				for (;;)
				{
					k -= displacement;
					
					if (k < 0)
					{
						k += MaxHashTable;
					}
					
					if (hash_code[k] == 0)
					{
						break;
					}
					
					if ((hash_prefix[k] == WaitingCode) AND
						(hash_suffix[k] == index))
					{
						WaitingCode = hash_code[k];
						next_pixel = true;
						break;
					}
				}
			
				if (next_pixel == true)
				{
					continue;
				}
			}
			#endif
			
			OutputCode(WaitingCode);
			if (NextCode < MaxGIFTable)
			{
				hash_code[k] = NextCode++;
				hash_prefix[k] = WaitingCode;
				hash_suffix[k] = index;
			}
			else
			{
				// Fill the hash table with empty entries.
				memset(hash_code, 0, sizeof(*hash_code)*MaxHashTable);

				// Reset compressor and issue a clear code.
				NextCode = ClearCode+2;
				OutputCode(ClearCode);
				CodeSize = 9;
				MaxCode = MaxCode(CodeSize);
			}

			WaitingCode=index;

			int v = (int) (p-Input);
			if (Meter AND (v%1000 == 0))
			{
				Meter->Value(v);
			}
		}

		// Flush out the buffered code.
		OutputCode(WaitingCode);
		OutputCode(EoiCode);

		// Flush accumulated data.
		Flush(true);
		Status = true;
	}
	
	// Free encoder memory.
	DeleteArray(hash_suffix);
	DeleteArray(hash_prefix);
	DeleteArray(hash_code);
	DeleteArray(packet);
	DeleteArray(Input);
	
	return Status;
}
