/*hdr
**      FILE:           Tga.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           16/8/2000
**      DESCRIPTION:    Tga file filter
**
**      Copyright (C) 2000, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "Lgi.h"

class GdcTga : public GdcFileFormat
{
public:
	GdcTga::GdcTga()
	{
		Props.Set("Type", "Targa");
		Props.Set("Extension", "TGA");
	}

	int GetCapabilites() { return FILTER_CAP_READ; }

	bool ReadImage(GdcBasePrimitives *pDC)
	{
		bool Status = false;
		
		if (pDC)
		{
			uchar IdFieldSize, ImgType, CMapType, CMapEntrySize, Bits, Descriptor;
			ushort CMapOrigin, CMapLength;
			ushort OrgX, OrgY, X, Y;
			
			*this >> IdFieldSize;
			*this >> CMapType;
			*this >> ImgType;
			*this >> CMapOrigin;
			*this >> CMapLength;
			*this >> CMapEntrySize;
			*this >> OrgX;
			*this >> OrgY;
			*this >> X;
			*this >> Y;
			*this >> Bits;
			*this >> Descriptor;

			Seek(18 + IdFieldSize, SEEK_SET);

			// load colour map
			if (CMapType)
			{
				Seek(CMapEntrySize * CMapLength, SEEK_CUR);
			}

			// load pixels
			if (pDC->Create(X, Y, Bits))
			{
				switch (ImgType)
				{
					case 2:
					{
						Status = true;
						switch (Bits)
						{
							case 16:
							case 24:
							case 32:
							{
								int PixSize = Bits / 8;
								int Line = PixSize * X;

								if (Meter)
								{
									Meter->SetDescription("Scanlines");
									Meter->SetLimits(0, Y-1);
								}

								for (int y=0; y<Y; y++)
								{
									uchar *p = (*pDC)[y];
									if (p)
									{
										Read(p, Line);
									}

									if (y % 10 == 0 AND
										Meter)
									{
										Meter->Value(y);
									}
								}
								break;
							}
							default:
							{
								Status = false;
								break;
							}
						}
						break;
					}
				}
			}
		}

		return Status;
	}

	bool WriteImage(GdcBasePrimitives *pDC)
	{
		return false;
	}
};

class GdcTgaFactory : public GdcFilterFactory
{
	bool CheckFile(char *File)
	{
		return (File) ? stristr(File, ".tga") != 0 : false;
	}

	GdcFileFormat *NewObject()
	{
		return NEW(GdcTga);
	}

} TgaFactory;
