/*hdr
**	FILE:			Region.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			22/8/97
**	DESCRIPTION:	GDC Region class
**
**	Copyright (C) 1997, Matthew Allen
**		fret@ozemail.com.au
*/

#include "Gdc2.h"

void GdcRegion::ZOff(int x, int y)
{
	x1 = y1 = 0;
	x2 = x;
	y2 = y;
}

void GdcRegion::Normal()
{
	if (x1 > x2)
	{
		swap(x1, x2);
	}
	
	if (y1 > y2)
	{
		swap(y1, y2);
	}
}

bool GdcRegion::Valid()
{
	return (x1 <= x2 AND y1 <= y2);
}

void GdcRegion::Offset(int x, int y)
{
	x1 += x;
	y1 += y;
	x2 += x;
	y2 += y;
}

void GdcRegion::Offset(GdcRegion *a)
{
	x1 += a->x1;
	y1 += a->y1;
	x2 += a->x2;
	y2 += a->y2;
}

void GdcRegion::Size(int x, int y)
{
	x1 += x;
	y1 += y;
	x2 -= x;
	y2 -= y;
}

void GdcRegion::Size(GdcRegion *a)
{
	x1 += a->x1;
	y1 += a->y1;
	x2 -= a->x2;
	y2 -= a->y2;
}

void GdcRegion::Dimension(int x, int y)
{
	x2 = x1 + x - 1;
	y2 = y1 + y - 1;
}

void GdcRegion::Dimension(GdcRegion *a)
{
	x2 = x1 + a->X() - 1;
	y2 = y1 + a->Y() - 1;
}

void GdcRegion::Bound(GdcRegion *b)
{
	x1 = max(x1,b->x1);		
	y1 = max(y1,b->y1);
	x2 = min(x2,b->x2);
	y2 = min(y2,b->y2);
}

bool GdcRegion::Overlap(int x, int y)
{
	return (x >= x1) AND (x <= x2) AND (y >= y1) AND (y <= y2);
}

bool GdcRegion::Overlap(GdcRegion *b)
{
	if (Valid() AND
		b->Valid() AND
		(x1 > b->x2 OR
		y1 > b->y2 OR
		x2 < b->x1 OR
		y2 < b->y1))
	{
		return FALSE;
	}

	return TRUE;
}

void GdcRegion::Union(int x, int y)
{
	x1 = min(x1, x);
	y1 = min(y1, y);
	x2 = max(x2, x);
	y2 = max(y2, y);
}

void GdcRegion::Union(GdcRegion *a)
{
	x1 = min(a->x1, x1);
	y1 = min(a->y1, y1);
	x2 = max(a->x2, x2);
	y2 = max(a->y2, y2);
}

void GdcRegion::Union(GdcRegion *a, GdcRegion *b)
{
	x1 = min(a->x1,b->x1);
	y1 = min(a->y1,b->y1);
	x2 = max(a->x2,b->x2);
	y2 = max(a->y2,b->y2);
}

void GdcRegion::Intersection(GdcRegion *a)
{
	if (Overlap(a))
	{
		x1 = max(a->x1, x1);
		y1 = max(a->y1, y1);
		x2 = min(a->x2, x2);
		y2 = min(a->y2, y2);
	}
	else
	{
		x1 = y1 = 0;
		x2 = y2 = -1;
	}
}

void GdcRegion::Intersection(GdcRegion *a, GdcRegion *b)
{
	if (a->Overlap(b))
	{
		x1 = max(a->x1,b->x1);
		y1 = max(a->y1,b->y1);
		x2 = min(a->x2,b->x2);
		y2 = min(a->y2,b->y2);
	}
	else
	{
		x1 = y1 = 0;
		x2 = y2 = -1;
	}
}

bool operator ==(GdcRegion &a, GdcRegion &b)
{
	return (a.x1 == b.x1 AND
		a.y1 == b.y1 AND
		a.x2 == b.x2 AND
		a.y2 == b.y2);
}

bool operator !=(GdcRegion &a, GdcRegion &b)
{
	return NOT ((a.x1 == b.x1 AND
		a.y1 == b.y1 AND
		a.x2 == b.x2 AND
		a.y2 == b.y2));
}

GdcComplexRegion::GdcComplexRegion()
{
	Size = Alloc = 0;
	a = 0;
	Current = 0;
}

GdcComplexRegion::GdcComplexRegion(int X1, int Y1, int X2, int Y2)
{
	Size = Alloc = 0;	
	Current = 0;
	a = 0;

	SetSize(1);
	if (a)
	{
		a[0].x1 = X1;
		a[0].y1 = Y1;
		a[0].x2 = X2;
		a[0].y2 = Y2;
	}
}

GdcComplexRegion::GdcComplexRegion(GdcRegion &r)
{
	Size = Alloc = 0;	
	Current = 0;
	a = 0;

	SetSize(1);
	if (a)
	{
		a[0].x1 = r.x1;
		a[0].y1 = r.y1;
		a[0].x2 = r.x2;
		a[0].y2 = r.y2;
	}
}

#if defined WIN32
GdcComplexRegion::GdcComplexRegion(RECT &r)
#else
GdcComplexRegion::GdcComplexRegion(BRect &r)
#endif
{
	Size = Alloc = 0;	
	Current = 0;
	a = 0;

	SetSize(1);
	if (a)
	{
		a[0].x1 = r.left;
		a[0].y1 = r.top;
		a[0].x2 = r.right-CornerOffset;
		a[0].y2 = r.bottom-CornerOffset;
	}
}

GdcComplexRegion::~GdcComplexRegion()
{
	Empty();
}

GdcRegion *GdcComplexRegion::First()
{
	Current = 0;
	return (Current >= 0 AND Current < Size) ? a+Current : 0;
}

GdcRegion *GdcComplexRegion::Last()
{
	Current = Size-1;
	return (Current >= 0 AND Current < Size) ? a+Current : 0;
}

GdcRegion *GdcComplexRegion::Next()
{
	Current++;
	return (Current >= 0 AND Current < Size) ? a+Current : 0;
}

GdcRegion *GdcComplexRegion::Prev()
{
	Current--;
	return (Current >= 0 AND Current < Size) ? a+Current : 0;
}

bool GdcComplexRegion::SetSize(int s)
{
	bool Status = TRUE;

	if (s > Alloc)
	{
		int NewAlloc = (s + 15) & (~0xF);
		GdcRegion *Temp = NEW(GdcRegion[NewAlloc]);
		if (Temp)
		{
			int Common = min(Size, s);
			memcpy(Temp, a, sizeof(GdcRegion)*Common);
			DeleteArray(a);
			a = Temp;
			Size = s;
			Alloc = NewAlloc;
		}
		else
		{
			Status = FALSE;
		}
	}
	else
	{
		Size = s;
		if (NOT Size)
		{
			DeleteArray(a);
			Alloc = 0;
		}
	}

	return Status;
}

bool GdcComplexRegion::Delete(int i)
{
	if (i >= 0 AND i < Size)
	{
		if (i != Size - 1)
		{
			memcpy(a + i, a + (i + 1), sizeof(GdcRegion) * (Size - i - 1));
		}
		Size--;
		return TRUE;
	}
	return FALSE;
}

void GdcComplexRegion::Empty()
{
	Size = Alloc = 0;
	DeleteArray(a);
}

void GdcComplexRegion::ZOff(int x, int y)
{
	if (SetSize(1))
	{
		a[0].ZOff(x, y);
	}
}

void GdcComplexRegion::Normal()
{
	for (int i=0; i<Size; i++)
	{
		a[i].Normal();
	}
}

bool GdcComplexRegion::Valid()
{
	bool Status = TRUE;

	for (int i=0; i<Size AND Status; i++)
	{
		Status &= a[i].Valid();
	}

	return Status AND Size > 0;
}

void GdcComplexRegion::Offset(int x, int y)
{
	for (int i=0; i<Size; i++)
	{
		a[i].Offset(x, y);
	}
}

void GdcComplexRegion::Bound(GdcRegion *b)
{
	if (b)
	{
		for (int i=0; i<Size; )
		{
			if (a[i].Overlap(b))
			{
				a[i++].Bound(b);
			}
			else
			{
				Delete(i);
			}
		}
	}
}

bool GdcComplexRegion::Overlap(GdcRegion *b)
{
	bool Status = FALSE;

	if (b)
	{
		for (int i=0; i<Size AND NOT Status; i++)
		{
			Status |= a[i].Overlap(b);
		}
	}

	return Status;
}

void GdcComplexRegion::Union(GdcRegion *b)
{
	if (b AND b->Valid())
	{
		Subtract(b);
		GdcRegion *n = NewOne();
		if (n)
		{
			*n = *b;
		}
	}
}

void GdcComplexRegion::Intersect(GdcRegion *b)
{
	if (b)
	{
		for (int i=0; i<Size; )
		{
			if (a[i].Overlap(b))
			{
				a[i++].Bound(b);
			}
			else
			{
				Delete(i);
			}
		}
	}
}

void GdcComplexRegion::Subtract(GdcRegion *b)
{
	if (b AND b->Valid())
	{
		GdcRegion Sub = *b;
		int StartSize = Size;

		Normal();
		Sub.Normal();

		for (int i=0; i<StartSize; )
		{
			GdcRegion c = *(a + i);
			if (c.Overlap(&Sub))
			{
				if (Sub.y1 > c.y1)
				{
					GdcRegion *n = NewOne();
					n->x1 = c.x1;
					n->y1 = c.y1;
					n->x2 = c.x2;
					n->y2 = Sub.y1 - 1;
				}

				if (Sub.y2 < c.y2)
				{
					GdcRegion *n = NewOne();
					n->x1 = c.x1;
					n->y1 = Sub.y2 + 1;
					n->x2 = c.x2;
					n->y2 = c.y2;
				}

				if (Sub.x1 > c.x1)
				{
					GdcRegion *n = NewOne();
					n->x1 = c.x1;
					n->y1 = max(Sub.y1, c.y1);
					n->x2 = Sub.x1 - 1;
					n->y2 = min(Sub.y2, c.y2);
				}

				if (Sub.x2 < c.x2)
				{
					GdcRegion *n = NewOne();
					n->x1 = Sub.x2 + 1;
					n->y1 = max(Sub.y1, c.y1);
					n->x2 = c.x2;
					n->y2 = min(Sub.y2, c.y2);
				}

				Delete(i);
				StartSize--;
			}
			else
			{
				i++;
			}
		}
	}	
}

bool operator ==(GdcComplexRegion &a, GdcComplexRegion &b)
{
	return FALSE;
}

bool operator !=(GdcComplexRegion &a, GdcComplexRegion &b)
{
	return FALSE;
}

/*
int StartSize = Size;
bool Merged = FALSE;

Normal();
b->Normal();

for (int i=0; i<StartSize; i++)
{
	GdcRegion c = *(a + i);
	if (c.Overlap(b))
	{
		GdcRegion *High, *Low;

		if (c.y1 < b->y1)
		{
			High = &c;
			Low = b;
		}
		else
		{
			High = b;
			Low = &c;
		}

		if (High->y1 < Low->y1)
		{
			GdcRegion *n = NewOne();
			n->x1 = High->x1;
			n->x2 = High->x2;
			n->y1 = High->y1;
			n->y2 = Low->y1 - 1;
		}

		if (Low->y2 > High->y2)
		{
			GdcRegion *n = NewOne();
			n->x1 = Low->x1;
			n->x2 = Low->x2;
			n->y1 = High->y2 + 1;
			n->y2 = Low->y2;
		}

		a[i].x1 = min(c.x1, b->x1);
		a[i].y1 = max(c.y1, b->y1);
		a[i].x2 = max(c.x2, b->x2);
		a[i].y2 = min(c.y2, b->y2);

		Merged = TRUE;
	}
}

if (NOT Merged)
{
	GdcRegion *n = NewOne();
	if (n)
	{
		*n = *b;
	}
}
*/

