/*hdr
**      FILE:           Contain.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           27/7/95
**      DESCRIPTION:    Continer classes
**
**      Copyright (C) 1996, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "LibDefs.h"
#include "VMem.h"
#include "Contain.h"

PtrArray::PtrArray()
{
	Used = 0;
	Size = 0;
	BankSize = DEFAULT_ARRAY_BANK;
	Array = 0;
}

PtrArray::~PtrArray()
{
	if (Array) delete [] Array;
}

bool PtrArray::SetSize(int NewSize)
{
	bool Status = FALSE;

	if (NewSize > Size OR abs(NewSize-Size) > BankSize)
	{
		int BankMod = NewSize % BankSize;
		
		if (BankMod)
		{
			NewSize += BankSize - BankMod;
		}
		
		void **a = NEW(void*[NewSize]);
		if (a)
		{
			int Common = min(Used, NewSize);
			memcpy(a, Array, Common);
			if (Common < NewSize) memset(a+Common, 0, sizeof(*a)*(NewSize-Common));

			DeleteArray(Array);
			Array = a;			

			Used = Common;			
			Size = NewSize;
			Status = TRUE;
		}
	}
	else
	{
		Used = max(NewSize, 0);
		Status = TRUE;
	}

	return Status;
}

bool PtrArray::Add(void *Ptr, int Where)
{
	if (Used >= Size - 1)
	{
		if (NOT SetSize(Size + BankSize))
		{
			return FALSE;
		}
	}

	Array[Used] = Ptr;
	Used++;

	return TRUE;
}

bool PtrArray::Remove(int Where)
{
	bool Status = FALSE;

	if (Array AND Where >= 0 AND Where < Used)
	{
		memmove(&Array[Where], &Array[Where+1], Used-Where);
		Used--;
		Status = TRUE;
	}
	
	return Status;
}

bool PtrArray::Empty()
{
	Used = 0;
	return TRUE;
}

VoidPtr &PtrArray::operator [](int i)
{
	void *Ptr = 0;
	return (Array AND i >= 0 AND i < Used) ? Array[i] : Ptr;
}

template <class StackType>
Stack<StackType>::Stack()
{
	Top = 0;
}

template <class StackType>
Stack<StackType>::~Stack()
{
	Empty();
}

template <class StackType>
void Stack<StackType>::Push(StackType *Data)
{
	StackType *Temp = NEW(StackNode);

	if (Temp)
	{
		Temp->Next = Top;
		Top = Temp;
		Top->Data = Data;
	}
}

template <class StackType>
StackType *Stack<StackType>::Pop()
{
	if (Top)
	{
		StackType *Data = Top->Data;
		StackNode *Temp = Top;
		Top = Top->Next;

		delete Temp;
		return Data;
	}

	return 0;
}

template <class StackType>
StackType *Stack<StackType>::Peek()
{
	return (Top) ? Top->Data : 0;
}

template <class StackType>
int Stack<StackType>::IsEmpty()
{
	return (Top) ? TRUE : FALSE;
}

template <class StackType>
void Stack<StackType>::Empty()
{
	while (!IsEmpty())
	{
		Pop();
	}
}

template <class QueType>
Que<QueType>::Que()
{
	First = Last = 0;
}

template <class QueType>
Que<QueType>::~Que()
{
	Empty();
}

template <class QueType>
void Que<QueType>::Insert(QueType *Data)
{
	if (Last)
	{
		Last->Next = NEW(QueNode);
		Last = Last->Next;
	}
	else
	{
		First = Last = NEW(QueNode);
	}

	Last->Next = 0;
	Last->Data = Data;
}

template <class QueType>
QueType *Que<QueType>::Delete()
{
	if (First)
	{
		QueNode *Temp = First;
		QueData *Data = First->Data;
		First = First->Next;
		if (!First) Last = 0;		

		delete Temp;
		return Data;
	}

	return 0;
}

template <class QueType>
QueType *Que<QueType>::Peek()
{
	return (First) ? First->Data : 0;
}

template <class QueType>
int Que<QueType>::IsEmpty()
{
	return (First) ? FALSE : TRUE;
}

template <class QueType>
void Que<QueType>::Empty()
{
	while (!IsEmpty())
	{
		Delete();
	}
}

template <class PQueType>
PriorityQue<PQueType>::PriorityQue()
{
	Nodes = 0;
	Top = 0;
	Size = 0;
	SetSize(DEFAULT_PQUE_SIZE);
}

template <class PQueType>
PriorityQue<PQueType>::~PriorityQue()
{
	Empty();
}

template <class PQueType>
void PriorityQue<PQueType>::SetSize(long NewSize)
{
	if (NewSize > 0)
	{
		PQueNode *Temp = NEW(PQueNode[NewSize]);
		
		if (Temp)
		{
			memset(Temp,0,sizeof(PQueNode)*NewSize);
			
			if (Top)
			{
				memcpy(Temp,Top,sizeof(PQueNode)*Nodes);
				delete [] Top;
			}
			
			Top = Temp;
			Size = NewSize;
		}
	}
	else
	{
		Empty();
	}
}

template <class PQueType>
void PriorityQue<PQueType>::BubbleUp(long Start)
{
	long Parent = (Start - 1) / 2;
	
	while (Start)
	{
		if (Top[Parent].Key > Top[Start].Key)
		{
			PQueNode Temp = Top[Parent];
			Top[Parent] = Top[Start];
			Top[Start] = Temp;			

			Start = Parent;
			Parent = (Start - 1) / 2;
		}
		else
		{
			Start = 0;
		}
	}
}

template <class PQueType>
void PriorityQue<PQueType>::BubbleDown()
{
	long Start = 0;
	long Child = 1;
	
	while (Child < Nodes)
	{
		long Swap = (Top[Child].Key < Top[Child+1].Key)
				? Child : Child + 1;
		Swap = min(Swap, Nodes-1);

		if (Top[Swap].Key < Top[Start].Key)
		{
			PQueNode Temp = Top[Swap];
			Top[Swap] = Top[Start];
			Top[Start] = Temp;			
			Start = Swap;
			Child = (Start * 2) + 1;
		}
		else
		{
			Child = Nodes;
		}
	}
}

template <class PQueType>
void PriorityQue<PQueType>::Insert(PQueType *Data,long Key)
{
	if (!Top || Nodes >= Size)
	{
		SetSize((Size) ? Size * 2 : DEFAULT_PQUE_SIZE);
	}

	if (Top && Nodes < Size)
	{
		Top[Nodes].Data = Data;
		Top[Nodes].Key = Key;
		Nodes++;

   		BubbleUp(Nodes-1);
	}
}

template <class PQueType>
PQueType *PriorityQue<PQueType>::DeleteMin()
{
	if (Top && Nodes > 0)
	{
		PQueType *Temp = Top->Data;
		
		Nodes--;

		if (Nodes >= 1)
		{
			Top[0] = Top[Nodes];			
			BubbleDown();
		}
		
		return Temp;
	}
	
	return 0;
}

template <class PQueType>
PQueType *PriorityQue<PQueType>::Peek()
{
	return (Top) ? Top->Data : 0;
}

template <class PQueType>
int PriorityQue<PQueType>::IsEmpty()
{
	return (Top) ? FALSE : TRUE;
}

template <class PQueType>
void PriorityQue<PQueType>::Empty()
{
	delete [] Top;
	Top = 0;
	Size = 0;
	Nodes = 0;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////

#define AssertValid()
// #define AssertValid()	ASSERT(IsValid())

DLinkList::DLinkList()
{
	Items = 0;
	Cur = FirstObj = LastObj = 0;
}

DLinkList::~DLinkList()
{
	Empty();
}

DLinkList &DLinkList::operator =(DLinkList &lst)
{
	Empty();

	for (Item *p = lst.FirstObj; p; p = p->Next)
	{
		Insert(p->Ptr);
	}
	
	return *this;
}

bool DLinkList::IsValid()
{
	if (FirstObj)
	{
		if (NOT LastObj) return false;
		int n=0;
		bool FoundCur = false;
		for (Item *i = FirstObj; i; i = i->Next)
		{
			if (NOT i->Ptr) return false;
			if (Cur AND Cur == i) FoundCur = true;
			if (i->Next AND i->Next->Prev != i) return false;
			if (i == LastObj)
			{
				if (i->Next != 0) return false;
			}
			n++;
		}

		return (NOT Cur OR FoundCur) AND (n == Items);
	}

	return Items == 0;
}

void DLinkList::Empty()
{
	AssertValid();

	Item *i = FirstObj;
	while (i)
	{
		Item *Ni = i->Next;
		delete i;
		i = Ni;
	}

	Cur = FirstObj = LastObj = 0;
	Items = 0;

	AssertValid();
}

void *DLinkList::operator [](int Index)
{
	AssertValid();

	if (Index >= 0 AND Index < Items)
	{
		Item *i = FirstObj;
		while (i AND Index)
		{
			i = i->Next;
			Index--;
		}

		if (Index == 0 AND i)
		{
			Cur = i;
			return i->Ptr;
		}
	}

	return 0;
}

int DLinkList::FindIndex(void *p)
{
	AssertValid();

	int i=0;
	Item *item = FirstObj;
	while (item)
	{
		if (item->Ptr == p)
		{
			return i;
		}
		
		i++;
		item = item->Next;

	}

	return -1;
}

bool DLinkList::Delete()
{
	bool Status = false;

	AssertValid();

	if (Cur)
	{
		Item *i = Cur->Next;

		if (Cur->Prev)
		{
			Cur->Prev->Next = Cur->Next;
		}
		else
		{
			FirstObj = Cur->Next;
		}

		if (Cur->Next)
		{
			Cur->Next->Prev = Cur->Prev;
		}
		else
		{
			LastObj = Cur->Prev;
		}

		DeleteObj(Cur);
		Cur = i;
		Items--;

		Status = true;
	}

	AssertValid();

	return Status;
}

bool DLinkList::Delete(int Index)
{
	bool Status = false;

	AssertValid();

	Item *i = FirstObj;
	while (i AND Index)
	{
		i = i->Next;
		Index--;
	}

	if (Index == 0 AND i)
	{
		Cur = i->Next;

		if (i->Prev)
		{
			i->Prev->Next = i->Next;
		}
		else
		{
			FirstObj = i->Next;
		}

		if (i->Next)
		{
			i->Next->Prev = i->Prev;
		}
		else
		{
			LastObj = i->Prev;
		}

		DeleteObj(i);
		Items--;

		Status = true;
	}

	AssertValid();

	return Status;
}

bool DLinkList::Delete(void *p)
{
	bool Status = false;

	AssertValid();

	Item *i = FirstObj;
	while (i AND i->Ptr != p)
	{
		i = i->Next;
	}

	if (i)
	{
		Cur = i->Next;

		if (i->Prev)
		{
			i->Prev->Next = i->Next;
		}
		else
		{
			FirstObj = i->Next;
		}

		if (i->Next)
		{
			i->Next->Prev = i->Prev;
		}
		else
		{
			LastObj = i->Prev;
		}

		DeleteObj(i);
		Items--;

		Status = true;
	}

	AssertValid();

	return Status;
}

bool DLinkList::Insert(void *p, int Index)
{
	AssertValid();

	Item *n = NEW(Item);
	if (n)
	{
		n->Ptr = p;
		if (NOT FirstObj)
		{
			FirstObj = LastObj = n;
			n->Prev = n->Next = 0;
			Items = 1;

			AssertValid();

			return TRUE;
		}
		else
		{
			if (Index < 0 OR Index >= Items)
			{
				Item *i = LastObj;
				LastObj = i->Next = n;
				n->Prev = i;
				n->Next = 0;
				Items++;

				AssertValid();

				return TRUE;
			}
			else if (Index == 0)
			{
				Item *i = FirstObj;
				FirstObj = i->Prev = n;
				n->Next = i;
				n->Prev = 0;
				Items++;

				AssertValid();

				return TRUE;
			}
			else
			{
				Item *i = FirstObj;
				while (Index > 1 AND i->Next)
				{
					i = i->Next;
					Index--;
				}

				n->Prev = i;
				n->Next = i->Next;
				i->Next = n;
				if (n->Next) n->Next->Prev = n;
				Items++;

				AssertValid();

				return TRUE;
			}
		}
	}

	AssertValid();

	return FALSE;
}

int DLinkList::IndexOf(void *p)
{
	AssertValid();

	if (p)
	{
		int Index = 0;
		Item *i = FirstObj;
		while (i)
		{
			if (i->Ptr == p)
			{
				return Index;
			}
			i = i->Next;
			Index++;
		}
	}

	return -1;
}

bool DLinkList::HasItem(void *p)
{
	AssertValid();

	if (p)
	{
		Item *i = FirstObj;
		while (i)
		{
			if (i->Ptr == p)
			{
				return TRUE;
			}
			i = i->Next;
		}
	}
	return FALSE;
}

void *DLinkList::ItemAt(int Index)
{
	AssertValid();

	if (Index >= 0 AND Index < Items)
	{
		Item *i = FirstObj;
		while (i AND Index > 0)
		{
			i = i->Next;
			Index--;
		}

		if (Index == 0 AND i)
		{
			Cur = i;
			return i->Ptr;
		}
	}

	return NULL;
}

void *DLinkList::First()
{
	AssertValid();

	Cur = FirstObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkList::Last()
{
	AssertValid();

	Cur = LastObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkList::Next()
{
	if (Cur)
	{
		Cur = Cur->Next;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkList::Prev()
{
	if (Cur)
	{
		Cur = Cur->Prev;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkList::Find(void *p)
{
	AssertValid();

	Cur = FirstObj;
	while (Cur AND Cur->Ptr != p)
	{
		Cur = Cur->Next;
	}
	return (Cur) ? Cur->Ptr : 0;
}

typedef int (*BTreeCompare)(void *a, void *b, int Data);
typedef void **AddrVoidPtr;

class BTreeNode
{
public:
	void *Node;
	BTreeNode *Left;
	BTreeNode *Right;

	BTreeNode()
	{
		Left = 0;
		Right = 0;
	}

	~BTreeNode()
	{
		DeleteObj(Left);
		DeleteObj(Right);
	}

	bool Add(void *Obj, BTreeCompare Proc, int Data)
	{
		int c = Proc(Obj, Node, Data);
		BTreeNode **Ptr = (c <= 0) ? &Left : &Right;
		if (*Ptr)
		{
			return (*Ptr)->Add(Obj, Proc, Data);
		}
		else
		{
			*Ptr = NEW(BTreeNode);
			if (*Ptr)
			{
				(*Ptr)->Node = Obj;
				return true;
			}
		}

		return false;
	}

	AddrVoidPtr *Index(AddrVoidPtr *Items)
	{
		if (Left)
		{
			Items = Left->Index(Items);
		}

		**Items = Node;
		Items++;

		if (Right)
		{
			Items = Right->Index(Items);
		}

		return Items;
	}
};

class BTree
{
	int Items;
	BTreeNode *Root;

public:
	BTree()
	{
		Root = 0;
		Items = 0;
	}

	~BTree()
	{
		DeleteObj(Root);
		Items = 0;
	}

	int GetItems() { return Items; }

	bool Add(void *Obj, BTreeCompare Proc, int Data)
	{
		bool Status = false;
		if (Obj)
		{
			if (NOT Root)
			{
				Root = NEW(BTreeNode);
				if (Root)
				{
					Root->Node = Obj;
					Status = true;
					Items = 1;
				}
			}
			else
			{
				Status = Root->Add(Obj, Proc, Data);
				if (Status)
				{
					Items++;
				}
			}
		}
		return Status;
	}

	void Index(void ***Items)
	{
		if (Root)
		{
			Root->Index(Items);
		}
	}
};

void DLinkList::Sort(int (*Compare)(void *a, void *b, int data), int Data)
{
	if (Items > 1)
	{
		BTree Tree;
		AddrVoidPtr *iLst = NEW(AddrVoidPtr[Items]);
		if (iLst)
		{
			int n=0;
			Item *i = FirstObj;
			while (i)
			{
				iLst[n++] = &i->Ptr;
				Tree.Add(i->Ptr, Compare, Data);
				i = i->Next;
			}

			Tree.Index(iLst);
			DeleteArray(iLst);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
DLinkIterator::DLinkIterator(DLinkList *list)
{
	List = list;
	Cur = 0;
}

DLinkIterator::~DLinkIterator()
{
	List = 0;
	Cur = 0;
}

void *DLinkIterator::First()
{
	Cur = List->FirstObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkIterator::Last()
{
	Cur = List->LastObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkIterator::Next()
{
	if (Cur)
	{
		Cur = Cur->Next;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkIterator::Prev()
{
	if (Cur)
	{
		Cur = Cur->Prev;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkIterator::Find(void *p)
{
	Cur = List->FirstObj;
	while (Cur AND Cur->Ptr != p)
	{
		Cur = Cur->Next;
	}
	return (Cur) ? Cur->Ptr : 0;
}

////////////////////////////////////////////////////////////////////////////////////////
BytePipe::BytePipe()
{
}

BytePipe::~BytePipe()
{
	Empty();
}

void BytePipe::Empty()
{
	for (Block *b = Mem.First(); b; b = Mem.Next())
	{
		DeleteObj(b);
	}
}

int BytePipe::Sizeof()
{
	int Size = 0;
	for (Block *b = Mem.First(); b; b = Mem.Next())
	{
		if (b->Ptr)
		{
			Size += b->Size - b->Next;
		}
	}
	return Size;
}

bool BytePipe::Peek(uchar *Ptr, int Size)
{
	bool Status = FALSE;

	if (Ptr AND Size <= Sizeof())
	{
		Block *b = 0;
		for (b = Mem.First(); b AND Size > 0; b = Mem.Next())
		{
			int Copy = min(Size, b->Size - b->Next);
			if (Copy > 0)
			{
				memcpy(Ptr, b->Ptr + b->Next, Copy);
				Ptr += Copy;
				Size -= Copy;
			}
		}

		Status = (Size == 0);
	}

	return Status;
}

bool BytePipe::Pop(uchar *Ptr, int Size)
{
	bool Status = FALSE;

	if (Ptr AND Size <= Sizeof())
	{
		Block *b = 0;
		for (b = Mem.First(); b AND Size > 0; b = Mem.Next())
		{
			int Copy = min(Size, b->Size - b->Next);
			if (Copy > 0)
			{
				memcpy(Ptr, b->Ptr + b->Next, Copy);
				Ptr += Copy;
				Size -= Copy;
				b->Next += Copy;
			}
		}

		for (b = Mem.First(); b AND b->Next >= b->Size; b = Mem.First())
		{
			Mem.Delete(b);
			DeleteObj(b);
		}

		Status = (Size == 0);
	}

	return Status;
}

bool BytePipe::Pop(short &i)
{
	short n;
	if (Pop((uchar*) &n, sizeof(n)))
	{
		i = n;
		return true;
	}
	return false;
}

bool BytePipe::Pop(int &i)
{
	int n;
	if (Pop((uchar*) &n, sizeof(n)))
	{
		i = n;
		return true;
	}
	return false;
}

bool BytePipe::Pop(double &i)
{
	double n;
	if (Pop((uchar*) &n, sizeof(n)))
	{
		i = n;
		return true;
	}
	return false;
}

bool BytePipe::Push(uchar *Ptr, int Size)
{
	bool Status = FALSE;

	if (Ptr AND Size > 0)
	{
		Block *b = NEW(Block);
		if (b)
		{
			b->Ptr = NEW(uchar[Size]);
			if (b->Ptr)
			{
				memcpy(b->Ptr, Ptr, Size);
				b->Size = Size;
				b->Next = 0;
				Mem.Insert(b);
				Status = TRUE;
			}
			else
			{
				DeleteObj(b);
			}
		}
	}

	return Status;
}

bool BytePipe::Push(short i)
{
	return Push((uchar*) &i, sizeof(i));
}

bool BytePipe::Push(int i)
{
	return Push((uchar*) &i, sizeof(i));
}

bool BytePipe::Push(double i)
{
	return Push((uchar*) &i, sizeof(i));
}

//////////////////////////////////////////////////////////////////////////
bool StringPipe::Push(char *Str, int Len)
{
	if (Str)
	{
		return BytePipe::Push((uchar*) Str, (Len>=0) ? Len : strlen(Str));
	}
	return 0;
}

char *StringPipe::NewStr()
{
	int Len = Sizeof();
	char *Str = NEW(char[Len+1]);
	if (Str)
	{
		Pop((uchar*) Str, Len);
		Str[Len] = 0;
	}
	return Str;
}

