/*hdr
**      FILE:           Contain.h
**      AUTHOR:         Matthew Allen
**      DATE:           27/11/96
**      DESCRIPTION:    Container class header
**
**      Copyright (C) 1996, Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef __CONTAIN_H
#define __CONTAIN_H

#define DEFAULT_ARRAY_BANK			16

typedef void *VoidPtr;

class LgiClass PtrArray {

	int Used;
	int Size;
	int BankSize;
	void **Array;

public:
	PtrArray();
	virtual ~PtrArray();

	int GetSize() { return Used; }
	bool SetSize(int NewSize);
	bool SetBankSize(int NewBank);
	virtual bool Add(void *Ptr, int Where);
	virtual bool Remove(int Where);
	virtual bool Empty();

	VoidPtr &operator [](int i);
};

template <class ArrayType>
class LgiClass Array : public PtrArray {
public:
	Array() {}
	~Array() {}

	bool SetSize(int Size, int Bank)
	{
		return PtrArray::SetSize(sizeof(ArrayType), Size, Bank);
	}

	bool Add(ArrayType *Obj, int Where)
	{
		return PtrArray::Add((char*) Obj, Where);
	}

	ArrayType *operator [](int i)
	{
		PtrArray *p = this;
		return (ArrayType*) (*p)[i];
	}
};

template <class StackType>
class LgiClass Stack {

	class StackNode {
	public:
		StackNode *Next;
		StackType *Data;
	} *Top;

public:
	Stack();
	~Stack();

	void Push(StackType *Data);
	StackType *Pop();
	StackType *Peek();
	int IsEmpty();
	void Empty();
};

template <class QueType>
class LgiClass Que {

	class QueNode {
	public:
		QueNode *Next;
		QueType *Data;
	} *First, *Last;

public:
	Que();
	~Que();

	void Insert(QueType *Data);
	QueType *Delete();
	QueType *Peek();
	int IsEmpty();
	void Empty();
};

#define DEFAULT_PQUE_SIZE			32

template <class PQueType>
class LgiClass PriorityQue {

	class PQueNode {
	public:
		long Key;
		PQueType *Data;
	} *Top;

	long Nodes;
	long Size;
	
	void SetSize(long NewSize);
	void BubbleUp(long Start);
	void BubbleDown();

public:
	PriorityQue();
	~PriorityQue();

	void NewSize(long NewSize);
	void Insert(PQueType *Data,long Key);
	PQueType *DeleteMin();
	PQueType *Peek();
	int IsEmpty();
	void Empty();
};

class DLinkList;
class DLinkInterator;

class LgiClass Item {
public:
	Item *Prev, *Next;
	void *Ptr;
};

class LgiClass DLinkList {

	friend class DLinkIterator;

protected:
	friend class ListIterator;

	int Items;
	Item *FirstObj, *LastObj, *Cur;

	int FindIndex(void *p);
	bool IsValid();

public:
	DLinkList();
	virtual ~DLinkList();

	int GetItems() { return Items; }
	virtual void Empty();
	void *operator [](int Index);
	DLinkList &operator =(DLinkList &lst);

	bool Delete();
	bool Delete(int Index);
	bool Delete(void *p);
	bool Insert(void *p, int Index = -1);
	void *First();
	void *Last();
	void *Next();
	void *Prev();
	void *Find(void *p);
	int IndexOf(void *p);
	bool HasItem(void *p);
	void *ItemAt(int i);
	void Sort(int (*Compare)(void *a, void *b, int Data), int Data);
};

class LgiClass DLinkIterator {
protected:
	DLinkList *List;
	Item *Cur;

public:
	DLinkIterator(DLinkList *list);
	~DLinkIterator();

	void *First();
	void *Last();
	void *Next();
	void *Prev();
	void *Find(void *p);
};

template <class Type>
class LgiClass List : public DLinkList {
public:
	virtual bool Delete() { return DLinkList::Delete(); }
	virtual bool Delete(Type *p) { return DLinkList::Delete((void*)p); }
	virtual bool Insert(Type *p, int Index = -1) { return DLinkList::Insert((void*)p, Index); }
	Type *First() { return (Type*) DLinkList::First(); }
	Type *Last() { return (Type*) DLinkList::Last(); }
	Type *Next() { return (Type*) DLinkList::Next(); }
	Type *Prev() { return (Type*) DLinkList::Prev(); }
	Type *Current() { return (Cur) ? (Type*) Cur->Ptr : 0; }
	Type *Find(Type *p) { return (Type*) DLinkList::Find((void*) p); }
	Type *operator [](int Index) { return ((Type*) ((*((DLinkList*)this))[Index])); }
	int IndexOf(Type *p) { return DLinkList::IndexOf((void*)p); }
	bool HasItem(Type *p) { return DLinkList::HasItem((void*)p); }
	Type *ItemAt(int i) { return (Type*) DLinkList::ItemAt(i); }
	void Sort(int (*Compare)(Type *a, Type *b, int data), int Data) { DLinkList::Sort( (int (*)(void *a, void *b, int data)) Compare, Data); }

	List &operator =(List<Type> &lst)
	{
		DLinkList *l = this;
		*l = (DLinkList&)lst;
		return *this;
	}
};

template <class Type>
class LgiClass Iterator : public DLinkIterator {
public:
	Iterator(DLinkList *list) : DLinkIterator(list) {}

	Type *First() { return (Type*) DLinkIterator::First(); }
	Type *Last() { return (Type*) DLinkIterator::Last(); }
	Type *Next() { return (Type*) DLinkIterator::Next(); }
	Type *Prev() { return (Type*) DLinkIterator::Prev(); }
	Type *Current() { return (Cur) ? (Type*) Cur->Ptr : 0; }
	Type *Find(Type *p) { return (Type*) DLinkIterator::Find((void*) p); }
	Type *operator [](int Index) { return ((Type*) ((*List)[Index])); }
};

#define DeleteObjList(list, type)									\
if (list)															\
{																	\
	for (type *_obj = (list)->First(); _obj; _obj = (list)->Next())	\
		DeleteObj(_obj);											\
	(list)->Empty();												\
}

#define DeleteArrayList(list, type)									\
if (list)															\
{																	\
	for (type *_obj = (list)->First(); _obj; _obj = (list)->Next())	\
		DeleteArray(_obj);											\
	(list)->Empty();												\
}

class LgiClass BytePipe
{
	class Block
	{
	public:
		int Next;
		int Size;
		uchar *Ptr;

		Block()
		{
			Next = 0;
			Size = 0;
			Ptr = 0;
		}

		~Block()
		{
			Next = 0;
			Size = 0;
			DeleteArray(Ptr);
		}
	};

	List<Block> Mem;

public:
	BytePipe();
	virtual ~BytePipe();

	virtual bool IsEmpty() { return Mem.GetItems() == 0; }
	virtual void Empty();
	virtual int Sizeof();

	virtual bool Peek(uchar *Ptr, int Size);

	virtual bool Pop(uchar *Ptr, int Size);
	bool Pop(short &i);
	bool Pop(int &i);
	bool Pop(double &i);

	virtual bool Push(uchar *Ptr, int Size);
	bool Push(short i);
	bool Push(int i);
	bool Push(double i);
};

class LgiClass StringPipe : public BytePipe {
public:
	bool Push(char *Str, int Len=-1);
	char *NewStr();
};

#endif
