/*
**	FILE:			DragAndDrop.h
**	AUTHOR:			Matthew Allen
**	DATE:			27/3/99
**	DESCRIPTION:	Drag and drop classes
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef __DRAG_AND_DROP
#define __DRAG_AND_DROP

#if defined WIN32
#include <shlobj.h>
#define CF_GDragDropSource					"LGI::GDragDropSource"
#elif defined BEOS
#define S_OK								0
#endif

//////////////////////////////////////////////////////////////////////////////
#ifndef WIN32
#define DROPEFFECT_NONE						0
#define DROPEFFECT_COPY						1
#define DROPEFFECT_MOVE						2
#define DROPEFFECT_LINK						3
#endif

class LgiClass GDragDropSource
#if defined WIN32
	: public IDataObject, public IDropSource, public IEnumFORMATETC
#endif
{
protected:
	#if defined WIN32
	int Index;
	UINT DdFormat;
	int CurrentFormat;

	// IUnknown
	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ppv);
	
	// IEnumFORMATETC
	HRESULT STDMETHODCALLTYPE Next(ULONG celt, FORMATETC *rgelt, ULONG *pceltFetched);
	HRESULT STDMETHODCALLTYPE Skip(ULONG celt);
	HRESULT STDMETHODCALLTYPE Reset();
	HRESULT STDMETHODCALLTYPE Clone(IEnumFORMATETC **ppenum);
	HRESULT STDMETHODCALLTYPE GetCanonicalFormatEtc(FORMATETC * pFormatetcIn, FORMATETC * pFormatetcOut);
	HRESULT STDMETHODCALLTYPE SetData(FORMATETC *pFormatetc, STGMEDIUM *pmedium, BOOL fRelease);
	HRESULT STDMETHODCALLTYPE DAdvise(FORMATETC *pFormatetc, DWORD advf, IAdviseSink *pAdvSink, DWORD *pdwConnection);
	HRESULT STDMETHODCALLTYPE DUnadvise(DWORD dwConnection);
	HRESULT STDMETHODCALLTYPE EnumDAdvise(IEnumSTATDATA **ppenumAdvise);

	// IDropSource
	HRESULT STDMETHODCALLTYPE QueryContinueDrag(BOOL fEscapePressed, DWORD InputState);
	HRESULT STDMETHODCALLTYPE GiveFeedback(DWORD dwEffect);

	// IDataObject
	HRESULT STDMETHODCALLTYPE GetData(FORMATETC *pFormatEtc, STGMEDIUM *PMedium);
	HRESULT STDMETHODCALLTYPE GetDataHere(FORMATETC *pFormatEtc, STGMEDIUM *PMedium);
	HRESULT STDMETHODCALLTYPE QueryGetData(FORMATETC *pFormatEtc);
	HRESULT STDMETHODCALLTYPE EnumFormatEtc(DWORD dwDirection, IEnumFORMATETC **ppFormatEtc);
	#endif

public:
	GDragDropSource();
	~GDragDropSource();

	// Start a drag operation
	int Drag(GView *SourceWnd, int Effect);

	// Override these
	virtual void OnDragInit(bool Success) {}
	virtual char *TypeOf() { return NULL; }							// string identifing the type of data
	virtual bool GetData(uchar **Data, int *Size) { return false; }	// get the data
	virtual bool GetFormats(List<int> &Formats);
};

class LgiClass GDragDropTarget
#if defined WIN32
	: public IDropTarget
#endif
{
private:
	GView *To;
	GDragDropSource *DragDropObj;
	uchar *DragDropData;
	int DragDropLength;
	List<int> Formats;

	#if defined WIN32
	// IUnknown
	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ppv);

	// IDropTarget
	HRESULT STDMETHODCALLTYPE DragEnter(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
	HRESULT STDMETHODCALLTYPE DragOver(DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
	HRESULT STDMETHODCALLTYPE DragLeave(void);
	HRESULT STDMETHODCALLTYPE Drop(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);

	// Misc
	HGLOBAL GetDataHandle(IDataObject *pDataObject);
	bool UnwrapData(HGLOBAL hMem);
	#elif defined BEOS
	#endif

public:
	GDragDropTarget();
	~GDragDropTarget();

	// call this when you have a HWND
	void SetWindow(GView *To);

	// Override these

	// Events
	virtual void OnDragInit(bool Success) {}
	virtual void OnDragEnter() {}
	virtual void OnDragExit() {}

	// Lgi apps D'n'd
	virtual int WillAccept(GDragDropSource *Source, GdcPt2 Pt, int KeyState) { return DROPEFFECT_NONE; }
	virtual int OnDrop(uchar *Data, int Length, GdcPt2 Pt, int KeyState) { return DROPEFFECT_NONE; }

	// OS D'n'd
	virtual int WillAccept(List<int> &Formats, GdcPt2 Pt, int KeyState) { return DROPEFFECT_NONE; }
};

#endif