/*hdr
**      FILE:           Filter2.h
**      AUTHOR:         Matthew Allen
**      DATE:           1/4/97
**      DESCRIPTION:    Graphics file filters
**
**      Copyright (C) 1997, Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef FILTER2_H
#define FILTER2_H

#include <string.h>
#include "Contain.h"
#include "Prop.h"

LgiFunc int FindHeader(int Offset, char *Str, VFile &f);

#define FILTER_CAP_READ			0x0001
#define FILTER_CAP_WRITE		0x0002
#define FILTER_CAP_INFO			0x0004

class LgiClass GdcFileFormat : public VFile
{
protected:
	GdcSurface *GetSurface(GdcDCBase *pDC) { return pDC->pMem; }
	GdcRegion *GetClip(GdcDCBase *pDC) { return &pDC->Clip; }
	int GetLineLength(GdcDCBase *pDC) { return (pDC->pMem) ? pDC->pMem->Line : 0; }

	Progress *Meter;

public:
	ObjProperties Props;

	GdcFileFormat() { Meter = 0; }
	virtual ~GdcFileFormat() {}

	// properties
	Progress *GetProgress() { return Meter; }
	void SetProgress(Progress *Prg) { Meter = Prg; }

	// image read write
	virtual int GetCapabilites() { return 0; }
	virtual int GetImages() { return 1; }
	virtual bool ReadImage(GdcBasePrimitives *pDC) { return FALSE; }
	virtual bool WriteImage(GdcBasePrimitives *pDC) { return FALSE; }
};

#define GDC_RLE_COLOUR				0x0001
#define GDC_RLE_MONO				0x0002
#define GDC_RLE_READONLY			0x0004
#define GDC_RLE_KEY					0x0008		// set if valid

class LgiClass GdcRleDC : public GdcMemDC {
protected:
	COLOUR	Key;

	int	Flags;
	int	Length;
	int	Alloc;
	uchar	*Data;
	uchar	**ScanLine;

	bool SetLength(int Len);
	bool FindScanLines();
	void Empty();

public:
	GdcRleDC();
	virtual ~GdcRleDC();

	void Move(GdcRleDC *pDC);

	bool Create(int x, int y, int Bits, int LineLen = 0);
	bool CreateInfo(int x, int y, int Bits);
	void ReadOnly(bool Read);
	bool ReadOnly();
	void Mono(bool Mono);
	bool Mono();

	void Update(int Flags);
	void Draw(GdcBasePrimitives *Dest, int x, int y);

	int SizeOf();
	bool Read(VFile &F);
	bool Write(VFile &F);
};


////////////////////////////////////////////////////////////////
class LgiClass GdcFilterFactory
{
	static class GdcFilterFactory *First;
	class GdcFilterFactory *Next;

	// Override these in your filter library
	virtual bool CheckFile(char *File) { return false; }
	virtual GdcFileFormat *NewObject() { return NULL; }

public:
	GdcFilterFactory();
	virtual ~GdcFilterFactory();

	static GdcFileFormat *New(char *File);
	static GdcFileFormat *NewAt(int i);
	static int GetItems();
};

#endif
