
#ifndef __GLIST_H
#define __GLIST_H

//////////////////////////////////////////////////////////////////////////////
#define GIC_OWNER_DRAW				0x0001
#define GIC_ASK_TEXT				0x0002
#define GIC_ASK_IMAGE				0x0004
#define GIC_OWN_LIST				0x0008

class LgiClass GItemContainer {
protected:
	int Flags;
	GImageList *ImageList;

public:
	GItemContainer();
	~GItemContainer();

	bool OwnerDraw() { return TestFlag(Flags, GIC_OWNER_DRAW); }
	void OwnerDraw(bool b) { if (b) SetFlag(Flags, GIC_OWNER_DRAW); else ClearFlag(Flags, GIC_OWNER_DRAW); }
	bool AskText() { return TestFlag(Flags, GIC_ASK_TEXT); }
	void AskText(bool b) { if (b) SetFlag(Flags, GIC_ASK_TEXT); else ClearFlag(Flags, GIC_ASK_TEXT); }
	bool AskImage() { return TestFlag(Flags, GIC_ASK_IMAGE); }
	void AskImage(bool b) { if (b) SetFlag(Flags, GIC_ASK_IMAGE); else ClearFlag(Flags, GIC_ASK_IMAGE); }
	bool OwnList() { return TestFlag(Flags, GIC_OWN_LIST); }
	void OwnList(bool b) { if (b) SetFlag(Flags, GIC_OWN_LIST); else ClearFlag(Flags, GIC_OWN_LIST); }
	GImageList *GetList() { return ImageList; }

	bool SetImageList(GImageList *List, bool Own = TRUE);
	bool Load(char *File);
};

class LgiClass GMeasureInfo
{
public:
	int x;
	int y;
};

class LgiClass GItem : public GEvent
{
public:
	// Events
	virtual void OnSelect() {}
	virtual void OnMouseClick(VMouse &m) {}
	virtual void OnPaint(GdcBasePrimitives *pDC, GdcRegion &r, COLOUR Back) {}
	virtual bool OnBeginDrag(VMouse &m) { return false; }
	virtual void OnMeasure(GMeasureInfo *Info) {}

	// Methods
	virtual void Update() {}
	virtual void ScrollTo() {}

	// Data
	virtual bool Selected() { return FALSE; }
	virtual void Selected(bool b) {}
	virtual char *GetText(int i=0) { return 0; }
	virtual bool SetText(char *s, int i=0) { return FALSE; }
	virtual int GetImage(int Flags = 0) { return -1; }
	virtual void SetImage(int i) {}
};

// List view
class LgiClass GListColumn
#ifdef LGI_RES
: public ResObject
#endif
{
	friend class GList;
	friend class GListItem;

	GdcRegion Pos;
	bool Down;

protected:
	GList *Parent;
	char *cName;
	int cWidth;
	int cType;
	GdcBasePrimitives *cIcon;
	int cImage;
	int cMark;
	bool OwnIcon;
	bool CanResize;

public:
	GListColumn(GList *parent, char *name, int width);
	virtual ~GListColumn();

	// properties
	void Name(char *n);
	char *Name();
	void Width(int i);
	int Width();
	void Type(int i);
	int Type();
	void Mark(int i);
	int Mark();
	void Icon(GdcBasePrimitives *i, bool Own = true);
	GdcBasePrimitives *Icon();
	
	void Image(int i) { cImage = i; }
	int Image() { return cImage; }
	bool Resizable() { return CanResize; }
	void Resizable(bool i) { CanResize = i; }

	// methods
	void OnPaint(GdcBasePrimitives *pDC, GdcRegion &r);
};

#define GLI_MARK_NONE				0
#define GLI_MARK_UP_ARROW			1
#define GLI_MARK_DOWN_ARROW			2

class LgiClass GListItem : public GItem
{
	friend class GList;

protected:
	GdcRegion Pos;
	GList *Parent;
	bool Selected;
	List<char> Data;
	int ListItem_Image;

	bool GridLines();

public:
	GListItem();
	virtual ~GListItem();

	// Inline
	GList *GetList() { return Parent; }

	// Properties
	bool SetText(char *s, int i=0);
	char *GetText(int i);
	int GetImage(int Flags = 0) { return ListItem_Image; }
	void SetImage(int i) { ListItem_Image = i; }
	bool IsSelected();
	void Select(bool b);

	// Methods
	void Update();
	void ScrollTo();

	// Events;
	void OnMeasure(GMeasureInfo *Info);
	void OnPaint(GdcBasePrimitives *pDC, GdcRegion &r, COLOUR Back);
	virtual void OnPaintColumn(GdcBasePrimitives *pDC, GdcRegion &r, int i, GListColumn *c);
};

typedef int (*GListCompareFunc)(GListItem *a, GListItem *b, int Data);

class LgiClass GList : public GLayout, public GItemContainer
#ifdef LGI_RES
, public ResObject
#endif
#ifndef BEOS
, public GSemaphore
#endif
{
	friend class GListItem;
	friend class GListColumn;

	#ifdef WIN32
	static GWin32Class WndClass;
	char *Class() { return LGI_LIST; }
	uint32 Style() { return	WS_CHILD |
							WS_TABSTOP |
							WS_VISIBLE |
							GView::Style(); }
	uint32 ExStyle() { return (Sunken()) ? WS_EX_CLIENTEDGE : 0; }
	#endif

	COLOUR _Selection;
	
protected:
	// Contents
	List<GListColumn> Columns;
	List<GListItem> Items;
	int Keyboard; // index of the item with keyboard focus

	// Flags
	bool ColumnHeaders;
	bool EditLabels;
	bool GridLines;
	bool MultiItemSelect;

	// Double buffered
	GdcBasePrimitives *Buf;

	// Drawing locations
	GdcRegion ItemsPos;
	GdcRegion ColumnHeader;
	GdcRegion ScrollX, ScrollY;
	int FirstVisible;
	int LastVisible;
	int DragMode;
	int DragData;
	GdcPt2 DragStart;
	GListColumn *IconCol;

	// Misc
	bool GetItemRect(GListItem *i, GdcRegion &r);
	GListItem *HitItem(int x, int y, int *Index = 0);
	GdcRegion &GetClientRect();
	void Pour();
	void UpdateScrollBars();
	void KeyScroll(int iTo, int iFrom, bool SelectItems);

public:
	GList(int id, int x, int y, int cx, int cy, char *name = "List");
	~GList();

	// Overridables
	virtual void OnItemClick(GListItem *Item, VMouse &m);
	virtual void OnItemBeginDrag(GListItem *Item, VMouse &m);
	virtual void OnItemSelect(GListItem *Item);
	virtual void OnColumnClick(int Col) {}
	
	// Events
	void OnPaint(GdcBasePrimitives *pDC);
	int OnEvent(GMessage *Msg);
	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
	int OnNotify(GView *Ctrl, int Flags);
	void OnPosChange();
	void OnKey(VKey &k);
	void OnMouseWheel(double Lines);
	void OnFocus(bool b);
	void OnPulse();

	// Columns
	GListColumn *AddColumn(char *Name, int Width = 50, int Where = -1);
	bool AddColumn(GListColumn *Col, int Where = -1);
	bool DeleteColumn(GListColumn *Col);
	void EmptyColumns();
	GListColumn *ColumnAt(int Index) { return Columns.ItemAt(Index); }
	int GetColumns() { return Columns.GetItems(); }

	// Properties
	bool ShowColumnHeader() { return ColumnHeaders; }
	void ShowColumnHeader(bool Show) { ColumnHeaders = Show; }
	bool AllowEditLabels() { return EditLabels; }
	void AllowEditLabels(bool b) { EditLabels = b; }
	bool DrawGridLines() { return GridLines; }
	void DrawGridLines(bool b) { GridLines = b; }
	bool MultiSelect() { return MultiItemSelect; }
	void MultiSelect(bool b) { MultiItemSelect = b; }

	// Methods
	int Value();
	void Value(int i);
	bool Select(GListItem *Obj);
	GListItem *GetSelection();
	bool GetSelection(List<GListItem> &l);
	void SelectAll();
	bool GetAll(List<GListItem> &l);
	void ScrollToSelection();

	// List
	GListItem *First() { return Items.First(); }
	GListItem *Last() { return Items.Last(); }
	GListItem *Next() { return Items.Next(); }
	GListItem *Prev() { return Items.Prev(); }
	GListItem *Current() { return Items.Current(); }
	int GetItems() { return Items.GetItems(); }

	bool IsEmpty() { return Items.GetItems() == 0; }
	bool Delete();
	bool Delete(int Index);
	bool Delete(GListItem *p);
	bool Insert(GListItem *p, int Index = -1);
	bool Remove(GListItem *Obj);
	bool HasItem(GListItem *Obj);
	int IndexOf(GListItem *Obj);
	GListItem *ItemAt(int Index);
	void Sort(GListCompareFunc Compare, int Data);

	void Empty();			// remove all items from list and delete objects
	void RemoveAll();		// remove all references to externally owned items

	// Locking
	bool LockLooper();
	void UnlockLooper();
};

#endif
