#ifndef __GMENU_H
#define __GMENU_H

///////////////////////////////////////////////////////////////////////////////////////////////
// Menu wrappers
class LgiClass GMenuLoader
{
	friend class GMenuItem;
	friend class GMenu;
	friend class GSubMenu;

protected:
#ifdef WIN32
	HMENU			hMenu;
#endif
	List<GMenuItem> Items;
	
public:
	#ifdef LGI_RES
	bool Load(List<XmlTag> &Tags);
	#endif
	
	virtual GMenuItem *AppendItem(char *Str, int Id, bool Enabled, int Where = -1) = 0;
	virtual GSubMenu *AppendSub(char *Str, int Where = -1) = 0;
	virtual GMenuItem *AppendSeparator(int Where = -1) = 0;
};

class LgiClass GSubMenu :
	public GObject,
	public GTarget,
	public GFlags,
	public GMenuLoader,
	public GItemContainer
{
	friend class GMenuItem;
	friend class GMenu;

#if defined WIN32

	HWND TrackHandle;
	#ifndef LGI_RES
	bool Load(HMENU hmenu);
	#endif

#elif defined BEOS

	BMenu *Info;
	void _CopyMenu(BMenu *To, GSubMenu *From);

#endif

protected:
	GMenuItem		*Parent;
	GMenu			*Menu;
	GView			*Window;

public:
	#ifdef WIN32
	GSubMenu(char *name = "", bool Popup = true);
	#else
	GSubMenu(char *name = (char*)"");
	#endif
	virtual ~GSubMenu();

	#ifdef WIN32
	HMENU Handle() { return hMenu; }
	#endif

	GMenuItem *AppendItem(char *Str, int Id, bool Enabled, int Where = -1);
	GSubMenu *AppendSub(char *Str, int Where = -1);
	GMenuItem *AppendSeparator(int Where = -1);
	void Empty();
	bool RemoveItem(int i);
	bool RemoveItem(GMenuItem *Item);
	GMenuItem *ItemAt(int i);
	GMenuItem *FindItem(int Id);

	int Float(GView *Parent, int x, int y, bool Left = false);
	GMenuItem *GetParent() { return Parent; }
	GMenu *GetMenu() { return Menu; }
};

class LgiClass GMenuItem :
	public GObject,
	public GTarget,
	public GFlags
{
	friend class GSubMenu;
	friend class GMenu;
	friend class GView;
	friend class LgiMenuItem;

private:
	#ifdef WIN32
	bool			Insert(int Pos);
	bool			Update();
	#endif

protected:
	GMenu			*Menu;
	GSubMenu		*Parent;
	GSubMenu		*Child;
	int				Position;
	int				_Icon;

	#if defined WIN32

	MENUITEMINFO	Info;

	#elif defined BEOS

	BMessage		*Msg;
	BMenuItem		*Info;

	#endif

	virtual void _Measure(GdcPt2 &Size);
	virtual void _Paint(GdcBasePrimitives *pDC, int Flags);
	virtual void _PaintText(GdcBasePrimitives *pDC, int x, int y, int Width);

public:
	GMenuItem();
	#ifdef BEOS
	GMenuItem(BMenuItem *item);
	GMenuItem(GSubMenu *p);
	#else if WIN32
	GMenuItem(GMenu *m, GSubMenu *p, int Pos);
	#endif
	virtual ~GMenuItem();

	GSubMenu *Create();
	bool Remove();
	GSubMenu *GetParent();
	#ifdef WIN32
	bool ScanForAccel();
	#endif

	// Set
	void Id(int i);
	void Separator(bool s);
	void Checked(bool c);
	bool Name(char *n);
	void Enabled(bool e);
	void Visible(bool v);
	void Focus(bool f);
	void Sub(GSubMenu *s);
	void Icon(int i);

	// Get
	int Id();
	char *Name();
	bool Separator();
	bool Checked();
	bool Enabled();
	bool Visible();
	bool Focus();
	GSubMenu *Sub();
	int Icon();
};

class LgiClass GAccelerator
{
	int Flags;
	int Key;
	int Id;

public:
	GAccelerator(int flags, int key, int id);
	
	int GetId() { return Id; }
	bool Match(int a, int b);
};

class LgiClass GMenu :
	public GSubMenu
{
	friend class GSubMenu;
	friend class GMenuItem;
	friend class GWindow;

	static GFont *_Font;

	#ifdef BEOS
	BMenuBar *Bar;
	#endif

protected:
	List<GAccelerator> Accel;

public:
	GMenu();
	virtual ~GMenu();

	GFont *GetFont();

	GView *GetWindow() { return Window; }
	bool Attach(GView *p);
	bool Detach();
	bool Load(GView *p, char *Res);

	#ifdef WIN32
	int OnEvent(GMessage *Msg);
	static int _OnEvent(GMessage *Msg);
	#endif
};

#endif