#ifndef __GTREE_H
#define __GTREE_H

#include "GList.h"

// Tree view
#if defined BEOS
extern int GTreeCompareFunc(const BListItem *a, const BListItem *b);
#endif

class LgiClass GTreeItem :
#if defined BEOS
	public BListItem,
#endif
	public GItem
{
	friend class GTree;
	#if defined BEOS
	friend int GTreeCompareFunc(const BListItem *a, const BListItem *b);
	#else if WIN32
	friend int GView::OnEvent(GMessage *Msg);
	#endif

protected:
	char *Str;
	int Sys_Image;
	GTree *Tree;
	GTreeItem *Parent;
	List<GTreeItem> Items;

	#if defined BEOS

	int Index;

	void DrawItem(BView *owner, BRect frame, bool complete);

	#elif defined WIN32

	TV_ITEM Item;
	HTREEITEM hTreeItem;

	bool Attach(GTreeItem *After);
	bool AttachAll();

	#endif

public:
	GTreeItem();
	virtual ~GTreeItem();

	char *GetText(int i=0);
	bool SetText(char *s, int i=0);
	int GetImage(int Flags = 0);
	void SetImage(int i);
	void Update();
	#if defined WIN32
	HTREEITEM ItemHandle() { return hTreeItem; }
	#endif

	// List
	GTreeItem *Insert(GTreeItem *Obj = NULL, int Pos = -1);
	void Remove();

	// States
	bool Selected();
	void Selected(bool b);
	bool Expanded();
	void Expanded(bool b);

	// State-Events
	virtual void OnExpand(bool b) {}

	// Relitives
	GTreeItem *GetChild();
	GTreeItem *GetParent() { return Parent; }
	GTreeItem *GetNext();
	int IndexOf();
};

class LgiClass GTree :
#if defined LGI_RES
	public ResObject,
#endif
#if defined BEOS
	public BOutlineListView,
#else if WIN32
	public GTreeItem,
#endif
	public GControl,
	public GItemContainer
{
	friend class GTreeItem;

	#if defined WIN32
	static GWin32Class WndClass;
	#elif defined BEOS
	void MouseDown(BPoint point);
	#else
	List<GTreeItem> Items;
	#endif

protected:
	#if defined BEOS
	BScrollView *ScrollView;
	bool Attach(GView *parent);
	void SelectionChanged();
	#elif defined WIN32
	char *Class() { return LGI_TREE; }
	uint32 Style() { return	WS_CHILD |
							WS_VISIBLE |
							WS_TABSTOP |
							((Lines) ? TVS_HASLINES : 0) |
							((Buttons) ? TVS_HASBUTTONS : 0) |
							((LinesAtRoot) ? TVS_LINESATROOT : 0) |
							((EditLabels) ? TVS_EDITLABELS : 0) |
							GView::Style(); }
	uint32 ExStyle() { return (Sunken()) ? WS_EX_CLIENTEDGE : 0; }
	#endif

	bool Lines;
	bool Buttons;
	bool LinesAtRoot;
	bool EditLabels;
	
public:
	GTree(int id, int x, int y, int cx, int cy, char *name);
	~GTree();

	// Events
	virtual void OnItemClick(GTreeItem *Item, VMouse &m);
	virtual void OnItemBeginDrag(GTreeItem *Item, int Flags);
	virtual void OnItemExpand(GTreeItem *Item, bool Expand);
	virtual void OnItemSelect(GTreeItem *Item);
	
	// Window
	void OnMouseClick(VMouse &m);
	int OnEvent(GMessage *Msg);

	// Adding/deleting
	GTreeItem *Insert(GTreeItem *Obj = 0, int Pos = -1);
	bool Delete(GTreeItem *Obj);
	bool Remove(GTreeItem *Obj);

	// Selection
	bool Select(GTreeItem *Obj);
	GTreeItem *Selection();

	// Container
	void Empty();		// delete all items
	void RemoveAll();	// remove reference to externally owned items

	// Impl
	#if defined BEOS
	bool SetPos(GdcRegion &p, bool Repaint = FALSE);
	bool Invalidate(GdcRegion *r = NULL, bool Repaint = FALSE) { return GView::Invalidate(r, Repaint); }
	bool GetMouse(VMouse &m, bool ScreenCoords = false) { return GView::GetMouse(m, ScreenCoords); }
	#else if WIN32
	int IndexOf(GTreeItem *Item);
	GTreeItem *ItemAt(int i);
	#endif
};

#endif