/*hdr
**	FILE:		Region.h
**	AUTHOR:		Matthew Allen
**	DATE:		22/8/97
**	DESCRIPTION:	GDC Region classes
**
**	Copyright (C) 1997, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef __GDCREGION_H
#define __GDCREGION_H

#if defined WIN32
#define CornerOffset 1
#else if BEOS
#define CornerOffset 0
#endif

#ifdef LINUX
class LgiClass BRect
{
public:
	int left, top, right, bottom;
};
#endif

class LgiClass GdcRegion {

	friend LgiClass bool operator ==(GdcRegion &a, GdcRegion &b);
	friend LgiClass bool operator !=(GdcRegion &a, GdcRegion &b);

public:
	int x1, y1, x2, y2;

	GdcRegion() {}
	GdcRegion(int X1, int Y1, int X2, int Y2)
	{
		x1 = X1;
		x2 = X2;
		y1 = Y1;
		y2 = Y2;
	}

	GdcRegion(GdcRegion *r)
	{
		x1 = r->x1;
		x2 = r->x2;
		y1 = r->y1;
		y2 = r->y2;
	}

	int X() { return x2 - x1 + 1; }
	int Y() { return y2 - y1 + 1; }
	void Set(int X1, int Y1, int X2, int Y2)
	{
		x1 = X1;
		x2 = X2;
		y1 = Y1;
		y2 = Y2;
	}

	void ZOff(int x, int y);
	void Normal();
	bool Valid();
	void Offset(int x, int y);
	void Offset(GdcRegion *a);
	void Size(int x, int y);
	void Size(GdcRegion *a);
	void Dimension(int x, int y);
	void Dimension(GdcRegion *a);
	void Bound(GdcRegion *b);
	bool Overlap(int x, int y);
	bool Overlap(GdcRegion *b);
	void Union(int x, int y);
	void Union(GdcRegion *a);
	void Union(GdcRegion *a, GdcRegion *b);
	void Intersection(GdcRegion *a);
	void Intersection(GdcRegion *a, GdcRegion *b);

	#if defined WIN32
	operator RECT()
	{
		RECT rc;
		rc.left = x1;
		rc.top = y1;
		rc.right = x2+CornerOffset;
		rc.bottom = y2+CornerOffset;
		return rc;
	}
	#else
	operator BRect()
	{
		BRect rc;
		rc.left = x1;
		rc.top = y1;
		rc.right = x2 + CornerOffset;
		rc.bottom = y2 + CornerOffset;
		return rc;
	}
	#endif

	#if defined WIN32
	GdcRegion(RECT r)
	{
		x1 = r.left;
		y1 = r.top;
		x2 = r.right - CornerOffset;
		y2 = r.bottom - CornerOffset;
	}
	#else
	GdcRegion(BRect r)
	{
		x1 = r.left;
		y1 = r.top;
		x2 = r.right - CornerOffset;
		y2 = r.bottom - CornerOffset;
	}
	#endif
};

LgiClass bool operator ==(GdcRegion &a, GdcRegion &b);
LgiClass bool operator !=(GdcRegion &a, GdcRegion &b);

class LgiClass GdcComplexRegion : public GdcRegion {

	int Size;
	int Alloc;
	int Current;
	GdcRegion *a;

	bool SetSize(int s);
	GdcRegion *NewOne() { return (SetSize(Size+1)) ? a+(Size-1) : 0; }
	bool Delete(int i);
	void Empty();

public:
	GdcComplexRegion();
	GdcComplexRegion(int X1, int Y1, int X2, int Y2);
	GdcComplexRegion(GdcRegion &r);
	#if defined WIN32
	GdcComplexRegion(RECT &r);
	#else
	GdcComplexRegion(BRect &r);
	#endif
	~GdcComplexRegion();

	int X() { return x2 - x1 + 1; }
	int Y() { return y2 - y1 + 1; }
	int GetSize() { return Size; }
	GdcRegion *operator [](int i) { return a+i; }
	GdcRegion *First();
	GdcRegion *Last();
	GdcRegion *Next();
	GdcRegion *Prev();

	void ZOff(int x, int y);
	void Normal();
	bool Valid();
	void Offset(int x, int y);
	void Bound(GdcRegion *b);
	bool Overlap(GdcRegion *b);
	
	// Set operations
	void Union(GdcRegion *a);
	void Intersect(GdcRegion *a);
	void Subtract(GdcRegion *a);

	friend bool operator ==(GdcComplexRegion &a, GdcComplexRegion &b);
	friend bool operator !=(GdcComplexRegion &a, GdcComplexRegion &b);
};

#endif