/*hdr
**      FILE:           Inet.h
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Internet sockets classes
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __INET_H
#define __INET_H

#include "VMem.h"
#include "Contain.h"

#if defined WIN32
#include <winsock.h>
#elif defined BEOS
#include <socket.h>
#include <netdb.h>
#else
typedef int SOCKET;
#endif

#define DEFAULT_TIMEOUT			30

// Parameters
#define PARAMETER_LOG			1
#define PARAMETER_PROGRESS		2

#if defined BEOS
#include "NetworkKit.h"
#define ValidSocket(s)			((s)>=0)
#define INVALID_SOCKET			0
typedef int SOCKET;
#elif defined WIN32
#define ValidSocket(s)			((s) != INVALID_SOCKET)
#endif

LgiFunc bool HaveNetConnection();
LgiFunc bool WhatsMyIp(char *IpAddr);
LgiFunc void MDStringToDigest(unsigned char digest[16], char *Str);

class GNetwork;
class IConnection;

class LgiClass GNetwork {

	bool SocketsOpen;

public:
	GNetwork();
	virtual ~GNetwork();

	bool IsValid() { return SocketsOpen; }

	IConnection *CreateConnection();
	bool EnumInterfaces(List<char> &Lst);
};

class LgiClass IConnection {

	friend class GNetwork;

protected:
	GNetwork *Parent;
	bool IsOK() { return (Parent) ? Parent->IsValid() : false; }

	bool		Async;
	#if defined WIN32
	SOCKET		Socket;
	#elif defined BEOS
	BNetEndpoint *Ep;
	#endif
	bool		Canceled;

public:
	int			BytesWritten;
	int			BytesRead;

	IConnection();
	virtual ~IConnection();
	SOCKET Handle();

	// Status
	bool IsReadable();
	#if defined WIN32
	bool IsOpen() { return Socket != INVALID_SOCKET; }
	#else
	bool IsOpen() { return Handle() >= 0; }
	#endif
	virtual bool Cancel() { return Canceled; }
	virtual void Cancel(bool i) { Canceled = i; }
	virtual bool IsAsync();
	virtual void IsAsync(bool i);

	// Client
	virtual bool Open(char *HostAddr, int port);
	virtual void Close();

	// Server
	virtual bool Listen(int Port = 0);
	virtual bool Accept(IConnection *c);

	// Data transfer
	virtual int Write(char *Data, int Len, int Flags);
	virtual int Read(char *Data, int Len, int Flags);
	virtual int Error();

	// Info
	bool WhatsMyIp(char *IpAddr);
	int WhatsMyPort();
	bool GetRemoteIp(char *IpAddr);

	// Events
	virtual void OnDisconnect();

	virtual void OnRead(char *Data, int Len) {}
	virtual void OnWrite(char *Data, int Len) {}
	virtual void OnError(int ErrorCode, char *ErrorDescription) {}
	virtual void OnInformation(char *Str) {}
	virtual void SetParameter(int Which, int What) {}
};

class LgiClass ISocks5Connection : public IConnection
{
	char *Proxy;
	int Port;
	char *UserName;
	char *Password;

protected:
	bool Socks5Connected;

public:
	ISocks5Connection(char *proxy, int port, char *username, char *password);
	~ISocks5Connection();

	// Connection
	bool Open(char *HostAddr, int port);

	// Server
	bool Listen(int Port) { return false; }
};

#endif
