/*hdr
**	FILE:			LgiClasses.h
**	AUTHOR:			Matthew Allen
**	DATE:			19/12/97
**	DESCRIPTION:	LGI Class definitions
**	PLATFROM:		Xp
**
**
**	Copyright (C) 1997-2000, Matthew Allen
**		fret@ozemail.com.au
*/

/////////////////////////////////////////////////////////////////////////////////////
// Includes
#ifndef __GUI_H
#define __GUI_H

#if defined BEOS
#include <string.h>
#endif

#include "GSemaphore.h"
#include "GLibrary.h"
#include "Xml.h"

/////////////////////////////////////////////////////////////////////////////////////
// Externs
extern long MouseWatcher(void *Ptr);

/////////////////////////////////////////////////////////////////////////////////////
// Classes
#if defined WIN32

	class GFont;

	class LgiClass GMessage {
	public:
		int Msg;
		int a;
		int b;

		GMessage()
		{
			Msg = a = b = 0;
		}

		GMessage(int M, int A = 0, int B = 0)
		{
			Msg = M;
			a = A;
			b = B;
		}
	};

#elif defined BEOS

	typedef BMessage GMessage;
	class GScrollView;

	class LgiClass BViewRedir : public BView
	{
		void AttachedToWindow();
		void DetachedFromWindow();
		void Draw(BRect UpdateRect);
		void FrameMoved(BPoint Point);
		void FrameResized(float width, float height);
		void Pulse();
		void MessageReceived(BMessage *message);
		void MakeFocus(bool f = true);
		void KeyDown(const char *bytes, int32 numBytes);
		void KeyUp(const char *bytes, int32 numBytes);
		void MouseDown(BPoint point);
		void MouseMoved(BPoint point, uint32 transit, const BMessage *message);
		bool QuitRequested();
	
		GView *Wnd;
		GScrollView *ScrollView;
		
	public:
		BViewRedir(GView *Wnd);

		GView *GetWindow() { return Wnd; }

		GScrollView *GetScrollView() { return ScrollView; }
		void SetScrollView(GScrollView *v) { ScrollView = v; }
	};

#endif

#ifdef BEOS

	// VKey flags
	#define LGI_VKEY_CTRL		B_CONTROL_KEY	// 0x0004
	#define LGI_VKEY_ALT		B_OPTION_KEY	// 0x0040
	#define LGI_VKEY_SHIFT		B_SHIFT_KEY		// 0x0001

	// VMouse flags
	#define LGI_VMOUSE_LEFT		B_PRIMARY_MOUSE_BUTTON		// 0x0001
	#define LGI_VMOUSE_MIDDLE	B_TERTIARY_MOUSE_BUTTON		// 0x0004
	#define LGI_VMOUSE_RIGHT	B_SECONDARY_MOUSE_BUTTON	// 0x0002

	#define LGI_VMOUSE_CTRL		0x08
	#define LGI_VMOUSE_ALT		0x10
	#define LGI_VMOUSE_SHIFT	0x20

	#define LGI_VMOUSE_DOWN		0x40
	#define LGI_VMOUSE_DOUBLE	0x80

#else

	// VKey flags
	#define LGI_VKEY_CTRL		MK_CTRL
	#define LGI_VKEY_ALT		MK_ALT
	#define LGI_VKEY_SHIFT		MK_SHIFT

	// VMouse flags
	#define LGI_VMOUSE_LEFT		MK_LBUTTON		// 0x0001
	#define LGI_VMOUSE_MIDDLE	MK_MBUTTON		// 0x0010
	#define LGI_VMOUSE_RIGHT	MK_RBUTTON		// 0x0002

	#define LGI_VMOUSE_SHIFT	MK_SHIFT		// 0x0004
	#define LGI_VMOUSE_CTRL		MK_CONTROL		// 0x0008
	#define LGI_VMOUSE_ALT		0x20

	#define LGI_VMOUSE_DOWN		0x40
	#define LGI_VMOUSE_DOUBLE	0x80

#endif

class LgiClass VKey {
public:
	uchar c;
	int Flags;
	int Data;
	bool Down;
	bool IsChar;

	// Keyboard modifiers
	bool Ctrl()		{ return TestFlag(Flags, LGI_VKEY_CTRL); }
	bool Alt()		{ return TestFlag(Flags, LGI_VKEY_ALT); }
	bool Shift()	{ return TestFlag(Flags, LGI_VKEY_SHIFT); }
};

class LgiClass VMouse {
public:
	int x;
	int y;
	int Flags;

	// Get
	bool Down()		{ return TestFlag(Flags, LGI_VMOUSE_DOWN); }
	bool Double()	{ return TestFlag(Flags, LGI_VMOUSE_DOUBLE); }
	bool Left()		{ return TestFlag(Flags, LGI_VMOUSE_LEFT); }
	bool Middle()	{ return TestFlag(Flags, LGI_VMOUSE_MIDDLE); }
	bool Right()	{ return TestFlag(Flags, LGI_VMOUSE_RIGHT); }
	bool Ctrl()		{ return TestFlag(Flags, LGI_VMOUSE_CTRL); }
	bool Alt()		{ return TestFlag(Flags, LGI_VMOUSE_ALT); }
	bool Shift()	{ return TestFlag(Flags, LGI_VMOUSE_SHIFT); }

	// Set
	#define AssignFlag(f, bit, to) ( (to) ) ? f |= bit : f &= ~(bit)
	void Down(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_DOWN, i); }
	void Double(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_DOUBLE, i); }
	void Left(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_LEFT, i); }
	void Middle(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_MIDDLE, i); }
	void Right(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_RIGHT, i); }
	void Ctrl(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_CTRL, i); }
	void Alt(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_ALT, i); }
	void Shift(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_SHIFT, i); }
};

class LgiClass GTarget {
public:
	virtual int OnEvent(GMessage *Msg) { return 0; }
};

class LgiClass GEvent {
public:
	// Mouse related
	virtual void OnMouseClick(VMouse &m) {}
	virtual void OnMouseEnter(VMouse &m) {}
	virtual void OnMouseExit(VMouse &m) {}
	virtual void OnMouseMove(VMouse &m) {}

	// Keyboard related
	virtual void OnKey(VKey &k) {}

	// Other
	virtual void OnPaint(GdcBasePrimitives *pDC) {}
	virtual void OnPosChange() {}
};

class LgiClass GFlags
{
protected:
	int WndFlags;

public:
	GFlags();

	// State
	virtual bool Enabled();
	virtual void Enabled(bool i);
	virtual bool Visible();
	virtual void Visible(bool i);
	virtual bool Focus();
	virtual void Focus(bool i);
	virtual bool DropTarget();
	virtual void DropTarget(bool t);

	virtual void OnUpdateState() {}

	// Border
	virtual bool Sunken();
	virtual void Sunken(bool i);
	virtual bool Flat();
	virtual void Flat(bool i);
	virtual bool Raised();
	virtual void Raised(bool i);

	virtual void OnUpdateBorder() {}
};

#ifdef BEOS

// redirector for the BWindow object back to the GView owner
class LgiClass GWnd : public BWindow {

	GView *Notify;

public:
	GWnd(GView *notify);
	GWnd(GView *notify, BRect frame, char *title, window_type type, uint32 flags, uint32 workspaces = B_CURRENT_WORKSPACE);
	bool QuitRequested();
	void MessageReceived(BMessage *Msg);
	void FrameMoved(BPoint origin);
	void FrameResized(float width, float height);
};

#elif defined WIN32

// Win32 window class
class LgiClass GWin32Class : public GObject {

	static LRESULT CALLBACK Redir(GWndRef hWnd, UINT m, WPARAM a, LPARAM b);
	static LRESULT CALLBACK SubClassRedir(GWndRef hWnd, UINT m, WPARAM a, LPARAM b);
	WNDPROC ParentProc;

public:
	WNDCLASS Class;

	GWin32Class();
	~GWin32Class();

	bool Create(char *ClassName);
	bool SubClass(char *ClassName, char *Parent);
	LRESULT CALLBACK CallParent(GWndRef hWnd, UINT m, WPARAM a, LPARAM b);
};

#endif

/////////////////////////////////////////////////////////////////////////////////
#define LgiApp						(GApp::ObjInstance())
#define SysFont						(LgiApp->SystemFont)
#define hAppInst()					(LgiApp?LgiApp->Instance():0)
#define LgiExitApp()				LgiApp->Exit(true)
#define LgiCloseApp()				LgiApp->Exit(false)

class LgiClass GApp
#ifdef BEOS
	: public BApplication
#endif
{
	friend class GView;

protected:
	// private member vars
	static GApp *TheApp;

	char *CmdLine;
	Xml *Config;

	#ifdef WIN32
	HINSTANCE hAppInst;
	HINSTANCE hPrevInst;
	int CmdShow;
	HCURSOR hNormalCursor;
	#endif

public:
	// public member vars
	#ifdef WIN32
	static GWndRef ActiveDialog;
	#endif
	GFont *SystemFont;
	GWindow *AppWnd;

	// public methods
	#ifdef WIN32

	GApp(HINSTANCE hInst, char *lpCmdLine, int nShow);

	HINSTANCE Instance() { return hAppInst; }
	HINSTANCE PrevInst() { return hPrevInst; }
	void SetPrevInst(HINSTANCE Prev) { hPrevInst = Prev; }
	int Show() { return CmdShow; }
	virtual bool Run(bool Loop = true);

	#elif defined BEOS

	GApp(char *MimeSig, char *lpCmdLine);
	void RefsReceived(BMessage *Msg);

	#else

	GApp(int Args, char **Arg);
	virtual bool Run(bool Loop = true);

	#endif

	virtual ~GApp();

	// The one and only instance
	static GApp *ObjInstance() { return TheApp; }

	// Options etc
	bool GetOption(char *Dst, char Option);
	XmlTag *GetConfig(char *Tag);

	// Override
	virtual void OnCmdLine(char *Cmd);
	virtual void OnReceiveFiles(List<char> &Files);
	virtual void Exit(int Code = 0);
};

class LgiClass GView :
	public GObject,
	public GTarget,
	public GEvent,
	public GFlags
{
	friend		class GWindow;
	friend		class GLayout;
	friend		class GDialog;
	friend		class GControl;
	friend		class GMenu;
	friend		class GSubMenu;
	friend		class GWnd;
	friend		class GScrollView;
	friend		class BViewRedir;

	friend		bool SysOnKey(GView *w, GMessage *m);
	friend		GView *GetNextTabStop(GView *Wnd, bool Back);

	#ifdef WIN32
	friend		LRESULT CALLBACK GWin32Class::Redir(GWndRef hWnd, UINT m, WPARAM a, LPARAM b);
	friend		LRESULT CALLBACK GWin32Class::SubClassRedir(GWndRef hWnd, UINT m, WPARAM a, LPARAM b);
	friend		LRESULT CALLBACK DlgRedir(GWndRef hWnd, UINT m, WPARAM a, LPARAM b);
	static		void CALLBACK TimerProc(GWndRef hwnd, UINT uMsg, UINT idEvent, uint32 dwTime);
	#else if BEOS
	friend		class GButtonRedir;
	friend		long _lgi_pulse_thread(void *ptr);
	#endif

protected:
	#if defined WIN32

	// Tracking handles
	static GView*Capturing;
	static GView*Over;
	static GWndRef	hPrevCapture;

	// Window data
	GWndRef			hWnd;					// Win32 HWND
	int				WndStyle;				// Win32 hWnd Style
	int				WndExStyle;				// Win32 hWnd ExStyle
	int				WndDlgCode;				// Win32 Dialog Code (WM_GETDLGCODE)
	char			*WndClass;
	UINT			_Win32_TimerId;

	// Methods
	virtual uint32	Style();
	virtual uint32	ExStyle();
	virtual char	*Class();
	virtual GView *FindReal(GdcPt2 *Offset = 0);
	virtual bool	HandleCapture(GView *Wnd, bool c);

	// Stubs
	virtual int		SysOnNotify(int Code) { return 0; }

	// Create a Win32 window
	virtual bool	Create(char *Class, GWndRef Parent);

	#elif defined BEOS

	BView			*Sys_View;				// BeOS view class
	thread_id		_CaptureThread;
	thread_id		_PulseThread;
	int				_PulseRate;

	void Sys_KeyDown(const char *bytes, int32 numBytes);
	void Sys_KeyUp(const char *bytes, int32 numBytes);
	void Sys_MouseDown(BPoint point);
	void Sys_MouseMoved(BPoint point, uint32 transit, const BMessage *message);
	void Sys_FrameMoved(BPoint p);
	virtual bool QuitRequested() {}

	#endif

	// General
	int				CtrlId;
	GdcRegion		Pos;
	GView		*Parent;
	GView		*Notify;
	GView		*WantPulses;
	GdcPt2			_MinimumSize;
	int				_BorderSize;

	// Impl
	virtual void _Paint(GdcBasePrimitives *pDC);

	// Complex Region searches
	GdcRegion *FindLargest(GdcComplexRegion &r);
	GdcRegion *FindSmallestFit(GdcComplexRegion &r, int Sx, int Sy);
	GdcRegion *FindLargestEdge(GdcComplexRegion &r, int Edge);

public:
	List<GView>	Children;

	#ifdef BEOS
	GView(BView *wnd);
	#else
	GView();
	#endif
	virtual ~GView();
	virtual void Quit(bool DontDelete = false);

	#ifdef BEOS
	
	// BeOS API
	virtual BWindow *GetBWindow() { return (Sys_View) ? Sys_View->Window() : 0; }
	virtual BView *GetBView() { return Sys_View; }
	virtual GWndRef Handle() { return Sys_View; }

	#elif defined WIN32

	// Win32 API
	virtual GWndRef Handle() { return hWnd; }
	operator GWndRef() { return hWnd; }

	#elif defined LINUX

	virtual GWndRef Handle() { return 0; } // FIXME: this should return an XWindow ref

	#endif

	// Attaching windows / heirarchy
	virtual bool Attach(GView *p);
	virtual bool AttachChildren();
	virtual bool Detach();
	virtual bool IsAttached();

	virtual GView *GetParent() { return Parent; }
	virtual void SetParent(GView *p) { Parent = p; }
	virtual GView *GetNotify() { return Notify; }		// the "notify" window gets told
	virtual void SetNotify(GView *n) { Notify = n; }	// via the OnNotify event handler
														// when this window has changed
	// Target
	int OnEvent(GMessage *Msg);							// generic message handler

	// Flags
	bool Enabled();
	void Enabled(bool e);
	bool Visible();
	void Visible(bool v);
	bool Focus();
	void Focus(bool f);

	// Border
	bool Sunken();
	void Sunken(bool i);
	bool Flat();
	void Flat(bool i);
	bool Raised();
	void Raised(bool i);

	// Thread
	bool InThread(); // return true if the calling thread owns the window
	bool PostEvent(int Cmd, int a, int b);

	// Non-implemented events
	virtual void OnMouseClick(VMouse &m) {}
	virtual void OnMouseEnter(VMouse &m) {}
	virtual void OnMouseExit(VMouse &m) {}
	virtual void OnMouseMove(VMouse &m) {}
	virtual void OnKey(VKey &k) {}
	virtual void OnMouseWheel(double Lines) {}
	virtual void OnCreate() {}
	virtual void OnDestroy() {}
	virtual void OnFocus(bool f) {}
	virtual void OnPulse() {}
	virtual void OnPosChange();
	virtual bool OnRequestClose() { return true; }
	virtual int OnHitTest(int x, int y) { return -1; }
	virtual int OnNotify(GView *Ctrl, int Flags);
	virtual void OnChildrenChanged(GView *Wnd, bool Attaching) {}
	virtual int OnCommand(int Cmd, int Event, GWndRef Wnd) { return 0; }
	virtual void OnPaint(GdcBasePrimitives *pDC) {}

	// Name
	bool Name(char *n);
	char *Name();

	// Font
	virtual GFont *GetFont();
	virtual void SetFont(GFont *Fnt);

	// Position
	virtual GdcRegion &GetPos() { return Pos; }
	virtual bool SetPos(GdcRegion &p, bool Repaint = FALSE);
	virtual int X() { return Pos.X(); }
	virtual int Y() { return Pos.Y(); }
	virtual GdcRegion &GetClient();

	void MoveOnScreen();
	void MoveToCenter();
	void MoveToMouse();

	// Mouse & Mouse capture capture
	virtual bool Capture(bool c);		// passing true to this function "captures"
										// the mouse, so that you will receive mouse
										// move and click events outside your windows
										// area. passing false stops this.
	virtual bool IsCapturing();			// returns whether capture is on.
	virtual bool GetMouse(VMouse &m, bool ScreenCoords = false);

	// Control related
	virtual int GetId() { return CtrlId; }
	virtual void SetId(int i);
	virtual int Value() { return 0; }
	virtual void Value(int i) {}
	virtual GView *FindControl(GWndRef hCtrl);
	virtual GView *FindControl(int Id);

	int GetCtrlValue(int CtrlId);
	void SetCtrlValue(int CtrlId, int i);
	char *GetCtrlName(int CtrlId);
	void SetCtrlName(int CtrlId, char *s);
	bool GetCtrlEnabled(int CtrlId);
	void SetCtrlEnabled(int Ctrld, bool Enabled);

	// Other
	virtual bool Pour(GdcComplexRegion &r) { return false; }
	virtual bool Invalidate(GdcRegion *r = NULL, bool Repaint = FALSE);
	virtual bool InvalidateFrame(GdcRegion *r = NULL, bool Repaint = FALSE);
	virtual bool IsOver(VMouse &m);
	virtual GView *WindowFromPoint(int x, int y);
	virtual void SetPulse(int Length = -1);
	#ifdef WIN32
	virtual bool WindowVirtualOffset(GdcPt2 *Offset);
	#endif
};

///////////////////////////////////////////////////////////////////////////////////////////////
// Control or View window
#define GV_EDGE_TOP				0x0001
#define GV_EDGE_RIGHT			0x0002
#define GV_EDGE_BOTTOM			0x0004
#define GV_EDGE_LEFT			0x0008

#define IDC_VSCROLL				14000
#define IDC_HSCROLL				14001

class LgiClass GLayout : public GView
{
protected:
	#ifdef WIN32

	static GWin32Class	Class;

	#else if BEOS

	bool XScroll, YScroll;
	BViewRedir *Redir;

	#endif

	// Private variables
	bool			_SettingScrollBars;
	bool			_PourLargest;
	GScrollBar		*VScroll;
	GScrollBar		*HScroll;

	// Private members
	bool SetScrollBars(bool x, bool y);

public:
	GLayout();
	~GLayout();

	// Overridables
	virtual void GetScrollPos(int &x, int &y);
	virtual void SetScrollPos(int x, int y);

	// Properties
	bool GetPourLargest();
	void SetPourLargest(bool i);

	// Events
	int OnEvent(GMessage *Msg);
	bool Pour(GdcComplexRegion &r);

	// Impl
	#ifdef BEOS
	bool Attach(GView *p);
	virtual GdcRegion &GetPos();
	bool SetPos(GdcRegion &p, bool Repaint = false);
	void Visible(bool i);
	bool Visible();
	#endif
};

////////////////////////////////////////////////////////////////////////////
// Dialog stuff
#include "LgiWidgets.h"

////////////////////////////////////////////////////////////////////////////
// Progress meters stuff
#include "Progress.h"
#include "GProgress.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
// Menus
#include "GMenu.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GWindow :
	public GView
{
	friend class BViewRedir;

	bool _QuitOnClose;

protected:
	#ifdef BEOS

	BWindow *Wnd;
	bool QuitRequested();

	#elif defined WIN32

	GdcRegion OldPos;
	HWND LastFocus;

	#endif
	
	GMenu *Menu;

public:
	GWindow();
	~GWindow();

	// Methods
	virtual void Pour();

	// Properties
	GMenu *GetMenu() { return Menu; }
	void SetMenu(GMenu *m) { Menu = m; }
	bool Visible();
	void Visible(bool v);
	bool GetQuitOnClose() { return _QuitOnClose; }
	void SetQuitOnClose(bool i) { _QuitOnClose = i; }

	// Events
	void OnPosChange();
	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	virtual void OnReceiveFiles(List<char> &Files) {}

	// BeOS impl
	#ifdef BEOS
	bool Name(char *n);
	char *Name();
	GdcRegion &GetPos();
	bool SetPos(GdcRegion &p, bool Repaint = FALSE);
	void OnChildrenChanged(GView *Wnd, bool Attaching);
	void OnCreate();
	virtual BWindow *GetBWindow() { return Wnd; }
	#endif
};

////////////////////////////////////////////////////////////////////////
#include "GFileSelect.h"
#include "GFindReplaceDlg.h"
#include "GToolBar.h"

////////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GSplitter : public GLayout {

	bool		Vertical;
	bool		Moving;
	int			SplitPos;
	int			SplitSet;
	bool		OverBar;
	GdcRegion	Bar;
	int			Offset;
	bool		SplitFollow;

	GdcRegion	PosA;
	GdcRegion	PosB;
	GView	*ViewA;
	GView	*ViewB;
	bool		BorderA;
	bool		BorderB;

	#ifdef WIN32
	HCURSOR		OldCursor;
	HCURSOR		Cursor;
	#endif

	void		CalcRegions(GdcRegion Rect, bool Follow = false);
	bool		OverSplit(int x, int y);

public:
	GSplitter();
	~GSplitter();

	bool Border() { return TestFlag(WndFlags, GWF_BORDER); }
	void Border(bool i) { (i) ? SetFlag(WndFlags, GWF_BORDER) : ClearFlag(WndFlags, GWF_BORDER); }
	bool IsVertical() { return Vertical; }
	void IsVertical(bool v);
	bool DoesSplitFollow() { return SplitFollow; }
	void DoesSplitFollow(bool i) { SplitFollow = i; }
	int Split() { return SplitPos; }
	void Split(int s);
	GView *GetViewA() { return ViewA; }
	GView *GetViewB() { return ViewB; }
	void SetViewA(GView *a, bool Border = true);
	void SetViewB(GView *b, bool Border = true);
	GView *FindControl(GWndRef hCtrl);

	bool Attach(GView *p);

	void OnPaint(GdcBasePrimitives *pDC);
	void OnPosChange();
	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
	#ifdef WIN32
	void OnMouseExit(VMouse &m);
	int OnHitTest(int x, int y);
	#endif
};

////////////////////////////////////////////////////////////////////////////////////////////////
#define STATUSBAR_SEPARATOR			4
#define GSP_SUNKEN					0x0001

typedef GStatusPane *GStatusPanePtr;

class LgiClass GStatusBar : public GLayout, private PtrArray {

	friend class GStatusPane;

protected:
	GFont *Font;
	void RePour();

public:
	GStatusBar();
	~GStatusBar();

	bool Pour(GdcComplexRegion &r);
	void OnPaint(GdcBasePrimitives *pDC);

	GStatusPane *AppendPane(char *Text, int Width);
	bool AppendPane(GStatusPane *Pane);
};

class LgiClass GStatusPane :
	public GView
{
	friend class GStatusBar;

protected:
	int		Flags;
	int		Width;
	GdcBasePrimitives *pDC;

public:
	GStatusPane();
	~GStatusPane();

	char *Name() { return GObject::Name(); }
	bool Name(char *n);
	void OnPaint(GdcBasePrimitives *pDC);

	int GetWidth();
	void SetWidth(int x);
	bool Sunken();
	void Sunken(bool i);
	GdcBasePrimitives *Bitmap();
	void Bitmap(GdcBasePrimitives *pdc);
};

/////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GCommand : public GObject, public GFlags {

	int			Flags;
	bool		PrevValue;

public:
	int			Id;
	GToolButton	*ToolButton;
	GMenuItem	*MenuItem;
	VKey		*Accelerator;
	char		*TipHelp;
	
	GCommand();
	~GCommand();

	bool Enabled();
	void Enabled(bool e);
	bool Value();
	void Value(bool v);
};

/////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GClipBoard {

	GView *Owner;
	bool Open;

public:
	GClipBoard(GView *o);
	~GClipBoard();

	bool IsOpen() { return true; }
	bool Empty();

	// Text
	bool Text(char *Str);
	char *Text();

	bool TextW(ushort *Str);
	ushort *TextW();

	// Bitmap
	bool Bitmap(GdcBasePrimitives *pDC);
	GdcBasePrimitives *Bitmap();
	#ifdef WIN32
	GdcBasePrimitives *ConvertFromPtr(void *Ptr);
	#endif

	// Binary
	bool Binary(int Format, uchar *Ptr, int Len);	// Set
	bool Binary(int Format, uchar **Ptr, int *Len);	// Get
};

//////////////////////////////////////////////////////////////////////////
#if defined BEOS
	#define PRIORITY_IDLE			B_LOW_PRIORITY
	#define PRIORITY_LOW			B_LOW_PRIORITY
	#define PRIORITY_NORMAL			B_NORMAL_PRIORITY
	#define PRIORITY_HIGH			B_DISPLAY_PRIORITY
#elif defined WIN32
	#define PRIORITY_IDLE			THREAD_PRIORITY_IDLE
	#define PRIORITY_LOW			THREAD_PRIORITY_BELOW_NORMAL
#else
	#define PRIORITY_IDLE			1
	#define PRIORITY_LOW			2
#endif

#define LGITHREAD_INIT			1
#define LGITHREAD_RUNNING		2
#define LGITHREAD_ASLEEP		3
#define LGITHREAD_EXITED		4

#if defined WIN32
typedef HANDLE ThreadType;
#elif defined BEOS
typedef thread_id ThreadType;
#else
typedef int ThreadType;
#endif

class LgiClass GThread
{
	int State; // LGITHREAD_???
	int ReturnValue;
	ThreadType hThread;

	#if defined WIN32
	friend uint WINAPI ThreadEntryPoint(int i);
	uint ThreadId;
	#else
	friend int32 ThreadEntryPoint(void *i);
	#endif

protected:
	bool DeleteOnExit;

public:
	GThread();
	virtual ~GThread();

	// Properties
	ThreadType Handle() { return hThread; }
	virtual int ExitCode();
	virtual bool IsExited();
	virtual bool IsRunning();

	// Methods
	virtual void Run(bool b);
	virtual void Terminate();
	virtual int Priority();
	virtual void Priority(int p);

	// Override to do something
	virtual int Main();

	// Events
	virtual void OnBeforeMain() {}
	virtual void OnAfterMain() {}
	virtual void OnRun(bool r) {}
};

/////////////////////////////////////////////////////////////////////////
class LgiClass GTrayIcon :
	public GObject,
	public GFlags
{
protected:
	static int Id;
	#ifdef WIN32
	static int TrayCreateMsg;
	#endif

	#ifdef WIN32
	HICON *hIcon;	// array of loaded icons
	#endif
	int Val;		// which icon visible
	int Icons;		// size of array
	int MyId;
	int Msg;
	GView *Notify;

public:
	#ifdef WIN32
	NOTIFYICONDATA	TrayIcon;
	#endif

	GTrayIcon(int Icons = 1);
	~GTrayIcon();

	bool LoadIcon(char *Str, int Index = 0);
	bool Visible();
	void Visible(bool v);

	int Value();
	void Value(int v);
	int NotifyMsg() { return Msg; }
	void NotifyMsg(int i) { Msg = i; }
	GView *GetNotify() { return Notify; }
	void SetNotify(GView *n) { Notify = n; }
	virtual int OnEvent(GMessage *Msg);
	virtual void OnMouseClick(VMouse &m);
};

//////////////////////////////////////////////////////////////////
#include "GInput.h"

//////////////////////////////////////////////////////////////////
typedef char *pchar;

class LgiClass GPrinter
{
	#ifdef WIN32
	PRINTDLG Info;
	#else if BEOS
	char Name[32];
	#endif

public:
	GPrinter();
	virtual ~GPrinter();

	#ifdef WIN32
	HDC StartDC(GView *Parent = 0);
	void EndDC();
	#else if BEOS
	void *StartDC(GView *Parent = 0);
	void EndDC();
	#endif
	bool Browse(GView *Parent);
	bool Serialize(pchar &Str, bool Write);
};

//////////////////////////////////////////////////////////////////
class LgiClass GAlert : public GDialog
{
public:
	GAlert(	GView *parent,
			char *Title,
			char *Text,
			char *Btn1,
			char *Btn2 = 0,
			char *Btn3 = 0);

	int OnNotify(GView *Ctrl, int Flags);
};

////////////////////////////////////////////////////////////////////////
#ifdef BEOS
class LgiClass GScroll : public BScrollBar {

	GScrollView *Notify;

public:
	GScroll(GScrollView *notify, orientation posture);
	void ValueChanged(float newValue);
};

class LgiClass GScrollView : public BView
{
	friend class GLayout;

	int Border;
	GView *Target;
	BScrollBar *X, *Y;
	GdcRegion Pos;

public:
	GScrollView(	char *name,
					GView *target,
					uint32 resizingMode = B_FOLLOW_LEFT | B_FOLLOW_TOP,
					uint32 flags = 0,
					bool horizontal = false,
					bool vertical = false,
					int border = SUNKEN);
	~GScrollView();

	void Draw(BRect update);
	void FrameResized(float width, float height);
	void AttachedToWindow();
	void SetPos(GdcRegion &r);
	GdcRegion &GetPos();
	void OnNotify(GScroll *Bar, float Value);
	BScrollBar *ScrollBar(orientation posture);
};

class LgiClass GSelectFont : public GDialog
{
	GCombo *CFont;
	GCombo *CSize;

public:
	char Font[64];
	int Size;

	GSelectFont(GView *parent, char *Face = 0);
	~GSelectFont();
	int OnNotify(GView *Ctrl, int Flags);
};

#endif

#endif
