/*hdr
**      FILE:           LgiWidgets.h
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog widgets / components
**		PLATFORM:		Xp
**
**      Copyright (C) 1998-2000 Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef __LGI_WIDGETS_H
#define __LGI_WIDGETS_H

/////////////////////////////////////////////////////////////////////////////////////////
#define IDM_STATIC				-1

/////////////////////////////////////////////////////////////////////////////////////////
class GDialog;
#if defined WIN32
class GDlgFunc;
#endif
class GListItem;
class GTreeItem;

class GButton;
class GEdit;
class GCheckBox;
class GText;
class GRadioGroup;
class GRadioButton;
class GTabControl;
class GTabPage;
class GSlider;
class GCombo;
class GBitmap;
class GList;
class GTree;

///////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GLgiRes
{
public:
	// Loading resources
	bool LoadFromResource(	char *Resource,
							List<GView> &Children,
							GdcRegion *Pos = 0,
							char *Name = 0);
};

class LgiClass GDialog :
#if defined LGI_RES
	public ResObject,
#endif
	public GView,
	public GLgiRes
{
	friend class GControl;

private:
	bool IsModal;

	#if defined WIN32

	GMem *Mem;

	#elif defined BEOS

	BWindow *Wnd;
	sem_id ModalSem;
	int ModalRet;
	bool QuitRequested();

	#endif

protected:
	bool LoadFromResource(char *Resource);
	
public:
	GDialog();
	~GDialog();

	#if defined BEOS
	BWindow *GetBWindow() { return Wnd; }
	#endif

	virtual int DoModal();
	virtual int DoModeless();
	virtual void EndModal(int Code = 0);
	virtual void EndModeless(int Code = 0);

	int OnEvent(GMessage *Msg);
};

class LgiClass GControl :
	public GView
{
	friend class GDialog;

protected:
	#if defined BEOS
	bigtime_t Sys_LastClick;
	void MouseClickEvent(bool Down);
	#elif defined WIN32
	GWin32Class *SubClass;
	#endif

	GdcPt2 SizeOfStr(char *Str);

public:
	#if defined BEOS
	GControl(BView *view);
	#elif defined WIN32
	GControl(GWin32Class *Sub = 0);
	#else
	GControl();
	#endif
	~GControl();

	int OnEvent(GMessage *Msg);
	#if defined BEOS // FIXME: Should be in GView?
	bool Attach(GView *parent);
	#endif
};

//////////////////////////////////////////////////////////////////////////////////////
// Controls
#if defined WIN32
#define LGI_BUTTON			"LGI_Button"
#define LGI_EDITBOX			"LGI_Editbox"
#define LGI_CHECKBOX		"LGI_CheckBox"
#define LGI_TEXT			"LGI_Text"
#define LGI_RADIOGROUP		"LGI_RadioGroup"
#define LGI_TABCONTROL		"LGI_TabControl"
#define LGI_TABPAGE			"LGI_TabPage"
#define LGI_SLIDER			"LGI_Slider"
#define LGI_COMBO			"LGI_Combo"
#define LGI_BITMAP			"LGI_Bitmap"
#define LGI_LIST			"LGI_ListView"
#define LGI_TREE			"LGI_TreeView"
#define LGI_PROGRESS		"LGI_Progress"
#define LGI_SCROLLBAR		"LGI_ScrollBar"
#endif

// Button
class LgiClass GButton : public GControl
#if defined LGI_RES
, public ResObject
#endif
{
	#if defined BEOS
	BButton *Button;
	#elif defined WIN32
	static GWin32Class WndClass;
	bool Def;
	char *Class();
	uint32 Style();
	#endif

public:
	GButton(int id, int x, int y, int cx, int cy, char *name);
	~GButton();
	
	bool Default();
	void Default(bool b);

	int OnEvent(GMessage *Msg);
	#if defined BEOS
	bool Name(char *s);
	#endif
};

// Edit
class LgiClass GEdit : public GControl
#if defined LGI_RES
, public ResObject
#endif
{
	#if defined BEOS
	BTextControl *Edit;
	BRect GetFrame(int x, int y, int cx, int cy);
	#elif defined WIN32
	static GWin32Class WndClass;
	#endif

protected:
	bool MultiLineEdit;
	bool PasswordEdit;
	bool Enter;
	#if defined WIN32
	int LastKey;

	char *Class() { return LGI_EDITBOX; }
	uint32 Style() { return	WS_CHILD |
							WS_VISIBLE |
							ES_AUTOHSCROLL |
							WS_TABSTOP |
							((Enter) ? ES_WANTRETURN : 0) |
							((MultiLineEdit) ? ES_MULTILINE | ES_AUTOVSCROLL | ES_WANTRETURN | WS_VSCROLL: 0) |
							((PasswordEdit) ? ES_PASSWORD : 0) |
							GView::Style(); }
	int SysOnNotify(int Code);
	#elif defined BEOS
	status_t Invoke(BMessage *message);
	#endif

public:
	GEdit(int id, int x, int y, int cx, int cy, char *name);
	~GEdit();

	bool WantReturn() { return Enter; }
	void WantReturn(bool b) { Enter = b; }
	bool MultiLine() { return MultiLineEdit; }
	void MultiLine(bool m);
	bool Password() { return PasswordEdit; }
	void Password(bool m);

	void Value(int i);
	int Value();

	#if defined BEOS
	char *Name();
	bool Name(char *s);
	bool SetPos(GdcRegion &p, bool Repaint = false);
	#endif
	int OnEvent(GMessage *Msg);
};

// Check box
class LgiClass GCheckBox : public GControl
#if defined LGI_RES
, public ResObject
#endif
{
	#if defined BEOS
	BCheckBox *Box;
	#elif defined WIN32
	static GWin32Class WndClass;
	char *Class() { return LGI_CHECKBOX; }
	uint32 Style() { return	WS_CHILD |
							WS_TABSTOP |
							WS_VISIBLE |
							((Three) ? BS_AUTO3STATE : BS_AUTOCHECKBOX) |
							GView::Style(); }
	#endif

	int Val;
	bool Three;

public:
	GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState = FALSE);
	~GCheckBox();

	// Methods
	bool ThreeState() { return Three; }
	void ThreeState(bool t) { Three = t; }
	int Value();
	void Value(int b);

	// Impl
	int OnEvent(GMessage *Msg);
	#if defined BEOS
	bool Name(char *s);
	#endif
};

// Static
class LgiClass GText :
#if defined LGI_RES
	public ResObject,
#endif
#if defined BEOS
	public BStringView,
#endif
	public GControl
{
	#if defined BEOS
	void Draw(BRect updateRect);
	#elif defined WIN32
	static GWin32Class WndClass;
	char *Class() { return LGI_TEXT; }
	uint32 Style() { return WS_CHILD | WS_VISIBLE | GView::Style(); }
	#endif

public:
	GText(int id, int x, int y, int cx, int cy, char *name);
	~GText();

	int OnEvent(GMessage *Msg);

	#if defined BEOS
	bool Name(char *n);
	char *Name();
	#endif
};

// Group of radio buttons
class LgiClass GRadioGroup : public GControl
#if defined LGI_RES
, public ResObject
#endif
{
	static int NextId;
	void OnCreate();

	#if defined BEOS
	BBox *Box;
	#elif defined WIN32
	static GWin32Class WndClass;
	#endif

protected:
	int InitialValue;

public:
	GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GRadioGroup();

	// Methods
	int Value();
	void Value(int i);
	GRadioButton *Append(int x, int y, char *name);

	// Impl
	int OnNotify(GView *Ctrl, int Flags);
	int OnEvent(GMessage *Msg);
	#if defined BEOS
	bool Name(char *s);
	#endif
};

class LgiClass GRadioButton : public GControl
#if defined LGI_RES
, public ResObject
#endif
{
	#if defined BEOS
	BRadioButton *Button;
	#elif defined WIN32
	friend class GRadioGroup;
	static GWin32Class WndClass;
	char *Class() { return LGI_BUTTON; }
	uint32 Style() { return BS_AUTORADIOBUTTON | BS_TEXT | WS_VISIBLE | WS_CHILD | ((First) ? WS_GROUP : 0) | GView::Style(); }
	bool First;
	bool InitVal;
	#endif

public:
	GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first = false);
	~GRadioButton();

	int Value();
	void Value(int i);

	int OnEvent(GMessage *Msg);
	#if defined BEOS
	bool Name(char *s);
	#endif
};

// Tab control
typedef List<GTabPage> TabPageList;

class LgiClass GTabControl :
#if defined LGI_RES
	public ResObject,
#endif
	public GControl
{
	friend class GTabPage;
	friend class GTabView;

	#if defined WIN32
	static GWin32Class W32Class;
	#endif

	int TabY();

	int Current;
	GdcRegion TabClient;

protected:
	List<GTabPage> Tabs;

public:
	GTabControl(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GTabControl();

	// Methods
	int Value();
	void Value(int i);

	// List modification
	bool Append(GTabPage *Page, int Where = -1);
	GTabPage *Append(char *name, int Where = -1);
	bool Delete(GTabPage *Page);

	// List access
	GTabPage *TabAt(int i) { return Tabs.ItemAt(i); }
	int GetTabs() { return Tabs.GetItems(); }
	
	// Impl
	bool Attach(GView *parent);
	int OnEvent(GMessage *Msg);
	GView *FindControl(int Id);
	int OnNotify(GView *Ctrl, int Flags);
	void OnChildrenChanged(GView *Wnd, bool Attaching);
	GdcRegion &GetTabClient();

	#if defined WIN32
	GView *FindControl(HWND hCtrl);
	#endif

	void OnPosChange();
	void OnPaint(GdcBasePrimitives *pDC);
	void OnMouseClick(VMouse &m);
	void OnKey(VKey &k);
	void OnFocus(bool f);
};

class LgiClass GTabPage :
#if defined BEOS
	public BView,
#endif
#if defined LGI_RES
	public ResObject,
#endif
	public GView,
	public GLgiRes
{

	friend class GTabControl;

	// Vars
	GTabControl *TabCtrl;
	GdcRegion Pos;
	GdcRegion TabPos;

	// Methods
	bool IsAttached();
	bool Attach(GView *parent);
	void PaintTab(GdcBasePrimitives *pDC, bool Selected);

	#ifdef WIN32
	char *Class();
	#endif

public:
	List<GView> Children;

	GTabPage(char *name);
	~GTabPage();

	bool Enabled();
	void Enabled(bool e);
	bool Visible();
	void Visible(bool v);
	char *Name();
	bool Name(char *n);
	GTabControl *GetTabControl() { return TabCtrl; }

	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	void OnKey(VKey &k);
	void OnFocus(bool b);

	void Append(GView *Wnd);
	bool Remove(GView *Wnd);
	bool AppendDialogFromResource(char *Name);
};

// Slider
class LgiClass GSlider :
	public GControl
#if defined LGI_RES
	, public ResObject
#endif
{
	#if defined WIN32
	static GWin32Class WndClass;

	char *Class() { return LGI_SLIDER; }
	uint32 ExStyle() { return 0; }
	uint32 Style() { return WS_VISIBLE | WS_CHILD | WS_TABSTOP | ((Vertical) ? TBS_VERT : TBS_HORZ) | GView::Style(); }
	int SysOnNotify(int Code);
	#endif

	bool Vertical;
	int Min, Max;
	int Val;

	GdcRegion Thumb;
	int Tx, Ty;
	bool Track;

public:
	GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert);
	~GSlider();

	void Value(int i);
	int Value();
	void GetLimits(int &x, int &y);
	void SetLimits(int x, int y);

	int OnEvent(GMessage *Msg);
	#if defined BEOS
	void OnPaint(GdcBasePrimitives *pDC);
	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
	#endif
};

// Bitmap
class LgiClass GBitmap :
#ifdef BEOS
	public BControl,
#endif
#ifdef LGI_RES
	public ResObject,
#endif
	public GControl
{
	#if defined BEOS
	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool f = true) { BView::MakeFocus(f); OnFocus(f); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }
	#elif defined WIN32
	static GWin32Class WndClass;
	char *Class() { return LGI_BITMAP; }
	uint32 Style() { return WS_CHILD | WS_VISIBLE | GView::Style(); }
	#endif

	GdcBasePrimitives *pDC;
	class GThread *pThread;

public:
	GBitmap(int id, int x, int y, char *FileName, bool Async = false);
	~GBitmap();

	virtual void SetDC(GdcBasePrimitives *pDC = 0);
	virtual GdcBasePrimitives *GetDC();

	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	void OnMouseClick(VMouse &m);
};

//////////////////////////////////////////////////////////////////////////////
// List view
#include "GCombo2.h"
#include "GList2.h"
#include "GTree2.h"

// Progress control
class LgiClass GProgress :
#if defined LGI_RES
	public ResObject,
#endif
#if defined BEOS
	public BView,
#endif
	public GControl,
	public Progress
{
	#if defined WIN32
	static GWin32Class WndClass;
	char *Class() { return LGI_PROGRESS; }
	uint32 Style() { return WS_CHILD | WS_TABSTOP | WS_VISIBLE | GView::Style(); }
	#elif defined BEOS
	COLOUR c;
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	#endif

public:
	GProgress(int id, int x, int y, int cx, int cy, char *name);
	~GProgress();

	void SetLimits(int l, int h);
	void Value(int v);
	int OnEvent(GMessage *Msg);

	#if defined BEOS	
	void GetLimits(int &l, int &h);
	int Value();
	void Colour(COLOUR Col);
	COLOUR Colour();
	void OnPaint(GdcBasePrimitives *pDC);
	#endif
};

// Scroll bars
class LgiClass GScrollBar : public GControl
#if defined LGI_RES
, public ResObject
#endif
{
	friend class GLayout;

	#if defined WIN32
	static GWin32Class WndClass;
	char *Class() { return LGI_SCROLLBAR; }
	uint32 Style();
	#endif

protected:
	#if defined BEOS
	BScrollBar *Bar;
	int High, Low, PageSize;
	#elif defined WIN32
	int			Type;		// vertical, horzontal or control
	SCROLLINFO	Info;
	void Update();
	void SetParentFlag(bool Bool);
	#endif

public:
	// call this constructor for embeded scrollbar
	// say in a window
	#if defined BEOS
	GScrollBar(BScrollBar *bar);
	#else
	GScrollBar();
	#endif

	// call this constructor for a control based
	// scrollbar, say in a dialog
	GScrollBar(int id, int x, int y, int cx, int cy, char *name);

	~GScrollBar();

	// data
	bool Vertical();
	void SetVertical(bool v);
	int Pos();
	void SetPos(int p);
	void Limits(int &Low, int &High);
	void SetLimits(int Low, int High);
	int Page();
	void SetPage(int p);

	#if defined WIN32
	void Invalidate();
	void SetParent(GView *p);
	#else
	bool Attach(GView *p);
	#endif
	
	// events
	int OnEvent(GMessage *Msg);

	// overridables
	virtual void OnChange(int Pos) {}
};

// Data Dlg
#include "GuiDataDlg.h"

#endif
