//
//      FILE:           LgiDefs.h
//      AUTHOR:         Matthew Allen
//      DATE:           24/9/99
//      DESCRIPTION:    Global LGI include
//
//      Copyright (C) 1999, Matthew Allen
//              fret@ozemail.com.au
//

#ifndef __LGIDEFS_H
#define __LGIDEFS_H

#include "LgiInc.h"

// unsafe typedefs, for backward compatibility
typedef			unsigned char		uchar;
typedef			unsigned short		ushort;
typedef			unsigned int		uint;
typedef			unsigned long		ulong;
typedef			unsigned short		HND;
typedef			unsigned short		MESSAGE;

// Length safe typedesf, use these in new code
typedef			signed char			int8;
typedef			unsigned char		uint8;
typedef			signed short		int16;
typedef			unsigned short		uint16;

// Boolean
#define TRUE						1
#define FALSE						0

// Basic macros
// #define abs(a)					(((a) > 0) ? (a) : -(a))
#define min(a,b)					(((a) < (b)) ? (a) : (b))
#define max(a,b)					(((a) > (b)) ? (a) : (b))
#define limit(i,l,u)				(((i)<(l)) ? (l) : (((i)>(u)) ? (u) : (i)))
#define makelong(a, b)				((a)<<16 | (b&0xFFFF))
#define loword(a)					(a&0xFFFF)
#define hiword(a)					(a>>16)
#define swap(a, b)					{ int n = a; a = b; b = n; }

// Operators
#define NOT							!
#define AND							&&
#define OR							||

// Good ol NULLy
#ifndef NULL
#define NULL						0
#endif

// Some objectish ones
#define ZeroObj(obj)				memset(&obj, 0, sizeof(obj))
#define CountOf(array)				(sizeof(array)/sizeof(array[0]))
#define StrEmpty(s)					(s)[0] = 0
#define IsStrEmpty(s)				((NOT s) OR ((s)[0] == 0))

// Flags
#define SetFlag(i, f)				(i) |= (f)
#define ClearFlag(i, f)				(i) &= ~(f)
#define TestFlag(i, f)				(((i) & (f)) != 0)

// Strings
LgiExtern char WhiteSpace[];

LgiFunc char *strnchr(char *s, char c, int Len);
LgiFunc char *strnstr(char *a, char *b, int n);
LgiFunc char *strnistr(char *a, char *b, int n);
LgiFunc char *stristr(char *a, char *b);
LgiFunc char *TrimStr(char *s, char *Delim = " \r\n\t");
LgiFunc bool ValidStr(char *s);
LgiFunc char *NewStr(char *s, int Len = -1);
LgiFunc bool MatchStr(char *a, char *b);

// Asserts
#include "assert.h"
#define ASSERT						assert

// Defines

// Os types
#define LGI_OS_UNKNOWN				0
#define LGI_OS_WIN9X				1
#define LGI_OS_WINNT				2
#define LGI_OS_BEOS					3
#define LGI_OS_LINUX				4
#define LGI_OS_ATHEOS				5
#define LGI_OS_MAX

// Colours
#define LC_BLACK					LgiColour(0)
#define LC_DKGREY					LgiColour(1)
#define LC_MIDGREY					LgiColour(2)
#define LC_LTGREY					LgiColour(3)
#define LC_WHITE					LgiColour(4)

#define LC_SHADOW					LgiColour(5)
#define LC_LOW						LgiColour(6)
#define LC_MED						LgiColour(7)
#define LC_HIGH						LgiColour(8)
#define LC_LIGHT					LgiColour(9)

#define LC_DIALOG					LgiColour(10)
#define LC_WORKSPACE				LgiColour(11)
#define LC_TEXT						LgiColour(12)
#define LC_SELECTION				LgiColour(13)
#define LC_SEL_TEXT					LgiColour(14)

#define LC_ACTIVE_TITLE				LgiColour(15)
#define LC_ACTIVE_TITLE_TEXT		LgiColour(16)
#define LC_INACTIVE_TITLE			LgiColour(17)
#define LC_INACTIVE_TITLE_TEXT		LgiColour(18)

#define LC_MAXIMUM					19

/////////////////////////////////////////////////////////////////////////////////////
// Class names

// Virtual input classes
class VKey;
class VMouse;

// General GUI classes
class GTarget;
class GComponent;
class GEvent;
class GId;
class GApp;
class GWindow;
class GWin32Class;
class GView;
class GLayout;
class GFileType;
class GFileSelect;
class GFindReplace;
class GSubMenu;
class GMenuItem;
class GMenu;
class GToolBar;
class GToolButton;
class GSplitter;
class GStatusPane;
class GStatusBar;
class GToolColour;
class GScrollBar;
class GImageList;
class GDialog;

// General parent object
class LgiClass GObject {

	char *ObjName;

public:
	GObject();	
	virtual ~GObject();

	virtual char *Name();
	virtual bool Name(char *n);
	virtual int Sizeof() { return 0; }
};

#endif