/*hdr
**      FILE:           Store.h
**      AUTHOR:         Matthew Allen
**      DATE:           29/10/98
**      DESCRIPTION:    Tree storage
**
**      Copyright (C) 1998 Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef __STORE_H
#define __STORE_H

#include "Gdc2.h"
#include "Contain.h"
#include "Progress.h"
#include "StoreCommon.h"

namespace Storage1
{
	#define STORAGE_ITEM_MAGIC		0x123400AA
	#define STORAGE_ITEM_NOSIZE		0x0001
	#define STORAGE_MAGIC			0x123400FF

	class StorageKitImpl;
	class StorageItemImpl;

	struct StorageItemHeader {

		int Magic;
		int Type;
		int ObjSize;
		int AllocatedSize;
		
		int Next;
		int Prev;
		int Parent;
		int Child;

	};

	class StorageItemImpl : public GObject, public StorageItem
	{
		friend class StorageKitImpl;

	private:
		int Sizeof();
		bool SetSize(int NewSize);
		bool Serialize(VFile &f, bool Write, int Flags = 0);
		bool MoveToLoc(int NewLoc);
		bool WriteHeader(VFile &f);

		StorageKitImpl *Tree;

	public:
		int StoreType;
		int StoreSize;
		int StoreAllocatedSize;
		int StoreLoc;
		int StoreNext;
		int StorePrev;
		int StoreParent;
		int StoreChild;

		StorageItemImpl *Next;
		StorageItemImpl *Prev;
		StorageItemImpl *Parent;
		StorageItemImpl *Child;

		StorageItemImpl();
		~StorageItemImpl();

		// Heirarchy
		StorageItem *GetNext();
		StorageItem *GetPrev();
		StorageItem *GetParent();
		StorageItem *GetChild();

		StorageItem *CreateNext(StorageObj *Obj);
		StorageItem *CreateChild(StorageObj *Obj);
		StorageItem *CreateSub(StorageObj *Obj);
		bool DeleteChild(StorageItem *Obj);
		bool DeleteAllChildren();

		// Properties
		int GetType();
		void SetType(int i);
		int GetTotalSize();
		int GetObjectSize();
		StorageKit *GetTree();

		// Impl
		bool Save();
		VFile *GotoObject();
		bool EndOfObj(VFile &f);
	};

	class Block {
	public:
		StorageItemImpl *Item;
		int Size;
	};

	struct StorageHeader {

		int Magic;
		int Version;
		int Reserved2;
		int FirstLoc;
		int Reserved3[12];
	};

	class StorageKitImpl : public GObject, public StorageKit
	{
		friend class StorageItemImpl;

	private:
		char *FileName;
		VFile File;
		bool Status;
		int StoreLoc;
		int Version;
		StorageItemImpl *Root;
		bool ReadOnly;

		uint UserData;
		ValidateItemFunc Func;

		void AddItem(	StorageItemImpl *Item,
						List<Block> &Blocks);
		StorageItemImpl *CreateItem(StorageObj *Obj);
		StorageItemImpl *LoadLocation(int Loc);
		bool Serialize(VFile &f, bool Write);
		bool SerializeItem(StorageItem *Item, bool Write, int Flags = 0)
		{
			return (Item) ? ((StorageItemImpl*)Item)->Serialize(File, Write, Flags) : 0;
		}

	public:
		StorageKitImpl(char *FileName);
		~StorageKitImpl();

		int GetStatus() { return Status; }
		int GetVersion() { return Version; }
		void SetVersion(int i);
		quad GetFileSize();
		bool GetPassword(GPassword *p) { return false; }
		bool SetPassword(GPassword *p) { return false; }
		
		StorageItem *GetRoot();
		StorageItem *CreateRoot(StorageObj *Obj);

		bool Compact(Progress *p, ValidateItemFunc func = 0, uint userdata = 0);
		bool DeleteItem(StorageItem *Item);
		bool SeparateItem(StorageItem *Item);
		bool ReparentItem(StorageItem *Item, StorageItem *NewParent, bool Next = false);
	};
}

#endif