#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "Gdc2.h"
#include "IHttp.h"
#include "Token.h"

///////////////////////////////////////////////////////////////////
class ILogProxy : public IConnection
{
	IConnection *Dest;

public:
	ILogProxy(IConnection *dest)
	{
		Dest = dest;
	}

	void OnError(int ErrorCode, char *ErrorDescription)
	{
		if (Dest AND ErrorDescription)
		{
			char Str[256];
			sprintf(Str, "[Data] %s", ErrorDescription);
			Dest->OnInformation(Str);
		}
	}

	void OnInformation(char *s)
	{
		if (Dest AND s)
		{
			char Str[256];
			sprintf(Str, "[Data] %s", s);
			Dest->OnInformation(Str);
		}
	}

	void OnDisconnect()
	{
		if (Dest)
		{
			Dest->OnInformation("[Data] Disconnect");
		}
	}
};

///////////////////////////////////////////////////////////////////
IHttp::IHttp()
{
	Socket = 0;
	Buffer[0] = 0;
	ResumeFrom = 0;
}

IHttp::~IHttp()
{
	Close();
}

bool IHttp::Open(IConnection *S, char *RemoteHost, int Port)
{
	Close();
	Socket = S;
	if (Socket AND
		Socket->Open(RemoteHost, (Port>0) ? Port : 80))
	{
		return true;
	}

	return false;
}

bool IHttp::Close()
{
	if (Socket)
	{
		Socket->Close();
		DeleteObj(Socket);
	}
	return 0;
}

bool IHttp::IsOpen()
{
	return Socket != 0;
}

bool IHttp::GetFile(char *File, BytePipe &Out, int Format, int *ProtocolStatus, int *DataLength)
{
	if (File AND Socket)
	{
		char Str[256];
		char *ContentStr = "Content-Length:";
		char Line[4096] = "", *Cur = Line;
		int ContentLength = 0;
		int DataToGo = 1000000000;

		char FilePart[256];
		char Host[256];
		char *f = strchr(File + 7, '/');
		if (NOT f) return false;

		strcpy(FilePart, f);
		char *h = (strnicmp(File, "http://", 7) == 0) ? File + 7 : File;
		if (NOT h) return false;

		int HostLen = (int)f-(int)h;
		memcpy(Host, h, HostLen);
		Host[HostLen] = 0;
		
		// get part
		if (Format == GET_TYPE_NORMAL)
		{
			sprintf(Str, "GET %s HTTP/1.0\r\n", File);
		}
		else if (Format == GET_TYPE_FILE_HOST)
		{
			sprintf(Str, "GET %s HTTP/1.0\r\nHost: %s\r\n", FilePart, Host);
		}
		else
		{
			// invalid argument
			return 0;
		}

		// resume info
		if (ResumeFrom > 0)
		{
			sprintf(Str+strlen(Str), "Range:Bytes=%i-\r\n", ResumeFrom);
		}

		// end of request
		strcat(Str, "\r\n");

		// write request
		Socket->SetParameter(PARAMETER_LOG, true);
		Socket->Write(Str, strlen(Str), 0);
		
		// read headers
		int LinesRead = 0;
		int StatusCode = 0;
		int Length = Socket->Read(Buffer, sizeof(Buffer), 0);
		while (	Length > 0 AND
				DataToGo > 0 AND
				(NOT Meter OR NOT Meter->Cancel()))
		{
			if (NOT ContentLength)
			{
				// search through data for feilds and blank line
				int Used = (int)(Cur-Line); // buffer bytes used
				int Left = sizeof(Line)-Used-1; // bytes left in buffer minus a NUL
				for (int i=0; i<Length AND i<Left; i++)
				{
					*Cur++ = Buffer[i];
					if (Cur > (Line+1) AND
						Cur[-2] == '\r' AND Cur[-1] == '\n')
					{
						// new line baby... reset
						Cur[-2] = 0;
						Cur = Line;

						// check data
						if (LinesRead == 0)
						{
							TokenParser T(Line, " ");
							if (T.Tokens >= 2)
							{
								StatusCode = atoi(T.Get(1));

								if (ProtocolStatus)
								{
									*ProtocolStatus = StatusCode;
								}

								if (ResumeFrom > 0 AND StatusCode != 206)
								{
									// Resume not supported
									ResumeFrom = 0;
									Socket->OnError(StatusCode, "Resume not supported on this server");
									return false;
								}
								
								if (ResumeFrom == 0 AND StatusCode != 200)
								{
									Socket->OnError(StatusCode, Line);
									return false;
								}
							}
						}
						else if (strnicmp(Line, ContentStr, strlen(ContentStr)) == 0)
						{
							// content length
							ContentLength = atoi(Line + strlen(ContentStr) + 1);

							if (DataLength)
							{
								*DataLength = ContentLength;
							}
						}
						else if (strlen(Line) < 2)
						{
							// end of headers
							i++;
							int DataInThisPacket = Length - i;
							DataToGo = ContentLength - DataInThisPacket;
							Out.Push((uchar*)Buffer+i, DataInThisPacket);

							if (Meter)
							{
								Meter->SetLimits(0, ContentLength-1);
								Meter->Value(DataInThisPacket);
							}
							break;
						}

						LinesRead++;
					}
				}
			}
			else
			{
				// push data onto output
				Out.Push((uchar*)Buffer, Length);
			}
			
			// get MORE data!!! HAHAHA
			Length = Socket->Read(Buffer, sizeof(Buffer), 0);
			Length = min(Length, DataToGo);
			DataToGo -= Length;

			if (Meter)
			{
				Meter->Value(ContentLength - DataToGo);
			}
		}

		if (ProtocolStatus AND
			Length <= 0 AND
			(NOT Meter OR NOT Meter->Cancel()) )
		{
			// socket error
			*ProtocolStatus = -1;
		}

		Out.Push((uchar*)Buffer, Length);
		if (Meter)
		{
			Meter->Value(0);
		}

		// Turn logging on
		Socket->SetParameter(PARAMETER_LOG, true);

		// Calculate status
		int Data = Out.Sizeof();
		bool Status = Data == ContentLength;
		ResumeFrom = 0;
		return Status;
	}

	return false;
}
