/*hdr
**      FILE:           ILdap.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           21/6/2000
**      DESCRIPTION:    Open LDAP plugin
**
**      Copyright (C) 2000, Matthew Allen
**              fret@ozemail.com.au
**
**		API doc URL:
**			http://docs.iplanet.com/docs/manuals/dirsdk/csdk30/functi99.htm#2215851
*/

#include <stdio.h>

#include "Lgi.h"
#include "ILdap.h"

////////////////////////////////////////////////////////////////
ILdapEntry::ILdapEntry()
{
	Name = 0;
	Email = 0;
	Type = 0;
}

ILdapEntry::~ILdapEntry()
{
	DeleteArray(Name);
	DeleteArray(Email);
}

////////////////////////////////////////////////////////////////
ILdap::ILdap()
{
	Ldap = 0;
}

ILdap::~ILdap()
{
}

bool ILdap::Open(char *RemoteServer, int Port)
{
	Ldap = ldap_open(RemoteServer, Port);
	if (Ldap)
	{
		Ldap->ld_sizelimit = LDAP_NO_LIMIT;
		Ldap->ld_deref = LDAP_DEREF_ALWAYS;

		if (ldap_simple_bind_s(Ldap, NULL, NULL) == LDAP_SUCCESS )
		{
			return true;
		}
	}

	return false;
}

bool ILdap::Close()
{
	if (Ldap)
	{
		ldap_unbind(Ldap);
		Ldap = 0;
	}

	return false;
}

ILdapEntry *ILdap::RetreiveOne(char *Name)
{
	if (Ldap AND Name)
	{
		// create a filter string
		char Filter[256];
		sprintf(Filter, "(|(cn=%s)(givenName=%s))", Name, Name);

		// call a sync search
		LDAPMessage *Msg = 0;
	    int Result = ldap_search_s(	Ldap,
									"",
									LDAP_SCOPE_SUBTREE,
									Filter,
									NULL, // attr's to return
									false,
									&Msg);
		switch (Result)
		{
			case LDAP_SUCCESS:
			{
				int Entries = ldap_count_entries(Ldap, Msg);

				for (LDAPMessage *m = ldap_first_entry(Ldap, Msg);
					m;
					m = ldap_next_entry(Ldap, m))
				{
					char **Type = ldap_get_values(Ldap, m, "objectClass");
					char **Name = ldap_get_values(Ldap, m, "cn");
					char **Email = ldap_get_values(Ldap, m, "rfc822Mailbox");

					if (Type AND Name AND Email)
					{
						ILdapEntry *e = NEW(ILdapEntry);
						if (e)
						{
							if (Type AND stricmp(Type[0], "groupOfNames") == 0)
							{
								e->Type = LDAP_ENTRY_GROUP;
							}
							else
							{
								e->Type = LDAP_ENTRY_PERSON;
							}

							if (Name)
							{
								e->Name = NewStr(Name[0]);
							}

							if (Email)
							{
								e->Email = NewStr(Email[0]);
							}

							return e;
						}
					}
				}
				break;
			}
			default:
			{
				// error
				break;
			}
		}
	}

	return 0;
}

bool ILdap::RetreiveList(List<ILdapEntry> &Lst)
{
	bool Status = false;
	if (Ldap)
	{
		for (char Start='A'; Start<='Z'; Start++)
		{
			// create a filter string
			char Str[256] = "";
			// sprintf(Str, "(&(cn>=%c!)(cn<=%c~))", Start, Start);
			sprintf(Str, "cn=%c*", Start);

			// call a sync search
			LDAPMessage *Msg = 0;
			int Result = ldap_search_s(	Ldap,
										"",
										LDAP_SCOPE_SUBTREE,
										Str,
										NULL, // attr's to return
										false,
										&Msg);
			switch (Result)
			{
				case LDAP_SUCCESS:
				{
					int Entries = ldap_count_entries(Ldap, Msg);

					for (LDAPMessage *m = ldap_first_entry(Ldap, Msg);
						m;
						m = ldap_next_entry(Ldap, m))
					{
						// objectClass: groupOfNames
						// objectClass: organizationalPerson
						
						char **Type = ldap_get_values(Ldap, m, "objectClass");
						char **Name = ldap_get_values(Ldap, m, "cn");
						char **Email = ldap_get_values(Ldap, m, "rfc822Mailbox");

						if (Type AND Name AND Email)
						{
							ILdapEntry *e = NEW(ILdapEntry);
							if (e)
							{
								if (Type AND stricmp(Type[0], "groupOfNames") == 0)
								{
									e->Type = LDAP_ENTRY_GROUP;
								}
								else
								{
									e->Type = LDAP_ENTRY_PERSON;
								}

								if (Name)
								{
									e->Name = NewStr(Name[0]);
								}

								if (Email)
								{
									e->Email = NewStr(Email[0]);
								}

								Lst.Insert(e);
								Status = true;
							}
						}
					}
					break;
				}
				default:
				{
					break;
				}
			}
		}
	}

	return Status;
}



