/*hdr
**      FILE:           INet.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Internet sockets
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#if defined WIN32
#include <winsock2.h>
#elif defined LINUX
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#include "INet.h"

#if 0	// just for LgiTrace !!! Remove later!!!
#include "Gdc2.h"
#include "LgiCommon.h"
FileDevice *  FileDevice::Instance = 0;
GdcDevice *  GdcDevice::pInstance = 0;
#endif

#ifdef BEOS
#define isdigit(c)			((c)>='0'AND(c)<='9')
#define SOMAXCONN			100
#define SOCKET_ERROR		0
#define INVALID_SOCKET		0
#endif

////////////////////////////////////////////////////////////////////////////////////////////////
GNetwork::GNetwork()
{
	#ifdef WIN32
	SocketsOpen = false;

	uchar Major = 1;
	uchar Minor = 1;
	ushort Version = MAKEWORD(Major, Minor);
	WSADATA WsaData;

	// Start sockets
	int Result = WSAStartup(Version, &WsaData);
	if (NOT Result)
	{
		if (LOBYTE(WsaData.wVersion) == Major AND HIBYTE(WsaData.wVersion) == Minor)
		{
			SocketsOpen = Result == 0;
		}
		else
		{
			WSACleanup();
		}
	}

	#else
	SocketsOpen = true;
	#endif
}

GNetwork::~GNetwork()
{
	#ifdef WIN32
	if (SocketsOpen)
	{
		WSACleanup();
		SocketsOpen = FALSE;
	}
	#endif
}

IConnection *GNetwork::CreateConnection()
{
	IConnection *C = NULL;

	if (IsValid())
	{
		C = NEW(IConnection);
		if (C)
		{
			C->Parent = this;
		}
	}

	return C;
}

#ifdef WIN32
#include "..\..\_Win32\INet\MibAccess.h"
#endif

bool GNetwork::EnumInterfaces(List<char> &Lst)
{
	bool Status = false;

	#ifdef WIN32
	UINT IPArray[50];
	UINT Size = 50;
	MibII m;

	if (m.Init() AND
		m.GetIPAddress(IPArray, Size))
	{
		/*
		UINT MaskArray[50];
		UINT MaskArraySize =50;
		
		if (m.GetIPMask(MaskArray,MaskArraySize))
		{
			m_Edit="";
			UINT i;
			for(i=0; i< MaskArraySize; i++)
			{
				m_Edit += "ip: " + TypeIP(IPArray[i])+" mask: " + TypeIP(MaskArray[i])+ "\r\n";
			}
		}
		*/

		for (int i=0; i<Size; i++)
		{
			if (IPArray[i])
			{
				char Str[256];
				uchar *Ip = (uchar*) &IPArray[i];
				sprintf(Str, "%i.%i.%i.%i", Ip[0], Ip[1], Ip[2], Ip[3]);
				Lst.Insert(NewStr(Str));

				Status = true;
			}
		}
	}	
	#endif

	return Status;
}

////////////////////////////////////////////////////////////////////////////////////////////////
IConnection::IConnection()
{
	Async = false;
	BytesWritten = 0;
	BytesRead = 0;
	Canceled = false;

	#if defined BEOS
	Ep = NEW(BNetEndpoint);
	ASSERT(Ep);
	#elif defined WIN32
	Socket = INVALID_SOCKET;
	#endif
}

IConnection::~IConnection()
{
	Close();
	#ifdef BEOS
	DeleteObj(Ep);
	#endif
}

void IConnection::OnDisconnect()
{
	#ifdef WIN32
	Socket = INVALID_SOCKET;
	#endif
}

SOCKET IConnection::Handle()
{
	#if defined WIN32
	return Socket;
	#elif defined BEOS
	return Ep->Socket();
	#endif
}

bool IConnection::IsAsync()
{
	return Async;
}

void IConnection::IsAsync(bool i)
{
	if (IsOpen())
	{
		Async = i;
		#if defined BEOS
		Ep->SetNonBlocking(Async);
		#elif defined WIN32
		ulong NonBlocking = i;
		ioctlsocket(Socket, FIONBIO, &NonBlocking);
		#endif
	}
}

bool IConnection::IsReadable()
{
	#if defined WIN32
	ulong Val = 0;
	if (ioctlsocket(Socket, FIONREAD, &Val) == 0)
	{
		return Val > 0;
	}
	return false;
	#elif defined BEOS
	return Ep->IsDataPending();
	#else
	return false;
	#endif
}

int IConnection::WhatsMyPort()
{
	#if defined WIN32

	struct sockaddr_in addr;
	int size;
	
	ZeroObj(addr);
	size = sizeof(addr);
	if ((getsockname(Handle(), (struct sockaddr *)&addr, &size)) >= 0)
	{
		return ntohs(addr.sin_port);
	}
	return 0;

	#elif defined BEOS

	char Buf[256];
	unsigned short Port;
	
	BNetAddress Addr = Ep->LocalAddr();
	if (Addr.GetAddr(Buf, &Port) == B_OK)
	{
		return Port;
	}
	return 0;

	#endif
}

bool IConnection::WhatsMyIp(char *IpAddr)
{
	if (IpAddr)
	{
		#ifndef LINUX
		struct sockaddr_in addr;
		int size;
		
		size = sizeof(addr);
		if ((getsockname(Handle(), (struct sockaddr *)&addr, &size)) < 0)
			return false;
		
		if (addr.sin_addr.s_addr == INADDR_ANY)
			return false;
		
		uchar *a = (uchar*)&addr.sin_addr.s_addr;
		sprintf(	IpAddr,
					"%i.%i.%i.%i",
					a[0],
					a[1],
					a[2],
					a[3]);
	
		return true;
		#endif
	}

	return false;
}

bool IConnection::GetRemoteIp(char *IpAddr)
{
	#ifdef WIN32
	if (IpAddr)
	{
		struct sockaddr_in a;
		int addrlen = sizeof(a);
		if (NOT getpeername(Handle(), (sockaddr*)&a, &addrlen))
		{
			uchar *addr = (uchar*)&a.sin_addr.s_addr;
			sprintf(IpAddr,
					"%u.%u.%u.%u",
					addr[0],
					addr[1],
					addr[2],
					addr[3]);
			return true;
		}
	}
	#endif

	return false;
}

bool IConnection::Open(char *HostAddr, int Port)
{
	bool Status = false;

	Close();

	#if defined WIN32
	sockaddr_in RemoteAddr;
	HOSTENT *Host = 0;
	long IpAddress = 0;

	Socket = socket(AF_INET, SOCK_STREAM, 0);
	if (Socket >= 0)
	{
		BytesWritten = 0;
		BytesRead = 0;

		if (isdigit(*HostAddr) AND strchr(HostAddr, '.'))
		{
			// Ip address
			IpAddress = inet_addr(HostAddr);
			// Host = gethostbyaddr((const char *)&IpAddress, 4, AF_INET);
		}
		else
		{
			// Name address
			Host = gethostbyname(HostAddr);
			if (NOT Host)
			{
				Error();
				return false;
			}
		}

		if (1)
		{
			RemoteAddr.sin_family = AF_INET;
			RemoteAddr.sin_port = htons(Port);
			if (Host)
			{
				memcpy(&RemoteAddr.sin_addr, Host->h_addr_list[0], sizeof(in_addr));
			}
			else
			{
				memcpy(&RemoteAddr.sin_addr, &IpAddress, sizeof(in_addr));
			}

			if (connect(Socket, (sockaddr*) &RemoteAddr, sizeof(sockaddr)))
			{
				Error();
			}
			else
			{
				char Info[256];
				sprintf(Info,
						"[INET] Socket Connect: %s [%i.%i.%i.%i], port: %i",
						HostAddr,
						(RemoteAddr.sin_addr.s_addr) & 0xFF,
						(RemoteAddr.sin_addr.s_addr >> 8) & 0xFF,
						(RemoteAddr.sin_addr.s_addr >> 16) & 0xFF,
						(RemoteAddr.sin_addr.s_addr >> 24) & 0xFF,
						Port);
				OnInformation(Info);

				Status = true;
			}
		}

		if (NOT Status)
		{
			closesocket(Socket);
			Socket = INVALID_SOCKET;
		}
	}
	else
	{
		Error();
	}
	#elif defined BEOS
	
	BNetAddress *Addr = NEW(BNetAddress(HostAddr, Port));
	//if (Addr AND Addr->InitCheck())
	{
		Status = Ep->Connect(*Addr) == B_OK;
	}
	
	#endif
	
	return Status;
}

bool IConnection::Listen(int Port)
{
	Close();

	#if defined WIN32
	Socket = socket(AF_INET, SOCK_STREAM, 0);
	if (Socket >= 0)
	{
		BytesWritten = 0;
		BytesRead = 0;

		sockaddr Addr;
		sockaddr_in *a = (sockaddr_in*) &Addr;
		ZeroObj(Addr);
		a->sin_family = AF_INET;
		a->sin_port = htons(Port);
		a->sin_addr.S_un.S_addr = INADDR_ANY;
		// a->sin_addr.s_addr = INADDR_ANY; // BeOS

		if (bind(Socket, &Addr, sizeof(Addr)) >= 0)
		{
			if (listen(Socket, SOMAXCONN) != SOCKET_ERROR)
			{
				return true;
			}
			else
			{
				Error();
			}
		}
		else
		{
			Error();
		}
	}
	else
	{
		Error();
	}
	return false;
	#elif defined BEOS
	if (Ep->Bind(Port) == B_OK)
	{
		return Ep->Listen() == B_OK;
	}
	return false;
	#endif
}

bool IConnection::Accept(IConnection *c)
{
	if (c)
	{
		#if defined WIN32
		
		SOCKET NewSocket = 0;
		sockaddr Address;
		int Length = sizeof(Address);
		NewSocket = accept(Socket, &Address, &Length);
		if (ValidSocket(NewSocket))
		{
			c->Socket = NewSocket;
			return true;
		}
		return false;
		
		#elif defined BEOS
		
		DeleteObj(c->Ep);
		c->Ep = Ep->Accept();
		return c->Ep != 0;
		
		#endif
	}

	return false;
}

void IConnection::Close()
{
	#if defined WIN32
	if (ValidSocket(Socket))
	{
		closesocket(Socket);
		OnDisconnect();
	}
	Socket = INVALID_SOCKET;
	#elif defined BEOS
	
	BNetAddress Addr = Ep->RemoteAddr();
	ushort Port = 0;
	if (Addr.GetAddr((char*)0, &Port) == B_OK AND Port != 0)
	{
		Ep->Close();
		OnDisconnect();
	}
	
	#endif
}

int IConnection::Write(char *Data, int Len, int Flags)
{
	int Total = 0;
	Canceled = false;
	while (	Data AND
			Total < Len AND
			NOT Canceled)
	{
		int Block = Len-Total;
		Block = min(Block, 4096);
		#if defined WIN32

		int i = send(Socket,	Data+Total, Block, Flags);
		if (i != SOCKET_ERROR)

		#elif defined BEOS

		int i = Ep->Send(		Data+Total, Block, Flags);
		if (i > 0)

		#else

		int i = 0;
		if (i > 0)

		#endif
		{
			OnWrite(Data+Total, i);
			BytesWritten += i;
			Total += i;
		}
		else
		{
			return 0;
		}
	}
	return Total;
}

int IConnection::Read(char *Data, int Len, int Flags)
{
	#if defined WIN32
	if (Data)
	{
		int Code = recv(Socket, Data, Len, Flags);
		if (Code <= 0)
		{
			Error();
		}
		else
		{
			if (Code < Len)
			{
				// Just in case it's a string lets be safe
				*(Data + Code) = 0;
			}

			BytesRead += Code;
			OnRead(Data, Code);
		}

		return Code;
	}
	return 0;
	#elif defined BEOS
	int i = Ep->Receive(Data, Len, Flags);
	if (i > 0)
	{
		BytesRead += i;
		OnRead(Data, i);
	}
	return i;
	#endif
}

int IConnection::Error()
{
	#if defined WIN32
	class ErrorMsg {
	public:
		int Code;
		char *Msg;
	} ErrorCodes[] = {
		{0,					"Socket disconnected."},
		{WSAEACCES,			"Permission denied."},
		{WSAEADDRINUSE,		"Address already in use."},
		{WSAEADDRNOTAVAIL,	"Cannot assign requested address."},
		{WSAEAFNOSUPPORT,	"Address family not supported by protocol family."},
		{WSAEALREADY,		"Operation already in progress."},
		{WSAECONNABORTED,	"Software caused connection abort."},
		{WSAECONNREFUSED,	"Connection refused."},
		{WSAECONNRESET,		"Connection reset by peer."},
		{WSAEDESTADDRREQ,	"Destination address required."},
		{WSAEFAULT,			"Bad address."},
		{WSAEHOSTDOWN,		"Host is down."},
		{WSAEHOSTUNREACH,	"No route to host."},
		{WSAEINPROGRESS,	"Operation now in progress."},
		{WSAEINTR,			"Interrupted function call."},
		{WSAEINVAL,			"Invalid argument."},
		{WSAEISCONN,		"Socket is already connected."},
		{WSAEMFILE,			"Too many open files."},
		{WSAEMSGSIZE,		"Message too long."},
		{WSAENETDOWN,		"Network is down."},
		{WSAENETRESET,		"Network dropped connection on reset."},
		{WSAENETUNREACH,	"Network is unreachable."},
		{WSAENOBUFS,		"No buffer space available."},
		{WSAENOPROTOOPT,	"Bad protocol option."},
		{WSAENOTCONN,		"Socket is not connected."},
		{WSAENOTSOCK,		"Socket operation on non-socket."},
		{WSAEOPNOTSUPP,		"Operation not supported."},
		{WSAEPFNOSUPPORT,	"Protocol family not supported."},
		{WSAEPROCLIM,		"Too many processes."},
		{WSAEPROTONOSUPPORT,"Protocol not supported."},
		{WSAEPROTOTYPE,		"Protocol wrong type for socket."},
		{WSAESHUTDOWN,		"Cannot send after socket shutdown."},
		{WSAESOCKTNOSUPPORT,"Socket type not supported."},
		{WSAETIMEDOUT,		"Connection timed out."},
		{WSAEWOULDBLOCK,	"Resource temporarily unavailable."},
		{WSAHOST_NOT_FOUND,	"Host not found."},
		{WSANOTINITIALISED,	"Successful WSAStartup not yet performed."},
		{WSANO_DATA,		"Valid name, no data record of requested type."},
		{WSANO_RECOVERY,	"This is a non-recoverable error."},
		{WSASYSNOTREADY,	"Network subsystem is unavailable."},
		{WSATRY_AGAIN,		"Non-authoritative host not found."},
		{WSAVERNOTSUPPORTED,"WINSOCK.DLL version out of range."},
		{WSAEDISCON,		"Graceful shutdown in progress."},
		{-1, 0}};

	int LastError = WSAGetLastError();
	ErrorMsg *Error = ErrorCodes;
	while (Error->Code >= 0 AND Error->Code != LastError)
	{
		Error++;
	}
	OnError(LastError, (Error->Code >= 0) ? Error->Msg : "<unknown error>");

	switch (LastError)
	{
		case WSAECONNABORTED:
		case WSAECONNRESET:
		case WSAENETRESET:
		{
			OnDisconnect();
			break;
		}
	}

	#elif defined BEOS
	
	int LastError = Ep->Error();
	OnError(LastError, Ep->ErrorStr());

	switch (LastError)
	{
		case ECONNABORTED:
		case ECONNRESET:
		case ENETRESET:
		{
			OnDisconnect();
			break;
		}
	}
	
	#else
	int LastError = 0;
	#endif
	
	return LastError;
}


//////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
#include <ras.h>
#endif

bool HaveNetConnection()
{
	bool Status = FALSE;

	// Check for dial up connection
	#if defined WIN32
	typedef DWORD (__stdcall *RasEnumConnections_Proc)(LPRASCONN lprasconn, LPDWORD lpcb, LPDWORD lpcConnections); 
	typedef DWORD (__stdcall *RasGetConnectStatus_Proc)(HRASCONN hrasconn, LPRASCONNSTATUS lprasconnstatus);

	HMODULE hRas = (HMODULE) LoadLibrary("rasapi32.dll");
	if (hRas)
	{
		RASCONN Con[10];
		DWORD Connections = 0;
		DWORD Bytes = sizeof(Con);

		ZeroObj(Con);
		Con[0].dwSize = sizeof(Con[0]);

		RasEnumConnections_Proc pRasEnumConnections = (RasEnumConnections_Proc) GetProcAddress(hRas, "RasEnumConnectionsA"); 
		RasGetConnectStatus_Proc pRasGetConnectStatus = (RasGetConnectStatus_Proc) GetProcAddress(hRas, "RasGetConnectStatusA"); 

		if (pRasEnumConnections AND
			pRasGetConnectStatus)
		{
			pRasEnumConnections(Con, &Bytes, &Connections);

			for (int i=0; i<Connections; i++)
			{
				RASCONNSTATUS Stat;
				char *Err = "";

				ZeroObj(Stat);
				Stat.dwSize = sizeof(Stat);

				DWORD Error = 0;				
				if (NOT (Error=pRasGetConnectStatus(Con[i].hrasconn, &Stat)))
				{
					if (Stat.rasconnstate == RASCS_Connected)
					{
						Status = TRUE;
					}
				}

				/*
				else
				{
					switch (Error)
					{
						case ERROR_NOT_ENOUGH_MEMORY:
						{
							Err = "Not enuf mem.";
							break;
						}
						case ERROR_INVALID_HANDLE:
						{
							Err = "Invalid Handle.";
							break;
						}
					}
				}

				int n=0;
				*/
			}
		}

		FreeLibrary(hRas);
	}
	#endif

	// Check for ethernet connection
	if (NOT Status)
	{
	}

	return Status;
}


///////////////////////////////////////////////////////////////////////////////////////////
#include <INet.h>
#if defined WIN32
typedef HOSTENT HostEnt;
#elif defined BEOS
typedef hostent HostEnt;
#endif

bool WhatsMyIp(char *IpAddr)
{
	bool Status = FALSE;

	strcpy(IpAddr, "(none)");
	GNetwork Sockets;
	if (Sockets.IsValid())
	{
		#if defined WIN32
		char hostname[256];
		HostEnt *e = NULL;
		if (gethostname(hostname, sizeof(hostname)) != SOCKET_ERROR)
		{
			e = gethostbyname(hostname);
		}

		if (e)
		{
			int Which = 0;
			for (; e->h_addr_list[Which]; Which++);
			Which--;

			sprintf(	IpAddr,
						"%i.%i.%i.%i",
						(uchar)e->h_addr_list[Which][0],
						(uchar)e->h_addr_list[Which][1],
						(uchar)e->h_addr_list[Which][2],
						(uchar)e->h_addr_list[Which][3]);
			Status = true;
		}
		#elif defined BEOS

		struct sockaddr_in addr;
		int sock, size;
		
		memset((char *)&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = 0;
		
		if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
			return false;
		
		if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
			return false;
		
		size = sizeof(addr);
		if ((getsockname(sock, (struct sockaddr *)&addr, &size)) < 0)
			return false;
		
		closesocket(sock);
		
		if (addr.sin_addr.s_addr == INADDR_ANY)
			return false;
		
		uchar *a = (uchar*)&addr.sin_addr.s_addr;
		sprintf(	IpAddr,
					"%i.%i.%i.%i",
					a[0],
					a[1],
					a[2],
					a[3]);
	
		Status = true;
		
		#endif
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////
#define SOCKS5_VER					5

#define SOCKS5_CMD_CONNECT			1
#define SOCKS5_CMD_BIND				2
#define SOCKS5_CMD_ASSOCIATE		3

#define SOCKS5_ADDR_IPV4			1
#define SOCKS5_ADDR_DOMAIN			3
#define SOCKS5_ADDR_IPV6			4

#define SOCKS5_AUTH_NONE			0
#define SOCKS5_AUTH_GSSAPI			1
#define SOCKS5_AUTH_USER_PASS		2

ISocks5Connection::ISocks5Connection(	char *proxy,
										int port,
										char *username,
										char *password)
{
	Socks5Connected = false;
	Proxy = NewStr(proxy);
	Port = port;
	UserName = NewStr(username);
	Password = NewStr(password);
}

ISocks5Connection::~ISocks5Connection()
{
	DeleteObj(Proxy);
	DeleteObj(UserName);
	DeleteObj(Password);
}

bool ISocks5Connection::Open(char *HostAddr, int port)
{
	bool Status = false;
	if (HostAddr)
	{
		char Msg[256];
		sprintf(Msg, "[SOCKS5] Connecting to proxy server '%s'", HostAddr);
		OnInformation(Msg);

		Status = IConnection::Open(Proxy, Port);
		if (Status)
		{
			char Buf[1024];
			Buf[0] = SOCKS5_VER;
			Buf[1] = 2; // methods
			Buf[2] = SOCKS5_AUTH_NONE;
			Buf[3] = SOCKS5_AUTH_USER_PASS;
			
			// No idea how to implement this.
			// AuthReq[3] = SOCKS5_AUTH_GSSAPI;

			OnInformation("[SOCKS5] Connected, Requesting authentication type.");
			IConnection::Write(Buf, 4, 0);
			if (IConnection::Read(Buf, 2, 0) == 2)
			{
				if (Buf[0] == SOCKS5_VER)
				{
					bool Authenticated = false;
					switch (Buf[1])
					{
						case SOCKS5_AUTH_NONE:
						{
							Authenticated = true;
							OnInformation("[SOCKS5] No authentication needed.");
							break;
						}
						case SOCKS5_AUTH_USER_PASS:
						{
							OnInformation("[SOCKS5] User/Pass authentication needed.");
							if (UserName AND Password)
							{
								char *b = Buf;
								*b++ = 1; // ver of sub-negotiation ??
								
								int NameLen = strlen(UserName);
								*b++ = NameLen;
								strcpy(b, UserName);
								b += NameLen;

								int PassLen = strlen(Password);
								*b++ = PassLen;
								strcpy(b, Password);
								b += PassLen;

								IConnection::Write(Buf, 3 + NameLen + PassLen, 0);
								if (IConnection::Read(Buf, 2, 0) == 2)
								{
									Authenticated = (Buf[0] == 1 AND Buf[1] == 0);
								}

								if (NOT Authenticated)
								{
									OnInformation("[SOCKS5] User/Pass authentication failed.");
								}
							}
							break;
						}
					}

					if (Authenticated)
					{
						OnInformation("[SOCKS5] Authentication successful.");

						int HostPort = htons(port);
						
						// Header
						char *b = Buf;
						*b++ = SOCKS5_VER;
						*b++ = SOCKS5_CMD_CONNECT;
						*b++ = 0; // reserved

						long IpAddr = inet_addr(HostAddr);
						if (IpAddr != -1)
						{
							// Ip
							*b++ = SOCKS5_ADDR_IPV4;
							memcpy(b, &IpAddr, 4);
							b += 4;
						}
						else
						{
							// Domain Name
							*b++ = SOCKS5_ADDR_DOMAIN;
							int Len = strlen(HostAddr);
							*b++ = Len;
							strcpy(b, HostAddr);
							b += Len;
						}

						// Port
						memcpy(b, &HostPort, 2);
						b += 2;

						IConnection::Write(Buf, (int)b - (int)Buf, 0);
						if (IConnection::Read(Buf, 10, 0) == 10)
						{
							if (Buf[0] == SOCKS5_VER)
							{
								switch (Buf[1])
								{
									case 0:
										Socks5Connected = true;
										OnInformation("[SOCKS5] Connected!");
										break;
									case 1:
										OnInformation("[SOCKS5] General SOCKS server failure");
										break;
									case 2:
										OnInformation("[SOCKS5] Connection not allowed by ruleset");
										break;
									case 3:
										OnInformation("[SOCKS5] Network unreachable");
										break;
									case 4:
										OnInformation("[SOCKS5] Host unreachable");
										break;
									case 5:
										OnInformation("[SOCKS5] Connection refused");
										break;
									case 6:
										OnInformation("[SOCKS5] TTL expired");
										break;
									case 7:
										OnInformation("[SOCKS5] Command not supported");
										break;
									case 8:
										OnInformation("[SOCKS5] Address type not supported");
										break;
									default:
										OnInformation("[SOCKS5] Unknown SOCKS server failure");
										break;
								}
							}
							else
							{
								OnInformation("[SOCKS5] Wrong socks version.");
							}
						}
						else
						{
							OnInformation("[SOCKS5] Connection request read failed.");
						}
					}
					else
					{
						OnInformation("[SOCKS5] Not authenticated.");
					}
				}
				else
				{
					OnInformation("[SOCKS5] Wrong socks version.");
				}
			}
			else
			{
				OnInformation("[SOCKS5] Authentication type read failed.");
			}

			Status = Socks5Connected;
			if (NOT Status)
			{
				IConnection::Close();
				OnInformation("[SOCKS5] Failure: Disconnecting.");
			}
		}
	}

	return Status;
}

