#include <string.h>

#include "VMem.h"
#include "INetTools.h"

// Gets a field value
char *InetGetField(char *s)
{
	char *e = s;

	while (*e)
	{
		if (*e == '\n')
		{
			if (NOT strchr(" \t", *(e+1)))
			{
				break;
			}
		}

		e++;
	}

	int Size = (int) e - (int) s;
	char *Str = NEW(char[Size+1]);
	if (Str)
	{
		char *In = s;
		char *Out = Str;

		while (In <= e)
		{
			if (*In == '\r')
			{
			}
			else if (*In == '\n')
			{
				if (*(In + 1) == 9)
				{
					In++;
				}
			}
			else
			{
				*Out++ = *In;
			}

			In++;
		}

		*Out++ = 0;
	}

	return Str;
}

char *SeekNextLine(char *s)
{
	if (s)
	{
		for (; *s AND *s != '\n'; s++);
		if (*s == '\n') s++;
	}

	return s;
}

// Search through headers for a field
char *InetGetHeaderField(	// Returns an allocated string or NULL on failure
	char *Headers,			// Pointer to all the headers
	char *Field,			// The field your after
	int Len)				// Maximum len to run, or -1 for NULL terminated
{
	if (Headers AND Field)
	{
		// for all lines
		for (char *s = Headers;
			*s AND
			(Len < 0 OR ((int)s-(int)Headers < Len));
			s = SeekNextLine(s))
		{
			if (*s != 9 AND
				strnicmp(s, Field, strlen(Field)) == 0)
			{
				// found a match
				s += strlen(Field);
				if (s[0] == ':' AND strchr(" \t", s[1]))
				{
					s += 2;
					return InetGetField(s);
				}
			}
		}
	}

	return 0;
}

void InetTokeniseHeaders(List<InetHeader> &Out, char *In)
{
	// loop through and extract all the headers
	char *e = 0;
	for (char *s=In; s AND *s; s = *e == '\n' ? e + 1 : e)
	{
		// skip whitespace
		while (strchr(" \t", *s)) s++;

		// get end-of-line
		char *e = strchr(s, '\n');
		if (NOT e) e = s + strlen(s);
		int Len = (int)e-(int)s;

		// process line
		char *eName = strnchr(s, ':', Len);
		if (eName)
		{
			InetHeader *h = NEW(InetHeader(NewStr(s, (int)eName-(int)s)));
			if (h)
			{
				// move on to text after the whitespace
				s = eName + 1;
				while (strchr(" \t", *s)) s++;

				// extract the data itself
				h->Str = InetGetField(s);
			}
		}
	}
}

char *InetRemoveField(char *Headers, char *Field)
{
	char *Status = Headers;

	if (Headers AND Field)
	{
		// loop through and look at all the headers
		char *e = 0;
		for (char *s=Headers; s AND *s; s = *e == '\n' ? e + 1 : e)
		{
			// skip whitespace
			while (strchr(" \t", *s)) s++;

			// get end-of-line
			e = strchr(s, '\n');
			if (NOT e) e = s + strlen(s);
			int Len = (int)e-(int)s;

			// process line
			char *eName = strnchr(s, ':', Len);
			if (eName)
			{
				if (strnicmp(s, Field, (int)eName-(int)s) == 0)
				{
					BytePipe Out;

					// found header... push headers before this one
					Out.Push((uchar*)Headers, (int)s-(int)Headers);

					// get end
					char *End = eName;
					for (char *eol = strchr(End, '\n'); eol; eol = strchr(eol+1, '\n'))
					{
						End = eol + 1;
						if (eol[1] != '\t') break;
					}

					// push text after this header
					Out.Push((uchar*)End, strlen(End));

					// replace text..
					DeleteArray(Headers);
					int Len = Out.Sizeof();
					Status = NEW(char[Len+1]);
					if (Status)
					{
						Out.Pop((uchar*)Status, Len);
						Status[Len] = 0;
					}

					// leave loop
					break;
				}
			}
		}
	}

	return Status;
}
