/*hdr
**      FILE:           GDocApp.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           20/4/01
**      DESCRIPTION:    Generic Document Application Impl
**
*/

#include "GDocApp.h"

#if defined WIN32
GWin32Class GDocApp::Win32Class;
#endif

GDocApp::GDocApp(char *appname, char *icon)
{
	// Default member vars
	_OptionsFile = 0;
	_Options = 0;
	_CurFile = 0;
	_FileMenu = 0;
	_Dirty = false;
	_AppName = NewStr(appname?appname:(char*)"LGI::GDocApp");

	// Default window pos
	GdcRegion r(100, 100, 600, 500);
	SetPos(r);

	// Setup class
	if (icon)
	{
		#if defined WIN32
		Win32Class.Class.hIcon = LoadIcon(hAppInst(), ((int)icon&0xffff0000)?icon:MAKEINTRESOURCE(icon));
		#endif
	}

}

GDocApp::~GDocApp()
{
	// Release memory
	DeleteArray(_OptionsFile);
	DeleteArray(_AppName);

	#ifdef BEOS
	LgiCloseApp();
	#endif
}

bool GDocApp::_DoSerialize(bool Write)
{
	bool Status = false;

	if (NOT _OptionsFile)
	{
		_OptionsFile = _OptionsFileName();
	}

	if (NOT _Options)
	{
		_Options = NEW(ObjProperties);
	}

	VFile f;
	if (f.Open(_OptionsFile, Write?O_WRITE:O_READ))
	{
		if (Write)
		{
			// clear existing contents
			f.SetSize(0);

			// Save window position
			#if defined WIN32
			if (NOT IsZoomed(Handle()) AND NOT IsIconic(Handle()))
			#endif
			{
				GdcRegion r = GetPos();
				_Options->Set("x", r.x1);
				_Options->Set("y", r.y1);
				_Options->Set("cx", r.X());
				_Options->Set("cy", r.Y());
			}
			
			// save misc options
			SerializeOptions(_Options, true);
			GMru::Serialize(_Options, "Mru", true);
		}

		// do the work
		Status = _SerializeFile(f, Write);

		if (NOT Write)
		{
			// read misc options
			GMru::Serialize(_Options, "Mru", false);
			SerializeOptions(_Options, false);

			// window pos
			GdcRegion r(100, 100, 500, 400);
			_Options->Get("x", r.x1);
			_Options->Get("y", r.y1);
			_Options->Get("cx", r.x2);
			_Options->Get("cy", r.y2);
			r.x2 += r.x1;
			r.y2 += r.y1;
			SetPos(r);
		}
	}

	return Status;
}

bool GDocApp::_SerializeFile(VFile &f, bool Write)
{
	return _Options->Serialize(f, Write);
}

void GDocApp::_Close()
{
	Empty();
	SetCurFile(0);
}

char *GDocApp::_OptionsFileName()
{
	// Get options file
	char *Status = 0;
	char Exe[256];
	LgiGetExeFile(Exe);
	char *File = strrchr(Exe, DIR_CHAR);
	if (File)
	{
		File++;
		#ifdef WIN32
		char *End = strchr(File, '.');
		if (End)
		{
			strcpy(End+1, "r");
		}
		else
		#endif
		{
			// unix apps have no '.' in their name
			strcat(File, ".r");
		}

		Status = LgiFindFile(File);
		if (NOT Status)
		{
			// No options file yet... so create one in the
			// app's directory
			Status = NewStr(Exe);
		}
	}

	return Status;
}

bool GDocApp::_Create()
{
	// Load options
	_DoSerialize(false);

	// Create and setup the main application window
	#ifdef WIN32
	Win32Class.Create(_AppName);
	if (Create(Win32Class.Name(), NULL))
	#endif
	{
		Name(_AppName);
		Visible(true);
	}

	return true;
}

bool GDocApp::_Destroy()
{
	// Save options
	_DoSerialize(true);

	return true;
}

bool GDocApp::_LoadMenu(char *Resource)
{
	Menu = NEW(GMenu);
	if (Menu)
	{
		#ifdef BEOS
		Menu->Attach(this);
		#endif

		if (Resource)
		{
			Menu->Load(this, Resource);
			_FileMenu = Menu->AppendSub("&File", 0);
		}
		else
		{
			_FileMenu = Menu->AppendSub("&File");
		}

		if (_FileMenu)
		{
			_FileMenu->AppendItem("&Open", IDM_OPEN, TRUE);
			_FileMenu->AppendItem("&Save", IDM_SAVE, TRUE);
			_FileMenu->AppendItem("Save &As", IDM_SAVEAS, TRUE);
			_FileMenu->AppendItem("&Close", IDM_CLOSE, TRUE);
			_FileMenu->AppendSeparator();

			GSubMenu *Recent = _FileMenu->AppendSub("Recent...");
			if (Recent)
			{
				Set(Recent);
				_FileMenu->AppendSeparator();
				GMru::Serialize(_Options, "Mru", false);
			}

			_FileMenu->AppendItem("E&xit", IDM_EXIT, TRUE);
		}
	}

	return Menu != 0;
}

bool GDocApp::_OpenFile(char *File)
{
	bool Status = false;
	if (SetDirty(false))
	{
		char RealPath[256];
		if (ResolveShortcut(File, RealPath, sizeof(RealPath)))
		{
			File = RealPath;
		}

		Status = GMru::_OpenFile(File);
		if (Status)
		{
			_Dirty = false;
			SetCurFile(File);
		}
	}

	return Status;
}

bool GDocApp::_SaveFile(char *File)
{
	char RealPath[256];
	if (ResolveShortcut(File, RealPath, sizeof(RealPath)))
	{
		File = RealPath;
	}

	bool Status = GMru::_SaveFile(File);
	if (Status)
	{
		_Dirty = false;
		SetCurFile(File);
	}

	return Status;
}

char *GDocApp::GetAppName()
{
	return _AppName;
}

void GDocApp::SetCurFile(char *f)
{
	if (f != _CurFile)
	{
		DeleteArray(_CurFile);
		_CurFile = NewStr(f);
	}

	char s[256];
	if (_CurFile)
	{
		sprintf(s, "%s [%s%s]", _AppName, _CurFile, _Dirty ? " changed" : "");
	}
	else if (_Dirty)
	{
		sprintf(s, "%s [changed]", _AppName);
	}
	else
	{
		strcpy(s, _AppName);
	}

	Name(s);
}

char *GDocApp::GetCurFile()
{
	return _CurFile;
}

bool GDocApp::SetDirty(bool Dirty)
{
	if (_Dirty ^ Dirty)
	{
		// Changing...
		if (Dirty)
		{
			// Setting dirty
			_Dirty = true;
			SetCurFile(_CurFile);
		}
		else
		{
			// Clearing dirty
			int Result = LgiMsg(this, "Do you want to save your changes?", _AppName, MB_YESNOCANCEL);
			if (Result == IDYES)
			{
				_SaveFile(_CurFile);
			}
			else if (Result == IDCANCEL)
			{
				return false;
			}

			_Dirty = false;
			SetCurFile(_CurFile);
		}
	}

	return true;
}

bool GDocApp::GetDirty()
{
	return _Dirty;
}

ObjProperties *GDocApp::GetOptions()
{
	return _Options;
}

void GDocApp::OnReceiveFiles(List<char> &Files)
{
	char *f = Files.First();
	if (f AND _OpenFile(f))
	{
		GMru::AddFile(f);
	}
}

bool GDocApp::OnRequestClose()
{
	return SetDirty(false);
}

int GDocApp::OnCommand(int Cmd, int Event, GWndRef Window)
{
	switch (Cmd)
	{
		case IDM_SAVE:
		{
			if (NOT GetCurFile())
			{
				GMessage m = CreateMsg(WM_COMMAND, IDM_SAVEAS, 0);
				GMru::OnEvent(&m);
			}
			else
			{
				_SaveFile(GetCurFile());
			}
			break;
		}
		case IDM_CLOSE:
		{
			_Close();
			break;
		}
		case IDM_EXIT:
		{
			LgiCloseApp();
			break;
		}
	}

	return 0;
}

int GDocApp::OnEvent(GMessage *m)
{
	GMru::OnEvent(m);
	return GWindow::OnEvent(m);
}



