/*hdr
**      FILE:           GList.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           14/2/2000
**      DESCRIPTION:    Lgi self-drawn listbox
**
**      Copyright (C) 2000 Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

#ifdef WIN32
GWin32Class GList::WndClass;
#endif

// #define DOUBLE_BUFFER_PAINT

#define DRAG_NONE					0
#define SELECT_ITEMS				1
#define DRAG_COLUMN					2
#define CLICK_COLUMN				3
#define TOGGLE_ITEMS				4
#define CLICK_ITEM					5

#define DRAG_THRESHOLD				4

#define IDC_VS						1000

#define ForAllItems(Var)			Iterator<GListItem> ItemIter(&Items); for (GListItem *Var = ItemIter.First(); Var; Var = ItemIter.Next())
#define ForAllItemsReverse(Var)		Iterator<GListItem> ItemIter(&Items); for (GListItem *Var = ItemIter.Last(); Var; Var = ItemIter.Prev())
#define VisibleItems()				(LastVisible - FirstVisible + 1)
#define MaxScroll()					(Items.GetItems() - VisibleItems())

////////////////////////////////////////////////////////////////////////////////////////////
// List column
GListColumn::GListColumn(GList *parent, char *name, int width)
#ifdef LGI_RES
	: ResObject(Res_Column)
#endif
{
	Parent = parent;
	cName = NewStr(name);
	cWidth = width;
	cIcon = 0;
	cType = GIC_ASK_TEXT;
	cImage = -1;
	cMark = GLI_MARK_NONE;
	Down = false;
	OwnIcon = false;
	CanResize = true;
}

GListColumn::~GListColumn()
{
	DeleteObj(cName);
	if (OwnIcon)
	{
		DeleteObj(cIcon);
	}
}

void GListColumn::Name(char *n)
{
	DeleteArray(cName);
	cName = NewStr(n);
	if (Parent)
	{
		Parent->Invalidate(&Parent->ColumnHeader);
	}
}

char *GListColumn::Name()
{
	return cName;
}

void GListColumn::Width(int i)
{
	cWidth = i;
	if (Parent)
	{
		Parent->Invalidate(&Parent->ColumnHeader);
	}
}

int GListColumn::Width()
{
	return cWidth;
}

void GListColumn::Mark(int i)
{
	cMark = i;
	if (Parent)
	{
		Parent->Invalidate(&Parent->ColumnHeader);
	}
}

int GListColumn::Mark()
{
	return cMark;
}

void GListColumn::Type(int i)
{
	cType = i;
	if (Parent)
	{
		Parent->Invalidate(&Parent->ColumnHeader);
	}
}

int GListColumn::Type()
{
	return cType;
}

void GListColumn::Icon(GdcBasePrimitives *i, bool Own)
{
	if (OwnIcon)
	{
		DeleteObj(cIcon);
	}
	cIcon = i;
	OwnIcon = Own;

	if (Parent)
	{
		Parent->Invalidate(&Parent->ColumnHeader);
	}
}

GdcBasePrimitives *GListColumn::Icon()
{
	return cIcon;
}

void GListColumn::OnPaint(GdcBasePrimitives *pDC, GdcRegion &Rgn)
{
	GdcRegion r = Rgn;

	int Off = 0;
	if (Down)
	{
		Off = 2;
		DrawMoveShallowEdge(pDC, r, SUNKEN);
	}
	else
	{
		DrawMoveDeepEdge(pDC, r, RAISED);
	}

	int Mx = r.x1 + 8, My = r.y1 + ((r.Y() - 8) / 2);
	if (cIcon)
	{
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle(&r);

		int x = (r.X()-cIcon->X()) / 2;
		pDC->Blt(	r.x1 + x + Off,
					r.y1 + ((r.Y()-cIcon->Y())/2) + Off,
					cIcon);

		if (cMark)
		{
			Mx += x + cIcon->X() + 4;
		}
	}
	else if (cImage >= 0 AND Parent)
	{
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle(&r);
		if (Parent->ImageList)
		{
			Parent->ImageList->Draw(pDC,
									r.x1 + Off,
									r.y1 + ((r.Y()-Parent->ImageList->TileY())/2) + Off,
									cImage, 0);
		}

		if (cMark)
		{
			Mx += Parent->ImageList->TileX() + 4;
		}
	}
	else if (ValidStr(cName))
	{
		SysFont->Transparent(false);
		SysFont->Colour(LC_TEXT, LC_MED);
		SysFont->Text(pDC,	r.x1 + Off + 3,
							r.y1 + Off, cName, -1, &r);

		if (cMark)
		{
			Mx += SysFont->X(cName);
		}
	}
	else
	{
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle(&r);
	}

	#define ARROW_SIZE	9
	pDC->Colour(0);
	Mx += Off;
	My += Off;

	switch (cMark)
	{
		case GLI_MARK_UP_ARROW:
		{
			pDC->Line(Mx + 2, My, Mx + 2, My + ARROW_SIZE - 1);
			pDC->Line(Mx, My + 2, Mx + 2, My);
			pDC->Line(Mx + 2, My, Mx + 4, My + 2);
			break;
		}
		case GLI_MARK_DOWN_ARROW:
		{
			pDC->Line(Mx + 2, My, Mx + 2, My + ARROW_SIZE - 1);
			pDC->Line(	Mx,
						My + ARROW_SIZE - 3,
						Mx + 2,
						My + ARROW_SIZE - 1);
			pDC->Line(	Mx + 2,
						My + ARROW_SIZE - 1,
						Mx + 4,
						My + ARROW_SIZE - 3);
			break;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
// List item
GListItem::GListItem()
{
	Parent = 0;
	Selected = 0;
	ListItem_Image = -1;
}

GListItem::~GListItem()
{
}

/*
Calling this to store your data is optional. Just override the
"GetText" function to return your own data to avoid duplication
in memory.
*/
bool GListItem::SetText(char *s, int i)
{
	if (s AND i >= 0)
	{
		// Insert missing columns
		while (Data.GetItems() < i)
		{
			Data.Insert(NewStr(""));
		}

		// Delete any existing column
		char *Old = Data.ItemAt(i);
		if (Old)
		{
			Data.Delete(Old);
			DeleteArray(Old);
		}

		// Add new string in
		Data.Insert(NewStr(s), i);

		return true;
	}

	return false;
}

// User can override this if they want to use their own data
char *GListItem::GetText(int i)
{
	return Data.ItemAt(i);
}

bool GListItem::IsSelected()
{
	return Selected;
}

void GListItem::Select(bool b)
{
	if (Selected != b)
	{
		Selected = b;
		Update();
		
		if (Parent AND Selected)
		{
			Parent->OnItemSelect(this);
		}
	}
}

void GListItem::ScrollTo()
{
	if (Parent AND Parent->VScroll)
	{
		int n = Parent->Items.IndexOf(this);
		if (n < Parent->FirstVisible)
		{
			Parent->VScroll->SetPos(n);
			Parent->Invalidate(&Parent->ItemsPos);
		}
		else if (n > Parent->LastVisible)
		{
			Parent->VScroll->SetPos(n - (Parent->LastVisible - Parent->FirstVisible));
			Parent->Invalidate(&Parent->ItemsPos);
		}
	}
}

void GListItem::Update()
{
	if (Parent)
	{
		GMeasureInfo Info;
		OnMeasure(&Info);
		
		GdcRegion r = Pos;
		if (r.Valid())
		{
			if (Info.y != r.Y())
			{
				r.y2 = Parent->ItemsPos.y2;
			}
			
			Parent->Invalidate(&r);
		}
	}
}

void GListItem::OnMeasure(GMeasureInfo *Info)
{
	if (Info)
	{
		Info->x = 1024;
		Info->y = max(16, SysFont->Y("A") + 2); // the default width
	}
}

bool GListItem::GridLines()
{
	return (Parent) ? Parent->GridLines : false;
}

void GListItem::OnPaintColumn(GdcBasePrimitives *pDC, GdcRegion &r, int i, GListColumn *c)
{
	if (pDC AND c)
	{
		GdcRegion ng = r; // non-grid area
		if (GridLines())
		{
			ng.x2--;
			ng.y2--;
		}

		if (c->Type() == GIC_ASK_TEXT)
		{
			char *Text = GetText(i);
			SysFont->TabSize(0);
			SysFont->Transparent(false);
			SysFont->Colour(Fore, Back);
			SysFont->Text(pDC, r.x1+1, r.y1+1, Text?Text:(char*)"", -1, &ng);
		}
		else
		{
			pDC->Colour(Back, 24);
			pDC->Rectangle(&ng);

			if (c->Type() == GIC_ASK_IMAGE AND
				Parent->ImageList)
			{
				int Img = GetImage();
				if (Img >= 0)
				{
					Parent->ImageList->Draw(pDC, r.x1+1, r.y1, GetImage(), 0);
				}
			}
		}

		if (GridLines())
		{
			pDC->Colour(LC_LOW, 24);
			pDC->Line(r.x1, r.y2, r.x2, r.y2);
			pDC->Line(r.x2, r.y1, r.x2, r.y2);
		}
	}
}

void GListItem::OnPaint(GdcBasePrimitives *pDC, GdcRegion &r)
{
	if (Parent)
	{
		int i=0;
		int x = r.x1;

		// Icon?
		if (Parent->IconCol)
		{
			GdcRegion cr(x, r.y1, x + Parent->IconCol->Width()-1, r.y2);

			// icon background
			pDC->Colour(Selected ? Parent->_Selection : LC_WORKSPACE, 24);
			pDC->Rectangle(&cr);

			// draw icon
			OnPaintColumn(pDC, cr, -1, Parent->IconCol);
			x = cr.x2 + 1;
		}

		// draw columns
		for (GListColumn *c = Parent->Columns.First(); c; c = Parent->Columns.Next(), i++)
		{
			GdcRegion cr(x, r.y1, x + c->Width()-1, r.y2);
			OnPaintColumn(pDC, cr, i, c);
			x = cr.x2 + 1;
		}

		// after columns
		if (x < r.x2)
		{
			pDC->Colour(Back, 24);
			pDC->Rectangle(x, r.y1, r.x2, r.y2);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// List control
GList::GList(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_ListView)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	Name(name);
	
	Buf = 0;
	GridLines = false;
	FirstVisible = -1;
	LastVisible = -1;
	ColumnHeaders = true;
	EditLabels = false;
	MultiItemSelect = true;
	DragMode = 0;
	DragData = 0;
	IconCol = 0;
	Keyboard = -1;
	Sunken(true);

	#if defined WIN32
	WndDlgCode = DLGC_WANTARROWS;
	WndClass.Create(Class());
	#elif defined BEOS
	GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
	#endif
}

GList::~GList()
{
	DeleteObj(Buf);

	Empty();
	EmptyColumns();
}

void GList::OnItemClick(GListItem *Item, VMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GList::OnItemBeginDrag(GListItem *Item, VMouse &m)
{
	if (Item)
	{
		Item->OnBeginDrag(m);
	}
}

void GList::OnItemSelect(GListItem *Item)
{
	if (Item)
	{
		Keyboard = Items.IndexOf(Item);
		Item->OnSelect();
	}
}

GListColumn *GList::AddColumn(char *Name, int Width, int Where)
{
	GListColumn *c = NEW(GListColumn(this, Name, Width));
	if (c)
	{
		Columns.Insert(c, Where);
	}
	return c;
}

bool GList::AddColumn(GListColumn *Col, int Where)
{
	if (Col)
	{
		return Columns.Insert(Col, Where);
	}
	return false;
}

bool GList::DeleteColumn(GListColumn *Col)
{
	if (Col AND
		Columns.HasItem(Col))
	{
		Columns.Delete(Col);
		DeleteObj(Col);
		Invalidate(&ColumnHeader);
		return true;
	}

	return false;
}

void GList::EmptyColumns()
{
	for (GListColumn *c = Columns.First(); c; c = Columns.Next())
	{
		DeleteObj(c);
	}
	Columns.Empty();
	DeleteObj(IconCol);
	Invalidate(&ColumnHeader);
}

int GList::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		case WM_CHANGE:
		{
			OnNotify((GView*)MessageA(Msg), MessageB(Msg));
			break;
		}
	}

	return GLayout::OnEvent(Msg);
}

int GList::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl->GetId() == IDC_VS AND VScroll)
	{
		Invalidate(&ItemsPos);
	}

	return 0;
}

GdcRegion &GList::GetClientRect()
{
	static GdcRegion r;

	r = Pos;
	r.Offset(-r.x1, -r.y1);
	#ifdef BEOS
	if (Sunken() OR Raised())
	{
		r.Size(2, 2);
	}
	#endif
	return r;
}

GListItem *GList::HitItem(int x, int y, int *Index)
{
	int n=0;
	ForAllItems(i)
	{
		if (y >= i->Pos.y1 AND y <= i->Pos.y2)
		{
			if (Index) *Index = n;
			return i;
		}
		n++;
	}

	return NULL;
}

void GList::KeyScroll(int iTo, int iFrom, bool SelectItems)
{
	int Start = -1, End = -1, i = 0;
	{
		ForAllItems(n)
		{
			if (n->IsSelected())
			{
				if (Start < 0)
				{
					Start = i;
				}
			}
			else if (Start >= 0 AND
					End < 0)
			{
				End = i - 1;
			}

			i++;
		}

		if (End < 0) End = i - 1;
	}

	iTo = limit(iTo, 0, Items.GetItems()-1);
	iFrom = limit(iFrom, 0, Items.GetItems()-1);
	GListItem *To = Items.ItemAt(iTo);
	GListItem *From = Items.ItemAt(iFrom);
	int Inc = (iTo < iFrom) ? -1 : 1;

	if (To AND From AND iTo != iFrom)
	{
		GListItem *Item = 0;
		if (SelectItems)
		{
			int OtherEnd = Keyboard == End ? Start : End;
			int Min = min(OtherEnd, iTo);
			int Max = max(OtherEnd, iTo);
			i = 0;

			ForAllItems(n)
			{
				n->Select(i>=Min AND i<=Max);
				i++;
			}
		}
		else
		{
			Select(To);
		}

		To->ScrollTo();

		Keyboard = iTo;
	}
}

void GList::OnMouseWheel(double Lines)
{
	if (VScroll)
	{
		int Max = MaxScroll();
		int NewPos = (int) limit(VScroll->Pos() + Lines, 0, Max);
		if (NewPos != VScroll->Pos())
		{
			VScroll->SetPos(NewPos);
			Invalidate(&ItemsPos);
		}
	}
}

void GList::OnKey(VKey &k)
{
	if (k.Down)
	{
		switch (k.c)
		{
			case VK_RETURN:
			{
				GView *Note = (GetNotify())?GetNotify():GetParent();
				if (Note)
				{
					Note->OnNotify(this, k.c);
				}
				break;
			}
			case 'A':
			case 'a':
			{
				if (k.Ctrl())
				{
					SelectAll();
				}
				break;
			}
			case VK_UP:
			{
				// int i = Value();
				KeyScroll(Keyboard-1, Keyboard, k.Shift());
				break;
			}
			case VK_DOWN:
			{
				KeyScroll(Keyboard+1, Keyboard, k.Shift());
				break;
			}
			case VK_PRIOR:
			{
				int Vis = VisibleItems();
				Vis = max(Vis, 0);

				KeyScroll(Keyboard-Vis, Keyboard, k.Shift());
				break;
			}
			case VK_NEXT:
			{
				int Vis = VisibleItems();
				Vis = max(Vis, 0);
				KeyScroll(Keyboard+Vis, Keyboard, k.Shift());
				break;
			}
			case VK_END:
			{
				// int i = Value();
				KeyScroll(Items.GetItems()-1, Keyboard, k.Shift());
				break;
			}
			case VK_HOME:
			{
				KeyScroll(0, Keyboard, k.Shift());
				break;
			}
		}
	}

	GListItem *Item = GetSelection();
	if (Item)
	{
		Item->OnKey(k);
	}
}

void GList::OnMouseClick(VMouse &m)
{
	if (LockLooper())
	{
		if (m.Down())
		{
			Focus(true);

			DragMode = DRAG_NONE;
			DragStart.x = m.x;
			DragStart.y = m.y;

			if (ColumnHeaders AND
				ColumnHeader.Overlap(m.x, m.y))
			{
				// Clicked on a column heading
				int x = ColumnHeader.x1 + ((IconCol) ? IconCol->Width() : 0);
				for (GListColumn *c = Columns.First(); c; c = Columns.Next())
				{
					x += c->Width();
					if (abs(m.x-x) < 6)
					{
						if (c->Resizable())
						{
							DragMode = DRAG_COLUMN;
							DragData = Columns.IndexOf(c);
							Capture(true);
						}
						break;
					}
					else if (c->Pos.Overlap(m.x, m.y))
					{
						DragMode = CLICK_COLUMN;
						DragData = Columns.IndexOf(c);
						
						c->Down = true;
						Invalidate(&c->Pos);
						
						Capture(true);
						break;
					}
				}
			}
			else if (ItemsPos.Overlap(m.x, m.y))
			{
				// Clicked in the items area
				int ItemIndex = -1;
				GListItem *Item = HitItem(m.x, m.y, &ItemIndex);
				DragData = ItemIndex;

				if (Item AND Item->IsSelected())
				{
					// Click on selected item
					if (m.Ctrl())
					{
						Item->Select(false);
						OnItemClick(Item, m);
					}
					else
					{
						// Could be drag'n'drop operation
						// Or just a select
						DragMode = CLICK_ITEM;

						SetPulse(100);
						Capture(true);
						OnItemClick(Item, m);

						if (NOT IsCapturing())
						{
							DragMode = DRAG_NONE;
						}
					}
				}
				else
				{
					// Selection change
					if (m.Shift() AND MultiSelect())
					{
						int n = 0;
						int a = min(ItemIndex, Keyboard);
						int b = max(ItemIndex, Keyboard);

						ForAllItems(i)
						{
							i->Select(n >= a AND n <= b);
							n++;
						}
					}
					else
					{
						ForAllItems(i)
						{
							if (Item == i) // clicked item
							{
								if (m.Ctrl())
								{
									// Toggle selected state
									if (NOT i->IsSelected())
									{
										Keyboard = Items.IndexOf(i);
									}

									i->Select(NOT i->IsSelected());
								}
								else
								{
									// Select this
									i->Select(true);
									Keyboard = Items.IndexOf(i);
								}
							}
							else if (NOT m.Ctrl() OR NOT MultiSelect())
							{
								i->Select(false);
							}
						}

						if (NOT m.Ctrl())
						{
							DragMode = SELECT_ITEMS;
							SetPulse(100);
							Capture(true);
						}
					}

					OnItemClick(Item, m);
				}
			}
		}
		else // Up Click
		{
			switch (DragMode)
			{
				case CLICK_COLUMN:
				{
					GListColumn *c = Columns.ItemAt(DragData);
					if (c)
					{
						c->Down = false;
						Invalidate(&c->Pos);

						if (c->Pos.Overlap(m.x, m.y))
						{
							OnColumnClick(Columns.IndexOf(c), m);
						}
					}
					break;
				}
				case CLICK_ITEM:
				{
					GListItem *Item = Items.ItemAt(DragData);
					if (Item)
					{
						ForAllItems(i)
						{
							i->Select(Item == i);
						}
					}
					break;
				}
			}

			if (IsCapturing())
			{
				Capture(false);
			}

			DragMode = DRAG_NONE;
		}

		UnlockLooper();
	}
}

void GList::OnPulse()
{
	if (LockLooper())
	{
		if (IsCapturing())
		{
			VMouse m;
			GetMouse(m);

			if (m.y < 0 OR m.y >= Y())
			{
				switch (DragMode)
				{
					case SELECT_ITEMS:
					{
						int Count = 0;
						if (m.y < 0)
						{
							int Space = -m.y;
							for (GListItem *i = Items.ItemAt(FirstVisible-1); i; i=Items.Prev())
							{
								GMeasureInfo Info;
								i->OnMeasure(&Info);
								if (Space > Info.y)
								{
									Count--;
									Space -= Info.y;
								}
								else
								{
									break;
								}
							}

						}
						else if (m.y >= Y())
						{
							int Space = m.y - Y();
							for (GListItem *i = Items.ItemAt(LastVisible+1); i; i=Items.Next())
							{
								GMeasureInfo Info;
								i->OnMeasure(&Info);
								if (Space > Info.y)
								{
									Count++;
									Space -= Info.y;
								}
								else
								{
									break;
								}
							}
						}

						if (Count != 0)
						{
							VScroll->SetPos(VScroll->Pos() + Count);
							Invalidate(&ItemsPos);
						}
						break;
					}
				}
			}
		}
		else
		{
			DragMode = DRAG_NONE;
		}

		UnlockLooper();
	}
}

void GList::OnMouseMove(VMouse &m)
{
	if (LockLooper())
	{
		switch (DragMode)
		{
			case DRAG_COLUMN:
			{
				GListColumn *c = Columns.ItemAt(DragData);
				if (c)
				{
					int OldWidth = c->Width();
					int NewWidth = m.x - c->Pos.x1;
					c->Width(max(NewWidth, 4));
					GdcRegion Up=ItemsPos;
					Up.x1 = c->Pos.x1 + min(OldWidth, c->Width()) - 1;
					Invalidate(&Up);
				}
				break;
			}
			case CLICK_COLUMN:
			{
				bool Update = false;
				GListColumn *c = Columns.ItemAt(DragData);
				if (c)
				{
					bool Over = c->Pos.Overlap(m.x, m.y);
					if (m.Down() AND Over != c->Down)
					{
						c->Down = Over;
						Invalidate(&c->Pos);
					}
				}
				break;
			}
			case SELECT_ITEMS:
			{
				int n=0;
				bool Selected = m.y < ItemsPos.y1;

				if (IsCapturing())
				{
					if (MultiItemSelect)
					{
						ForAllItems(i)
						{
							bool Over = (m.y >= i->Pos.y1) AND (m.y <= i->Pos.y2);
							int EndPoint = (Over ? 1 : 0) +
										   ((n == DragData) ? 1 : 0);

							if (Over)
							{
								Keyboard = n;
							}
							if (EndPoint == 1)
							{
								Selected = NOT Selected;
								i->Select(true);
							}
							else if (EndPoint == 2)
							{
								i->Select(true);
								Keyboard = n;
							}
							else
							{
								i->Select(Selected);
							}
							n++;
						}
					}
					else
					{
						ForAllItems(i)
						{
							i->Select(m.y >= i->Pos.y1 AND m.y <= i->Pos.y2);
						}
					}
				}
				break;
			}
			case CLICK_ITEM:
			{
				if (abs(DragStart.x-m.x) > DRAG_THRESHOLD OR
					abs(DragStart.y-m.y) > DRAG_THRESHOLD)
				{
					OnItemBeginDrag(Items.ItemAt(DragData), m);
					DragMode = DRAG_NONE;
					Capture(false);
				}
				break;
			}
		}

		UnlockLooper();
	}
}

int GList::Value()
{
	int n=0;
	ForAllItems(i)
	{
		if (i->IsSelected())
		{
			return n;
		}
		n++;
	}
	return -1;
}

void GList::Value(int Index)
{
	int n=0;
	ForAllItems(i)
	{
		if (n == Index)
		{
			i->Select(true);
			Keyboard = n;
		}
		else
		{
			i->Select(false);
		}
		n++;
	}
}

void GList::SelectAll()
{
	if (LockLooper())
	{
		ForAllItems(i)
		{
			i->Selected = true;
		}

		UnlockLooper();

		Invalidate();
	}
}

bool GList::Select(GListItem *Obj)
{
	bool Status = false;
	ForAllItems(i)
	{
		i->Select(Obj == i);
		if (Obj == i) Status = true;
	}
	return true;
}

GListItem *GList::GetSelection()
{
	ForAllItems(i)
	{
		if (i->IsSelected())
		{
			return i;
		}
	}

	return 0;
}

bool GList::GetSelection(List<GListItem> &l)
{
	ForAllItems(i)
	{
		if (i->IsSelected())
		{
			l.Insert(i);
		}
	}

	return l.GetItems() > 0;
}

bool GList::GetAll(List<GListItem> &l)
{
	ForAllItems(i)
	{
		l.Insert(i);
	}

	return true;
}

bool GList::GetItemRect(GListItem *i, GdcRegion &r)
{
	int Index = Items.IndexOf(i);
	if (Index >= -1 AND
		Index >= FirstVisible AND
		Index <= LastVisible)
	{
		r = i->Pos;
		return true;
	}

	return false;
}

bool GList::LockLooper()
{
	if (IsAttached())
	{
		#ifdef BEOS
		return GetBView()->LockLooper();
		#else
		return Lock();
		#endif
	}
	return true;
}

void GList::UnlockLooper()
{
	if (IsAttached())
	{
		#ifdef BEOS
		GetBView()->UnlockLooper();
		#else
		Unlock();
		#endif
	}
}

bool GList::Insert(GListItem *i, int Index, bool Update)
{
	bool Status = false;

	if (i AND LockLooper())
	{
		// Insert
		i->Parent = this;
		i->Select(false);
		Items.Insert(i, Index);

		if (Update)
		{
			// Update screen
			Pour();
			GdcRegion Up;
			if (GetItemRect(i, Up))
			{
				Up.y2 = Pos.Y();
				Invalidate(&Up);
			}

			// Notify
			GView *Note = (GetNotify())?GetNotify():GetParent();
			if (Note)
			{
				Note->OnNotify(this, 0);
			}
		}

		Status = true;

		UnlockLooper();
	}

	return Status;
}

bool GList::Delete()
{
	return Delete(Items.Current());
}

bool GList::Delete(int Index)
{
	return Delete(Items.ItemAt(Index));
}

bool GList::Delete(GListItem *i)
{
	bool Status = false;
	if (LockLooper())
	{
		if (Remove(i))
		{
			// Delete
			DeleteObj(i);
			Status = true;
		}
		UnlockLooper();
	}

	return Status;
}

bool GList::Remove(GListItem *i)
{
	bool Status = false;

	if (LockLooper())
	{
		if (i)
		{
			GdcRegion Up;
			bool Visible = GetItemRect(i, Up);
			int Index = Items.IndexOf(i);
			int Pos = (VScroll) ? VScroll->Pos() : 0;

			// Remove from list
			Items.Delete(i);
			UpdateScrollBars();

			// Update screen
			if ((VScroll AND VScroll->Pos() != Pos) OR
				Index < FirstVisible)
			{
				Invalidate(&ItemsPos);
			}
			else if (Visible)
			{
				Up.y2 = ItemsPos.y2;
				Invalidate(&Up);
			}

			// Notify
			GView *Note = (GetNotify())?GetNotify():GetParent();
			if (Note)
			{
				Note->OnNotify(this, 0);
			}

			Status = true;
		}

		UnlockLooper();
	}

	return Status;
}

bool GList::HasItem(GListItem *Obj)
{
	return Items.HasItem(Obj);
}

int GList::IndexOf(GListItem *Obj)
{
	return Items.IndexOf(Obj);
}

GListItem *GList::ItemAt(int Index)
{
	return Items.ItemAt(Index);
}

void GList::ScrollToSelection()
{
	if (VScroll)
	{
		int n=0;
		int Vis = VisibleItems();
		ForAllItems(i)
		{
			if (i->IsSelected())
			{
				if (n < FirstVisible OR n > LastVisible)
				{
					int k = n - (Vis/2);
					VScroll->SetPos(max(k, 0));
					Invalidate(&ItemsPos);
					break;
				}
			}
			n++;
		}
	}
}

void GList::Sort(GListCompareFunc Compare, int Data)
{
	if (LockLooper())
	{
		if (Compare)
		{
			Items.Sort(Compare, Data);
			Invalidate(&ItemsPos);
		}

		UnlockLooper();
	}
}

void GList::Empty()
{
	if (LockLooper())
	{
		for (GListItem *i = Items.First(); i; i = Items.Next())
		{
			DeleteObj(i);
		}
		Items.Empty();

		FirstVisible = LastVisible = -1;

		if (VScroll)
		{
			VScroll->SetPos(0);
			VScroll->SetLimits(0, -1);
		}

		Invalidate();
		UnlockLooper();
	}
}

void GList::RemoveAll()
{
	if (LockLooper())
	{
		Items.Empty();
		FirstVisible = LastVisible = -1;

		if (VScroll)
		{
			// these have to be in this order because
			// "SetLimits" can cause the VScroll object to
			// be deleted and becoming NULL
			VScroll->SetPos(0);
			VScroll->SetLimits(0, -1);
		}

		Invalidate();
		UnlockLooper();
	}
}

void GList::OnPosChange()
{
	Pour();
}

void GList::UpdateScrollBars()
{
	if (VScroll)
	{
		int Vis = VisibleItems();
		int Max = MaxScroll();
		if (VScroll->Pos() > max(Max, 0))
		{
			VScroll->SetPos(Max);
		}

		VScroll->SetId(IDC_VS);
		VScroll->SetPage(Vis);
		VScroll->SetLimits(0, Items.GetItems() - 1);
	}
}

void GList::Pour()
{
	// Layout all the elements
	GdcRegion Client = GetClient();
	Client.Offset(-Client.x1, -Client.y1);

	// printf("GList::Pour(%i, %i, %i, %i)\n", Client.x1, Client.y1, Client.x2, Client.y2);

	if (ColumnHeaders)
	{
		ColumnHeader = Client;
		ColumnHeader.y2 = ColumnHeader.y1 + SysFont->Y("A") + 4;
		ItemsPos = Client;
		ItemsPos.y1 = ColumnHeader.y2 + 1;
	}
	else
	{
		ItemsPos = Client;
	}

	int n = 0;
	int y = ItemsPos.y1;
	int Max = MaxScroll();
	if (Max < 0) Max = 0;
	FirstVisible = (VScroll) ? VScroll->Pos() : 0;
	if (FirstVisible < 0) FirstVisible = 0;
	if (FirstVisible > Max) FirstVisible = Max;
	LastVisible = 0x7FFFFFFF;
	bool SomeHidden = false;
	
	ForAllItems(i)
	{
		if (n < FirstVisible OR n > LastVisible)
		{
			i->Pos.Set(-1, -1, -2, -2);
			SomeHidden = true;
		}
		else
		{
			GMeasureInfo Info;
			
			i->OnMeasure(&Info);
			i->Pos.Set(ItemsPos.x1, y, ItemsPos.x2, y+Info.y-1);
			y = y+Info.y;
			if (i->Pos.y2 > ItemsPos.y2)
			{
				LastVisible = n - 1;
			}
		}

		n++;
	}

	if (LastVisible >= Items.GetItems())
	{
		LastVisible = Items.GetItems() - 1;
	}

	SetScrollBars(false, SomeHidden);
	UpdateScrollBars();
}

void GList::OnPaint(GdcBasePrimitives *pDC)
{
	if (LockLooper())
	{
		COLOUR Back = Enabled() ? LC_WORKSPACE : LC_MED;
		GdcRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
		Pour();

		// Check icon column status then draw
		if (AskImage() AND NOT IconCol)
		{
			IconCol = NEW(GListColumn(this, 0, 18));
			if (IconCol)
			{
				IconCol->Resizable(false);
				IconCol->Type(GIC_ASK_IMAGE);
			}
		}
		else if (NOT AskImage() AND IconCol) DeleteObj(IconCol);

		// Draw column headings
		if (ColumnHeaders)
		{
			// Draw columns
			GdcRegion cr = ColumnHeader;
			int cx = cr.x1;
			pDC->ClipRgn(&cr);

			if (IconCol)
			{
				cr.x1 = cx;
				cr.x2 = cr.x1 + IconCol->Width() - 1;
				IconCol->Pos = cr;
				IconCol->OnPaint(pDC, cr);
				cx += IconCol->Width();
			}

			// Draw other columns
			for (GListColumn *c = Columns.First(); c; c = Columns.Next())
			{
				cr.x1 = cx;
				cr.x2 = cr.x1 + c->Width() - 1;
				c->Pos = cr;
				c->OnPaint(pDC, cr);
				cx += c->Width();
			}

			// Draw ending peice
			cr.x1 = cx;
			cr.x2 = ColumnHeader.x2 + 2;
			DrawMoveDeepEdge(pDC, cr, RAISED);
			pDC->Colour(LC_MED, 24);
			pDC->Rectangle(&cr);

			pDC->ClipRgn(0);
		}

		// Draw items
		if (NOT Buf)
		{
			Buf = NEW(GdcMemDC);
		}

		r = ItemsPos;
		int n = FirstVisible;
		int LastY = r.y1;
		if (NOT Focus())
		{
			_Selection = GdcMixColour(Back, LC_SELECTION);
		}
		else
		{
			_Selection = LC_SELECTION;
		}

		for (GListItem *i = Items.ItemAt(n); i; i = Items.Next(), n++)
		{
			if (i->Pos.Valid())
			{
				// tell the item what colour to use
				if (i->IsSelected())
				{
					i->Fore = LC_SEL_TEXT;
					i->Back = _Selection;
				}
				else
				{
					i->Fore = LC_TEXT;
					i->Back = Back;
				}

				#ifdef DOUBLE_BUFFER_PAINT
				if (Buf->X() < i->Pos.X() OR
					Buf->Y() < i->Pos.Y())
				{
					Buf->Create(i->Pos.X(), i->Pos.Y(), GdcD->GetBits());
				}

				GdcRegion r = i->Pos;
				r.Offset(-r.x1, -r.y1);
				i->OnPaint(Buf, r);
				pDC->Blt(i->Pos.x1, i->Pos.y1, Buf, &r);
				#else
				i->OnPaint(pDC, i->Pos);
				#endif

				LastY = i->Pos.y2 + 1;
			}
			else
			{
				break;
			}
		}

		// Draw blank space under items
		if (LastY <= r.y2)
		{
			pDC->Colour(Back, 24);
			pDC->Rectangle(r.x1, LastY, r.x2, r.y2);
		}

		UnlockLooper();
	}
}

void GList::OnFocus(bool b)
{
	for (GListItem *i = Items.ItemAt(FirstVisible); i; i = Items.Next())
	{
		if (i->Pos.Valid() AND
			i->Selected)
		{
			Invalidate(&i->Pos);
		}
	}
}
