#include <stdio.h>

#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
GSemaphore::GSemaphore()
{
	_Thread = 0;
	_Count = 0;

	#if defined WIN32
	
	_Mutex = CreateMutex(NULL, FALSE, NULL);
	
	#elif defined BEOS
	
	_Sem = create_sem(1, "LGI.GSemaphore");

	#elif defined LINUX

	// FIXME: Implement
	
	#endif
}

GSemaphore::~GSemaphore()
{
	#if defined WIN32

	CloseHandle(_Mutex);

	#elif defined BEOS

	if (_Sem > 0)
	{
		delete_sem(_Sem);
		_Sem = 0;
	}

	#endif
}

bool GSemaphore::_Lock()
{
	#if defined WIN32

	return WaitForSingleObject(_Mutex, INFINITE) == WAIT_OBJECT_0;

	#elif defined BEOS

	// if (acquire_sem_etc(_Sem, 1, B_RELATIVE_TIMEOUT, Timeout) == B_OK)
	return acquire_sem(_Sem) == B_OK;

	#elif defined LINUX

	// FIXME: Implement
	return false;

	#endif	
}

void GSemaphore::_Unlock()
{
	#if defined WIN32
	
	ReleaseMutex(_Mutex);

	#elif defined BEOS

	release_sem(_Sem);

	#elif defined LINUX

	// FIXME: implement

	#endif
}

bool GSemaphore::Lock()
{
	bool Status = false;

	while (NOT Status)
	{
		if (_Lock())
		{
			LgiThreadId CurrentThread = LgiGetCurrentThread();
			if (NOT _Thread OR
				_Thread == CurrentThread)
			{
				_Thread = CurrentThread;
				_Count++;
				Status = true;
			}
			_Unlock();
		}

		if (NOT Status)
		{
			LgiSleep(5);
		}
	}

	return Status;
}

bool GSemaphore::LockWithTimeout(int Timeout)
{
	int Start = LgiCurrentTime();
	bool Status = false;

	while (NOT Status AND
			LgiCurrentTime() < Start + Timeout)
	{
		if (_Lock())
		{
			LgiThreadId CurrentThread = LgiGetCurrentThread();
			if (NOT _Thread OR
				_Thread == CurrentThread)
			{
				_Thread = CurrentThread;
				_Count++;
				Status = true;
			}
			_Unlock();
		}
		
		if (NOT Status)
		{
			LgiSleep(5);
		}
	}

	return Status;
}

void GSemaphore::Unlock()
{
	if (_Lock())
	{
		// if this assert fails then you tryed to unlock an object that
		// wasn't locked in the first place
		LgiAssert(_Count > 0);

		// decrement the lock counter
		if (_Count > 0)
		{
			_Count--;
		}
		if (_Count < 1)
		{
			_Thread = 0;
		}

		_Unlock();
	}
}

