#include <stdio.h>

#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
GSemaphore::GSemaphore()
{
	_Thread = 0;
	_Count = 0;

	#if defined WIN32
	
	_Mutex = CreateMutex(NULL, FALSE, NULL);
	
	#elif defined BEOS
	
	_Sem = create_sem(1, "LGI.GSemaphore");
	
	#endif
}

GSemaphore::~GSemaphore()
{
	#if defined WIN32

	CloseHandle(_Mutex);

	#elif defined BEOS

	if (_Sem > 0)
	{
		delete_sem(_Sem);
		_Sem = 0;
	}

	#endif
}

bool GSemaphore::_Lock()
{
	bool Status = false;

	#if defined WIN32

	int ThisThread = GetCurrentThreadId();
	DWORD Result = WaitForSingleObject(_Mutex, 10); // INFINITE
	if (Result == WAIT_OBJECT_0)
	{

	#elif defined BEOS

	// status_t Ret;
	// if ( (Ret = acquire_sem_etc(_Sem, 1, B_RELATIVE_TIMEOUT, 10 * 1000000)) == B_OK)
	if (acquire_sem(_Sem) == B_OK)
	{
		thread_id ThisThread = find_thread(0);
	
	#endif

		if (ThisThread == _Thread OR
			_Count == 0)
		{
			_Count++;
			_Thread = ThisThread;
			Status = true;
		}

		#if defined WIN32
		
		ReleaseMutex(_Mutex);

		#elif defined BEOS
	
		release_sem(_Sem);
	
		#endif
	}
	
	return Status;
}

bool GSemaphore::Lock()
{
	while (NOT _Lock())
	{
		LgiSleep(50);
	}

	return true;
}

int GSemaphore::LockWithTimeout(int Timeout)
{
	int Start = LgiCurrentTime();
	while (LgiCurrentTime() < Start + Timeout)
	{
		if (_Lock())
		{
			return true;
		}
	}

	return false;
}

void GSemaphore::Unlock()
{
	#if defined WIN32

	int ThisThread = GetCurrentThreadId();
	DWORD Result = WaitForSingleObject(_Mutex, 10); // INFINITE
	if (Result == WAIT_OBJECT_0)
	{

	#elif defined BEOS

	// status_t Ret;
	// if ( (Ret = acquire_sem_etc(_Sem, 1, B_RELATIVE_TIMEOUT, 10 * 1000000)) == B_OK)
	if (acquire_sem(_Sem) == B_OK)
	{
		thread_id ThisThread = find_thread(0);

	#endif

		if (_Count > 0)
		{
			if (ThisThread == _Thread)
			{
				_Count--;
			}
			if (_Count == 0)
			{
				_Thread = 0;
			}
		}

		#if defined WIN32
	
		ReleaseMutex(_Mutex);

		#elif defined BEOS
	
		release_sem(_Sem);
	
		#endif
	}
}

/*
//////////////////////////////////////////////////////////////////////////////
GSemaphore::GSemaphore()
{
	_Thread = 0;
	_Count = 0;

	#ifdef WIN32
	InitializeCriticalSection(&_Section);
	#endif
}

GSemaphore::~GSemaphore()
{
	#ifdef WIN32
	DeleteCriticalSection(&_Section);
	#endif
}

bool GSemaphore::_Lock()
{
	bool Status = false;
	#ifdef WIN32
	EnterCriticalSection(&_Section);
	int ThisThread = GetCurrentThreadId();
	if (ThisThread == _Thread OR
		_Count == 0)
	{
		_Count++;
		_Thread = ThisThread;
		Status = true;
	}
	LeaveCriticalSection(&_Section);
	#else
	ASSERT(0);
	#endif
	return Status;
}

bool GSemaphore::Lock()
{
	while (NOT _Lock())
	{
		LgiSleep(10);
	}

	return true;
}

int GSemaphore::LockWithTimeout(int Timeout)
{
	int Start = LgiCurrentTime();
	while (LgiCurrentTime() < Start + Timeout)
	{
		if (_Lock())
		{
			return true;
		}
	}

	return false;
}

void GSemaphore::Unlock()
{
	#ifdef WIN32
	EnterCriticalSection(&_Section);
	int ThisThread = GetCurrentThreadId();
	if (_Count > 0)
	{
		if (ThisThread == _Thread)
		{
			_Count--;
		}
		if (_Count == 0)
		{
			_Thread = 0;
		}
	}
	LeaveCriticalSection(&_Section);
	#else
	ASSERT(0);
	#endif
}

*/