/*
**	FILE:			GuiDataDlg.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			30/10/98
**	DESCRIPTION:	Data serialisation classes
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/


#include <stdio.h>
#include "Lgi.h"

////////////////////////////////////////////////////////////////////////////////////////////////
DataDlgField::DataDlgField(int type, int ctrlid, char *opt, char *desc)
{
	Type = type;
	CtrlId = ctrlid;
	Option = NewStr(opt);
	Desc = NewStr(desc);
}

DataDlgField::~DataDlgField()
{
	DeleteArray(Option);
	DeleteArray(Desc);
}

/////////////////////////////////////////////////////////////////////////////////
DataDlgTools::DataDlgTools()
{
	Dlg = 0;
	Options = 0;
}

void DataDlgTools::Set(GView *dlg, ObjProperties *options)
{
	Dlg = dlg;
	Options = options;
}

ObjProperties *DataDlgTools::GetOptions()
{
	return Options;
}

void DataDlgTools::SaveInt(GEdit *Ed, char *Name)
{
	if (Ed AND Name)
	{
		Options->Set(Name, Ed->Value());
	}
}

void DataDlgTools::InitInt(GEdit *Ed, char *Name)
{
	if (Ed AND Name)
	{
		int n = 0;
		Options->Get(Name, n);
		Ed->Value(n);
	}
}

void DataDlgTools::SaveEdit(GEdit *Ed, char *Name)
{
	if (Ed AND Name)
	{
		char *s = Ed->GEdit::Name();
		if (s)
		{
			Options->Set(Name, s);
		}
	}
}

void DataDlgTools::InitEdit(GEdit *Ed, char *Name)
{
	if (Ed AND Name)
	{
		char *n = "";
		Options->Get(Name, n);
		Ed->Name(n);
	}
}

void DataDlgTools::SaveCheck(GCheckBox *Chk, char *Name)
{
	if (Chk AND Name)
	{
		int v = Chk->Value();
		Options->Set(Name, v);
	}
}

void DataDlgTools::InitCheck(GCheckBox *Chk, char *Name)
{
	if (Chk AND Name)
	{
		int v;
		if (Options->Get(Name, v))
		{
			Chk->Value(v);
		}
	}
}

bool DataDlgTools::ProcessField(DataDlgField *f, bool Write, char *OptionOverride)
{
	bool Processed = false;

	if (f)
	{
		char *Opt = (OptionOverride) ? OptionOverride : f->Option;

		switch (f->Type)
		{
			case DATA_STR:
			{
				GEdit *e = dynamic_cast<GEdit*>(Dlg->FindControl(f->CtrlId));
				if (e)
				{
					if (Write)
					{
						SaveEdit(e, Opt);
					}
					else
					{
						InitEdit(e, Opt);
					}
					Processed = true;
				}
				break;
			}
			case DATA_BOOL:
			{
				GCheckBox *e = dynamic_cast<GCheckBox*>(Dlg->FindControl(f->CtrlId));
				if (e)
				{
					if (Write)
					{
						SaveCheck(e, Opt);
					}
					else
					{
						InitCheck(e, Opt);
					}
					Processed = true;
				}
				break;
			}
			case DATA_INT:
			{
				GView *w = Dlg->FindControl(f->CtrlId);
				GEdit *e = dynamic_cast<GEdit*>(w);
				if (e)
				{
					if (Write)
					{
						SaveInt(e, Opt);
					}
					else
					{
						InitInt(e, Opt);
					}
					Processed = true;
				}
				else
				{
					GCombo *c = dynamic_cast<GCombo*>(w);
					if (c)
					{
						if (Write)
						{
							Options->Set(Opt, c->Index());
						}
						else
						{
							int n;
							if (Options->Get(Opt, n))
							{
								c->Index(n);
							}
						}
					}
					else
					{
						GRadioGroup *g = dynamic_cast<GRadioGroup*>(w);
						if (g)
						{
							if (Write)
							{
								Options->Set(Opt, g->Value());
							}
							else
							{
								int n;
								if (Options->Get(Opt, n))
								{
									g->Value(n);
								}
							}
						}
					}
				}
				break;
			}
		}
	}

	return Processed;
}

bool DataDlgTools::ProcessFields(DataDlgFieldList &Field, bool Write)
{
	int Processed = 0;
	
	for (DataDlgField *f = Field.First(); f; f = Field.Next())
	{
		if (ProcessField(f, Write))
		{
			Processed++;
		}
	}

	return Processed == Field.GetItems();
}

bool DataDlgTools::LoadFields(DataDlgFieldList &Field)
{
	return ProcessFields(Field, false);
}

bool DataDlgTools::SaveFields(DataDlgFieldList &Field)
{
	return ProcessFields(Field, true);
}

////////////////////////////////////////////////////////////////////////////////
DRecordSetCtrls::DRecordSetCtrls(	GView *window,
									DataDlgFieldList *fields,
									List<DRecordSetObj> *records,
									int DescId,
									int ScrollId,
									int NewId,
									int DelId)

{
	Current = 0;
	Dlg = window;
	Fields = fields;
	Records = records;
	NewRecordId = NewId;
	DeleteRecordId = DelId;

	if (Dlg)
	{
		Description = dynamic_cast<GText*>(Dlg->FindControl(DescId));
		Scroll = dynamic_cast<GScrollBar*>(Dlg->FindControl(ScrollId));
		OnMoveRecord(Records->First());
	}
}

int DRecordSetCtrls::GetCurrentIndex()
{
	return (Records AND Current) ? Records->IndexOf(Current) : -1;
}

void DRecordSetCtrls::Serialize(bool Write)
{
	if (Fields AND Dlg AND Current)
	{
		ObjProperties Temp;
		Options = &Temp;

		if (Write)
		{
			// Move data from controls to propertiy list
			SaveFields(*Fields);

			// Move data from property list to record
			Current->Serialize(Temp, false);
		}
		else
		{
			// Move data from record to property list
			Current->Serialize(Temp, true);

			// Load the new data
			LoadFields(*Fields);
		}

		Options = 0;
	}
}

void DRecordSetCtrls::OnMoveRecord(DRecordSetObj *r)
{
	if (Dlg AND Fields)
	{
		Serialize(true);
		Current = r;
		Serialize(false);

		// Set controls enabled flag correctly
		for (DataDlgField *f = Fields->First(); f; f = Fields->Next())
		{
			GView *Ctrl = dynamic_cast<GView*>(Dlg->FindControl(f->GetCtrl()));
			if (Ctrl)
			{
				if (NOT Current)
				{
					// clear controls
				}

				Ctrl->Enabled(Current != 0);
			}
		}

		// Set the scrollbar
		if (Scroll)
		{
			Scroll->SetLimits(0, Records->GetItems()-1);
			Scroll->SetPos((Current) ? Records->IndexOf(Current) : 0);
			Scroll->SetPage(1);
		}

		// Set the description
		if (Description)
		{
			char Str[256];
			sprintf(Str, "Record: %i of %i", GetCurrentIndex()+1, (Records) ? Records->GetItems() : 0);
			Description->Name(Str);
		}
	}
}

int DRecordSetCtrls::OnNotify(GView *Col, int Flags)
{
	if (Fields AND Dlg AND Records)
	{
		if (Col == Scroll)
		{
			int CurrentIndex = (Current) ? Records->IndexOf(Current) : -1;
			int NewIndex = Flags;
			if (CurrentIndex != NewIndex)
			{
				OnMoveRecord(Records->ItemAt(NewIndex));
			}
		}

		if (Col->GetId() == NewRecordId)
		{
			DRecordSetObj *r = NewObj();
			if (r)
			{
				Records->Insert(r);
				OnMoveRecord(r);
			}
		}

		if (Col->GetId() == DeleteRecordId)
		{
			int Index = GetCurrentIndex();
			DRecordSetObj *r = Records->ItemAt(Index);
			if (r)
			{
				Records->Delete(r);
				OnMoveRecord(Records->ItemAt(limit(Index, 0, Records->GetItems()-1)));
				DelObj(r);
			}
		}
	}

	return 0;
}
