/*hdr
**      FILE:           GuiRes.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Resource interfaces
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <time.h>

#include "Lgi.h"
#include "Token.h"

// If this isn't defined these functions load win32
// resources from the executable
//
// If it is defined it will use the cross platform 
// "res" library distributed with the LGI library.

#ifdef LGI_RES

#include "xml.h"

#define DeleteAllListObjects(pList, Type)	{ for (Type *o = (pList)->First(); o; o = (pList)->Next()) { DeleteObj(o); } }
#define CastToGWnd(RObj)					((RObj != 0) ? dynamic_cast<GView*>((GView*)RObj) : 0)

LgiStringRes::LgiStringRes()
{
	Ref = 0;
	Id = 0;
	Str = 0;
}

LgiStringRes::~LgiStringRes()
{
	DeleteArray(Str);
}

bool LgiStringRes::Read(List<XmlTag> &Tags)
{
	XmlTag *t = Tags.Current();
	if (t AND stricmp(t->Name, "string") == 0)
	{
		char *n = 0;
		if (t->Get("Id", n))
		{
			Id = atoi(n);
		}
		if (t->Get("Ref", n))
		{
			Ref = atoi(n);
		}
		if (t->Get("Define", n))
		{
			if (strcmp(n, "IDOK") == 0)
			{
				Id = IDOK;
			}
			else if (strcmp(n, "IDCANCEL") == 0)
			{
				Id = IDCANCEL;
			}
			else if (strcmp(n, "IDC_STATIC") == 0)
			{
				Id = -1;
			}
		}

		int CurLang = LgiGetLanguageId();
		char Name[256];
		sprintf(Name, "Text(%i)", CurLang);
		if (t->Get(Name, n))
		{
			DeleteArray(Str);
			Str = NewStr(n);
			if (Str)
			{
				char *d = Str;
				for (char *s = Str; s AND *s;)
				{
					if (*s == '\\' AND *(s+1) == 'n')
					{
						*d++ = '\n';
						s += 2;
					}
					else 
					{
						*d++ = *s++;
					}
				}
				*d++ = 0;
			}
		}

		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////
// Lgi resource factory
LgiResources *Resources = 0;
LgiResources::LgiResources()
{
	// global pointer init
	Resources = this;

	// find the resources file, extension ".lr"
	char Str[256];
	if (LgiGetExeFile(Str))
	{
		char *File = 0;
		{
			for (char *p = Str+strlen(Str)-1; p > Str; p--)
			{
				if (*p == DIR_CHAR)
				{
					File = p + 1;
					break;
				}
			}
		}

		if (File)
		{
			char *Period = strchr(File, '.');
			if (Period)
			{
				strcpy(Period+1, "lr");
			}
			else
			{
				strcat(File, ".lr");
			}

			char *FileName = LgiFindFile(File);
			if (FileName)
			{
				Load(FileName);
				DeleteArray(FileName);
			}
		}
	}
}

LgiResources::~LgiResources()
{
	DeleteAllListObjects(&Dialogs, LgiDialogRes);
	DeleteAllListObjects(&Strings, LgiStringRes);
}

bool LgiResources::Load(char *FileName)
{
	bool Status = false;
	if (FileName)
	{
		VFile F;
		if (F.Open(FileName, O_READ))
		{
			Xml X;
			if (X.ParseXmlFile(F) > 0)
			{
				Status = true;

				for (XmlTag *t = X.Values.First(); t; t = X.Values.Next())
				{
					if (t->Name)
					{
						if (stricmp(t->Name, "string-group") == 0)
						{
							// search through group, adding strings to database
							for (	t = X.Values.Next();
									t AND t->Name AND stricmp(t->Name, "/string-group") != 0;
									t = X.Values.Next())
							{
								LgiStringRes *s = NEW(LgiStringRes);
								if (s AND s->Read(X.Values))
								{
									Strings.Insert(s);
								}
								else
								{
									DeleteObj(s);
								}
							}
						}
						else if (stricmp(t->Name, "dialog") == 0)
						{
							LgiDialogRes *d = NEW(LgiDialogRes);
							if (d AND d->Read(X.Values))
							{
								Dialogs.Insert(d);
							}
							else
							{
								DeleteObj(d);
							}
						}
					}
				}
			}
		}
	}

	return Status;
}

LgiStringRes *LgiResources::StrFromRef(int Ref)
{
	for (LgiStringRes *s = Strings.First(); s; s = Strings.Next())
	{
		if (s->Ref == Ref)
		{
			return s;
		}
	}

	return 0;
}

char *LgiResources::StringFromRef(int Ref)
{
	for (LgiStringRes *s = Strings.First(); s; s = Strings.Next())
	{
		if (s->Ref == Ref)
		{
			return s->Str;
		}
	}

	return 0;
}

ResObject *LgiResources::CreateObject(char *Name, ResObject *Parent)
{
	ResObject *Wnd = 0;
	if (Name)
	{
		if (stricmp(Name, Res_StaticText) == 0)
		{
			Wnd = NEW(GText(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_EditBox) == 0)
		{
			Wnd = NEW(GEdit(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_CheckBox) == 0)
		{
			Wnd = NEW(GCheckBox(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Button) == 0)
		{
			Wnd = NEW(GButton(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Group) == 0)
		{
			Wnd = NEW(GRadioGroup(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_RadioBox) == 0)
		{
			Wnd = NEW(GRadioButton(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Tab) == 0)
		{
			if (Parent)
			{
				GRadioGroup *Grp = dynamic_cast<GRadioGroup*>((GView*)Parent);
				if (Grp)
				{
					Wnd = Grp->Append(0, 0, "");
				}
			}
		}
		else if (stricmp(Name, Res_TabView) == 0)
		{
			Wnd = NEW(GTabControl(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_ListView) == 0)
		{
			Wnd = NEW(GList(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Column) == 0)
		{
			if (Parent)
			{
				GList *Lst = dynamic_cast<GList*>((GView*)Parent);
				if (Lst)
				{
					Wnd = Lst->AddColumn("");
				}
			}
		}
		else if (stricmp(Name, Res_ComboBox) == 0)
		{
			Wnd = NEW(GCombo(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_TreeView) == 0)
		{
			Wnd = NEW(GTree(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Bitmap) == 0)
		{
			Wnd = NEW(GBitmap(0, 0, 0, 0));
		}
		else if (stricmp(Name, Res_Progress) == 0)
		{
			Wnd = NEW(GProgress(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Slider) == 0)
		{
			Wnd = NEW(GSlider(0, 0, 0, -1, -1, "", false));
		}
		else if (stricmp(Name, Res_ScrollBar) == 0)
		{
			Wnd = NEW(GScrollBar(0, 0, 0, 20, 100, ""));
		}
	}

	return Wnd;
}

void LgiResources::Res_SetPos(ResObject *Obj, int x1, int y1, int x2, int y2)
{
	GView *w = CastToGWnd(Obj);
	if (w)
	{
		w->SetPos(GdcRegion(x1, y1, x2, y2));
	}
}

void LgiResources::Res_SetPos(ResObject *Obj, char *s)
{
	if (Obj AND s)
	{
		TokenParser T(s, ",");
		if (T.Tokens == 4)
		{
			Res_SetPos(Obj, atoi(T.Get(0)), atoi(T.Get(1)), atoi(T.Get(2)), atoi(T.Get(3)));
		}
	}
}

GdcRegion LgiResources::Res_GetPos(ResObject *Obj)
{
	GView *w = CastToGWnd(Obj);
	if (w)
	{
		return w->GetPos();
	}

	return GdcRegion(0, 0, 0, 0);
}

int LgiResources::Res_GetStrRef(ResObject *Obj)
{
	return 0;
}

void LgiResources::Res_SetStrRef(ResObject *Obj, int Ref)
{
	for (LgiStringRes *s = Strings.First(); s; s = Strings.Next())
	{
		if (s->Ref == Ref)
		{
			GView *w = CastToGWnd(Obj);
			if (w)
			{
				w->Name(s->Str);
				w->SetId(s->Id);
			}
			break;
		}
	}
}

void LgiResources::Res_Attach(ResObject *Obj, ResObject *Parent)
{
	GView *o = CastToGWnd(Obj);
	GView *p = CastToGWnd(Parent);
	if (o AND p)
	{
		if (NOT o->Attach(p))
		{
			p->Children.Insert(o);
		}
	}
}

bool LgiResources::Res_GetChildren(ResObject *Obj, List<ResObject> *l, bool Deep)
{
	GView *o = CastToGWnd(Obj);
	if (o AND l)
	{
		for (GView *w = o->Children.First(); w; w = o->Children.Next())
		{
			ResObject *n = dynamic_cast<ResObject*>(w);
			if (n)
			{
				l->Insert(n);
			}
		}
		return true;
	}
	return false;
}

void LgiResources::Res_Append(ResObject *Obj, ResObject *Parent)
{
	if (Obj AND Parent)
	{
		GListColumn *Col = dynamic_cast<GListColumn*>(Obj);
		GList *Lst = dynamic_cast<GList*>(Parent);
		if (Lst AND Col)
		{
			Lst->AddColumn(Col);
		}

		GTabPage *Tab = dynamic_cast<GTabPage*>(Obj);
		GTabControl *Tabs = dynamic_cast<GTabControl*>(Parent);
		if (Tab AND Tabs)
		{
			Tabs->Append(Tab);
		}
	}
}

bool LgiResources::Res_GetItems(ResObject *Obj, List<ResObject> *l)
{
	if (Obj AND l)
	{
		GList *Lst = dynamic_cast<GList*>(Obj);
		if (Lst)
		{
			for (int i=0; i<Lst->GetColumns(); i++)
			{
				l->Insert(Lst->ColumnAt(i));
			}
			return true;
		}

		GTabControl *Tabs = dynamic_cast<GTabControl*>(Obj);
		if (Tabs)
		{
			for (int i=0; i<Tabs->GetTabs(); i++)
			{
				l->Insert(Tabs->TabAt(i));
			}
			return true;
		}
	}
	return false;
}

///////////////////////////////////////////////////////
LgiDialogRes::LgiDialogRes()
{
	Dialog = 0;
	Str = 0;
}

LgiDialogRes::~LgiDialogRes()
{
	DeleteAllListObjects(&Tags, XmlTag);
}

bool LgiDialogRes::Read(List<XmlTag> &l)
{
	XmlTag *t = Dialog = l.Current();
	if (Dialog)
	{
		char *n = 0;
		if (Dialog->Get("ref", n))
		{
			int Ref = atoi(n);
			Str = Resources->StrFromRef(Ref);
		}
		if (Dialog->Get("pos", n))
		{
			TokenParser T(n, ",");
			if (T.Tokens == 4)
			{
				Pos.x1 = atoi(T.Get(0));
				Pos.y1 = atoi(T.Get(1));
				Pos.x2 = atoi(T.Get(2));
				Pos.y2 = atoi(T.Get(3));
			}
		}
	}

	for (	t = l.Next();
			t AND t->Name AND
			stricmp(t->Name, "/dialog") != 0;
			t = l.Next())
	{
		Tags.Insert(t);
	}

	for (	XmlTag *d = Tags.First();
			d;
			d = Tags.Next())
	{
		l.Delete(d);
	}

	if (t)
	{
		l.Find(t);
	}

	return true;
}
#endif

/////////////////////////////////////////////////////////////////////////////////////////
// Classes
/*
short *GDlgFunc::StrCopy(short *A, char *N)
{
	uchar *n = (uchar*) N;
	if (n)
	{
		while (*n)
		{
			*A++ = *n++;
		}
	}
	*A++ = 0;
	return A;
}

short *GDlgFunc::Push(short *A, int n)
{
	int *Ptr = (int*) A;
	*Ptr++ = n;
	return (short*)Ptr;
}

short *GDlgFunc::PadToDWord(short *A, bool Seek)
{
	char *c = (char*) A;
	while ((int) c & 3)
	{
		if (Seek)
		{
			c++;
		}
		else
		{
			*c++ = 0;
		}
	}
	return (short*) c;
}

short *GDlgFunc::StrSkip(short *Ptr)
{
	if (*Ptr == -1)
	{
		Ptr++;
		Ptr++; // skip global atom
	}
	else
	{
		while (*Ptr)
		{
			Ptr++;
		}
		Ptr++;
	}
	return Ptr;
}

short *GDlgFunc::StrExtract(short *Ptr, char *Str, short *Atom)
{
	if (*Ptr == -1)
	{
		Ptr++;
		if (Atom)
		{
			*Atom = *Ptr++;
		}
		else
		{
			Ptr++; // skip global atom
		}
	}
	else
	{
		while (*Ptr)
		{
			*Str++ = (char) *Ptr;
			Ptr++;
		}
		Ptr++;
	}
	*Str++ = 0;
	return Ptr;
}

void GDlgFunc::PixelToDialog(GdcRegion &r)
{
	r.x1 = (double) r.x1 / DIALOG_X;
	r.y1 = (double) r.y1 / DIALOG_Y;
	r.x2 = (double) r.x2 / DIALOG_X;
	r.y2 = (double) r.y2 / DIALOG_Y;
}

void GDlgFunc::DialogToPixel(GdcRegion &r)
{
	r.x1 = (double) r.x1 * DIALOG_X;
	r.y1 = (double) r.y1 * DIALOG_Y;
	r.x2 = (double) r.x2 * DIALOG_X;
	r.y2 = (double) r.y2 * DIALOG_Y;
}
*/

bool GDlgFunc::LoadFromResource(char *Resource, List<GView> &Children, GdcRegion *Pos, char *Name)
{
	bool Status = false;
	#ifdef LGI_RES
	if (NOT Resources)
	{
		NEW(LgiResources);
	}
	if (Resources)
	{
		for (LgiDialogRes *Dlg = Resources->Dialogs.First(); Dlg; Dlg = Resources->Dialogs.Next())
		{
			if (Dlg->Id() == ((int) Resource))
			{
				// found the dialog to load, set properties
				if (Name AND Dlg->Name())
				{
					strcpy(Name, Dlg->Name());
				}
				if (Pos)
				{
					Pos->ZOff(Dlg->X()-1, Dlg->Y()-1);
				}

				// instantiate control list
				for (XmlTag *t = Dlg->Tags.First(); t; t = Dlg->Tags.Current())
				{
					ResObject *Obj = Resources->CreateObject(t->Name, 0);
					if (Obj)
					{
						if (Resources->Res_Read(Obj, Dlg->Tags))
						{
							GView *w = dynamic_cast<GView*>((GView*)Obj);
							if (w)
							{
								GdcRegion r = w->GetPos();
								r.Offset(-7, -17);
								w->SetPos(r);

								Children.Insert(w);
							}
						}
					}
				}

				Status = true;
			}
		}
	}
	#else
	HRSRC hRsrc = FindResourceEx(NULL, RT_DIALOG, Resource, LgiGetLanguageId());

	if (NOT hRsrc)
	{
		hRsrc = FindResource(NULL, Resource, RT_DIALOG);
	}

	if (hRsrc)
	{
		int Size = SizeofResource(NULL, hRsrc);
		HGLOBAL hMem = LoadResource(NULL, hRsrc);
		if (hMem)
		{
			uchar *Ptr = (uchar*) LockResource(hMem);
			if (Ptr)
			{
				DLGTEMPLATE *Template = (DLGTEMPLATE*) Ptr;

				if (Pos)
				{
					Pos->ZOff((double) Template->cx * DIALOG_X, (double) Template->cy * DIALOG_Y);
					Pos->Offset((double) Template->x * DIALOG_X, (double) Template->y * DIALOG_Y);
				}

				char Temp[256];
				short *A = (short*) (Template+1);

				// menu
				A = StrExtract(A, Temp);
				// class
				A = StrExtract(A, Temp);
				// title
				A = StrExtract(A, Temp);
				if (strlen(Temp) > 0 AND Name)
				{
					strcpy(Name, Temp);
				}

				if (Template->style & DS_SETFONT)
				{
					// font
					int PointSize = *A++;
					// name
					A = StrExtract(A, Temp);
				}

				A = PadToDWord(A, TRUE);

				// loop thru controls
				for (int i=0; i<Template->cdit; i++)
				{
					GControlInfo *CtrlInfo = NEW(GControlInfo);
					if (CtrlInfo)
					{
						DLGITEMTEMPLATE *Item = (DLGITEMTEMPLATE*) A;
						memcpy(&CtrlInfo->Item, Item, sizeof(DLGITEMTEMPLATE));
						A = (short*) (Item+1);

						// class
						A = StrExtract(A, Temp, &CtrlInfo->Atom);
						if (strlen(Temp) > 0)
						{
							CtrlInfo->Class = NewStr(Temp);
						}

						// title
						A = StrExtract(A, Temp);
						if (strlen(Temp) > 0)
						{
							CtrlInfo->Text = NewStr(Temp);
						}

						// creation data
						int Size = *A++;
						if (Size > 0)
						{
							A = PadToDWord(A, TRUE);
							A = (short*) (((uchar*)A)+Size);
						}
						A = PadToDWord(A, TRUE);

						GView *c = CreateControl(CtrlInfo);
						if (c)
						{
							c->Visible(CtrlInfo->Item.style & WS_VISIBLE);
							Children.Insert(c);
						}

						DeleteObj(CtrlInfo);
					}
				}

				Status = TRUE;
				GlobalUnlock(hMem);
			}
		}
	}
	#endif
	return Status;
}

/*
GView *GDlgFunc::CreateControl(GControlInfo *Info)
{
	GView *Ctrl = 0;

	Info->Item.x	= (double) Info->Item.x * CTRL_X;
	Info->Item.cx	= (double) Info->Item.cx * CTRL_X;
	Info->Item.y	= (double) Info->Item.y * CTRL_Y;
	Info->Item.cy	= (double) Info->Item.cy * CTRL_Y;

	switch (Info->Atom)
	{
		case 0x0080: // Button
		{
			int Type = Info->Item.style & 0xF;
			static GRadioGroup *LastGroup = NULL;

			if (Type == BS_CHECKBOX OR Type == BS_AUTOCHECKBOX)
			{
				Ctrl = NEW(GCheckBox(	Info->Item.id,
										Info->Item.x,
										Info->Item.y,
										Info->Item.cx,
										Info->Item.cy,
										Info->Text));
			}
			else if (Type == BS_GROUPBOX)
			{
				Ctrl = LastGroup = NEW(GRadioGroup(
										Info->Item.id,
										Info->Item.x,
										Info->Item.y,
										Info->Item.cx,
										Info->Item.cy,
										Info->Text));
			}
			else if (Type == BS_RADIOBUTTON OR Type == BS_AUTORADIOBUTTON)
			{
				if (LastGroup)
				{	
					GdcRegion p = LastGroup->GetPos();
					LastGroup->Append(	Info->Item.x-p.x1,
										Info->Item.y-p.y1,
										Info->Text);
				}
			}
			else
			{
				GButton *Btn = NEW(GButton(	Info->Item.id,
											Info->Item.x,
											Info->Item.y,
											Info->Item.cx,
											Info->Item.cy,
											Info->Text));
				Ctrl = Btn;
				if (Btn)
				{
					if (Info->Item.style & BS_DEFPUSHBUTTON)
					{
						Btn->Default(true);
					}
				}
			}
			break;
		}
		case 0x0081: // Edit
		{
			Ctrl = NEW(GEdit(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			if (Ctrl)
			{
				GEdit *Ed = (GEdit*) Ctrl;

				if (Info->Item.style & ES_PASSWORD)
				{
					Ed->Password(TRUE);
				}

				if (Info->Item.style & ES_MULTILINE)
				{
					Ed->MultiLine(TRUE);
				}

				if (Info->Item.style & ES_READONLY)
				{
					Ed->Enabled(FALSE);
				}
			}
			break;
		}
		case 0x0082: // Static
		{
			Ctrl = NEW(GText(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			break;
		}
		case 0x0083: // List box
		{
			// not implemented
			// use list view
			break;
		}
		case 0x0084: // Scroll bar
		{
			Ctrl = NEW(GScrollBar(
								Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			if (Ctrl)
			{
				((GScrollBar*) Ctrl)->SetVertical((Info->Item.style & SBS_VERT) != 0);
			}
			break;
		}
		case 0x0085: // Combo box
		{
			Ctrl = NEW(GCombo(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			break;
		}
		default:
		{
			if (Info->Class)
			{
				if (NOT strcmp(Info->Class, TRACKBAR_CLASS))
				{
					Ctrl = NEW(GSlider(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text,
								(Info->Item.style & TBS_VERT) ? TRUE : FALSE));
				}
				else if (NOT strcmp(Info->Class, WC_TABCONTROL))
				{
					Ctrl = NEW(GTabControl(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
				}
				else if (NOT strcmp(Info->Class, WC_LISTVIEW))
				{
					Ctrl = NEW(GList(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
				}
			}
			break;
		}
	}

	return Ctrl;
}

///////////////////////////////////////////////////////////////////////////////////////////
GControlInfo::GControlInfo()
{
	ZeroObj(Item);
	Atom = 0;
	Class = 0;
	Text = 0;
	Ctrl = 0;
}

GControlInfo::~GControlInfo()
{
	DeleteArray(Class);
	DeleteArray(Text);
	Ctrl = 0;
}
*/

////////////////////////////////////////////////////////////////////////
// Get language
int LgiGetLanguageId()
{
	#ifdef LGI_RES

		// Check for command line override
		char Buf[64];
		if (GApp::GetOption(Buf, 'i'))
		{
			LangInfo *i = LangFind(0, atoi(Buf));
			if (i)
			{
				return i->Id;
			}
		}
	
		// Use a system setting
		WORD Lang = PRIMARYLANGID(GetSystemDefaultLCID());
		for (LangInfo *i = AllLanguages; i->Id; i++)
		{
			if (i->Win32Id == Lang)
			{
				return i->Id;
			}
		}

		// Return a default of English
		return LangIdEnglish;
	
	#else
	
		return MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT);
		// return MAKELANGID(LANG_FRENCH, SUBLANG_FRENCH);
	
	#endif
}

///////////////////////////////////////////////////////////////////////////////////////
// Load string
char *LgiLoadString(WORD Res)
{
	#ifdef LGI_RES

	if (NOT Resources)
	{
		NEW(LgiResources);
	}
	if (Resources)
	{
		for (LgiStringRes *s = Resources->Strings.First(); s; s = Resources->Strings.Next())
		{
			if (s->Id == Res)
			{
				return s->Str;
			}
		}
	}
	else
	{
		LgiMsg(NULL, "Resource file missing.", "Fatal LGI Error", MB_OK);
		LgiExitApp();
	}
	return 0;

	#else

	static char Str[512];

	Str[0] = 0;
	HRSRC hRes = FindResourceEx(NULL, RT_STRING, MAKEINTRESOURCE(Res), LgiGetLanguageId());
	if (hRes)
	{
			HGLOBAL hMem = LoadResource(hAppInst(), hRes);
			if (hMem)
			{
				char *Ptr = (char*) LockResource(hMem);
				if (Ptr)
				{
					char *d = Str;
					short *s = (short*) (Ptr + 2);
					int Len = *s++;

					while (Len--)
					{
						*d++ = *s++;
					}
					*d++ = 0;
				}
			}
	}

	if (strlen(Str) < 1)
	{
		// can't find the language dependent string
		LoadString(hAppInst(), Res, Str, sizeof(Str));
	}

	return Str;
	#endif
}

//////////////////////////////////////////////////////////////////////
// Menus
bool GSubMenu::Load(HMENU hmenu)
{
	bool Status = FALSE;
	hMenu = hmenu;
	if (hMenu)
	{
		int Count = GetMenuItemCount(hMenu);
		for (int i=0; i<Count; i++)
		{
			MENUITEMINFO Info;
			char Name[256];

			ZeroObj(Info);
			Info.cbSize = sizeof(Info);
			Info.fMask = MIIM_CHECKMARKS | MIIM_DATA | MIIM_ID | MIIM_STATE | MIIM_SUBMENU | MIIM_TYPE;
			Info.dwTypeData = Name;
			Info.cch = sizeof(Name)-1;

			if (GetMenuItemInfo(hMenu,
								i,
								TRUE, 
								&Info))
			{
				GMenuItem *Item = NEW(GMenuItem);
				if (Item)
				{
					Info.fMask = 0;
					memcpy(&Item->Info, &Info, sizeof(Info));
					Item->Menu = Menu;
					Item->Parent = this;
					Item->Position = i;

					if (Info.fType == MFT_STRING)
					{
						Item->GObject::Name((char*) Info.dwTypeData);
						Item->ScanForAccel();
					}

					if (Info.hSubMenu)
					{
						Item->Child = NEW(GSubMenu("", true));
						if (Item->Child)
						{
							Item->Child->Parent = Item;
							Item->Child->Menu = Menu;
							Item->Child->Load(Info.hSubMenu);
						}
					}

					Items.Insert(Item);
				}
			}
		}

		Status = TRUE;
	}

	return Status;
}

bool GMenu::Load(GView *w, char *Res)
{
	bool Status = FALSE;

	#ifdef LGI_RES
	if (NOT Resources)
	{
		NEW(LgiResources);
	}
	if (Resources)
	{
		
	}
	#else
	if (Res)
	{
		HMENU hNewMenu = 0;

		// try looking for a language specific menu first
		HRSRC hRes = FindResourceEx(hAppInst(), RT_MENU, Res, LgiGetLanguageId());
		if (hRes)
		{
			HGLOBAL hMem = LoadResource(hAppInst(), hRes);
			if (hMem)
			{
				void *Ptr = LockResource(hMem);
				if (Ptr)
				{
					hNewMenu = LoadMenuIndirect((MENUTEMPLATE*) Ptr);
				}
			}
		}

		if (NOT hNewMenu)
		{
			// if the language specific load fails we do a normal load here
			hNewMenu = LoadMenu(hAppInst(), Res);
		}

		if (hNewMenu)
		{
			hMenu = hNewMenu;
			Menu = this;
			Status = GSubMenu::Load(hMenu);
		}
	}
	#endif

	return Status;
}

