// Lgi.cpp
#include "Lgi.h"

//#define XP_CTRLS

//////////////////////////////////////////////////////////////////////////
COLOUR _LgiColours[LC_MAXIMUM];

#define ReadColourConfig(def)	_lgi_read_colour_config("Colour."#def, _LgiColours+i++)

int _hex_to_int(char h)
{
	if (h >= '0' AND h <= '9')
	{
		return h - '0';
	}
	else if (h >= 'A' AND h <= 'F')
	{
		return h - 'A' + 10;
	}
	else if (h >= 'a' AND h <= 'f')
	{
		return h - 'a' + 10;
	}

	LgiAssert(0);
	return 0;
}

void _lgi_read_colour_config(char *Tag, COLOUR *c)
{
	XmlTag *Col = LgiApp->GetConfig(Tag);
	if (Col)
	{
		char *h;
		if (Col->Get("Hex", h))
		{
			int n = 0;
			if (*h == '#') h++;
			
			for (int i=0; i<5; i++)
			{
				n |= _hex_to_int(*h++);
				n <<= 4;
			}
			n |= _hex_to_int(*h++);

			if (c)
			{
				*c = Rgb24( n>>16, n>>8, n );
			}
		}
	}
}

void LgiInitColours()
{
	int i = 0;

	// Basic colours
	_LgiColours[i++] = Rgb24(0, 0, 0); // LC_BLACK
	_LgiColours[i++] = Rgb24(0x40, 0x40, 0x40); // LC_DKGREY
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_MIDGREY
	_LgiColours[i++] = Rgb24(0xc0, 0xc0, 0xc0); // LC_LTGREY
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_WHITE

	// Variable colours
	#if defined XP_CTRLS
	_LgiColours[i++] = Rgb24(0x42, 0x27, 0x63); // LC_SHADOW
	_LgiColours[i++] = Rgb24(0x7a, 0x54, 0xa9); // LC_LOW
	_LgiColours[i++] = Rgb24(0xbc, 0xa9, 0xd4); // LC_MED
	_LgiColours[i++] = Rgb24(0xdd, 0xd4, 0xe9); // LC_HIGH
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_LIGHT
	_LgiColours[i++] = Rgb24(0xbc, 0xa9, 0xd4); // LC_DIALOG
	_LgiColours[i++] = Rgb24(0xeb, 0xe6, 0xf2); // LC_WORKSPACE
	_LgiColours[i++] = Rgb24(0x35, 0x1f, 0x4f); // LC_TEXT
	_LgiColours[i++] = Rgb24(0xbf, 0x67, 0x93); // LC_SELECTION
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_SEL_TEXT
	_LgiColours[i++] = Rgb24(0x70, 0x3a, 0xec); // LC_ACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_ACTIVE_TITLE_TEXT
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_INACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0x40, 0x40, 0x40); // LC_INACTIVE_TITLE_TEXT
	#elif defined WIN32
	_LgiColours[i++] = GetSysColor(COLOR_3DDKSHADOW); // LC_SHADOW
	_LgiColours[i++] = GetSysColor(COLOR_3DSHADOW); // LC_LOW
	_LgiColours[i++] = GetSysColor(COLOR_3DFACE); // LC_MED
	_LgiColours[i++] = GetSysColor(COLOR_3DLIGHT); // LC_HIGH, 
	_LgiColours[i++] = GetSysColor(COLOR_3DHIGHLIGHT); // LC_LIGHT
	_LgiColours[i++] = GetSysColor(COLOR_3DFACE); // LC_DIALOG
	_LgiColours[i++] = GetSysColor(COLOR_WINDOW); // LC_WORKSPACE
	_LgiColours[i++] = GetSysColor(COLOR_BTNTEXT); // LC_TEXT
	_LgiColours[i++] = GetSysColor(COLOR_HIGHLIGHT); // LC_SELECTION
	_LgiColours[i++] = GetSysColor(COLOR_HIGHLIGHTTEXT); // LC_SEL_TEXT
	_LgiColours[i++] = GetSysColor(COLOR_ACTIVECAPTION); // LC_ACTIVE_TITLE
	_LgiColours[i++] = GetSysColor(COLOR_CAPTIONTEXT); // LC_ACTIVE_TITLE_TEXT
	_LgiColours[i++] = GetSysColor(COLOR_INACTIVECAPTION); // LC_INACTIVE_TITLE
	_LgiColours[i++] = GetSysColor(COLOR_INACTIVECAPTIONTEXT); // LC_INACTIVE_TITLE_TEXT
	#else // defaults for non-windows, plain greys
	_LgiColours[i++] = Rgb24(128, 128, 128); // LC_SHADOW
	_LgiColours[i++] = Rgb24(164, 164, 164); // LC_LOW
	_LgiColours[i++] = Rgb24(216, 216, 216); // LC_MED
	_LgiColours[i++] = Rgb24(236, 236, 236); // LC_HIGH
	_LgiColours[i++] = Rgb24(255, 255, 255); // LC_LIGHT
	_LgiColours[i++] = Rgb24(216, 216, 216); // LC_DIALOG
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_WORKSPACE
	_LgiColours[i++] = Rgb24(0, 0, 0); // LC_TEXT
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_SELECTION
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_SEL_TEXT
	_LgiColours[i++] = Rgb24(0, 0, 0xff); // LC_ACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_ACTIVE_TITLE_TEXT
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_INACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0x40, 0x40, 0x40); // LC_INACTIVE_TITLE_TEXT
	#endif

	// Tweak
	if (LgiGetOs() == LGI_OS_WINNT)
	{
		// NT/2K doesn't seem to get this right, so we just tweak it here
		#define MixComp(a)		( (a(_LgiColours[7])+a(_LgiColours[9])) / 2 )
		_LgiColours[8] = Rgb24(MixComp(R24), MixComp(G24), MixComp(B24));
	}

	// Read any settings out of config
	i = 5;
	ReadColourConfig(LC_SHADOW);
	ReadColourConfig(LC_LOW);
	ReadColourConfig(LC_MED);
	ReadColourConfig(LC_HIGH);
	ReadColourConfig(LC_LIGHT);
	ReadColourConfig(LC_DIALOG);
	ReadColourConfig(LC_WORKSPACE);
	ReadColourConfig(LC_TEXT);
	ReadColourConfig(LC_SELECTION);
	ReadColourConfig(LC_SEL_TEXT);
	ReadColourConfig(LC_ACTIVE_TITLE);
	ReadColourConfig(LC_ACTIVE_TITLE_TEXT);
	ReadColourConfig(LC_INACTIVE_TITLE);
	ReadColourConfig(LC_INACTIVE_TITLE_TEXT);
}

COLOUR LgiColour(int i)
{
	return i<LC_MAXIMUM ? _LgiColours[i] : 0;
}

//////////////////////////////////////////////////////////////////////////
int LgiGetOs(int *Ver, int *Rev)
{
	#if defined WIN32

	OSVERSIONINFO v;
	v.dwOSVersionInfoSize=sizeof(v);
	GetVersionEx(&v);

	if (Ver) *Ver = v.dwMajorVersion;
	if (Rev) *Rev = v.dwMinorVersion;

	return (v.dwPlatformId == VER_PLATFORM_WIN32_NT) ? LGI_OS_WINNT : LGI_OS_WIN9X;

	#elif defined BEOS

	if (Ver) *Ver = 5;
	if (Rev) *Rev = 0;
	
	return LGI_OS_BEOS;

	#elif defined LINUX

	return LGI_OS_LINUX;

	#elif defined ATHEOS

	if (Ver) *Ver = 0;
	if (Rev) *Rev = 0x305;
	
	return LGI_OS_ATHEOS;

	#else

	return LGI_OS_UNKNOWN;

	#endif
}

char *LgiGetOsName()
{
	char *Str[LGI_OS_MAX] =
	{
		"Unknown",
		"Win32",
		"Win32",
		"BeOS",
		"Linux",
		"AtheOS"
	};

	return Str[LgiGetOs()];
}

//////////////////////////////////////////////////////////////////////////
GObject::GObject()
{
	ObjName = 0;
}

GObject::~GObject()
{
	DeleteArray(ObjName);
}

char *GObject::Name()
{
	return ObjName;
}

bool GObject::Name(char *n)
{
	if (n == ObjName) return TRUE;
	DeleteObj(ObjName);
	if (n)
	{
		ObjName = NEW(char[strlen(n)+1]);
		if (ObjName)
		{
			strcpy(ObjName, n);
			return TRUE;
		}
	}
	return FALSE;
}


