/*
**	FILE:			Res.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			25/9/1999
**	DESCRIPTION:	Dialog Resources
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail->com->au
*/

////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>

#include "Lgi.h"
#include "Xml.h"
#include "Res.h"

////////////////////////////////////////////////////////////////////
#ifdef WIN32
#define DL(l) l
#else if BEOS
#define DL(l) 0
#endif

LangInfo AllLanguages[] = {
	{1,  "Afrikaans",	DL(LANG_AFRIKAANS),		"LANG_AFRIKAANS"},
	{2,  "Albanian",	DL(LANG_ALBANIAN),		"LANG_ALBANIAN"},
	{3,  "Arabic",		DL(LANG_ARABIC),		"LANG_ARABIC"},
	{4,  "Basque",		DL(LANG_BASQUE),		"LANG_BASQUE"},
	{5,  "Belarusian",	DL(LANG_BELARUSIAN),	"LANG_BELARUSIAN"},
	{6,  "Bulgarian",	DL(LANG_BULGARIAN),		"LANG_BULGARIAN"},
	{7,  "Catalan",		DL(LANG_CATALAN),		"LANG_CATALAN"},
	{8,  "Chinese",		DL(LANG_CHINESE),		"LANG_CHINESE"},
	{9,  "Croatian",	DL(LANG_CROATIAN),		"LANG_CROATIAN"},
	{10, "Czech",		DL(LANG_CZECH),			"LANG_CZECH"},
	{11, "Danish",		DL(LANG_DANISH),		"LANG_DANISH"},
	{12, "Dutch",		DL(LANG_DUTCH),			"LANG_DUTCH"},
	{13, "English",		DL(LANG_ENGLISH),		"LANG_ENGLISH"}, // English
	{14, "Estonian",	DL(LANG_ESTONIAN),		"LANG_ESTONIAN"},
	{15, "Faeroese",	DL(LANG_FAEROESE),		"LANG_FAEROESE"},
	{16, "Farsi",		DL(LANG_FARSI),			"LANG_FARSI"},
	{17, "Finnish",		DL(LANG_FINNISH),		"LANG_FINNISH"},
	{18, "French",		DL(LANG_FRENCH),		"LANG_FRENCH"}, // French
	{38, "German",		DL(LANG_GERMAN),		"LANG_GERMAN"},
	{40, "Greek",		DL(LANG_GREEK),			"LANG_GREEK"},
	{42, "Hebrew",		DL(LANG_HEBREW),		"LANG_HEBREW"},
	{44, "Hungarian",	DL(LANG_HUNGARIAN),		"LANG_HUNGARIAN"},
	{45, "Icelandic",	DL(LANG_ICELANDIC),		"LANG_ICELANDIC"},
	{46, "Indonesian",	DL(LANG_INDONESIAN),	"LANG_INDONESIAN"},
	{47, "Italian",		DL(LANG_ITALIAN),		"LANG_ITALIAN"},
	{48, "Japanese",	DL(LANG_JAPANESE),		"LANG_JAPANESE"},
	{49, "Korean",		DL(LANG_KOREAN),		"LANG_KOREAN"},
	{50, "Latvian",		DL(LANG_LATVIAN),		"LANG_LATVIAN"},
	{51, "Lithuanian",	DL(LANG_LITHUANIAN),	"LANG_LITHUANIAN"},
	{52, "Neutral",		DL(LANG_NEUTRAL),		"LANG_NEUTRAL"},
	{53, "Norwegian",	DL(LANG_NORWEGIAN),		"LANG_NORWEGIAN"},
	{54, "Polish",		DL(LANG_POLISH),		"LANG_POLISH"},
	{55, "Portuguese",	DL(LANG_PORTUGUESE),	"LANG_PORTUGUESE"},
	{56, "Romanian",	DL(LANG_ROMANIAN),		"LANG_ROMANIAN"},
	{57, "Russian",		DL(LANG_RUSSIAN),		"LANG_RUSSIAN"},
	{58, "Serbian",		DL(LANG_SERBIAN),		"LANG_SERBIAN"},
	{59, "Slovak",		DL(LANG_SLOVAK),		"LANG_SLOVAK"},
	{60, "Slovenian",	DL(LANG_SLOVENIAN),		"LANG_SLOVENIAN"},
	{61, "Spanish",		DL(LANG_SPANISH),		"LANG_SPANISH"},
	{62, "Swedish",		DL(LANG_SWEDISH),		"LANG_SWEDISH"},
	{63, "Thai",		DL(LANG_THAI),			"LANG_THAI"},
	{64, "Turkish",		DL(LANG_TURKISH),		"LANG_TURKISH"},
	{65, "Vietnamese",	DL(LANG_VIETNAMESE),	"LANG_VIETNAMESE"},
	{0,  0,				0,						0} // Null end of file marker
};

LangInfo *LangFind(char *Define, int Lang, char *Name)
{
	for (LangInfo *l = AllLanguages; l->Id; l++)
	{
		if (l->Id == Lang OR
			(Define AND stricmp(Define, l->Win32Define) == 0) OR
			(Name AND stricmp(Name, l->Name) == 0))
		{
			return l;
		}
	}

	return 0;
}

////////////////////////////////////////////////////////////////////
char Res_Dialog[]		= "Dialog";
char Res_StaticText[]	= "StaticText";
char Res_EditBox[]		= "EditBox";
char Res_CheckBox[]		= "CheckBox";
char Res_Button[]		= "Button";
char Res_Group[]		= "Group";
char Res_RadioBox[]		= "RadioBox";
char Res_TabView[]		= "TabView";
char Res_Tab[]			= "Tab";
char Res_ListView[]		= "ListView";
char Res_Column[]		= "Column";
char Res_ComboBox[]		= "ComboBox";
char Res_TreeView[]		= "TreeView";
char Res_Bitmap[]		= "Bitmap";
char Res_Progress[]		= "Progress";
char Res_Slider[]		= "Slider";
char Res_ScrollBar[]	= "ScrollBar";

////////////////////////////////////////////////////////////////////
class ResObjectImpl
{
protected:
	static int TabDepth;
	ResFactory *Factory;

	void TabString(char *Str);
	char *ReadInt(char *s, int &Value);
	bool IsEndTag(XmlTag *Tag, char *ObjName);

public:
	ResObject *Object;
	ResObjectImpl *GetImpl(ResObject *o)
	{
		return (o) ? o->GetObjectImpl(Factory) : 0;
	}

	ResObjectImpl(ResFactory *factory, ResObject *object);
	ResObjectImpl *CreateCtrl(char *Name, ResObject *Parent);

	// Store
	virtual bool Res_Read(List<XmlTag> &Tags) { return false; }
	virtual bool Res_Write(VFile &f) { return false; }

	// Ctrl
	void Res_SetPos(int x1, int y1, int x2, int y2);
	void Res_SetPos(char *s);
	GdcRegion Res_GetPos();

	void Res_SetStrRef(int Ref);
	char *Res_GetStrRef();

	// Children
	void Res_Attach(ResObjectImpl *Parent);
	bool Res_GetChildren(List<ResObjectImpl> *Children, bool Deep);

	// Items (Columns, Tabs etc)
	void Res_Append(ResObjectImpl *Parent);
	bool Res_GetItems(List<ResObjectImpl> *Items);
};


// Each object needs it's own loaded/saver
class ResDialogObj : public ResObjectImpl
{
public:
	ResDialogObj(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResStaticText : public ResObjectImpl
{
public:
	ResStaticText(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResEditBox : public ResObjectImpl
{
public:
	ResEditBox(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResCheckBox : public ResObjectImpl
{
public:
	ResCheckBox(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResButton : public ResObjectImpl
{
public:
	ResButton(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResGroup : public ResObjectImpl
{
public:
	ResGroup(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResRadio : public ResObjectImpl
{
public:
	ResRadio(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResTabView : public ResObjectImpl
{
public:
	ResTabView(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResTab : public ResObjectImpl
{
public:
	ResTab(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResColumn : public ResObjectImpl
{
public:
	ResColumn(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResListView : public ResObjectImpl
{
public:
	ResListView(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResComboBox : public ResObjectImpl
{
public:
	ResComboBox(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResScrollBar : public ResObjectImpl
{
public:
	ResScrollBar(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResTreeView : public ResObjectImpl
{
public:
	ResTreeView(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResBitmap : public ResObjectImpl
{
public:
	ResBitmap(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResSlider : public ResObjectImpl
{
public:
	ResSlider(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

class ResProgress : public ResObjectImpl
{
public:
	ResProgress(ResFactory *f, ResObject *o) : ResObjectImpl(f, o) {}
	bool Res_Read(List<XmlTag> &Tags);
	bool Res_Write(VFile &f);
};

////////////////////////////////////////////////////////////////////
ResObject::ResObject(char *Name)
{
	_ObjName = Name;
	_ObjImpl = 0;
}

ResObjectImpl *ResObject::GetObjectImpl(ResFactory *f)
{
	if (_ObjName AND
		NOT _ObjImpl AND
		f)
	{
		#define Create(CtrlName, Obj) \
			if (stricmp(_ObjName, CtrlName) == 0) \
			{ \
				_ObjImpl = NEW(Obj(f, this)); \
			}

		Create(Res_Dialog, ResDialogObj);
		Create(Res_StaticText, ResStaticText);
		Create(Res_EditBox, ResEditBox);
		Create(Res_CheckBox, ResCheckBox);
		Create(Res_Button, ResButton);
		Create(Res_Group, ResGroup);
		Create(Res_RadioBox, ResRadio);
		Create(Res_Tab, ResTab);
		Create(Res_TabView, ResTabView);
		Create(Res_ListView, ResListView);
		Create(Res_Column, ResColumn);
		Create(Res_ComboBox, ResComboBox);
		Create(Res_ScrollBar, ResScrollBar);
		Create(Res_TreeView, ResTreeView);
		Create(Res_Bitmap, ResBitmap);
		Create(Res_Slider, ResSlider);
		Create(Res_Progress, ResProgress);
	}

	LgiAssert(_ObjImpl);

	return _ObjImpl;
}

////////////////////////////////////////////////////////////////////
bool ResFactory::Res_Read(ResObject *Obj, List<XmlTag> &Tags)
{
	if (Obj)
	{
		ResObjectImpl *Impl = Obj->GetObjectImpl(this);
		if (Impl)
		{
			return Impl->Res_Read(Tags);
		}
	}
	return false;
}

bool ResFactory::Res_Write(ResObject *Obj, VFile &f)
{
	if (Obj)
	{
		ResObjectImpl *Impl = Obj->GetObjectImpl(this);
		if (Impl)
		{
			return Impl->Res_Write(f);
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////
ResObjectImpl::ResObjectImpl(ResFactory *factory, ResObject *object)
{
	Object = object;
	Factory = factory;

	LgiAssert(Object AND Object != (ResObject*)0xcdcdcdcd);
}

ResObjectImpl *ResObjectImpl::CreateCtrl(char *Name, ResObject *Parent)
{
	ResObject *o = Factory->CreateObject(Name, Parent);
	return (o) ? o->GetObjectImpl(Factory) : 0;
}

void ResObjectImpl::Res_SetPos(int x1, int y1, int x2, int y2)
{
	Factory->Res_SetPos(Object, x1, y1, x2, y2);
}

GdcRegion ResObjectImpl::Res_GetPos()
{
	return Factory->Res_GetPos(Object);
}

void ResObjectImpl::Res_SetStrRef(int Ref)
{
	Factory->Res_SetStrRef(Object, Ref);
}

char *ResObjectImpl::Res_GetStrRef()
{
	int Ref = Factory->Res_GetStrRef(Object);
	static char Buf[64];
	if (Ref >= 0)
	{
		sprintf(Buf, "Ref=%i", Ref);
	}
	else
	{
		Buf[0] = 0;
	}
	return Buf;
}

void ResObjectImpl::Res_Attach(ResObjectImpl *Parent)
{
	Factory->Res_Attach(Object, Parent->Object);
}

bool ResObjectImpl::Res_GetChildren(List<ResObjectImpl> *Children, bool Deep)
{
	List<ResObject> l;
	bool Status = Factory->Res_GetChildren(Object, &l, Deep);
	if (Status AND Children)
	{
		for (ResObject *o = l.First(); o; o = l.Next())
		{
			Children->Insert(GetImpl(o));
		}
	}

	return Status;
}

void ResObjectImpl::Res_Append(ResObjectImpl *Parent)
{
	Factory->Res_Append(Object, Parent->Object);
}

bool ResObjectImpl::Res_GetItems(List<ResObjectImpl> *Items)
{
	List<ResObject> l;
	bool Status = Factory->Res_GetItems(Object, &l);
	if (Status AND Items)
	{
		for (ResObject *o = l.First(); o; o = l.Next())
		{
			Items->Insert(GetImpl(o));
		}
	}

	return Status;
}

int ResObjectImpl::TabDepth = 0;
void ResObjectImpl::TabString(char *Str)
{
	if (Str)
	{
		for (int i=0; i<TabDepth; i++)
		{
			*Str++ = 9;
		}
		*Str = 0;
	}
}

char *ResObjectImpl::ReadInt(char *s, int &Value)
{
	char *c = strchr(s, ',');
	if (c)
	{
		*c = 0;
		Value = atoi(s);
		return c+1;
	}
	Value = atoi(s);
	return 0;
}

bool ResObjectImpl::IsEndTag(XmlTag *Tag, char *ObjName)
{
	return	Tag AND
			Tag->Name AND
			ObjName AND
			(Tag->Name[0] == '/') AND
			(stricmp(ObjName, Tag->Name+1) == 0);
}

void ResObjectImpl::Res_SetPos(char *Str)
{
	if (Str)
	{
		char *s = NewStr(Str);
		if (s)
		{
			int x1, y1, x2, y2;
			char *p = ReadInt(s, x1);
			if (p) p = ReadInt(p, y1);
			if (p) p = ReadInt(p, x2);
			if (p) p = ReadInt(p, y2);
			DeleteArray(s);

			Res_SetPos(x1, y1, x2, y2);
		}
	}
}

////////////////////////////////////////////////////////////////////
bool ResDialogObj::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Dialog) == 0)
		{
			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			for (XmlTag *t = Tags.Next(); t; t = Tags.Current())
			{
				if (IsEndTag(t, Res_Dialog))
				{
					Tags.Next();
					Status = true;
					break;
				}
				else
				{
					ResObjectImpl *Ctrl = CreateCtrl(t->Name, Object);
					if (Ctrl AND Ctrl->Res_Read(Tags))
					{
						Ctrl->Res_Attach(this);
					}
					else
					{
						LgiAssert(0);
						break;
					}
				}
			}
		}
	}

	return Status;
}

bool ResDialogObj::Res_Write(VFile &f)
{
	TabDepth = 1;
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();
	
	f.Print(	"<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
				Res_Dialog,
				Pos.x1,
				Pos.y1,
				Pos.x2,
				Pos.y2,
				Res_GetStrRef());


	List<ResObjectImpl> Children;
	if (Res_GetChildren(&Children, false))
	{
		for (ResObjectImpl *c = Children.First(); c; c = Children.Next())
		{
			c->Res_Write(f);
		}
	}

	f.Print("</%s>\r\n", Res_Dialog);
	return true;
}

////////////////////////////////////////////////////////////////////////////
bool ResStaticText::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_StaticText) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResStaticText::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print(Tabs);
	f.Print("<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Res_StaticText,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

//////////////////////////////////////////////////////////////////////////////////
bool ResEditBox::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_EditBox) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			ObjProperties Props;
			if (Tag->Get("pw", s))
			{
				Props.Set("password", atoi(s));
			}
			if (Tag->Get("enabled", s))
			{
				Props.Set("enabled", atoi(s));
			}
			if (Tag->Get("multiline", s))
			{
				Props.Set("multiline", atoi(s));
			}
			if (Props.FirstKey())
			{
				if (Factory->Res_SetProperties(Object, &Props))
				{
				}
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResEditBox::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	int Password = false;
	int Enabled = true;
	int MultiLine = false;
	ObjProperties Props;
	if (Factory->Res_GetProperties(Object, &Props))
	{
		Props.Get("Password", Password);
		Props.Get("Enabled", Enabled);
		Props.Get("MultiLine", MultiLine);
	}

	char PasswordStr[32] = "";
	if (Password)
	{
		sprintf(PasswordStr, " pw=%i", Password);
	}

	char *EnabledStr = (Enabled) ? (char*)"" : (char*)" enabled=0";
	char *MultiLineStr = (MultiLine) ? (char*)" multiline=1" : (char*)"";

	f.Print(Tabs);
	f.Print("<%s pos=\"%i,%i,%i,%i\"%s%s%s %s>\r\n",
			Res_EditBox,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			PasswordStr,
			EnabledStr,
			MultiLineStr,
			Res_GetStrRef());

	return true;
}

////////////////////////////////////////////////////////////////////////////////
bool ResCheckBox::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_CheckBox) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResCheckBox::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_CheckBox,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

/////////////////////////////////////////////////////////////////////////////////
bool ResButton::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Button) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResButton::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_Button,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

///////////////////////////////////////////////////////////////////////////////
bool ResGroup::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Group) == 0)
		{
			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			for (Tag = Tags.Next(); Tag; Tag = Tags.Current())
			{
				if (IsEndTag(Tag, Res_Group))
				{
					Tags.Next();
					Status = true;
					break;
				}
				else
				{
					ResObjectImpl *Ctrl = CreateCtrl(Tag->Name, Object);
					if (Ctrl AND Ctrl->Res_Read(Tags))
					{
						Ctrl->Res_Attach(this);
					}
					else
					{
						break;
					}
				}
			}
		}
	}

	return Status;
}

bool ResGroup::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_Group,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());
	TabDepth++;

	List<ResObjectImpl> Children;
	if (Res_GetChildren(&Children, false))
	{
		for (ResObjectImpl *c = Children.First(); c; c = Children.Next())
		{
			c->Res_Write(f);
		}
	}

	TabDepth--;
	f.Print("%s</%s>\r\n", Tabs, Res_Group);

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////
bool ResRadio::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_RadioBox) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResRadio::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_RadioBox,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

///////////////////////////////////////////////////////////////////////////////////
bool ResTab::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Tab) == 0)
		{
			char *s = 0;
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			for (Tag = Tags.Next(); Tag; Tag = Tags.Current())
			{
				if (IsEndTag(Tag, Res_Tab))
				{
					Tags.Next();
					Status = true;
					break;
				}
				else
				{
					ResObjectImpl *Ctrl = CreateCtrl(Tag->Name, Object);
					if (Ctrl AND Ctrl->Res_Read(Tags))
					{
						Ctrl->Res_Attach(this);
					}
					else
					{
						break;
					}
				}
			}
		}
	}

	return Status;
}

bool ResTab::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s %s>\r\n",
			Tabs,
			Res_Tab,
			Res_GetStrRef());
	TabDepth++;

	List<ResObjectImpl> Children;
	if (Res_GetChildren(&Children, false))
	{
		for (ResObjectImpl *c = Children.First(); c; c = Children.Next())
		{
			c->Res_Write(f);
		}
	}

	TabDepth--;
	f.Print("%s</%s>\r\n", Tabs, Res_Tab);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////
/*
void CtrlTabs::ListChildren(List<ResDialogCtrl> &l)
{
	int n=0;
	for (CtrlTab *t = Tabs->First(); t; t = Tabs->Next(), n++)
	{
		List<GView> *CList = (Current == n) ? &Children : &t->Children;
		for (GView *w = CList->First(); w; w = CList->Next())
		{
			ResDialogCtrl *c = dynamic_cast<ResDialogCtrl*>(w);
			if (c)
			{
				l->Insert(c);
				c->ListChildren(l);
			}
		}
	}
}
*/

bool ResTabView::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_TabView) == 0)
		{
			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			for (Tag = Tags.Next(); Tag; Tag = Tags.Current())
			{
				if (IsEndTag(Tag, Res_TabView))
				{
					Tags.Next();
					Status = true;
					break;
				}
				else if (stricmp(Tag->Name, Res_Tab) == 0)
				{
					ResObjectImpl *Tab = CreateCtrl(Res_Tab, Object);
					if (Tab AND Tab->Res_Read(Tags))
					{
						Res_Append(Tab);
					}
					else
					{
						break;
					}
				}
				else
				{
					break;
				}
			}
		}
	}

	return Status;
}

bool ResTabView::Res_Write(VFile &f)
{
	char TabStr[256];
	TabString(TabStr);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			TabStr, 
			Res_TabView,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());
	TabDepth++;

	List<ResObjectImpl> Items;
	if (Res_GetItems(&Items))
	{
		for (ResObjectImpl *Tab = Items.First(); Tab; Tab = Items.Next())
		{
			Tab->Res_Write(f);
		}
	}

	TabDepth--;
	f.Print("%s</%s>\r\n", TabStr, Res_TabView);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////
bool ResColumn::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Column) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("width", s))
			{
				Res_SetPos(0, 0, atoi(s) - 1, 20);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResColumn::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s %s width=%i>\r\n", Tabs, Res_Column, Res_GetStrRef(), Pos.X());

	return true;
}

//////////////////////////////////////////////////////////////////////////////
/*
void CtrlList::ListChildren(List<ResDialogCtrl> &l)
{
	for (GView *w = Cols->First(); w; w = Cols->Next())
	{
		ResDialogCtrl *c = dynamic_cast<ResDialogCtrl*>(w);
		if (c)
		{
			l->Insert(c);
			c->ListChildren(l);
		}
	}
}
*/

bool ResListView::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_ListView) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			for (Tag = Tags.Next(); Tag; Tag = Tags.Current())
			{
				if (IsEndTag(Tag, Res_ListView))
				{
					Tags.Next();
					Status = true;
					break;
				}
				else if (stricmp(Tag->Name, Res_Column) == 0)
				{
					ResObjectImpl *Col = CreateCtrl(Res_Column, Object);
					if (Col AND Col->Res_Read(Tags))
					{
						Res_Append(Col);
					}
					else
					{
						break;
					}
				}
				else
				{
					break;
				}
			}
		}
	}

	return Status;
}

bool ResListView::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_ListView,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());
	TabDepth++;

	List<ResObjectImpl> Items;
	if (Res_GetItems(&Items))
	{
		for (ResObjectImpl *c = Items.First(); c; c = Items.Next())
		{
			c->Res_Write(f);
		}
	}

	TabDepth--;
	f.Print("%s</%s>\r\n", Tabs, Res_ListView);

	return true;
}

////////////////////////////////////////////////////////////////////
bool ResComboBox::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_ComboBox) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResComboBox::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_ComboBox,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

////////////////////////////////////////////////////////////////////
bool ResScrollBar::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_ScrollBar) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResScrollBar::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_ScrollBar,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

/////////////////////////////////////////////////////////////
bool ResTreeView::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_TreeView) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResTreeView::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_TreeView,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

/////////////////////////////////////////////////////////////
bool ResBitmap::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Bitmap) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResBitmap::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_Bitmap,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

/////////////////////////////////////////////////////////////
bool ResSlider::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Slider) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResSlider::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_Slider,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

/////////////////////////////////////////////////////////////
bool ResProgress::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *Tag = Tags.Current();
	if (Tag)
	{
		if (stricmp(Tag->Name, Res_Progress) == 0)
		{
			Status = true;

			char *s = 0;
			if (Tag->Get("pos", s))
			{
				Res_SetPos(s);
			}
			if (Tag->Get("ref", s))
			{
				Res_SetStrRef(atoi(s));
			}

			Tags.Next();
		}
	}

	return Status;
}

bool ResProgress::Res_Write(VFile &f)
{
	char Tabs[256];
	TabString(Tabs);
	GdcRegion Pos = Res_GetPos();

	f.Print("%s<%s pos=\"%i,%i,%i,%i\" %s>\r\n",
			Tabs,
			Res_Progress,
			Pos.x1,
			Pos.y1,
			Pos.x2,
			Pos.y2,
			Res_GetStrRef());

	return true;
}

////////////////////////////////////////////////////////////////////
/*
bool ResDialog::Res_Read(List<XmlTag> &Tags)
{
	bool Status = false;

	// turn symbol creation off so that ctrls find their
	// symbol via Id matching instead of creating their own
	CreateSymbols = false;

	ResDialogCtrl *Dlg = dynamic_cast<ResDialogCtrl*>(Children->First());
	if (Dlg)
	{
		Symbols->DeleteStr(Dlg->Str);
		Dlg->Str = 0;
	}

	// Load the resource
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children->First());
	if (Ctrl)
	{
		Status = Ctrl->Read(Tags);
		Item->Update();
	}

	CreateSymbols = true;
	return Status;
}

bool ResDialog::Write(VFile &f)
{
	bool Status = false;
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children->First());
	if (Ctrl)
	{
		// duplicates symbols should have been removed before the 
		// strings were written out
		// so we don't have to do it here->

		// write the resources out
		if (Ctrl)
		{
			Status = Ctrl->Write(f);
		}
	}

	return Status;
}

*/
