#include "Lgi.h"
#include "GZipLibrary.h"

//////////////////////////////////////////////////////////////////////////////
// Zip lib support

GLogTarget *GZipLibrary::Log = 0;

GZipLibrary::GZipLibrary(GLogTarget *log) : GLibrary("Zip32")
{
	#ifdef _DEBUG
	if (NOT IsLoaded())
	{
		// try some other paths...
		if (NOT Load("\\CodeLib\\Zip23\\windll\\Debug\\Zip32"))
		{
			Load("..\\Zip\\Zip32");
		}
	}
	#endif

	Log = log;
	UserFuncs.print = DllPrint;
	UserFuncs.comment = DllComment;
	UserFuncs.password = DllPassword;
	UserFuncs.ServiceApplication = DllService;
	InitResult = 0;

	if (IsLoaded())
	{
		ZpVersion =		(_ZpVersion)	GetAddress("ZpVersion");
		ZpInit =		(_ZpInit)		GetAddress("ZpInit");
		ZpSetOptions =	(_ZpSetOptions)	GetAddress("ZpSetOptions");
		ZpGetOptions =	(_ZpGetOptions)	GetAddress("ZpGetOptions");
		ZpArchive =		(_ZpArchive)	GetAddress("ZpArchive");

		if (IsOk())
		{
			InitResult = ZpInit(&UserFuncs);
		}
	}
}

bool GZipLibrary::IsOk()
{
	return	IsLoaded() AND
			ZpVersion AND
			ZpInit AND
			ZpSetOptions AND
			ZpGetOptions AND
			ZpArchive;
}

bool GZipLibrary::Zip(char *ZipFile, char *BaseDir, List<char> &Extensions, char *AfterDate)
{
	bool Status = false;
	if (ZipFile AND
		BaseDir AND
		IsOk())
	{
		ZPOPT Opt;

		ZeroObj(Opt);
		Opt.fRecurse = 2;
		Opt.szRootDir =  BaseDir;
		Opt.Date = AfterDate;
		Opt.fIncludeDate = AfterDate != 0;

		ZpSetOptions(&Opt);

		char **Ext = NEW(char*[Extensions.GetItems()]);
		if (Ext)
		{
			int n=0;
			for (char *e=Extensions.First(); e; e=Extensions.Next())
			{
				Ext[n++] = e;
			}

			ZCL Cmd;
			Cmd.argc = Extensions.GetItems();
			Cmd.lpszZipFN = ZipFile;
			Cmd.FNV = Ext;

			Status = NOT ZpArchive(Cmd);
		}
	}

	return Status;
}

int WINAPI GZipLibrary::DllPrint(LPSTR str, unsigned long i)
{
	if (Log)
	{
		char Buf[256];
		char *Out = Buf;
		for (char *In = str; 1; In++)
		{
			if (NOT *In OR *In == '\n')
			{
				*Out++ = 0;

				if (Buf[0] AND
					NOT stristr(Buf, "adding:") AND
					NOT stristr(Buf, "deflated"))
				{
					Log->Log(Buf, stristr(Buf, "zip error") ? Rgb24(128, 0, 0) : Rgb24(0x80, 0x80, 0x80));
				}

				Out = Buf;

				if (NOT *In) break;
			}
			else if (*In == '\t')
			{
				*Out++ = ' ';
				*Out++ = ' ';
				*Out++ = ' ';
			}
			else
			{
				*Out++ = *In;
			}
		}
	}
	return 0;
}

int WINAPI GZipLibrary::DllComment(LPSTR s)
{
	return 0;
}

int WINAPI GZipLibrary::DllPassword(LPSTR str, int i, LPCSTR a, LPCSTR b)
{
	return 0;
}

int WINAPI GZipLibrary::DllService(LPCSTR str, unsigned long i)
{
	return 0;
}
