/*hdr
**      FILE:           Device.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           10/4/98
**      DESCRIPTION:    Device class
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

/****************************** Includes ********************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#include "LibDefs.h"
#include "LgiOsDefs.h"
#include "Device.h"

/****************************** Classes *********************************************/

bool GDevice::Verbose = true;
GDevice *GDevice::First = NULL;

bool GDevice::AddDevice(GDevice *pDev)
{
	if (First)
	{
		GDevice *Temp = First;
		while (Temp->Next)
		{
			Temp = Temp->Next;
		}
		Temp->Next = pDev;
		pDev->Next = 0;
		pDev->Prev = Temp;
	}
	else
	{
		First = pDev;
		pDev->Next = 0;
		pDev->Prev = 0;
	}
	
	return TRUE;
}

bool GDevice::RemoveDevice(GDevice *pDev)
{
	if (First && pDev)
	{
		GDevice *Temp = First;
		
		while (Temp && Temp != pDev)
		{
			Temp = Temp->Next;
		}
		
		if (Temp == pDev)
		{
			if (pDev == First) First = pDev->Next;
			if (pDev->Next) pDev->Next->Prev = pDev->Prev;
			if (pDev->Prev) pDev->Prev->Next = pDev->Next;
			return TRUE;
		}
	}
	
	return FALSE;
}

GDevice *GDevice::FindDevice(char *Name)
{
	GDevice *Temp = First;
	while (Temp)
	{
		if (stricmp(Name, Temp->Name) == 0)
		{
			return Temp;
		}
		Temp = Temp->Next;
	}
	
	return 0;
}

bool GDevice::Print(char *str, ...)
{
	if (str)
	{
		if (Verbose)
		{
			// char buffer[256];
			va_list arg;

			va_start(arg ,str);
			// vprintf(str, arg);
			va_end(arg);
		}
		
		return TRUE;
	}
	
	return FALSE;
}

bool GDevice::SetVerbose(bool v)
{
	bool Prev = Verbose;
	Verbose = v;
	return Prev;
}

void GDevice::CriticalExit(char *str, ...)
{
	if (str)
	{
		va_list arg;

		va_start(arg ,str);
		vprintf(str, arg);
		va_end(arg);
		
	}

	exit(-1);
}

GDevice::GDevice()
{
	AddDevice(this);
}

GDevice::~GDevice()
{
	RemoveDevice(this);
}

//////////////////////////////////////////////////////////////////////////////
DoEvery::DoEvery(int p) // p = timeout in ms
{
	Init(p);
}

void DoEvery::Init(int p) // p = timeout in ms
{
	LastTime = clock();
	if (p > 0)
	{
		Period = (p * CLOCKS_PER_SEC) / 1000;
	}
}

bool DoEvery::DoNow()
{
	int Now = clock();
	if (LastTime + Period < Now)
	{
		LastTime = Now;
		return TRUE;
	}
	return FALSE;
}

