/*hdr
**	FILE:			File.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			8/10/2000
**	DESCRIPTION:	The new file subsystem
**
**	Copyright (C) 2000, Matthew Allen
**		fret@ozemail.com.au
*/

/****************************** Includes **********************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>

#include "LibDefs.h"
#include "VFile.h"

/****************************** Defines ***********************************/

// #define FILEDEBUG

#define FLOPPY_360K				0x0001
#define FLOPPY_720K				0x0002
#define FLOPPY_1_2M				0x0004
#define FLOPPY_1_4M				0x0008
#define FLOPPY_5_25				(FLOPPY_360K | FLOPPY_1_2M)
#define FLOPPY_3_5				(FLOPPY_720K | FLOPPY_1_4M)

/****************************** Globals ***********************************/

FileDevice *FileDevice::Instance = 0;

/****************************** Helper Functions **************************/

int FileSize(char *FileName)
{
	VFile F;
	if (FileName AND F.Open(FileName, O_READ))
	{
		return F.GetSize();
	}
	return 0;
}

bool FileExists(char *Name)
{
	VFile F;
	return F.Open(Name, O_READ);
}

bool ResolveShortcut(char *LinkFile, char *Path, int Len) 
{
	bool Status = FALSE;

	return Status;
}

void WriteStr(VFile &f, char *s)
{
	ulong Len = (s) ? strlen(s) : 0;
	f << Len;
	if (Len > 0)
	{
		f.Write(s, Len);
	}
}

char *ReadStr(VFile &f)
{
	ulong Len;
	f >> Len;
	char *s = 0;

	if (Len > 0)
	{
		s = NEW(char[Len+1]);
		if (s)
		{
			f.Read(s, Len);
			s[Len] = 0;
		}
		else
		{
			f.Seek(Len, SEEK_CUR);
		}
	}

	return s;
}

int SizeofStr(char *s)
{
	return sizeof(ulong) + ((s) ? strlen(s) : 0);
}

/****************************** Classes *************************************************************************************/

FileDevice::FileDevice()
{
	Instance = this;
	strcpy(Name, "File");
	Print("Starting file...\n");

	FillDiskList();
	Print("\t%i disks found\n", NumDrive);
}

FileDevice::~FileDevice()
{
	DeleteArray(DriveList);
}

int FloppyType(int Letter)
{
	uchar MaxTrack;
	uchar SecPerTrack;

	/*
	_asm {
		mov eax, 0800h
		mov edx, Letter
		int 13h
		mov MaxTrack, ch
		mov SecPerTrack, cl
	}
	
	if (MaxTrack > 39)
	{
		switch (SecPerTrack)
		{
			case 9:
			{
				return FLOPPY_720K;
			}
			case 15:
			{
				return FLOPPY_1_2M;
			}
			case 18:
			{
				return FLOPPY_1_4M;
			}
		}
	}
	else
	{
		return FLOPPY_360K;
	}
	*/

	return 0;
}

DriveInfo *FileDevice::GetDriveInfo(int Drive)
{
	/*
	if (DriveList AND Drive >= 0 AND Drive < 26)
	{
		DriveInfo *Info = DriveList + Drive;

		if (Info->Type == DIF_NONE)
		{
			char DriveName[] = "A:\\";
			char VolumeName[256];
			ulong SerialNum;
			ulong MaxComponentLen;
			ulong Flags;
			char FileSysName[256];

			DriveName[0] = 'A' + Drive;

			if (GetVolumeInformation(	DriveName,
										VolumeName,
										sizeof(VolumeName),
										&SerialNum,
										&MaxComponentLen,
										&Flags,
										FileSysName,
										sizeof(FileSysName)))
			{
				NumDrive++;

				switch (GetDriveType(DriveName))
				{
					case DRIVE_REMOVABLE:
					{
						if (Drive < 2)
						{
							int Type = FloppyType(Drive);

							switch (Type)
							{
								case FLOPPY_360K:
								case FLOPPY_1_2M:
									Info->Type = DIF_5_25FLOPPY;
									break;
								case FLOPPY_720K:
								case FLOPPY_1_4M:
									Info->Type = DIF_3_5FLOPPY;
									break;
							}
						}
						else
						{
							Info->Type = DIF_REMOVABLE;
						}
						break;
					}

					case DRIVE_UNKNOWN:
					case DRIVE_NO_ROOT_DIR:
					case DRIVE_FIXED:
					{
						Info->Type = DIF_HARDDISK;
						break;
					}

					case DRIVE_REMOTE:
					{
						Info->Type = DIF_NET;
						break;
					}

					case DRIVE_CDROM:
					{
						Info->Type = DIF_CDROM;
						break;
					}

					case DRIVE_RAMDISK:
					{
						Info->Type = DIF_RAM;
						break;
					}

					default:
					{
						Info->Type = DIF_NONE;
						break;
					}
				}
			}
		}

		return Info;
	}
	*/

	return NULL;
}

int FileDevice::FillDiskList()
{
	NumDrive = 0;
	DriveList = 0;
	/*
	DriveList = NEW(DriveInfo[26]);
	if (DriveList)
	{
		memset(DriveList, 0, sizeof(DriveInfo[26]));

		for (int i=2; i<26; i++)
		{
			GetDriveInfo(i);
		}
	}
	*/

	return TRUE;
}

VDirectory *FileDevice::GetDir(char *DirName)
{
	return NEW(VLinuxDir);
}

bool FileDevice::DeleteFile(char *FileName)
{
	return unlink(FileName) == 0;
}

bool FileDevice::CreateDirectory(char *PathName)
{
	return mkdir(PathName, S_IRWXU | S_IXGRP | S_IXOTH);
}

bool FileDevice::RemoveDirectory(char *PathName)
{
	return unlink(PathName) == 0;
}

bool FileDevice::SetCurrentDirectory(char *PathName)
{
	return chdir(PathName) == 0;
}

bool FileDevice::GetCurrentDirectory(char *PathName, int Length)
{
	return getcwd(PathName, Length) != 0;
}

bool FileDevice::MoveFile(char *OldName, char *NewName)
{
	return rename(OldName, NewName) != -1;
}

bool FileDevice::GetVolumeInfomation(char Drive, VolumeInfo *pInfo)
{
	bool Status = FALSE;

	/*
	if (pInfo)
	{
		char Name[] = "A:\\";
		ulong Serial;

		Name[0] = Drive + 'A';

		if (GetVolumeInformation(	Name,
						pInfo->Name,
						sizeof(pInfo->Name),
						&Serial,
						&pInfo->MaxPath,
						&pInfo->Flags,
						pInfo->System,
						sizeof(pInfo->System)))
		{
			pInfo->Drive = Drive;
			Status = TRUE;
		}
	}
	*/

	return Status;
}

/*
void GetFlagsStr(char *flagstr, short flags)
{
	if (flagstr)
	{
		flagstr[0] = 0;

		if (flags & O_WRONLY)
		{
			strcat(flagstr,"w");
		}
		else if (flags & O_APPEND)
		{
			strcat(flagstr,"a");
		}
		else
		{
			strcat(flagstr,"r");
		}

		if ((flags & O_RDWR) == O_RDWR)
		{
			strcat(flagstr,"+");
		}

		strcat(flagstr,"b");
	}
}
*/

bool Match(char *Name, char *Mask)
{
	strupr(Name);
	strupr(Mask);

	while (*Name AND *Mask)
	{
		if (*Mask == '*')
		{
			if (*Name == *(Mask+1))
			{
				Mask++;
			}
			else
			{
				Name++;
			}
		}
		else if (*Mask == '?' OR *Mask == *Name)
		{
			Mask++;
			Name++;
		}
		else
		{
			return FALSE;
		}
	}

	while (*Mask AND ((*Mask == '*') OR (*Mask == '.'))) Mask++;

	return (*Name == 0 AND *Mask == 0);
}

short DaysInMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
int LeapYear(int year)
{
	if (year & 3)
	{
		return 0;
	}
	if ((year % 100 == 0) && !(year % 400 == 0))
	{
		return 0;
	}
	
	return 1;
}

/////////////////////////////////////////////////////////////////////////////////
bool VDirectory::ConvertToTime(char *Str, quad Time)
{
	return false;
}

bool VDirectory::ConvertToDate(char *Str, quad Time)
{
	return false;
}

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// Directory //////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

VLinuxDir::VLinuxDir()
{
	Dir = 0;
	De = 0;
	Path = 0;
}

VLinuxDir::~VLinuxDir()
{
	DeleteArray(Path);
	Close();
}

bool VLinuxDir::First(char *Name, char Must, char Search)
{
	Close();

	if (Name)
	{
		char *End = 0;
		for (End=Name+strlen(Name)-1; End>Name AND *End!=DIR_CHAR; End--);
		if (End > Name)
		{
			char Buffer[260];

			if (End[-1] == ':')
			{
				// raw drive
				strcpy(Buffer, Name);
			}
			else
			{
				// dir	
				*End = 0;
				strcpy(Buffer, End);
				if (Buffer[strlen(Buffer)-1] != DIR_CHAR) strcat(Buffer, DIR_STR);
				strcat(Buffer, End+1);
			}

			Dir = opendir(Buffer);
			if (Dir)
			{
				De = readdir(Dir);

				char *Ast = strchr(Buffer, '*');
				if (Ast)
				{
					*Ast = 0;
					Path = NewStr(Buffer);
					strcpy(Ast, De->d_name);
					stat(Buffer, &Stat);
				}
			}
		}
	}

	return Dir != 0 AND De != 0;
}

bool VLinuxDir::Next()
{
	int Status = FALSE;
	
	if (Dir)
	{
		De = readdir(Dir);
		if (De AND Path)
		{
			char Str[260];
			sprintf(Str, "%s%s", Path, De->d_name);
		}
		Status = De != 0;
	}

	return Status;
}

bool VLinuxDir::Close()
{
	if (Dir)
	{
		closedir(Dir);
		Dir = 0;
	}
	De = 0;

	return true;
}

bool VLinuxDir::IsDir()
{
	return S_ISDIR(GetAttributes());
}

long VLinuxDir::GetAttributes()
{
	return Stat.st_mode;
}

char *VLinuxDir::GetName()
{
	return (De) ? De->d_name : 0;
}

void VLinuxDir::GetName(char *Buf, int Len)
{
	if (De)
	{
		strncpy(Buf, De->d_name, Len);
	}
}

const quad VLinuxDir::GetCreationTime()
{
	return Stat.st_ctime;
}

const quad VLinuxDir::GetLastAccessTime()
{
	return Stat.st_atime;
}

const quad VLinuxDir::GetLastWriteTime()
{
	return Stat.st_mtime;
}

const quad VLinuxDir::GetSize()
{
	return Stat.st_size;
}

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// File ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

VFile::VFile()
{
	Handle = INVALID_HANDLE;
	Name = 0;
	Swap = false;
	Status = false;
	Attributes = 0;
}

VFile::~VFile()
{
	if (ValidHandle(Handle))
	{
		Close();
	}
	DeleteArray(Name);
}

bool VFile::Open()
{
	return ValidHandle(Handle);
}

bool VFile::Open(char *File, int Mode)
{
	int Status = FALSE;

	if (File)
	{
		if ((Mode & O_WRITE) != 0)
		{
			Mode |= O_CREAT;
		}

		Close();
		Handle = open(File, Mode, 0);

		if (ValidHandle(Handle))
		{
			Attributes = Mode;
			Name = NEW(char[strlen(File)+1]);
			if (Name)
			{
				strcpy(Name, File);
			}
			Status = true;
			VFile::Status = false;
		}
	}

	return Status;
}

bool VFile::Close()
{
	if (ValidHandle(Handle))
	{
		close(Handle);
		Handle = INVALID_HANDLE;
		DeleteArray(Name);
	}

	return true;
}

int VFile::Print(char *Format, ...)
{
	int Chars = 0;

	if (Format)
	{
		char Buffer[1024];
		va_list Arg;

		va_start(Arg, Format);
		Chars = vsprintf(Buffer, Format, Arg);
		va_end(Arg);

		if (Chars > 0)
		{
			Write(Buffer, Chars);
		}
	}

	return Chars;
}

#define CHUNK		0xFFF0

int VFile::Read(void *Buffer, int Size)
{
	int Red = 0;

	if (Buffer AND Size > 0)
	{
		Red = read(Handle, Buffer, Size);
	}
	Status = Red == Size;

	return max(Red, 0);
}

int VFile::Write(void *Buffer, int Size)
{
	int Written = 0;

	if (Buffer AND Size > 0)
	{
		Written = write(Handle, Buffer, Size);
	}
	Status = Written == Size;

	return max(Written, 0);
}

int VFile::Seek(int To, int Whence)
{
	return lseek(Handle, To, Whence);
}

int VFile::GetPosition()
{
	return lseek(Handle, 0, SEEK_CUR);
}

int VFile::GetSize()
{
	int Here = GetPosition();
	int Size = lseek(Handle, 0, SEEK_END);
	lseek(Handle, Here, SEEK_SET);
	return Size;
}

bool VFile::SetSize(int Size)
{
	if (Name AND ValidHandle(Handle))
	{
		int Pos = GetPosition();
		close(Handle);
		truncate(Name, Size);
		Handle = open(Name, Attributes, 0);
		if (Handle)
		{
			lseek(Handle, Pos, SEEK_SET);
		}
	}

	return GetSize();
}

bool VFile::Eof()
{
	return GetPosition() >= GetSize();
}

int VFile::SwapRead(uchar *Buf, int Size)
{
	int i = 0;
	int r = 0;
	Buf = &Buf[Size-1];
	while (Size--)
	{
		r = read(Handle, Buf--, 1);
		i += r;
	}
	return i;
}

int VFile::SwapWrite(uchar *Buf, int Size)
{
	int i = 0;
	int w = 0;
	Buf = &Buf[Size-1];
	while (Size--)
	{
		w = write(Handle, Buf--, 1);
		i += w;
	}
	return i;
}

int VFile::ReadStr(char *Buf, int Size)
{
	int i = 0;
	int r = 0;
	if (Buf AND Size > 0)
	{
		char c;

		Size--;

		do
		{
			r = read(Handle, &c, 1);
			if (Eof())
			{
				break;
			}
			
			*Buf++ = c;
			i++;
		
		} while (i < Size - 1 AND c != '\n');

		*Buf = 0;
	}

	return i;
}

int VFile::WriteStr(char *Buf, int Size)
{
	int i = 0;
	int w;

	while (i <= Size)
	{
		w = write(Handle, Buf, 1);
		Buf++;
		i++;
		if (*Buf == '\n') break;
	}

	return i;
}

void VFile::SetStatus(bool s)
{
	Status = s;
}

bool VFile::GetStatus()
{
	return Status;
}

#define RdIO { Status |= ((Swap) ? SwapRead((uchar*) &i, sizeof(i)) : Read(&i, sizeof(i))) != sizeof(i); return *this; }
#define WrIO { Status |= ((Swap) ? SwapWrite((uchar*) &i, sizeof(i)) : Write(&i, sizeof(i))) != sizeof(i); return *this; }

VFile &VFile::operator >> (char &i) RdIO
VFile &VFile::operator >> (signed char &i) RdIO
VFile &VFile::operator >> (unsigned char &i) RdIO
VFile &VFile::operator >> (signed short &i) RdIO
VFile &VFile::operator >> (unsigned short &i) RdIO
VFile &VFile::operator >> (signed int &i) RdIO
VFile &VFile::operator >> (unsigned int &i) RdIO
VFile &VFile::operator >> (signed long &i) RdIO
VFile &VFile::operator >> (unsigned long &i) RdIO
VFile &VFile::operator >> (float &i) RdIO
VFile &VFile::operator >> (double &i) RdIO
VFile &VFile::operator >> (long double &i) RdIO
VFile &VFile::operator >> (quad &i) RdIO

VFile &VFile::operator << (char i) WrIO
VFile &VFile::operator << (signed char i) WrIO
VFile &VFile::operator << (unsigned char i) WrIO
VFile &VFile::operator << (signed short i) WrIO
VFile &VFile::operator << (unsigned short i) WrIO
VFile &VFile::operator << (signed int i) WrIO
VFile &VFile::operator << (unsigned int i) WrIO
VFile &VFile::operator << (signed long i) WrIO
VFile &VFile::operator << (unsigned long i) WrIO
VFile &VFile::operator << (float i) WrIO
VFile &VFile::operator << (double i) WrIO
VFile &VFile::operator << (long double i) WrIO
VFile &VFile::operator << (quad i) WrIO

VFileBuf::VFileBuf()
{
	BufSize = 0;
	BufPos = 0;
	Data = 0;
}

VFileBuf::~VFileBuf()
{
	Close();
}

bool VFileBuf::Open(char *Name, int Attrib)
{
	bool Status = FALSE;

	if (VFile::Open(Name, Attrib))
	{
		BufPos = 0;
		BufSize = VFile::GetSize();
		Data = NEW(uchar[BufSize]);
		if (Data)
		{
			VFile::Read(Data, BufSize);
			Status = TRUE;
		}
	}

	return Status;
}

bool VFileBuf::Open()
{
	return (Data != 0) AND ValidHandle(Handle);
}

bool VFileBuf::Close()
{
	BufSize = 0;
	BufPos = 0;
	DeleteArray(Data);
	VFile::Close();
	return TRUE;
}

int VFileBuf::Read(void *Buffer, int Size)
{
	int Len = 0;

	if (Buffer AND Data AND BufPos >= 0 AND BufPos + 1 < BufSize)
	{
		Len = min(Size, BufSize - BufPos);
		memcpy(Buffer, Data + BufPos, Len);
		BufPos += Len;
	}

	return Len;
}

int VFileBuf::Write(void *Buffer, int Size)
{
	int Len = 0;

	if (Buffer AND BufPos >= 0 AND BufPos + 1 < BufSize)
	{
		Len = min(Size, BufSize - BufPos);
		memcpy(Data + BufPos, Buffer, Len);
		BufPos += Len;

		Seek(BufPos, SEEK_SET);
		Len = VFile::Write(Buffer, Size);
	}

	return Len;
}

int VFileBuf::Seek(int To, int Whence)
{
	int Start = BufPos;

	switch (Whence)
	{
		case SEEK_SET:
		{
			BufPos = To;
			break;
		}
		case SEEK_CUR:
		{
			BufPos += To;
			break;
		}
		case SEEK_END:
		{
			BufPos = BufSize + To;
			break;
		}
	}

	if (BufPos != Start)
	{
		VFile::Seek(To, Whence);
	}
	
	return BufPos;
}

int VFileBuf::GetPosition()
{
	return BufPos;
}

int VFileBuf::GetSize()
{
	return BufSize;
}

bool VFileBuf::SetSize(int Size)
{
	bool Status = FALSE;

	if (Size != BufSize)
	{
		int Common = min(Size, BufSize);
		uchar *Temp = NEW(uchar[Size]);
		if (Temp)
		{
			memcpy(Temp, Data, Common);
			if (Common < Size)
			{
				memset(Temp + Common, 0, Size - Common);
			}

			DeleteArray(Data);
			Data = Temp;
			BufSize = Size;
			Status = VFile::SetSize(Size);
		}
	}

	return Status;
}

bool VFileBuf::Eof()
{
	return (BufPos >= 0 AND BufPos < BufSize);
}

int VFileBuf::ReadStr(char *Buf, int Size)
{
	int i;

	for (i=0; i<Size-1 AND Data[BufPos++] != '\n'; i++);
	if (Buf)
	{
		memcpy(Buf, Data+BufPos, i);
		Buf[Size-1] = 0;
		return i;
	}
	return 0;
}

int VFileBuf::WriteStr(char *Buf, int Size)
{
	if (Buf)
	{
		int i;
		uchar *p = Data + BufPos;
		for (i=0; i<Size AND Buf[BufPos] != '\n'; i++)
			*p++ = Buf[BufPos++];

		Seek(BufPos, SEEK_SET);
		VFile::Write(Buf, i);

		return i;
	}
	return 0;
}

int VFileBuf::SwapRead(uchar *Buf, int Size)
{
	int i=0;
	Buf = &Buf[Size-1];
	while (Size-- AND BufPos < BufSize)
	{
		*Buf-- = Data[BufPos++];
		i++;
	}
	return i;
}

int VFileBuf::SwapWrite(uchar *Buf, int Size)
{
	int i=0;
	int Start = BufPos;

	Buf = &Buf[Size-1];
	while (Size-- AND BufPos < BufSize)
	{
		Data[BufPos++] = *Buf--;
		i++;
	}

	Seek(Start, SEEK_SET);
	VFile::Write(Data + Start, i);

	return i;
}

#define BufRdIO { Status |= ((Swap) ? SwapRead((uchar*) &i, sizeof(i)) : Read(&i, sizeof(i))) != sizeof(i); return *this; }
#define BufWrIO { Status |= ((Swap) ? SwapWrite((uchar*) &i, sizeof(i)) : Write(&i, sizeof(i))) != sizeof(i); return *this; }

VFile &VFileBuf::operator >> (char &i) BufRdIO
VFile &VFileBuf::operator >> (signed char &i) BufRdIO
VFile &VFileBuf::operator >> (unsigned char &i) BufRdIO
VFile &VFileBuf::operator >> (signed short &i) BufRdIO
VFile &VFileBuf::operator >> (unsigned short &i) BufRdIO
VFile &VFileBuf::operator >> (signed int &i) BufRdIO
VFile &VFileBuf::operator >> (unsigned int &i) BufRdIO
VFile &VFileBuf::operator >> (signed long &i) BufRdIO
VFile &VFileBuf::operator >> (unsigned long &i) BufRdIO
VFile &VFileBuf::operator >> (float &i) BufRdIO
VFile &VFileBuf::operator >> (double &i) BufRdIO
VFile &VFileBuf::operator >> (long double &i) BufRdIO

VFile &VFileBuf::operator << (char i) BufWrIO
VFile &VFileBuf::operator << (signed char i) BufWrIO
VFile &VFileBuf::operator << (unsigned char i) BufWrIO
VFile &VFileBuf::operator << (signed short i) BufWrIO
VFile &VFileBuf::operator << (unsigned short i) BufWrIO
VFile &VFileBuf::operator << (signed int i) BufWrIO
VFile &VFileBuf::operator << (unsigned int i) BufWrIO
VFile &VFileBuf::operator << (signed long i) BufWrIO
VFile &VFileBuf::operator << (unsigned long i) BufWrIO
VFile &VFileBuf::operator << (float i) BufWrIO
VFile &VFileBuf::operator << (double i) BufWrIO
VFile &VFileBuf::operator << (long double i) BufWrIO

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// VDirView ///////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

VDirView::VDirView()
{
	// File = (FileDevice*) Device::FindDevice("File");
	// Keyboard = (KeyboardDevice*) Device::FindDevice("Keyboard");

	Items = 0;
	Files = 0;
	Dirs = 0;
	strcpy(CurrentDir, ".");


	Index = 0;
	Root = 0;
}

VDirView::~VDirView()
{
	DeleteObj(Root);
	DeleteArray(Index);
}

bool VDirView::Read(char *Dir)
{
	if (NOT Dir) Dir = ".";
	FileDev->SetCurrentDirectory(Dir);
	FileDev->GetCurrentDirectory(CurrentDir, sizeof(CurrentDir));

	if (Root)
	{
		delete Root;
		Root = 0;
	}

	Items = 0;
	Files = 0;
	Dirs = 0;
	DeleteArray(Index);

	Node *Current;
	if (strlen(CurrentDir) > 3)
	{
		Current = NEW(Node("..", FA_DIRECTORY, 0));
		if (Current)
		{
			if (NOT Root)
			{
				Root = Current;
			}
			else
			{
				Root->Add(Current);
			}

			Dirs++;
			Items++;
		}
	}

	bool Done = First("*.*", 0x10, 0x1F);
	while (NOT Done)
	{
		if (strcmp(GetName(), ".") AND strcmp(GetName(), ".."))
		{
			Current = NEW(Node(GetName(), GetAttributes(), GetSize()));
			if (Current)
			{
				if (NOT Root)
				{
					Root = Current;
				}
				else
				{
					Root->Add(Current);
				}

				Dirs++;
				Items++;
			}
		}
		Done = Next();
	}

	Done = First("*.*", 0x0, 0x2F);
	while (NOT Done)
	{
		Current = NEW(Node(GetName(), GetAttributes(), GetSize()));
		if (Current)
		{
			if (NOT Root)
			{
				Root = Current;
			}
			else
			{
				Root->Add(Current);
			}

			Files++;
			Items++;
		}
		Done = Next();
	}

	if (Root AND Items)
	{
		Index = NEW(Node*[Items]);
		if (Index)
		{
			Node **Ptr = Root->Traverse(Index, TRUE);
			Root->Traverse(Ptr, FALSE);
		}
	}

	return TRUE;
}

VDirView::Node::Node()
{
	Left = Right = 0;
}

VDirView::Node::Node(char *n, int a, quad s) : NodeInfo(n, a, s)
{
	Left = Right = 0;
}

VDirView::Node::~Node()
{
	DeleteObj(Left);
	DeleteObj(Right);
}

VDirView::Node **VDirView::Node::Traverse(VDirView::Node **ppNode, int Dir)
{
	if (Left)
	{
		ppNode = Left->Traverse(ppNode, Dir);
	}

	if (Dir)
	{
		if (GetAttributes() & FA_DIRECTORY)
		{
			*ppNode++ = this;
		}
	}
	else
	{
		if (NOT (GetAttributes() & FA_DIRECTORY))
		{
			*ppNode++ = this;
		}
	}

	if (Right)
	{
		ppNode = Right->Traverse(ppNode, Dir);
	}

	return ppNode;
}

void VDirView::Node::Add(Node *Info)
{
	int Result = *this ^ *Info;

	if (Result < 0)
	{
		if (Right)
		{
			Right->Add(Info);
		}
		else
		{
			Right = Info;
		}
	}
	else if (Result > 0)
	{
		if (Left)
		{
			Left->Add(Info);
		}
		else
		{
			Left = Info;
		}
	}
}

NodeInfo::NodeInfo()
{
	Name = 0;
	Attributes = 0;
	Size = 0;
}

NodeInfo::NodeInfo(char *n, int a, quad s)
{
	if (n)
	{
		Name = NEW(char[strlen(n)+1]);
		if (Name)
		{
			strcpy(Name, n);
		}
	}
	else
	{
		Name = 0;
	}
	Attributes = a;
	Size = s;
}

NodeInfo::~NodeInfo()
{
	DeleteArray(Name);
}

int NodeInfo::operator ^(NodeInfo &N)
{
	return stricmp(Name, N.Name);
}
