/*hdr
**	FILE:		Gdc2.h
**	AUTHOR:		Matthew Allen
**	DATE:		20/2/97
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#ifndef _GDC2_H_
#define _GDC2_H_

// OS headers
#include <X11/X.h>
#include <X11/Xlib.h>
#undef Status
#undef Success
#undef Below
#undef Above

// sub-system headers
#include "LibDefs.h"				// Xp
#include "LgiOsDefs.h"				// Platform specific
#include "Progress.h"				// Xp
#include "VFile.h"					// Platform specific
#include "VMem.h"					// Platform specific
#include "Device.h"					// Xp
#include "Core.h"					// Platform specific
#include "Contain.h"				// Xp

//				Defines
#define DEFAULT_GAMMA				1.0
#define DegToRad(i)					((i)*PI/180)
#define RadToDeg(i)					((i)*180/PI)
// #define GDC_USE_ASM				// Use intel assembly instructions
									// comment out for porting

// GDC2 Compile time options
#define ADJUST_ORIGIN				// allows you the set the origin of a DC

// Drawing operation types
#define GDC_SET						0
#define GDC_AND						1
#define GDC_OR						2
#define GDC_XOR						3
#define GDC_ALPHA					4
#define GDC_REMAP					5
#define GDC_MAXOP					6
#define GDC_CACHE_SIZE				4

// Channel types
#define GDCC_MONO					0
#define GDCC_GREY					1
#define GDCC_INDEX					2
#define GDCC_R						3
#define GDCC_G						4
#define GDCC_B						5
#define GDCC_ALPHA					6

// Native data formats
#define GDC_8BIT					0
#define GDC_16BIT					1
#define GDC_24BIT					2
#define GDC_32BIT					3
#define GDC_MAXFMT					4

// Colour spaces
#define GDC_8I						0	// 8 bit paletted
#define GDC_5R6G5B					1	// 16 bit
#define GDC_8B8G8B					2	// 24 bit
#define GDC_8A8R8G8B				3	// 32 bit
#define GDC_MAXSPACE				4

// Update types
#define GDC_PAL_CHANGE				0x1
#define GDC_BITS_CHANGE				0x2

// Flood fill types
#define GDC_FILL_TO_DIFFERENT		0
#define GDC_FILL_TO_BORDER			1
#define GDC_FILL_NEAR				2

// Gdc options
#define GDC_REDUCE_TYPE				0
	#define REDUCE_NONE					0
	#define REDUCE_NEAREST				1
	#define REDUCE_HALFTONE				2
	#define REDUCE_ERROR_DIFFUSION		3
	#define REDUCE_DL1					4
#define GDC_HALFTONE_BASE_INDEX		1
#define GDC_PALETTE_TYPE			2
	#define PALTYPE_ALLOC				0
	#define PALTYPE_RGB_CUBE			1
	#define PALTYPE_HSL					2
#define GDC_IMPORT_8_AS_16			3
#define GDC_MAX_OPTION				4

//	Flags
// Surface
#define GDC_OWN_MEMORY				0x0001
#define GDC_ON_SCREEN				0x0002
#define GDC_ALPHA_CHANNEL			0x0004
#define GDC_UPDATED_PALETTE			0x0008

// BaseDC
#define GDC_OWN_APPLICATOR			0x0001
#define GDC_CACHED_APPLICATOR		0x0002
#define GDC_OWN_PALETTE				0x0004
#define GDC_DRAW_ON_ALPHA			0x0008

// Region types
#define GDC_RGN_NONE				0	// No clipping
#define GDC_RGN_SIMPLE				1	// Single rectangle
#define GDC_RGN_COMPLEX				2	// Many rectangles

// Error codes
#define GDCERR_NONE					0
#define GDCERR_ERROR				1
#define GDCERR_CANT_SET_SCAN_WIDTH	2
#define GDC_INVALIDMODE				-1

// Font display flags
#define GDCFNT_TRANSPARENT			0x0001	// not set - SOLID
#define GDCFNT_UNDERLINE			0x0002

// Palette file types
#define GDCPAL_JASC					1
#define GDCPAL_MICROSOFT			2

// Misc
#define BMPWIDTH(bits)				((((bits)+31)/32)*4)


//				Typedefs

//	COLOUR typedef description
//
//				31              24 23              16 15              8  7               0
//				|-|-|-|-|-|-|-|-|  |-|-|-|-|-|-|-|-|  |-|-|-|-|-|-|-|-|  |-|-|-|-|-|-|-|-|
//
//  Standard:
//	8bit mode:	|------------------------ unused ---------------------|  |---- index ----|
//	15bit mode:	|----------------- unused ------------|T|-- red --|-- green ---|-- blue -|
//	16bit mode:	|------------- unused -------------|  |-- red --|--- green ----|-- blue -|
//	24bit mode:	|---- unused ---|  |----- blue ----|  |---- green ----|  |----- red -----|
//	32bit mode:	|---- alpha ----|  |----- red -----|  |---- green ----|  |----- blue ----|
//		
//  Note:		In 15bit mode the T bit can toggle between 8-bit indexing and RGB
//	
typedef unsigned long				COLOUR;

// alpha value, 0 - transparent, 255 - solid
typedef unsigned char				ALPHA;

// colour conversion defines
// RGB colour space
#define BitWidth(bits, cropbits)		( (((bits)+(cropbits)-1)/(cropbits)) * 4 )
#define Rgb32(r, g, b)				( (b&0xFF) | ((g&0xFF)<<8) | ((r&0xFF)<<16) )
#define Rgb24(r, g, b)				( (r&0xFF) | ((g&0xFF)<<8) | ((b&0xFF)<<16) )
#define Rgb15(r, g, b)				( ((r&0xF8)<<7) | ((g&0xF8)<<2) | ((b&0xF8)>>3))
#define Rgb16(r, g, b)				( ((r&0xF8)<<8) | ((g&0xFC)<<3) | ((b&0xF8)>>3))

#define Rgb32To24(c32)				( ((c32&0xFF0000)>>16) | ((c32&0xFF00)) | ((c32&0xFF)<<16))
#define Rgb32To16(c32)				( ((c32&0xF8)>>3) | ((c32&0xFC00)>>5) | ((c32&0xF80000)>>8) )
#define Rgb32To15(c32)				( ((c32&0xF8)>>3) | ((c32&0xF800)>>6) | ((c32&0xF80000)>>9) )
#define Rgb24To32(c24)				( ((c24&0xFF0000)>>16) | ((c24&0xFF00)) | ((c24&0xFF)<<16))
#define Rgb24To16(c24)				( ((c24>>19)&0x1F) | ((c24>>5)&0x7E0) | ((c24<<8)&0xF800) )
#define Rgb24To15(c24)				( ((c24>>19)&0x1F) | ((c24>>6)&0x3E0) | ((c24<<7)&0x7C00) )
#define Rgb15To32(c15)				( ((c15&0x7C00)<<9) | ((c15&0x3E0)<<6) | ((c15&0x1F)<<3) )
#define Rgb15To16(c15)				( ((c15&0x7FE0)<<1) | (c15&0x1F) )
#define Rgb15To24(c15)				( ((c15&0x7C00)>>7) | ((c15&0x3E0)<<6) | ((c15&0x1F)<<19) )
#define Rgb16To32(c16)				( ((c16&0xF800)<<8) | ((c16&0x7E0)<<5) | ((c16&0x1F)<<3) )
#define Rgb16To15(c16)				( ((c16&0xFFC0)>>1) | (c16&0x1F) )
#define Rgb16To24(c16)				( ((c16&0xF800)>>8) | ((c16&0x7E0)<<5) | ((c16&0x1F)<<19) )

#define R32(c32)					( (uchar) (c32) )
#define G32(c32)					( (uchar) (c32>>8) )
#define B32(c32)					( (uchar) (c32>>16) )
#define B24(c24)					( (uchar) (c24>>16) )
#define G24(c24)					( (uchar) (c24>>8) )
#define R24(c24)					( (uchar) (c24) )
#define R16(c24)					( (uchar) ((c24>>11)&0x1F) )
#define G16(c24)					( (uchar) ((c24>>5)&0x3F) )
#define B16(c24)					( (uchar) ((c24)&0x1F) )
#define R15(c24)					( (uchar) ((c24>>10)&0x1F) )
#define G15(c24)					( (uchar) ((c24>>5)&0x1F) )
#define B15(c24)					( (uchar) ((c24)&0x1F) )

// HLS colour Space
#define Hls32(h, l, s)				( ((h)<<16) | ((l)<<8) | (s) )
#define H32(c)						( ((c)>>16)&0xFFFF )
#define L32(c)						( ((c)>>8)&0xFF )
#define S32(c)						( (c)&0xFF )
#define HUE_UNDEFINED				1024
#define HlsIsUndefined(Hls)			(H32(Hls) == HUE_UNDEFINED)

COLOUR RgbToHls(COLOUR Rgb24);
COLOUR HlsToRgb(COLOUR Hsl32);

class Pixel24 {
public:
	uchar B;
	uchar G;
	uchar R;
};

class Pixel32 {
public:
	uchar A;
	uchar R;
	uchar G;
	uchar B;
};

//				Functions
LgiExtern void MemSet(void *d, int c, uint size);
LgiExtern void MemCpy(void *d, void *s, uint size);
LgiExtern void MemAnd(void *d, void *s, uint size);
LgiExtern void MemOr(void *d, void *s, uint size);
LgiExtern void MemXor(void *d, void *s, uint size);

// Prompt the user in text mode for a vesa screen mode
// returns -1 on error/cancel
extern int ScreenModeOptions();

//				Classes
class GdcFileFormat;
class GdcBasePrimitives;

#include "GdcRegion.h"
#include "GdcFont.h"

class GdcPt2 {
public:
	int x, y;

	GdcPt2(int Ix = 0, int Iy = 0)
	{
		x = Ix;
		y = Iy;
	}

	GdcPt2(const GdcPt2 &p)
	{
		x = p.x;
		y = p.y;
	}

	bool Inside(GdcRegion &r)
	{
		return	(x >= r.x1) AND
				(x <= r.x2) AND
				(y >= r.y1) AND
				(y <= r.y2);
	}
};

class GdcPt3 {
public:
	int x, y, z;
};

class GdcRGB {
public:
	uchar R, G, B, Flags;
	void Set(uchar r, uchar g, uchar b, uchar f = 0) { R = r; G = g; B = b; Flags = f; }
};

class GdcPalette {
protected:
	int Size;
	GdcRGB *Data;

public:
	GdcPalette();
	GdcPalette(GdcPalette *pPal);
	GdcPalette(uchar *pPal, int s = 256);
	virtual ~GdcPalette();
	
	void Set(GdcPalette *pPal);
	void Set(uchar *pPal, int s = 256);

	int GetSize() { return Size; }
	GdcRGB *operator [](int i) { return (i>=0 AND i<Size) ? Data + i : 0; }
	bool Update();
	bool SetSize(int s = 256);
	void SwapRAndB();
	int MatchRgb(COLOUR Rgb);
	void CreateCube();
	bool Load(VFile &F);
	bool Save(VFile &F, int Format);

	bool operator ==(GdcPalette &p);
	bool operator !=(GdcPalette &p);
};

class GdcSurface {
public:
	uchar *Base;
	int x, y, Bits, Line;
	int Flags;

	GdcSurface();
	~GdcSurface();
};

class GdcApplicator {
protected:
	GdcSurface *Dest;
	GdcSurface *Alpha;
	GdcPalette *Pal;
	int Op;

public:
	COLOUR c;				// main colour

	GdcApplicator() { c = 0; }
	GdcApplicator(COLOUR Colour) { c = Colour; }

	virtual int OnEvent(int m, int a, int b) { return 0; }
	void SetOp(int o) { Op = o; }
	int GetOp() { return Op; }
	int GetBits() { return (Dest) ? Dest->Bits : 0; }
	int GetFlags() { return (Dest) ? Dest->Flags : 0; }
	GdcPalette *GetPal() { return Pal; }

	virtual bool SetSurface(GdcSurface *d, GdcPalette *p = 0, GdcSurface *a = 0) = 0; // sets Dest, returns FALSE on error
	virtual void SetPtr(int x, int y) = 0;	// calculates Ptr from x, y
	virtual void IncX() = 0;
	virtual void IncY() = 0;
	virtual void IncPtr(int X, int Y) = 0;

	virtual void Set() = 0;			// sets pixel at Ptr
	virtual COLOUR Get() = 0;		// get pixel at Ptr
	virtual void VLine(int height) = 0;	// vertical line from Ptr down
						// rectangle from Ptr down
	virtual void Rectangle(int x, int y) = 0;
						// blt from Src to Dest
	virtual bool Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha = 0) = 0;
};

class GdcApplicatorServer {
public:
	GdcApplicatorServer() {}
	virtual ~GdcApplicatorServer() {}

	virtual int Op() { return 0; }
	virtual GdcApplicator *Create(int Bits) { return NULL; }
};

typedef List<GdcApplicatorServer> GdcApplicatorServerList;

#define OrgX(x)			x += OriginX
#define OrgY(y)			y += OriginY
#define OrgXy(x, y)		x += OriginX; y += OriginY
#define OrgPt(p)		p.x += OriginX; p.y += OriginY
#define OrgRgn(r)		r.x1 += OriginX; r.y1 += OriginY; r.x2 += OriginX; r.y2 += OriginY

class GdcDCBase {

	friend class GdcFileFormat;
	friend class GdcBasePrimitives;
	friend class GdcDibSection;

protected:
	int		Flags;
	int		PrevOp;
	GdcSurface	*pMem;
	GdcBasePrimitives *pAlphaDC;
	GdcRegion	Clip;
	GdcPalette	*pPalette;
	GdcApplicator	*pApp;
	GdcApplicator	*pAppCache[GDC_CACHE_SIZE];

	#ifdef ADJUST_ORIGIN
	int		OriginX, OriginY;
	#endif

	GdcDCBase	*Parent, *Next, *Prev, *Child;
	bool		AddChild(GdcDCBase *c, int Pos = 0);
	void		RemoveFromTree();

	GdcApplicator	*CreateApplicator(int Op = GDC_SET, int Bits = 0);

public:
	GdcDCBase();
	virtual ~GdcDCBase();

	virtual bool Create(int x, int y, int Bits, int LineLen = 0, bool KeepData = FALSE) { return FALSE; }
	virtual void Update(int Flags) {}

	#ifdef ADJUST_ORIGIN
	// Origin
	virtual void GetOrigin(int &x, int &y) { x = OriginX; y = OriginY; }
	virtual void SetOrigin(int x, int y) { OriginX = x; OriginY = y; }
	#endif

	// Alpha channel
	virtual bool IsAlpha() { return pAlphaDC != 0; }
	virtual bool IsAlpha(bool b);
	bool DrawOnAlpha() { return ((Flags & GDC_DRAW_ON_ALPHA) != 0); }
	bool DrawOnAlpha(bool Draw);
	GdcBasePrimitives *AlphaDC() { return pAlphaDC; }

	// Applicator
	virtual bool Applicator(GdcApplicator *pApp);
	virtual GdcApplicator *Applicator();

	// Palette
	virtual GdcPalette *Palette();
	virtual void Palette(GdcPalette *pPal, bool bOwnIt = TRUE);

	// Clip region
	virtual GdcRegion ClipRgn(GdcRegion *Rgn);
	virtual GdcRegion ClipRgn();

	// Set properties
	virtual COLOUR Colour(COLOUR c, int Bits = 0);
	virtual int Op(int Op);

	// Get properties
	virtual COLOUR Colour() { return pApp->c; }
	virtual int Op() { return (pApp) ? pApp->GetOp() : 0; }
	virtual int X() { return (pMem) ? pMem->x : 0; }
	virtual int Y() { return (pMem) ? pMem->y : 0; }
	virtual int GetBits() { return (pMem) ? pMem->Bits : 0; }
	virtual int GetFlags() { return Flags; }
	virtual bool IsScreen() { return (pApp) ? pApp->GetFlags() & GDC_ON_SCREEN : 0; }
	virtual uchar *operator[](int y) { return (pMem) ? pMem->Base + (pMem->Line * y) : 0; }
};

class GdcBasePrimitives : public GdcDCBase {
protected:
	uint		LineBits;
	uint		LineMask;

public:
	GdcBasePrimitives();
	GdcBasePrimitives(GdcBasePrimitives *pDC);

	// Pixel
	virtual void Set(int x, int y);
	virtual COLOUR Get(int x, int y);

	// Line
	virtual void HLine(int x1, int x2, int y);
	virtual void VLine(int x, int y1, int y2);
	virtual void Line(int x1, int y1, int x2, int y2);
	
	virtual uint LineStyle(uint Bits) { uint B = LineBits; LineBits = Bits; return B; }
	virtual uint LineStyle() { return LineBits; }

	// Curve
	virtual void Circle(double cx, double cy, double radius);
	virtual void FilledCircle(double cx, double cy, double radius);
	virtual void Arc(double cx, double cy, double radius, double start, double end);
	virtual void FilledArc(double cx, double cy, double radius, double start, double end);
	virtual void Ellipse(double cx, double cy, double x, double y);
	virtual void FilledEllipse(double cx, double cy, double x, double y);

	// Rectangular
	virtual void Box(int x1, int y1, int x2, int y2);
	virtual void Box(GdcRegion *a = NULL);
	virtual void Rectangle(int x1, int y1, int x2, int y2);
	virtual void Rectangle(GdcRegion *a = NULL);
	virtual void Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a = NULL);
	virtual void StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s);

	// Other
	virtual void Polygon(int Points, GdcPt2 *Data);
	virtual void Text(GdcFont *pFont, int x, int y, char *Str, int Len);
	virtual void Bezier(int Threshold, GdcPt2 *Pt);
	virtual void FloodFill(int x, int y, int Mode, COLOUR Border = 0, GdcRegion *Bounds = NULL);
};

class GdcViewDC : public GdcBasePrimitives {

	int BitDepth;

public:
	GdcViewDC(void *io);
	~GdcViewDC();

	// BaseDC
	#ifdef ADJUST_ORIGIN
	void GetOrigin(int &x, int &y);
	void SetOrigin(int x, int y);
	#endif

	GdcRegion ClipRgn();
	GdcRegion ClipRgn(GdcRegion *Rgn);
	COLOUR Colour();
	COLOUR Colour(COLOUR c, int Bits = 0);
	int Op();
	int Op(int Op);
	int X();
	int Y();
	int GetBits();
	int GetFlags();
	bool IsScreen() { return TRUE; }
	uchar *operator[](int y) { return NULL; }

	// Primitives
	void Set(int x, int y);
	COLOUR Get(int x, int y);
	void HLine(int x1, int x2, int y);
	void VLine(int x, int y1, int y2);
	void Line(int x1, int y1, int x2, int y2);
	void Circle(double cx, double cy, double radius);
	void FilledCircle(double cx, double cy, double radius);
	void Arc(double cx, double cy, double radius, double start, double end);
	void FilledArc(double cx, double cy, double radius, double start, double end);
	void Ellipse(double cx, double cy, double x, double y);
	void FilledEllipse(double cx, double cy, double x, double y);
	void Box(int x1, int y1, int x2, int y2);
	void Box(GdcRegion *a = NULL);
	void Rectangle(int x1, int y1, int x2, int y2);
	void Rectangle(GdcRegion *a = NULL);
	void Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a = NULL);
	void StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s);
	void Polygon(int Points, GdcPt2 *Data);
	void Text(GdcFont *pFont, int x, int y, char *Str, int Len);
	void Bezier(int Threshold, GdcPt2 *Pt);
	void FloodFill(int x, int y, int Mode, COLOUR Border = 0, GdcRegion *Bounds = NULL);
};

class GdcMemDC : public GdcBasePrimitives {
private:

public:
	GdcMemDC();
	GdcMemDC(GdcBasePrimitives *pDC);
	~GdcMemDC();

	void *Handle() { return 0; }

	bool IsScreen() { return FALSE; }
	bool Create(int x, int y, int Bits, int LineLen = 0, bool KeepData = FALSE);
	void Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a = NULL);
	void StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s);
	
	void HLine(int x1, int x2, int y, COLOUR a, COLOUR b);
	void VLine(int x, int y1, int y2, COLOUR a, COLOUR b);
};

class GdcSprite {
protected:
	// Dimensions
	int Sx, Sy, Bits;
	
	// Screen relation
	bool Visible;
	int PosX, PosY;
	int HotX, HotY;

	// Mode
	int DrawMode;			// See access functions for description

	// Destination DC, usually the screen
	// Don't own and must not be NULL
	GdcBasePrimitives *pScreen;
	
	// Internal sprite DC's
	// TODO: combine all these into one DC 3 times bigger
	GdcBasePrimitives *pBack;
	GdcBasePrimitives *pMask;
	GdcBasePrimitives *pColour;
	GdcBasePrimitives *pTemp;

	// This function gets called to update any sprite position dependent
	// graphics when the sprite moves.
	virtual void DrawOnMove(	GdcBasePrimitives *pDC,
					int Nx, int Ny,
					int Ox, int Oy) {}

public:
	GdcSprite();
	virtual ~GdcSprite();

	int X() { return Sx; }
	int Y() { return Sy; }
	int GetPosX() { return PosX; }
	int GetPosY() { return PosY; }
	int GetHotX() { return HotX; }
	int GetHotY() { return HotY; }
	int GetBits() { return Bits; }
	bool IsVisible() { return Visible; }

	void SetXY(int x, int y);

	bool Create(GdcBasePrimitives *pScreen, int x, int y, int Bits, uchar *Colour, uchar *Mask);
	bool CreateKey(GdcBasePrimitives *pScreen, int x, int y, int Bits, uchar *Colour, int Key);
	bool Create(GdcBasePrimitives *pScreen, GdcSprite *pSpr);
	bool SetSize(int x, int y, int Bits, GdcBasePrimitives *pScr = NULL);
	void Delete();

	int Mode() { return DrawMode; }
	int Mode(int m) { int Old = DrawMode; DrawMode = m; return Old; }
	// Mode:
	//	0 - Transparent background, AND mask, XOR colour
	//	1 - Transparent background, AND mask, OR colour
	//	3 - COPY colour

	void SetHotPoint(int x, int y);
	void SetVisible(bool v);
	void Move(int x, int y, bool WaitRetrace);

	void Draw(int x, int y, COLOUR Back, int Mode = 0, GdcBasePrimitives *pDC = NULL);
	// Mode:
	//	0 - Solid background, AND colour, XOR mask
	//	1 - Solid background, AND mask, OR colour
	//	2 - Transparent background, AND mask, OR colour
	//	3 - COPY colour

	void MoveToScreen(GdcBasePrimitives *Scr)
	{
		bool v = Visible;
		SetVisible(FALSE);
		pScreen = Scr;
		SetVisible(v);
	}
	GdcBasePrimitives *GetMask() { return pMask; }
	GdcBasePrimitives *GetColour() { return pColour; }
};

class GdcDevice : public GDevice {

	static GdcDevice *pInstance;

protected:
	bool PrevVerbose;

	// Lookups
	ulong *CharSquareData;

	// Palette
	double GammaCorrection;
	uchar GammaTable[256];
	GdcPalette *pSysPal;

	// Current mode info
	int ScrX;
	int ScrY;
	int ScrBits;

	// Options
	int OptVal[GDC_MAX_OPTION];

	// Applicator server
	GdcApplicatorServerList AppServerList;

public:
	GdcDevice();
	~GdcDevice();

	static GdcDevice *GetInst() { return pInstance; }
	static Display *Dsp;

	// Mode
	int GetBits();
	int X();
	int Y();

	// Options
	int GetOption(int Opt) { return (Opt >= 0 AND Opt < GDC_MAX_OPTION) ? OptVal[Opt] : 0; }
	int SetOption(int Opt, int Value)
	{
		int Prev = OptVal[Opt];
		if (Opt >= 0 AND Opt < GDC_MAX_OPTION)
			OptVal[Opt] = Value;
		return Prev;
	}

	// Lookups
	ulong *GetCharSquares() { return (CharSquareData) ? CharSquareData + 255 : 0; }

	// Applicators
	bool AddAppServer(GdcApplicatorServer *pServer);
	GdcApplicatorServer *GetAppServer(int Op);

	// Palette/Colour
	void SetGamma(double Gamma);
	double GetGamma();

	void SetSystemPalette(int Start, int Size, GdcPalette *Pal);
	GdcPalette *GetSystemPalette();

	void SetColourPaletteType(int Type);	// Type = PALTYPE_xxx define
	COLOUR GetColour(COLOUR Rgb24, GdcDCBase *pDC = NULL);
};

// file filter support
#include "Filter2.h"

// applicators
#include "GdcApp.h"

// globals
#define GdcD			GdcDevice::GetInst()
extern COLOUR CBit(int DstBits, COLOUR c, int SrcBits = 24, GdcPalette *Pal = 0);
extern GdcBasePrimitives *LoadDC(char *Name);


#endif
