
#ifndef __LGI_OS_DEFS_H
#define __LGI_OS_DEFS_H

#include "assert.h"

//////////////////////////////////////////////////////////////////
// Typedefs
typedef	long long int				quad;
typedef int							int32;
typedef unsigned int				uint32;
typedef	signed long long int		int64;
typedef unsigned long long int		uint64;

/////////////////////////////////////////////////////////////////
// Defines
#define _DEBUG 1

#define LINUX						1
#define MAKEINTRESOURCE(i)			((char*)(i))

// assert
#ifdef _DEBUG
extern void _lgi_assert(bool b, char *test, char *file, int line);
#define LgiAssert(b)				_lgi_assert(b, #b, __FILE__, __LINE__)
#else
#define LgiAssert(b)
#endif

#define LgiSleep(i)					ASSERT(0)
extern void _lgi_yield();
#define LgiYield()					

#define K_CHAR						0x0

#define SND_ASYNC 1

#define DOUBLE_CLICK_THRESHOLD		5

#define MB_OK						1
#define MB_YESNO					2
#define MB_YESNOCANCEL				3
#define IDYES						4
#define IDNO						5

/*
#define MK_LEFT						0x01
#define MK_RIGHT					0x02
#define MK_MIDDLE					0x04
#define MK_CTRL						0x08
#define MK_ALT						0x10
#define MK_SHIFT					0x20
*/

// Window flags
#define GWF_VISIBLE					0x00000001
#define GWF_ENABLED					0x00000002
#define GWF_FOCUS					0x00000004
#define GWF_OVER					0x00000008
#define GWF_DROP_TARGET				0x00000010
#define GWF_SUNKEN					0x00000020
#define GWF_FLAT					0x00000040
#define GWF_RAISED					0x00000080
#define GWF_BORDER					0x00000100
#define GWF_DIALOG					0x00000200
#define GWF_DESTRUCTOR				0x00000400
#define GWF_QUIT_WND				0x00000800

// Edge types
#define SUNKEN						1
#define RAISED						2
#define CHISEL						3
#define FLAT						4

// Colours
#define TOOL_VLOW					Rgb24(128, 128, 128)
#define TOOL_LOW					Rgb24(164, 164, 164)
#define TOOL_MED					Rgb24(216, 216, 216)
#define TOOL_HIGH					Rgb24(236, 236, 236)
#define TOOL_VHIGH					Rgb24(255, 255, 255)
#define C_WORKSPACE					Rgb24(255, 255, 255)

// Directories
#define DIR_CHAR					'/'
#define DIR_STR						"/"

#define IsSlash(c)					(((c)=='/')OR((c)=='\\'))
#define IsQuote(c)					(((c)=='\"')OR((c)=='\''))

// Messages
#define WM_USER						(0xF0000000) // how strange, this isn't defined in BeOS?? :P

// wParam = bool Inside; // is the mouse inside the client area?
// lParam = MAKELONG(x, y); // mouse location
#define WM_MOUSEENTER				(WM_USER+0x100)
#define WM_MOUSEEXIT				(WM_USER+0x101)

// wParam = (GView*) Wnd;
// lParam = (int) Flags;
#define WM_CHANGE					(WM_USER+0x102)

// wParam = (GView*) Wnd;
// lParam = (char*) Text; // description from window
#define WM_DESCRIBE					(WM_USER+0x103)

// return (bool)
#define WM_WANT_DIALOG_PROC			(WM_USER+0x104)

#define WM_MENU						(WM_USER+0x105)
#define WM_COMMAND					(WM_USER+0x106)
#define WM_DRAG_DROP				(WM_USER+0x107)
#define WM_VSCROLL					(WM_USER+0x108)
#define WM_HSCROLL					(WM_USER+0x109)

// Dialog stuff
#define IDOK						1
#define IDCANCEL					2

// Keys
#define VK_DELETE					1
#define VK_SHIFT					2
#define VK_ESCAPE					127
#define VK_RETURN					13
#define VK_BACK						8
#define VK_RIGHT					3
#define VK_LEFT						4
#define VK_UP						5
#define VK_DOWN						6
#define VK_PRIOR					7
#define VK_NEXT						19
#define VK_HOME						18
#define VK_END						17
#define VK_INSERT					16

// The BeOS defines actually overlap some other keys
// so I don't use them, instead I just defined my own block
// of codes
#define VK_F1						0x11
#define VK_F2						0x12
#define VK_F3						0x13
#define VK_F4						0x14
#define VK_F5						0x15
#define VK_F6						0x16
#define VK_F7						0x17
#define VK_F8						0x18
#define VK_F9						0x19
#define VK_F10						0x01
#define VK_F11						0x02
#define VK_F12						0x03

/////////////////////////////////////////////////////////////////////////////////////
struct GMessage
{
	int Msg, a, b;
};
#define Message(m) m->Msg
#define MessageA(m) m->a
#define MessageB(m) m->b
extern GMessage CreateMsg(int m, int a, int b);
typedef void *GWndRef;
typedef void *GFontRef;

/////////////////////////////////////////////////////////////////////////////////////
// Externs
LgiFunc int stricmp(char *a, char *b);
LgiFunc char *strnistr(char *a, char *b, int n);
LgiFunc int strnicmp(char *a, char *b, int i);
LgiFunc void FormatSize(char *Str, quad Size);
LgiFunc void strupr(char *a);
LgiFunc void strlwr(char *a);
LgiFunc int stricmp(char *a, char *b);

#endif