/*
**	FILE:		GuiToolbar.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Toolbar classes
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

////////////////////////////////////////////////////////////////////////
GToolBar *LgiLoadToolbar(GView *Parent, char *File, int x, int y)
{
	GToolBar *Toolbar = NEW(GToolBar);
	if (Toolbar)
	{
		char *FileName = LgiFindFile(File);
		bool Success = FileName AND Toolbar->SetBitmap(FileName, x, y);
		if (NOT Success)
		{
			LgiMsg(	Parent,
					"Can't find the graphic '%s' for the toolbar.\n"
					"You can find it in this program's archive.",
					"LgiLoadToolbar",
					MB_OK,
					File);
			DeleteObj(Toolbar);
		}
		DeleteArray(FileName);
	}
	return Toolbar;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GImageList::GImageList(int x, int y, GdcBasePrimitives *pDC)
{
	Sx = x;
	Sy = y;
	pCreate = pDC;
}

GImageList::~GImageList()
{
}

bool GImageList::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	bool Status = GdcMemDC::Create(x, y, (Bits < 8) ? 24 : Bits, LineLen, KeepData);
	if (Status)
	{
		if (pCreate)
		{
			Blt(0, 0, pCreate);
		}
	}
	return Status;
}

void GImageList::Update(int Flags)
{
}

void GImageList::Draw(GdcBasePrimitives *pDest, int Dx, int Dy, int Image, int Flags)
{
	if (Flags & IMGLST_GDC)
	{
		if (pDest)
		{
			COLOUR Key = Get(0, 0);
			GdcRegion r;
			r.ZOff(Sx-1, Sy-1);
			r.Offset(Image * Sx, 0);

			if (Flags & IMGLST_DISABLED)
			{
				COLOUR Low = CBit(pDest->GetBits(), TOOL_LOW);
				COLOUR High = CBit(pDest->GetBits(), TOOL_HIGH);
				
				for (int y=0; y<r.Y(); y++)
				{
					for (int x=0; x<r.X(); x++)
					{
						if (Get(r.x1 + x, r.y1 + y) != Key)
						{
							pDest->Colour(Low);
							pDest->Set(Dx + x, Dy + y);
							pDest->Colour(High);
							pDest->Set(Dx + x + 1, Dy + y + 1);
						}
					}
				}
			}
			else
			{
				for (int y=0; y<r.Y(); y++)
				{
					for (int x=0; x<r.X(); x++)
					{
						COLOUR c = Get(r.x1 + x, r.y1 + y);
						if (c != Key)
						{
							pDest->Colour(c);
							pDest->Set(Dx + x, Dy + y);
						}
					}
				}
			}
		}
	}
	else
	{
		// ??? Fix me
		ASSERT(0);
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolButton::GToolButton(int Bx, int By)
{
	Name("");
	Type = TBT_PUSH;
	SetId(IDM_NONE);
	SetDown(FALSE);
	Clicked = FALSE;
	ImgIndex = 0;

	Pos.ZOff(Bx+1, By+1);
	Parent = 0;
	TipId = -1;
}

GToolButton::~GToolButton()
{
}

void GToolButton::OnPaint(GdcBasePrimitives *pDC)
{
	GToolBar *Par = (GToolBar*) Parent;
	if (Par AND Par->ImgList)
	{
		GdcMemDC *pMemDC = NEW(GdcMemDC);
		int MemBits = GdcD->GetBits();
		if (pMemDC AND pMemDC->Create(Pos.X(), Pos.Y(), (MemBits < 8) ? 24 : MemBits))
		{
			int Px = Pos.X()-1;
			int Py = Pos.Y()-1;

			// Draw Background
			COLOUR Grey = CBit(pMemDC->GetBits(), (pMemDC->GetBits()==16) ? Rgb24(R24(Grey)-1, G24(Grey)-1, B24(Grey)-1) : TOOL_MED);
			pMemDC->Colour(Grey, 24);
			pMemDC->Rectangle();

			if (((Type == TBT_RADIO) OR (Type == TBT_TOGGLE)) AND Down)
			{
				// Draw highlight
				pMemDC->Colour(TOOL_HIGH, 24);

				for (int y=0; y<pMemDC->Y(); y++)
				{
					for (int x=y&1; x<pMemDC->X(); x+=2)
					{
						pMemDC->Set(x, y);
					}
				}
			}

			int Off = BORDER_BUTTON + ((Down) ? 1 : 0);
			
			if (GetId() >= 0)
			{
				// button
				int Flags = IMGLST_GDC;

				if (Enabled())
				{
					// draw normal
					Par->ImgList->Draw(pMemDC, Off, Off, ImgIndex, Flags);
				}
				else
				{
					// draw disabled
					Par->ImgList->Draw(pMemDC, Off, Off, ImgIndex, IMGLST_DISABLED | Flags);
				}

				// draw border
				if (Down)
				{
					pMemDC->Colour((Down) ? TOOL_HIGH : TOOL_LOW, 24);
					pMemDC->Line(Px, Py, Px, 0);
					pMemDC->Line(Px, Py, 0, Py);

					pMemDC->Colour((Down) ? TOOL_LOW : TOOL_HIGH, 24);
					pMemDC->Line(0, 0, 0, Py);
					pMemDC->Line(0, 0, Px, 0);
				}
			}
			else
			{
				// separator
				if (X() > Y())
				{
					int c = Y()/2-1;
					pMemDC->Colour(TOOL_LOW, 24);
					pMemDC->Line(2, c, Px-2, c);
					pMemDC->Colour(TOOL_HIGH, 24);
					pMemDC->Line(2, c+1, Px-2, c+1);
				}
				else
				{
					int c = X()/2-1;
					pMemDC->Colour(TOOL_LOW, 24);
					pMemDC->Line(c, 2, c, Py-2);
					pMemDC->Colour(TOOL_HIGH, 24);
					pMemDC->Line(c+1, 2, c+1, Py-2);
				}
			}
			pDC->Blt(0, 0, pMemDC);
		}

		DeleteObj(pMemDC);
	}
}

void GToolButton::Value(int b)
{
	switch (Type)
	{
		case TBT_PUSH:
		{
			// do nothing... can't set value
			break;
		}

		case TBT_TOGGLE:
		{
			SetDown(b);
			Invalidate();
			break;
		}

		case TBT_RADIO:
		{
			if (Parent AND b)
			{
				// Clear any other radio buttons that are down
				GToolButton *But;
				for (	But = (GToolButton*) Parent->Children.Find(this);
						But AND But->GetId() >= 0;
						But = (GToolButton*) Parent->Children.Next())
				{
					if (But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(false);
						But->Invalidate();
					}
				}

				for (	But = (GToolButton*) Parent->Children.Find(this);
						But AND But->GetId() >= 0;
						But = (GToolButton*) Parent->Children.Prev())
				{
					if (But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(false);
						But->Invalidate();
					}
				}
			}

			SetDown(b);
			Invalidate();
			break;
		}
	}
}

void GToolButton::OnCommand(VMouse &m)
{
	if (Parent)
	{
		((GToolBar*) Parent)->OnButtonClick(this);
	}
}

void GToolButton::SendNotify(int Flags)
{
	GView *To = (Notify) ? Notify : Parent;
	if (To)
	{
		GMessage m = CreateMsg(WM_CHANGE, (int)((GView*)this), Flags);
		To->OnEvent(&m);
	}
}

void GToolButton::OnMouseClick(VMouse &m)
{
	if (GetId() >= 0 AND Enabled())
	{
		switch (Type)
		{
			case TBT_PUSH:
			{
				bool Old = Down;

				Clicked = m.Down();
				Capture(m.Down());

				if (Old AND IsOver(m))
				{
					if (m.Flags & MK_LEFT)
					{
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}

				SetDown(m.Down());
				if (Old != Down)
				{
					Invalidate();
				}
				break;
			}

			case TBT_TOGGLE:
			{
				if (m.Down())
				{
					if (m.Left())
					{
						Value(NOT Down);
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}
				break;
			}

			case TBT_RADIO:
			{
				if (m.Down())
				{
					if (NOT Down AND m.Left())
					{
						Value(true);
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}
				break;
			}
		}
	}
}

void GToolButton::OnMouseEnter(VMouse &m)
{
	if (Clicked)
	{
		SetDown(TRUE);
		Invalidate();
	}
	else
	{
		if (TipId < 0)
		{
			GToolBar *Bar = (GToolBar*) Parent;
		}


		((GToolBar*) Parent)->PostDescription(this, Name());
	}
}

void GToolButton::SetDown(bool d)
{
	Down = d;
}

void GToolButton::OnMouseMove(VMouse &m)
{
}

void GToolButton::OnMouseExit(VMouse &m)
{
	if (Clicked)
	{
		SetDown(FALSE);
		Invalidate();
	}
	else
	{
		((GToolBar*) Parent)->PostDescription(this, "");
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolBar::GToolBar()
{
	Name("LGI_Toolbar");
	Bx = By = 0;
	Sx = Sy = 10;
	Vertical = FALSE;

	LastIndex = 0;
	OwnImgList = false;
	ImgList = 0;
}

GToolBar::~GToolBar()
{
}

bool GToolBar::Pour(GdcComplexRegion &r)
{
	if (ImgList)
	{
		int PosX = BORDER_SHADE + BORDER_SPACER;
		int PosY = BORDER_SHADE + BORDER_SPACER;
		int SrcX = 0;
		int SrcY = 0;
		int EndX = 0;
		int EndY = 0;

		GdcRegion ButPos;
		GView *But = Children.First();
		while (But)
		{
			ButPos = But->GetPos();

			if (But->GetId() == IDM_SEPARATOR)
			{
				// separator
				if (Vertical)
				{
					ButPos.ZOff(Bx+(2*BORDER_BUTTON)-1, BORDER_SEPARATOR+1);
				}
				else
				{
					ButPos.ZOff(BORDER_SEPARATOR+1, By+(2*BORDER_BUTTON)-1);
				}
			}
			else if (But->GetId() >= 0)
			{
				// other control
			}

			ButPos.Offset(PosX - ButPos.x1, PosY - ButPos.y1);

			if (But->GetId() == IDM_BREAK)
			{
				ButPos.ZOff(0, 0);
				if (Vertical)
				{
					PosX = EndX;
					PosY = BORDER_SHADE + BORDER_SPACER;
				}
				else
				{
					PosX = BORDER_SHADE + BORDER_SPACER;
					PosY = EndY;
				}
			}
			else
			{
				if (Vertical)
				{
					EndX = max(EndX, ButPos.x2 + 1);
					PosY = ButPos.y2 + 1;
				}
				else
				{
					PosX = ButPos.x2 + 1;
					EndY = max(EndY, ButPos.y2 + 1);
				}
			}

			But->SetPos(ButPos);
			But = Children.Next();
		}

		if (Vertical)
		{
			PosX = EndX;
		}
		else
		{
			PosY = EndY;
		}

		Sx = PosX + BORDER_SHADE + BORDER_SPACER;
		Sy = PosY + BORDER_SHADE + BORDER_SPACER;
		Pos.ZOff(Sx-1, Sy-1);

		// GdcRegion *Best = FindLargest(r);
		GdcRegion *Best = FindLargestEdge(r, GV_EDGE_TOP);
		if (Best)
		{
			Pos.Offset(Best->x1, Best->y1);
			Pos.Bound(Best);
			SetPos(Pos);

			return TRUE;
		}
	}

	return FALSE;
}

void GToolBar::OnButtonClick(GToolButton *Btn)
{
	GView *w = (Notify)?Notify:GView::Parent;
	if (w AND Btn)
	{
		GMessage m = CreateMsg(WM_COMMAND, Btn->GetId(), (int) Handle());
		w->OnEvent(&m);
	}
}

int GToolBar::PostDescription(GView *Ctrl, char *Text)
{
	if (GView::Parent)
	{
		GMessage m = CreateMsg(WM_DESCRIBE, (int) Ctrl, (int) Text);
		return GView::Parent->OnEvent(&m);
	}
	return 0;
}

int GToolBar::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		case WM_CHANGE:
		{
			return GView::Parent->OnEvent(Msg);
		}
	}

	return GView::OnEvent(Msg);
}

void GToolBar::OnPaint(GdcBasePrimitives *pScr)
{
	GdcRegion r(0, 0, X()-1, Y()-1);

	// Draw a border
	DrawMoveShallowEdge(pScr, r, RAISED);
	pScr->Colour(TOOL_MED, 24);
	pScr->Box(&r);
	r.Size(1, 1);

	// Draw all the children
	GView::OnPaint(pScr);
}

bool GToolBar::SetBitmap(char *File, int bx, int by)
{
	bool Status = false;
	
	GdcBasePrimitives *pDC = LoadDC(File);
	if (pDC)
	{
		Status = SetDC(pDC, bx, by);
		DeleteObj(pDC);
	}

	return Status;
}

bool GToolBar::SetDC(GdcBasePrimitives *pNewDC, int bx, int by)
{
	if (OwnImgList)
	{
		DeleteObj(ImgList);
	}
	Bx = bx;
	By = by;

	if (pNewDC)
	{
		ImgList = NEW(GImageList(bx, by, pNewDC));
		if (ImgList AND
			ImgList->Create(pNewDC->X(), pNewDC->Y(), GdcD->GetBits()))
		{
			return true;
		}
	}
	return false;
}

bool GToolBar::SetImageList(GImageList *l, int bx, int by, bool Own)
{
	if (OwnImgList)
	{
		DeleteObj(ImgList);
	}
	OwnImgList = Own;
	Bx = bx;
	By = by;

	ImgList = l;
	return ImgList != 0;
}

GToolButton *GToolBar::AppendButton(char *Tip, int Id, int Type, int Enabled, int IconId)
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->Name(Tip);
		But->SetId(Id);
		But->Type = Type;
		But->Parent = this;
		But->Enabled(Enabled);
		
		if (IconId >= 0)
		{
			But->ImgIndex = IconId;
		}
		else
		{
			But->ImgIndex = LastIndex++;
		}
		Children.Insert(But);
	}
	return But;
}

bool GToolBar::AppendSeparator()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->SetId(IDM_SEPARATOR);
		But->Parent = this;
		Children.Insert(But);
		return TRUE;
	}
	return FALSE;
}

bool GToolBar::AppendBreak()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->SetId(IDM_BREAK);
		But->Parent = this;
		Children.Insert(But);
		return TRUE;
	}
	return FALSE;
}

bool GToolBar::AppendControl(GView *Ctrl)
{
	bool Status = FALSE;
	if (Ctrl)
	{
		Ctrl->SetParent(this);
		Children.Insert(Ctrl);
		Status = TRUE;
	}
	return Status;
}

void GToolBar::Empty()
{
	for (GView *But = Children.First(); But; But = Children.Next())
	{
		DeleteObj(But);
	}
}

void GToolBar::OnMouseClick(VMouse &m)
{
}

void GToolBar::OnMouseMove(VMouse &m)
{
}

void GToolBar::OnMouseEnter(VMouse &m)
{
}

void GToolBar::OnMouseExit(VMouse &m)
{
}
