/*
**	FILE:		GuiViews.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Standard Views
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
GLayout::GLayout()
{
	_PourLargest = false;
	VScroll = 0;
	HScroll = 0;
}

GLayout::~GLayout()
{
	DeleteObj(VScroll);
	DeleteObj(HScroll);
}

bool GLayout::GetPourLargest()
{
	return _PourLargest;
}

void GLayout::SetPourLargest(bool i)
{
	_PourLargest = i;
}

void GLayout::GetScrollPos(int &x, int &y)
{
	if (HScroll)
	{
		x = HScroll->Pos();
	}
	else
	{
		x = 0;
	}

	if (VScroll)
	{
		y = VScroll->Pos();
	}
	else
	{
		y = 0;
	}
}

void GLayout::SetScrollPos(int x, int y)
{
	if (HScroll)
	{
		HScroll->SetPos(x);
	}

	if (VScroll)
	{
		VScroll->SetPos(y);
	}
}

bool GLayout::SetScrollBars(bool x, bool y)
{
	bool Status = FALSE;
	bool Processing = false;

	if (NOT Processing)
	{
		Processing = true;


		Processing = false;
	}

	return Status;
}

int GLayout::OnEvent(GMessage *Msg)
{
	if (Message(Msg) == WM_CHANGE AND
		Parent)
	{
		Parent->OnEvent(Msg);
	}

	if (VScroll) VScroll->OnEvent(Msg);
	if (HScroll) HScroll->OnEvent(Msg);
	return GView::OnEvent(Msg);
}

bool GLayout::Pour(GdcComplexRegion &r)
{
	if (_PourLargest)
	{
		GdcRegion *Best = FindLargest(r);
		if (Best)
		{
			SetPos(*Best);
			return TRUE;
		}
	}
	return FALSE;
}



//////////////////////////////////////////////////////////////////////////////////////
GScrollBar::GScrollBar()
#ifdef LGI_RES
	: ResObject(Res_ScrollBar)
#endif
{
	Parent = 0;
}

GScrollBar::GScrollBar(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_ScrollBar)
#endif
{
	SetId(id);
	GView::Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);
}

GScrollBar::~GScrollBar()
{
}

bool GScrollBar::Attach(GView *p)
{
	return false;
}

bool GScrollBar::Vertical()
{
	return 0;
}

void GScrollBar::SetVertical(bool v)
{
}

bool IsWindowThread(GView *w)
{
	return 0;
}

int GScrollBar::Pos()
{
	return 0;
}

void GScrollBar::SetPos(int p)
{
}

void GScrollBar::Limits(int &Low, int &High)
{
}

void GScrollBar::SetLimits(int Low, int High)
{
}

int GScrollBar::Page()
{
	return 0;
}

void GScrollBar::SetPage(int p)
{
}

int GScrollBar::OnEvent(GMessage *Msg)
{
	return FALSE;
}

//////////////////////////////////////////////////////////////////////////////////////
GStatusBar::GStatusBar()
{
	Name("LGI_StatusBar");

	Font = NEW(GFont);
	if (Font)
	{
		if (NOT Font->Create("Ms Sans Serif", 8))
		{
			DeleteObj(Font);
		}
		else
		{
			Font->Transparent(true);
		}
	}
}

GStatusBar::~GStatusBar()
{
	for (int i=0; i<GetSize(); i++)
	{
		DeleteObj((*this)[0]);
	}

	DeleteObj(Font);
}

bool GStatusBar::Pour(GdcComplexRegion &r)
{
	GdcRegion *Best = FindLargestEdge(r, GV_EDGE_BOTTOM);
	if (Best)
	{
		GdcRegion Take = *Best;
		if (Take.Y() > 21)
		{
			Take.y1 = max((Take.y2 - 21), Take.y1);
			SetPos(Take);
			RePour();
			return TRUE;
		}
	}
	return FALSE;
}

void GStatusBar::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, X()-1, Y()-1);

	DrawMoveShallowEdge(pDC, r, RAISED);

	GdcComplexRegion Cr(r);
	#ifdef WIN32
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		GdcRegion n = c->GetPos();
		Cr.Subtract(&n);
	}
	#endif
	pDC->Colour(TOOL_MED, 24);
	for (GdcRegion *i = Cr.First(); i; i = Cr.Next())
	{
		pDC->Rectangle(i);
	}

	GView::OnPaint(pDC);
}

void GStatusBar::RePour()
{
	int X = Pos.X() - 5;
	int i = Children.GetItems()-1;

	for (	GView *w = Children.Last();
			w;
			w = Children.Prev(), i--)
	{
		GStatusPane *Pane = dynamic_cast<GStatusPane*>(w);
		if (Pane)
		{
			if (i)
			{
				// not first
				X -= Pane->Width; 
				Pane->Pos.ZOff(Pane->Width, Pos.Y()-7);
				Pane->Pos.Offset(X, 3);
				X -= STATUSBAR_SEPARATOR;
			}
			else
			{
				// first
				Pane->Pos.ZOff(X-2, Pos.Y()-4);
				Pane->Pos.Offset(2, 2);
				X = 0;
			}
		}
	}
}

GStatusPane *GStatusBar::AppendPane(char *Text, int Width)
{
	GStatusPane *Pane = 0;
	if (Text)
	{
		Pane = NEW(GStatusPane);
		if (Pane)
		{
			Pane->Parent = this;
			Pane->Name(Text);
			Pane->Width = Width;
			Children.Insert(Pane);
		}
	}
	return Pane;
}

bool GStatusBar::AppendPane(GStatusPane *Pane)
{
	if (Pane)
	{
		Pane->Parent = this;
		Children.Insert(Pane);
	}
	return TRUE;
}

// Pane
GStatusPane::GStatusPane()
{
	Parent = 0;
	Flags = 0;
	Width = 32;
	Pos.ZOff(Width-1, 20);
	pDC = 0;
}

GStatusPane::~GStatusPane()
{
	DeleteObj(pDC)
}

bool GStatusPane::Name(char *n)
{
	bool Status = GObject::Name(n);
	GdcRegion p(0, 0, Pos.X()-1, Pos.Y()-1);
	p.Size(1, 1);
	Invalidate(&p);
	return Status;
}

void GStatusPane::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, X()-1, Y()-1);
	int Off = 1;

	if (Sunken())
	{
		Off = 0;
		DrawMoveShallowEdge(pDC, r, SUNKEN);
	}

	GStatusBar *Par = (GStatusBar*) Parent;
	if (Par AND Par->Font)
	{
		char *t = Name();
		if (t AND strlen(t)>0)
		{
			Par->Font->Colour(Rgb24(0, 0, 0), TOOL_MED);
			Par->Font->Transparent(false);
			Par->Font->Text(pDC, 2+Off, 1+Off, Name(), -1, &r);
			return;
		}
	}

	pDC->Colour(TOOL_MED, 24);
	pDC->Rectangle(&r);
}

int GStatusPane::GetWidth()
{
	return Width;
}

void GStatusPane::SetWidth(int x)
{
	Width = x;
	if (Parent)
	{
		((GStatusBar*) Parent)->RePour();
	}
}

bool GStatusPane::Sunken()
{
	return TestFlag(Flags, GSP_SUNKEN);
}

void GStatusPane::Sunken(bool s)
{
	bool Old = Sunken();
	if (s) SetFlag(Flags, GSP_SUNKEN);
	else ClearFlag(Flags, GSP_SUNKEN);
	if (Old != Sunken() AND Parent)
	{
		Parent->Invalidate(&Pos);
	}
}

GdcBasePrimitives *GStatusPane::Bitmap()
{
	return pDC;
}

void GStatusPane::Bitmap(GdcBasePrimitives *pNewDC)
{
	DeleteObj(pDC);
	if (pNewDC)
	{
		pDC = NEW(GdcMemDC);
		if (pDC)
		{
			if (pDC->Create(pNewDC->X(), pNewDC->Y(), pNewDC->GetBits()))
			{
				pDC->Blt(0, 0, pNewDC);
			}
			else
			{
				DeleteObj(pDC);
			}
		}
	}

	if (Parent)
	{
		Parent->Invalidate(&Pos);
	}
}

//////////////////////////////////////////////////////////////////////////////
#define SPLITER_BORDER			4
#define SPLITER_BAR				5
#define SPLITER_INSET			2
#define SPLITER_MIN_X			(2*SPLITER_INSET)
#define SPLITER_MIN_Y			(2*SPLITER_INSET)

#define SPLITER_FULL_DRAG

GSplitter::GSplitter()
{
	Name("LGI_Spliter");
	SplitSet = SplitPos = -1;
	Moving = FALSE;

	PosA.ZOff(0, 0);
	PosB.ZOff(0, 0);
	ViewA = 0;
	ViewB = 0;
	BorderA = true;
	BorderB = true;
	OverBar = false;
	SplitFollow = false;

	IsVertical(true);
	Border(true);
	Raised(true);
	Split(0);
}

GSplitter::~GSplitter()
{
	DeleteObj(ViewA);
	DeleteObj(ViewB);
}

void GSplitter::IsVertical(bool v)
{
	Vertical = v;
}

void GSplitter::SetViewA(GView *a, bool Border)
{
	BorderA = Border;
	ViewA = a;
	if (a)
	{
		a->Attach(this);
		a->Visible(true);

		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

void GSplitter::SetViewB(GView *b, bool Border)
{
	BorderB = Border;
	ViewB = b;
	if (b)
	{
		b->Attach(this);
		b->Visible(true);

		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

void GSplitter::Split(int s)
{
	SplitSet = s;
	int Limit = ((Vertical) ? Pos.X() : Pos.Y()) - 18;
	int NewPos = (s > 0) ? max(s, 4) : max(4, Limit+s);
	if (NewPos != SplitPos)
	{
		SplitPos = NewPos;
		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

void GSplitter::CalcRegions(GdcRegion Rect, bool Follow)
{
	PosA = Rect;
	PosA.Offset(-PosA.x1, -PosA.y1);
	if (Border())
	{
		PosA.Size(SPLITER_BORDER, SPLITER_BORDER);
	}

	PosB = PosA;

	if (SplitFollow AND Follow)
	{
		SplitPos = max(1, ((Vertical) ? Pos.X()-10 : Pos.Y()-10) + SplitSet);
	}

	if (Vertical)
	{
		PosA.x2 = PosA.x1 + SplitPos;
		PosB.x1 = PosA.x2 + SPLITER_BAR;
	}
	else
	{
		PosA.y2 = PosA.y1 + SplitPos;
		PosB.y1 = PosA.y2 + SPLITER_BAR;
	}

	if (ViewA)
	{
		GdcRegion r = PosA;
		if (BorderA)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		ViewA->SetPos(r);
		
		GdcComplexRegion Rgn(r);
		ViewA->Pour(Rgn);
	}
	if (ViewB)
	{
		GdcRegion r = PosB;
		if (BorderB)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		ViewB->SetPos(r);

		GdcComplexRegion Rgn(r);
		ViewB->Pour(Rgn);
	}
}

void GSplitter::OnPosChange()
{
	CalcRegions(Pos, true);
}

void GSplitter::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r;
	r.ZOff(pDC->X()-1, pDC->Y()-1);

	if (Border())
	{
		if (Sunken())
		{
			DrawMoveShallowEdge(pDC, r, SUNKEN);
		}
		else if (Raised())
		{
			DrawMoveShallowEdge(pDC, r, RAISED);
		}

		pDC->Colour(TOOL_MED, 24);
		DrawMoveFill(pDC, r, SPLITER_BORDER);
	}
	else
	{
		pDC->Colour(TOOL_MED, 24);
	}

	// bar
	if (Vertical)
	{
		pDC->Rectangle(PosA.x2, PosA.y1, PosB.x1, PosA.y2);
	}
	else
	{
		pDC->Rectangle(PosA.x1, PosA.y2, PosA.x2, PosB.y1);
	}

	// PosA region
	r = PosA;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderA)
		{
			DrawMoveDeepEdge(pDC, r, SUNKEN);
		}

		if (ViewA)
		{
			if (NOT ViewA->Handle())
			{
				pDC->SetOrigin(PosA.x1, PosA.y1);
				ViewA->OnPaint(pDC);
				pDC->SetOrigin(0, 0);
			}
			else
			{
				// View is real and will paint itself
			}
		}
		else
		{
			pDC->Colour(C_WORKSPACE, 24);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	r = PosB;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderB)
		{
			DrawMoveDeepEdge(pDC, r, SUNKEN);
		}

		if (ViewB)
		{
			if (NOT ViewB->Handle())
			{
				pDC->SetOrigin(PosB.x1, PosB.y1);
				ViewB->OnPaint(pDC);
				pDC->SetOrigin(0, 0);
			}
			else
			{
				// View is real and will paint itself
			}
		}
		else
		{
			pDC->Colour(C_WORKSPACE, 24);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

}

void GSplitter::OnMouseClick(VMouse &m)
{

	if (m.Down())
	{
		if (OverSplit(m.x, m.y))
		{
			if (Vertical)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x2;
				Bar.y1 = PosA.y1;
				Bar.x2 = PosB.x1;
				Bar.y2 = PosA.y2;
				Offset = m.x - Bar.x1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
			else
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x1;
				Bar.y1 = PosA.y2;
				Bar.x2 = PosA.x2;
				Bar.y2 = PosB.y1;
				Offset = m.y - Bar.y1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
	}
	else
	{
		#ifdef SPLITER_FULL_DRAG
		Capture(FALSE);
		Moving = FALSE;
		#else
		if (Moving)
		{
			HashRectangle(this, &Bar);
			Capture(FALSE);
			Moving = FALSE;

			int NewPos;
			if (Vertical)
			{
				NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
			}
			else
			{
				NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
			}

			if (NewPos != SplitPos)
			{
				SplitPos = NewPos;
				CalcRegions(Pos);
				Invalidate();
			}
		}
		#endif
	}
}

bool GSplitter::OverSplit(int x, int y)
{
	if (Vertical)
	{
		return (x > PosA.x2 AND x < PosB.x1 AND y > PosA.y1 AND y < PosA.y2);
	}
	else
	{
		return (x > PosA.x1 AND x < PosA.x2 AND y > PosA.y2 AND y < PosB.y1);
	}
}

void GSplitter::OnMouseMove(VMouse &m)
{
	if (Moving)
	{
		#ifdef SPLITER_FULL_DRAG
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}

		int NewPos;
		if (Vertical)
		{
			NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
		}
		else
		{
			NewPos = limit(Bar.y1, 4, Pos.Y()-18);
		}

		if (NewPos != SplitPos)
		{
			SplitPos = NewPos;
			CalcRegions(Pos);
			Invalidate();
		}
		#else
		HashRectangle(this, &Bar);
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}
		HashRectangle(this, &Bar);
		#endif
	}
}

GView *GSplitter::FindControl(GWndRef hCtrl)
{
	return 0;
}

bool GSplitter::Attach(GView *p)
{
	bool Status = GLayout::Attach(p);
	if (Status)
	{
		if (ViewA AND NOT ViewA->Handle())
		{
			ViewA->Attach(this);
		}

		if (ViewB AND NOT ViewB->Handle())
		{
			ViewB->Attach(this);
		}
	}

	return Status;
}
