/*hdr
**	FILE:		Gdc2.h
**	AUTHOR:		Matthew Allen
**	DATE:		20/2/97
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 1997, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include <math.h>

#include "Gdc2.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////
void GdcGdiScreen::Init()
{
	OwnDC = false;
	hWnd = 0;
	hPen = 0;
	hBrush = 0;
	hNullPen = 0;
	hNullBrush = 0;
	hOldBitmap = 0;
	Mode = GDC_SET;
	Col = 0;
	Sx = Sy = 0;

	LOGBRUSH LogBrush;
	LogBrush.lbStyle = BS_NULL;
	LogBrush.lbColor = 0;
	LogBrush.lbHatch = 0;
	hNullBrush = CreateBrushIndirect(&LogBrush);
	hNullPen = CreatePen(PS_NULL, 1, 0);
}

GdcGdiScreen::GdcGdiScreen()
{
	Init();
}

GdcGdiScreen::GdcGdiScreen(HWND hWindow)
{
	Init();

	hWnd = hWindow;
	OwnDC = true;
	Create(BeginPaint(hWnd, &Ps));

	RECT rc;
	GetClientRect(hWnd, &rc);
	SetSize(rc.right-rc.left, rc.bottom-rc.top);
}

GdcGdiScreen::GdcGdiScreen(HDC hdc, HWND hwnd)
{
	Init();

	hWnd = hwnd;
	Create(hdc);

	RECT rc;
	GetWindowRect(hWnd, &rc);
	SetSize(rc.right-rc.left, rc.bottom-rc.top);
}

GdcGdiScreen::GdcGdiScreen(HBITMAP hbmp, int Sx, int Sy)
{
	Init();

	Create(CreateCompatibleDC(0));
	hBmp = hbmp;
	hOldBitmap = (HBITMAP) SelectObject(hDC, hBmp);
	SetSize(Sx, Sy);
}

GdcGdiScreen::~GdcGdiScreen()
{
	if (hRgn)
	{
		DeleteObject(hRgn);
	}
	if (hDC)
	{
		if (hPen)
		{
			hPen = (HPEN) SelectObject(hDC, hPen);
			DeleteObject(hPen);
		}
		if (hBrush)
		{
			hBrush = (HBRUSH) SelectObject(hDC, hBrush);
			DeleteObject(hBrush);
		}

		if (hOldBitmap)
		{
			SelectObject(hDC, hOldBitmap);
			DeleteDC(hDC);
			hDC = 0;
		}
	}

	DeleteObject(hNullPen);
	DeleteObject(hNullBrush);

	if (OwnDC)
	{
		if (hWnd)
		{
			EndPaint(hWnd, &Ps);
		}
		else
		{
			ReleaseDC(hWnd, hDC);
		}
	}
}

void GdcGdiScreen::SetSize(int x, int y)
{
	Sx = x;
	Sy = y;

	Clip.ZOff(Sx-1, Sy-1);
}

bool GdcGdiScreen::Create(HDC hdc)
{
	bool Status = FALSE;

	hDC = hdc;
	if (hdc)
	{
		LOGBRUSH LogBrush;
		LogBrush.lbStyle = BS_SOLID;
		LogBrush.lbColor = 0;
		LogBrush.lbHatch = 0;
		hBrush = (HBRUSH) SelectObject(hDC, CreateBrushIndirect(&LogBrush));
		hPen = (HPEN) SelectObject(hDC, CreatePen(PS_SOLID, 1, 0));

		Status = TRUE;
	}

	return Status;
}

void GdcGdiScreen::GetOrigin(int &x, int &y)
{
	POINT pt;
	if (GetWindowOrgEx(hDC, &pt))
	{
		x = pt.x;
		y = pt.y;
	}
	else
	{
		x = y = 0;
	}
}

void GdcGdiScreen::SetOrigin(int x, int y)
{
	OriginX = x;
	OriginY = y;
	SetWindowOrgEx(hDC, x, y, NULL);
}

bool GdcGdiScreen::Applicator(GdcApplicator *pApp)
{
	return FALSE;
}

GdcApplicator *GdcGdiScreen::Applicator()
{
	return NULL;
}

GdcPalette *GdcGdiScreen::Palette()
{
	return GdcDCBase::Palette();
}

void GdcGdiScreen::Palette(GdcPalette *pPal, bool bOwnIt)
{
	GdcDCBase::Palette(pPal, bOwnIt);
}

GdcRegion GdcGdiScreen::ClipRgn(GdcRegion *Rgn)
{
	GdcRegion Prev = Clip;

	if (hRgn)
	{
		DeleteObject(hRgn);
		hRgn = 0;
	}

	if (Rgn)
	{
		Clip.x1 = max(Rgn->x1 - OriginX, 0);
		Clip.y1 = max(Rgn->y1 - OriginY, 0);
		Clip.x2 = min(Rgn->x2 - OriginX, X()-1);
		Clip.y2 = min(Rgn->y2 - OriginY, Y()-1);

		hRgn = CreateRectRgn(Clip.x1, Clip.y1, Clip.x2+1, Clip.y2+1);
		SelectClipRgn(hDC, hRgn);
	}
	else
	{
		Clip.x1 = 0;
		Clip.y1 = 0;
		Clip.x2 = X()-1;
		Clip.y2 = Y()-1;

		SelectClipRgn(hDC, NULL);
	}

	return Prev;
}

GdcRegion GdcGdiScreen::ClipRgn()
{
	return Clip;
}

COLOUR GdcGdiScreen::Colour(COLOUR c, int Bits)
{
	COLOUR Prev = Col;

	if (hDC)
	{
		hPen = (HPEN) SelectObject(hDC, hPen);
		hBrush = (HBRUSH) SelectObject(hDC, hBrush);
		DeleteObject(hPen);
		DeleteObject(hBrush);

		if (Bits)
		{
			Col = CBit(24, c, Bits, pPalette);
		}
		else
		{
			Col = CBit(24, c, GetBits(), pPalette);
		}

		LOGBRUSH LogBrush;
		LogBrush.lbStyle = BS_SOLID;
		LogBrush.lbColor = Col;
		LogBrush.lbHatch = 0;
		hBrush = (HBRUSH) SelectObject(hDC, CreateBrushIndirect(&LogBrush));
		hPen = (HPEN) SelectObject(hDC, CreatePen(PS_SOLID, 1, Col));
	}

	return Prev;
}

int GdcGdiScreen::Op(int Op)
{
	int Prev = Mode;

	switch (Op)
	{
		case GDC_SET:
		{
			Mode = R2_COPYPEN;
			break;
		}
		case GDC_AND:
		{
			Mode = R2_MASKPEN;
			break;
		}
		case GDC_OR:
		{
			Mode = R2_MERGEPEN;
			break;
		}
		case GDC_XOR:
		{
			Mode = R2_XORPEN;
			break;
		}
		default:
		{
			return Prev;
		}
	}

	SetROP2(hDC, Mode);

	return Prev;
}

COLOUR GdcGdiScreen::Colour()
{
	return Col;
}

int GdcGdiScreen::Op()
{
	return Mode;
}

int GdcGdiScreen::X()
{
	// return GetDeviceCaps(hDC, HORZSIZE);
	return Sx;
}

int GdcGdiScreen::Y()
{
	// return GetDeviceCaps(hDC, VERTSIZE);
	return Sy;
}

int GdcGdiScreen::GetBits()
{
	return GetDeviceCaps(hDC, BITSPIXEL) * GetDeviceCaps(hDC, PLANES);
}

bool GdcGdiScreen::IsScreen()
{
	return TRUE;
}

void GdcGdiScreen::Set(int x, int y)
{
	SetPixel(hDC, x, y, Col);
}

COLOUR GdcGdiScreen::Get(int x, int y)
{
	return GetPixel(hDC, x, y);
}

uint GdcGdiScreen::LineStyle(uint Bits)
{
	return LineBits;
}

uint GdcGdiScreen::LineStyle()
{
	return LineBits;
}

void GdcGdiScreen::HLine(int x1, int x2, int y)
{
	MoveToEx(hDC, x1, y, NULL);
	LineTo(hDC, x2, y);
	SetPixel(hDC, x2, y, Col);
}

void GdcGdiScreen::VLine(int x, int y1, int y2)
{
	MoveToEx(hDC, x, y1, NULL);
	LineTo(hDC, x, y2);
	SetPixel(hDC, x, y2, Col);
}

void GdcGdiScreen::Line(int x1, int y1, int x2, int y2)
{
	MoveToEx(hDC, x1, y1, NULL);
	LineTo(hDC, x2, y2);
	SetPixel(hDC, x2, y2, Col);
}

void GdcGdiScreen::Circle(double cx, double cy, double radius)
{
	HBRUSH hTemp = (HBRUSH) SelectObject(hDC, hNullBrush);
	::Ellipse(hDC, cx - radius, cy - radius, cx + radius, cy + radius);
	SelectObject(hDC, hTemp);
}

void GdcGdiScreen::FilledCircle(double cx, double cy, double radius)
{
	::Ellipse(hDC, cx - radius, cy - radius, cx + radius, cy + radius);
}

void GdcGdiScreen::Arc(double cx, double cy, double radius, double start, double end)
{
	int StartX = cx + (cos(start) * radius);
	int StartY = cy + (sin(start) * radius);
	int EndX = cx + (cos(end) * radius);
	int EndY = cy + (sin(end) * radius);

	::Arc(hDC, cx - radius, cy - radius, cx + radius, cy + radius, StartX, StartY, EndX, EndY);
}

void GdcGdiScreen::FilledArc(double cx, double cy, double radius, double start, double end)
{
	int StartX = cx + (cos(start) * radius);
	int StartY = cy + (sin(start) * radius);
	int EndX = cx + (cos(end) * radius);
	int EndY = cy + (sin(end) * radius);

	Pie(hDC, cx - radius, cy - radius, cx + radius, cy + radius, StartX, StartY, EndX, EndY);
}

void GdcGdiScreen::Ellipse(double cx, double cy, double x, double y)
{
	HBRUSH hTemp = (HBRUSH) SelectObject(hDC, hNullBrush);
	::Ellipse(hDC, cx - x, cy - y, cx + x, cy + y);
	SelectObject(hDC, hTemp);
}

void GdcGdiScreen::FilledEllipse(double cx, double cy, double x, double y)
{
	::Ellipse(hDC, cx - x, cy - y, cx + x, cy + y);
}

void GdcGdiScreen::Box(int x1, int y1, int x2, int y2)
{
	HBRUSH hTemp = (HBRUSH) SelectObject(hDC, hNullBrush);
	::Rectangle(hDC, x1, y1, x2+1, y2+1);
	SelectObject(hDC, hTemp);
}

void GdcGdiScreen::Box(GdcRegion *a)
{
	HBRUSH hTemp = (HBRUSH) SelectObject(hDC, hNullBrush);
	if (a)
	{
		::Rectangle(hDC, a->x1, a->y1, a->x2+1, a->y2+1);
	}
	else
	{
		::Rectangle(hDC, 0, 0, X(), Y());
	}
	SelectObject(hDC, hTemp);
}

void GdcGdiScreen::Rectangle(int x1, int y1, int x2, int y2)
{
	::Rectangle(hDC, x1, y1, x2+1, y2+1);
}

void GdcGdiScreen::Rectangle(GdcRegion *a)
{
	GdcRegion b;
	if (a)
	{
		b = *a;
	}
	else
	{
		b.ZOff(X()-1, Y()-1);
	}

	::Rectangle(hDC, b.x1, b.y1, b.x2+1, b.y2+1);
}

GdcPalette *GdcGdiScreen::LastRealized = 0;

void GdcGdiScreen::Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a)
{
	if (Src)
	{
		GdcRegion b;
		if (a)
		{
			b = *a;
		}
		else
		{
			b.ZOff(Src->X()-1, Src->Y()-1);
		}

		int RowOp;

		switch (Mode)
		{
			case GDC_SET:
			{
				RowOp = SRCCOPY;
				break;
			}
			case GDC_AND:
			{
				RowOp = SRCAND;
				break;
			}
			case GDC_OR:
			{
				RowOp = SRCPAINT;
				break;
			}
			case GDC_XOR:
			{
				RowOp = SRCINVERT;
				break;
			}
			default:
			{
				return;
			}
		}

		HDC hDestDC = StartDC();
		HDC hSrcDC = Src->StartDC();

		GdcPalette *Pal = Src->Palette();
		HPALETTE sPal = 0, dPal = 0;
		if (Pal)
		{
			// Pal->Update();
			Src->Update(-1);
			dPal = SelectPalette(hDestDC, Pal->Handle(), false);
		}

		BitBlt(hDestDC, x, y, b.X(), b.Y(), hSrcDC, b.x1, b.y1, RowOp);

		if (Pal)
		{
			dPal = SelectPalette(hDestDC, dPal, false);
		}
		
		Src->EndDC();
		EndDC();
	}
}

void GdcGdiScreen::StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s)
{
	if (Src)
	{
		GdcRegion DestR;
		if (d)
		{
			DestR = *d;
		}
		else
		{
			DestR.ZOff(X()-1, Y()-1);
		}

		GdcRegion SrcR;
		if (s)
		{
			SrcR = *s;
		}
		else
		{
			SrcR.ZOff(Src->X()-1, Src->Y()-1);
		}

		int RowOp;

		switch (Mode)
		{
			case GDC_SET:
			{
				RowOp = SRCCOPY;
				break;
			}
			case GDC_AND:
			{
				RowOp = SRCAND;
				break;
			}
			case GDC_OR:
			{
				RowOp = SRCPAINT;
				break;
			}
			case GDC_XOR:
			{
				RowOp = SRCINVERT;
				break;
			}
			default:
			{
				return;
			}
		}

		HDC hDestDC = StartDC();
		HDC hSrcDC = Src->StartDC();
		::StretchBlt(hDestDC, DestR.x1, DestR.y1, DestR.X(), DestR.Y(), hSrcDC, SrcR.x1, SrcR.y1, SrcR.X(), SrcR.Y(), RowOp);
		Src->EndDC();
		EndDC();
	}
}

void GdcGdiScreen::Polygon(int Points, int *Data)
{
	if (Points > 0 AND Data)
	{
		PPOINT pt = NEW(POINT[Points]);
		if (pt)
		{
			int *d = Data;

			for (int i=0; i<Points; i++)
			{
				pt[i].x = *d++;
				pt[i].y = *d++;
			}

			::Polygon(hDC, pt, Points);
			DeleteArray(pt);
		}
	}
}

void GdcGdiScreen::Text(GdcFont *pFont, int x, int y, char *Str, int Len)
{
}

void GdcGdiScreen::Bezier(int Threshold, GdcPt2 *Pt)
{
}

void GdcGdiScreen::FloodFill(int x, int y, int Mode, COLOUR Border)
{
	::FloodFill(hDC, x, y, Col);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
GdcMemDC::GdcMemDC()
{
	pBits = 0;
	hBmp = 0;
	hDC = 0;
	Info = 0;
	OldPal = 0;
}

GdcMemDC::~GdcMemDC()
{
	if (hBmp)
	{
		DeleteObject(hBmp);
		hBmp = 0;
	}
	DeleteObj(Info);
}

void GdcMemDC::Update(int Flags)
{
	if (Info AND pPalette AND (Flags & GDC_PAL_CHANGE))
	{
		GdcRGB *p = (*pPalette)[0];
		if (p)
		{
			for (int i=0; i<pPalette->GetSize(); i++, p++)
			{
				Info->bmiColors[i].rgbRed = p->R;
				Info->bmiColors[i].rgbGreen = p->G;
				Info->bmiColors[i].rgbBlue = p->B;
			}

			if (hDC)
			{
				SetDIBColorTable(hDC, 0, 256, Info->bmiColors);
			}
		}
	}

}

HDC GdcMemDC::StartDC()
{
	hDC = CreateCompatibleDC(NULL);
	if (hDC)
	{
		hBmp = (HBITMAP) SelectObject(hDC, hBmp);
		SetWindowOrgEx(hDC, -OriginX, -OriginY, NULL);
	
		/*
		if (GetBits() <= 8 AND pPalette)
		{
			OldPal = SelectPalette(hDC, pPalette->Handle(), FALSE);
		}
		*/
	}

	return hDC;
}

void GdcMemDC::EndDC()
{
	if (hDC)
	{
		/*
		if (GetBits() <= 8 AND pPalette)
		{
			SelectPalette(hDC, OldPal, FALSE);
		}
		*/

		hBmp = (HBITMAP) SelectObject(hDC, hBmp);
		DeleteDC(hDC);
		hDC = 0;
	}
}

GdcRegion GdcMemDC::ClipRgn(GdcRegion *Rgn)
{
	// tell the HDC about the new clipping area
	// mainly for the OS text functions to clip correctly..
	GdcRegion Prev = Clip;

	if (hRgn)
	{
		DeleteObject(hRgn);
		hRgn = 0;
	}

	if (Rgn)
	{
		Clip.x1 = max(Rgn->x1, 0);
		Clip.y1 = max(Rgn->y1, 0);
		Clip.x2 = min(Rgn->x2, X()-1);
		Clip.y2 = min(Rgn->y2, Y()-1);

		hRgn = CreateRectRgn(Clip.x1, Clip.y1, Clip.x2+1, Clip.y2+1);
		SelectClipRgn(hDC, hRgn);
	}
	else
	{
		Clip.x1 = 0;
		Clip.y1 = 0;
		Clip.x2 = X()-1;
		Clip.y2 = Y()-1;

		SelectClipRgn(hDC, NULL);
	}

	return Prev;
}

bool GdcMemDC::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	bool Status = FALSE;
	HBITMAP hOldBmp = hBmp;
	BITMAPINFO *OldInfo = Info;
	GdcSurface *pOldMem = pMem;

	DrawOnAlpha(FALSE);
	DeleteObj(pAlphaDC);

	hBmp = 0;
	Info = 0;
	pMem = 0;

	if (Bits == 15) Bits = 16;

	if ((x > 0 AND y > 0) AND (Bits == 8 OR Bits == 16 OR Bits == 24 OR Bits == 32))
	{
		int Colours = 1 << min(Bits, 8);
		int SizeOf = sizeof(BITMAPINFO)+(sizeof(RGBQUAD)*Colours);
		Info = (PBITMAPINFO) NEW(char[SizeOf]);
		if (Info)
		{
			LineLen = max((((x * Bits) + 31) / 32) * 4, LineLen);

			Info->bmiHeader.biSize = sizeof(Info->bmiHeader);
			Info->bmiHeader.biWidth = x;
			Info->bmiHeader.biHeight = -y;
			Info->bmiHeader.biPlanes = 1;
			Info->bmiHeader.biBitCount = Bits;
			Info->bmiHeader.biCompression = (Bits == 16 OR Bits == 32) ? BI_BITFIELDS : BI_RGB;
			Info->bmiHeader.biSizeImage = LineLen * y;
			Info->bmiHeader.biXPelsPerMeter = 3000;
			Info->bmiHeader.biYPelsPerMeter = 3000;
			Info->bmiHeader.biClrUsed = 0;
			Info->bmiHeader.biClrImportant = 0;

			if (Bits == 16)
			{
				int *BitFeilds = (int*) Info->bmiColors;
				BitFeilds[0] = 0xF800;
				BitFeilds[1] = 0x07E0;
				BitFeilds[2] = 0x001F;
			}
			else if (Bits == 32)
			{
				int *BitFeilds = (int*) Info->bmiColors;
				BitFeilds[0] = 0x00FF0000;
				BitFeilds[1] = 0x0000FF00;
				BitFeilds[2] = 0x000000FF;
			}

			HDC hDC = GetDC(NULL);
			if (hDC)
			{
				if (GdcD AND
					GdcD->GetBits() <= 8)
				{
					PALETTEENTRY Pal[256];
					int Cols = 1 << GdcD->GetBits();
					GetSystemPaletteEntries(hDC, 0, Cols, Pal);
					for (int i=0; i<Cols; i++)
					{
						Info->bmiColors[i].rgbRed = Pal[i].peRed;
						Info->bmiColors[i].rgbGreen = Pal[i].peGreen;
						Info->bmiColors[i].rgbBlue = Pal[i].peBlue;
					}
				}

				hBmp = CreateDIBSection(hDC, Info, DIB_RGB_COLORS, &pBits, NULL, 0);
				if (hBmp)
				{
					pMem = NEW(GdcSurface);
					if (pMem)
					{
						pMem->Base = (uchar*) pBits;
						pMem->x = x;
						pMem->y = y;
						pMem->Bits = Bits;
						pMem->Line = LineLen;
						pMem->Flags = 0;

						Status = TRUE;

						int NewOp = (pApp) ? Op() : GDC_SET;

						if (	(Flags & GDC_OWN_APPLICATOR) AND
							NOT (Flags & GDC_CACHED_APPLICATOR))
						{
							DeleteObj(pApp);
						}

						for (int i=0; i<GDC_CACHE_SIZE; i++)
						{
							DeleteObj(pAppCache[i]);
						}

						if (NewOp < GDC_CACHE_SIZE AND NOT DrawOnAlpha())
						{
							pApp = (pAppCache[NewOp]) ? pAppCache[NewOp] : pAppCache[NewOp] = CreateApplicator(NewOp);
							Flags &= ~GDC_OWN_APPLICATOR;
							Flags |= GDC_CACHED_APPLICATOR;
						}
						else
						{
							pApp = CreateApplicator(NewOp);
							Flags &= ~GDC_CACHED_APPLICATOR;
							Flags |= GDC_OWN_APPLICATOR;
						}

						Clip.ZOff(X()-1, Y()-1);
					}
				}
				else
				{
					DWORD Error = GetLastError();
					int k=0;
				}
			}

			ReleaseDC(NULL, hDC);
		}

		if (Status)
		{
			if (KeepData)
			{
				GdcApplicator *MyApp = CreateApplicator(GDC_SET, Bits);
				if (MyApp)
				{
					GdcSurface Temp = *pOldMem;
					Temp.x = min(pMem->x, pOldMem->x);
					Temp.y = min(pMem->y, pOldMem->y);

					MyApp->SetSurface(pMem);
					MyApp->c = 0;
					MyApp->SetPtr(0, 0);
					MyApp->Rectangle(pMem->x, pMem->y);
					// MyApp->Blt(&Temp, pPalette, pPalette);

					DeleteObj(MyApp);
				}
				else
				{
					Status = FALSE;
				}
			}
		}

		if (NOT Status)
		{
			if (hBmp)
			{
				DeleteObject(hBmp);
				hBmp = 0;
			}
			DeleteObj(Info);
			DeleteObj(pMem)
		}
	}

	if (hOldBmp)
	{
		DeleteObject(hOldBmp);
		hOldBmp = 0;
	}
	DeleteObj(OldInfo);
	DeleteObj(pOldMem);

	return Status;
}


void GdcMemDC::Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a)
{
	if (Src->IsScreen())
	{
		if (Src)
		{
			GdcRegion b;
			if (a)
			{
				b = *a;
			}
			else
			{
				b.ZOff(Src->X()-1, Src->Y()-1);
			}

			int RowOp;

			switch (Op())
			{
				case GDC_SET:
				{
					RowOp = SRCCOPY;
					break;
				}
				case GDC_AND:
				{
					RowOp = SRCAND;
					break;
				}
				case GDC_OR:
				{
					RowOp = SRCPAINT;
					break;
				}
				case GDC_XOR:
				{
					RowOp = SRCINVERT;
					break;
				}
				default:
				{
					return;
				}
			}

			HDC hDestDC = StartDC();
			HDC hSrcDC = Src->StartDC();

			/*
			GdcPalette *Pal = Src->Palette();
			HPALETTE hSPal = 0;
			HPALETTE hDPal = 0;
			if (Pal)
			{
				PALETTEENTRY Entries[256];
				
				GetPaletteEntries(Pal->Handle(), 0, 256, Entries);
				
				hDPal = SelectPalette(hDestDC, Pal->Handle(), FALSE);
				hSPal = SelectPalette(hSrcDC, Pal->Handle(), FALSE);
			}
			*/

			BitBlt(hDestDC, x, y, b.X(), b.Y(), hSrcDC, b.x1, b.y1, RowOp);

			/*
			if (Pal)
			{
				hDPal = SelectPalette(hDestDC, hDPal, FALSE);
				hSPal = SelectPalette(hSrcDC, hSPal, FALSE);
			}
			*/
			
			Src->EndDC();
			EndDC();
		}
	}
	else
	{
		GdcBasePrimitives::Blt(x, y, Src, a);
	}
}

void GdcMemDC::StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s)
{
	if (Src)
	{
		GdcRegion DestR;
		if (d)
		{
			DestR = *d;
		}
		else
		{
			DestR.ZOff(X()-1, Y()-1);
		}

		GdcRegion SrcR;
		if (s)
		{
			SrcR = *s;
		}
		else
		{
			SrcR.ZOff(Src->X()-1, Src->Y()-1);
		}

		int RowOp;

		switch (Op())
		{
			case GDC_SET:
			{
				RowOp = SRCCOPY;
				break;
			}
			case GDC_AND:
			{
				RowOp = SRCAND;
				break;
			}
			case GDC_OR:
			{
				RowOp = SRCPAINT;
				break;
			}
			case GDC_XOR:
			{
				RowOp = SRCINVERT;
				break;
			}
			default:
			{
				return;
			}
		}

		HDC hDestDC = StartDC();
		HDC hSrcDC = Src->StartDC();
		::StretchBlt(hDestDC, DestR.x1, DestR.y1, DestR.X(), DestR.Y(), hSrcDC, SrcR.x1, SrcR.y1, SrcR.X(), SrcR.Y(), RowOp);
		Src->EndDC();
		EndDC();
	}
}

void GdcMemDC::HLine(int x1, int x2, int y, COLOUR a, COLOUR b)
{
	if (x1 > x2) swap(x1, x2);

	if (x1 < Clip.x1) x1 = Clip.x1;
	if (x2 > Clip.x2) x2 = Clip.x2;
	if (	x1 <= x2 AND
		y >= Clip.y1 AND
		y <= Clip.y2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x1, y);
		for (; x1 <= x2; x1++)
		{
			if (x1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncX();
		}

		pApp->c = Prev;
	}
}

void GdcMemDC::VLine(int x, int y1, int y2, COLOUR a, COLOUR b)
{
	if (y1 > y2) swap(y1, y2);
	
	if (y1 < Clip.y1) y1 = Clip.y1;
	if (y2 > Clip.y2) y2 = Clip.y2;
	if (	y1 <= y2 AND
		x >= Clip.x1 AND
		x <= Clip.x2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x, y1);
		for (; y1 <= y2; y1++)
		{
			if (y1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncY();
		}

		pApp->c = Prev;
	}
}

#ifdef ADJUST_ORIGIN
void GdcMemDC::SetOrigin(int x, int y)
{
	GdcDCBase::SetOrigin(x, y);
	if (hDC)
	{
		SetWindowOrgEx(hDC, -OriginX, -OriginY, NULL);
	}
}
#endif
