//
//      FILE:           LgiOsDefs.h (Win32)
//      AUTHOR:         Matthew Allen
//      DATE:           24/9/99
//      DESCRIPTION:    Lgi Win32 OS defines
//
//      Copyright (C) 1999, Matthew Allen
//              fret@ozemail.com.au
//

#ifndef __LGI_OS_DEFS_H
#define __LGI_OS_DEFS_H

#include <string.h>

//////////////////////////////////////////////////////////////////
// Includes
#include "windows.h"
#include "commctrl.h"
#include "mmsystem.h"

#include "LgiInc.h"

//////////////////////////////////////////////////////////////////
// Typedefs
typedef			__int64				quad;
typedef			signed __int64		int64;
typedef			unsigned __int64	uint64;
typedef			signed int			int32;
typedef			unsigned int		uint32;

//////////////////////////////////////////////////////////////////
// Defines
#define Message(m)					(m->Msg)
#define MessageA(m)					(m->a)
#define MessageB(m)					(m->b)
#define CreateMsg(m, a, b)			GMessage(m, a, b)
#define LgiAssert(a)				ASSERT(a)

// Sleep the current thread
#define LgiSleep(i)					::Sleep(i)

// Run the message loop to process any pending messages
#define LgiYield()					GApp::ObjInstance()->Run(false)

#define K_CHAR						0x0

#define MK_LEFT						MK_LBUTTON
#define MK_RIGHT					MK_RBUTTON
#define MK_MIDDLE					MK_MBUTTON
#define MK_CTRL						MK_CONTROL
	
// Edge types
#define SUNKEN						1
#define RAISED						2
#define CHISEL						3
#define FLAT						4

// Stupid mouse wheel defines don't work. hmmm
#define WM_MOUSEWHEEL				0x020A
#define WHEEL_DELTA					120
#define SPI_GETWHEELSCROLLLINES		104

// Window flags
#define GWF_VISIBLE					0x00000001
#define GWF_ENABLED					0x00000002
#define GWF_FOCUS					0x00000004
#define GWF_OVER					0x00000008
#define GWF_DROP_TARGET				0x00000010
#define GWF_SUNKEN					0x00000020
#define GWF_FLAT					0x00000040
#define GWF_RAISED					0x00000080
#define GWF_BORDER					0x00000100
#define GWF_DIALOG					0x00000200
#define GWF_DESTRUCTOR				0x00000400
#define GWF_QUIT_WND				0x00000800

/*
// Colours
#define TOOL_LOW					GetSysColor(COLOR_3DSHADOW)
#define TOOL_MED					GetSysColor(COLOR_3DFACE)
#define TOOL_HIGH					GetSysColor(COLOR_3DHILIGHT)
#define C_WORKSPACE					GetSysColor(COLOR_3DHILIGHT)
*/

// Menus
#define PostMenuCmd(hWnd, Cmd)		PostMessage(hWnd, WM_COMMAND, MAKELONG(Cmd, 0), (int) hWnd)

// Widgets
#define DialogToPixelX(i)			(((i)*Bx)/4)
#define DialogToPixelY(i)			(((i)*By)/8)
#define PixelToDialogX(i)			(((i)*4)/Bx)
#define PixelToDialogY(i)			(((i)*8)/By)

#define DIALOG_X					1.56
#define DIALOG_Y					1.85
#define CTRL_X						1.50
#define CTRL_Y						1.64

// Messages

// Quite a lot of windows stuff uses WM_USER+n where
// n < 0x1A0 or so... so stay out of their way.

	// wParam = bool Inside; // is the mouse inside the client area?
	// lParam = MAKELONG(x, y); // mouse location
	#define WM_MOUSEENTER				(WM_USER+0x1000)
	#define WM_MOUSEEXIT				(WM_USER+0x1001)

	// wParam = (GView*) Wnd;
	// lParam = (int) Flags;
	#define WM_CHANGE					(WM_USER+0x1002)

	// wParam = (GView*) Wnd;
	// lParam = (char*) Text; // description from window
	#define WM_DESCRIBE					(WM_USER+0x1003)

	// return (bool)
	#define WM_WANT_DIALOG_PROC			(WM_USER+0x1004)

	// wParam = void
	// lParam = (MSG*) Msg;
	#define WM_TRANSLATE_ACCELERATOR	(WM_USER+0x1005)

	// None
	#define WM_TRAY_NOTIFY				(WM_USER+0x1006)

// Directories
#define DIR_CHAR					'\\'
#define DIR_STR						"\\"

#define IsSlash(c)					(((c)=='/')OR((c)=='\\'))
#define IsQuote(c)					(((c)=='\"')OR((c)=='\''))

/////////////////////////////////////////////////////////////////////////////////////
// Typedefs
typedef HWND GWndRef;

/////////////////////////////////////////////////////////////////////////////////////
// Externs
LgiFunc GView *GWindowFromHandle(GWndRef hWnd);
LgiFunc int GetMouseWheelLines();
LgiFunc void FormatSize(char *Str, quad Size);

////////////////////////////////////////////////////////////////////
// Classes

// Win32 Memory Handler
class LgiClass GMem {

	int Size;
	HGLOBAL hMem;
	void *pMem;

public:
	GMem(int size)
	{
		Size = size;
		hMem = GlobalAlloc(GMEM_FIXED, Size);
		pMem = 0;
	}

	GMem(HGLOBAL hmem)
	{
		hMem = hmem;
		Size = GlobalSize(hMem);
		pMem = 0;
	}

	~GMem()
	{
		if (pMem)
		{
			GlobalUnlock(hMem);
			pMem = 0;
		}
		GlobalFree(hMem);
		hMem = 0;
	}

	void *Lock()
	{
		return pMem = GlobalLock(hMem);
	}

	void UnLock()
	{
		if (pMem)
		{
			GlobalUnlock(hMem);
			pMem = 0;
		}
	}

	int GetSize()
	{
		return Size;
	}

	void Detach()
	{
		pMem = 0;
		hMem = 0;
		Size = 0;
	}

	HGLOBAL Handle() { return hMem; }
};

#endif