#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////
GClipBoard::GClipBoard(GView *o)
{
	Owner = o;
	if (Owner)
	{
		Open = OpenClipboard(Owner->Handle());
	}
	else
	{
		Open = FALSE;
	}
}

GClipBoard::~GClipBoard()
{
	if (Open)
	{
		CloseClipboard();
		Open = FALSE;
	}
}

bool GClipBoard::Empty()
{
	return EmptyClipboard();
}

// Text
bool GClipBoard::Text(char *Str)
{
	return Binary(CF_TEXT, (uchar*) Str, strlen(Str)+1);
}

char *GClipBoard::Text()
{
	int Len = 0;
	char *Str = NULL;
	if (Binary(CF_TEXT, (uchar**) &Str, &Len))
	{
		return Str;
	}

	DeleteArray(Str);
	return 0;
}

bool GClipBoard::TextW(ushort *Str)
{
	return Binary(CF_UNICODETEXT, (uchar*) Str, (wcslen(Str)+1) * sizeof(ushort));
}

ushort *GClipBoard::TextW()
{
	int Len = 0;
	ushort *Str = NULL;
	if (Binary(CF_UNICODETEXT, (uchar**) &Str, &Len))
	{
		return Str;
	}

	DeleteArray(Str);
	return 0;
}

// Bitmap
bool GClipBoard::Bitmap(GdcBasePrimitives *pDC)
{
	bool Status = FALSE;

	Empty();
	if (pDC)
	{
		int Bytes = BMPWIDTH(pDC->X() * pDC->GetBits());
		int Colours = (pDC->GetBits()>8) ? ((pDC->GetBits() != 24) ? 3 : 0) : 1 << pDC->GetBits();
		int HeaderSize = sizeof(BITMAPINFOHEADER);
		int PaletteSize = Colours * sizeof(RGBQUAD);
		int MapSize = Bytes * pDC->Y();

		HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, HeaderSize + PaletteSize + MapSize);
		if (hMem)
		{
			BITMAPINFO *Info = (BITMAPINFO*) GlobalLock(hMem);
			if (Info)
			{
				// Header
				Info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				Info->bmiHeader.biWidth = pDC->X();
				Info->bmiHeader.biHeight = pDC->Y();
				Info->bmiHeader.biPlanes = 1;
				Info->bmiHeader.biBitCount = pDC->GetBits();
				
				if (pDC->GetBits() == 16 OR
					pDC->GetBits() == 32)
				{
					Info->bmiHeader.biCompression = BI_BITFIELDS;
				}
				else
				{
					Info->bmiHeader.biCompression = BI_RGB;
				}

				Info->bmiHeader.biSizeImage = 0;
				Info->bmiHeader.biXPelsPerMeter = 3000;
				Info->bmiHeader.biYPelsPerMeter = 3000;
				Info->bmiHeader.biClrUsed = 0;
				Info->bmiHeader.biClrImportant = 0;

				if (pDC->GetBits() <= 8)
				{
					// Palette
					GdcPalette *Pal = pDC->Palette();
					RGBQUAD *Rgb = Info->bmiColors;
					if (Pal)
					{
						GdcRGB *p = (*Pal)[0];
						if (p)
						{
							for (int i=0; i<Colours; i++, p++, Rgb++)
							{
								Rgb->rgbRed = p->R;
								Rgb->rgbGreen = p->G;
								Rgb->rgbBlue = p->B;
								Rgb->rgbReserved = p->Flags;
							}
						}
					}
					else
					{
						memset(Rgb, 0, Colours * sizeof(RGBQUAD));
					}
				}
				else
				{
					int *Primaries = (int*) Info->bmiColors;
					
					switch (pDC->GetBits())
					{
						case 16:
						{
							Primaries[0] = Rgb16(255, 0, 0);
							Primaries[1] = Rgb16(0, 255, 0);
							Primaries[2] = Rgb16(0, 0, 255);
							break;
						}
						case 32:
						{
							Primaries[0] = Rgb32(255, 0, 0);
							Primaries[1] = Rgb32(0, 255, 0);
							Primaries[2] = Rgb32(0, 0, 255);
							break;
						}
					}
				}

				// Bits
				uchar *Dest = ((uchar*)Info) + HeaderSize + PaletteSize;
				for (int y=pDC->Y()-1; y>=0; y--)
				{
					uchar *s = (*pDC)[y];
					if (s)
					{
						memcpy(Dest, s, Bytes);
					}
					Dest += Bytes;
				}

				Status = SetClipboardData(CF_DIB, hMem) != 0;
			}
			else
			{
				GlobalFree(hMem);
			}
		}
	}
	return Status;
}

GdcBasePrimitives *GClipBoard::ConvertFromPtr(void *Ptr)
{
	GdcBasePrimitives *pDC = NULL;
	if (Ptr)
	{
		BITMAPINFO *Info = (BITMAPINFO*) Ptr;
		if (Info)
		{
			pDC = NEW(GdcMemDC);
			if (pDC AND
				(Info->bmiHeader.biCompression == BI_RGB OR
				Info->bmiHeader.biCompression == BI_BITFIELDS))
			{
				if (pDC->Create(Info->bmiHeader.biWidth,
						Info->bmiHeader.biHeight,
						max(Info->bmiHeader.biPlanes * Info->bmiHeader.biBitCount, 8)))
				{
					int Colours = 0;
					char *Source = (char*) Info->bmiColors;

					// do palette
					if (pDC->GetBits() <= 8)
					{
						if (Info->bmiHeader.biClrUsed > 0)
						{
							Colours = Info->bmiHeader.biClrUsed;
						}
						else
						{
							Colours = 1 << pDC->GetBits();
						}

						GdcPalette *Pal = NEW(GdcPalette(NULL, Colours));
						if (Pal)
						{
							GdcRGB *d = (*Pal)[0];
							RGBQUAD *s = (RGBQUAD*) Source;
							if (d)
							{
								for (int i=0; i<Colours; i++, d++, s++)
								{
									d->R = s->rgbRed;
									d->G = s->rgbGreen;
									d->B = s->rgbBlue;
									d->Flags = s->rgbReserved;
								}
							}
							Source = (char*) s;
							pDC->Palette(Pal);
						}
					}

					if (Info->bmiHeader.biCompression == BI_BITFIELDS)
					{
						Source += sizeof(DWORD) * 3;
					}

					// do pixels
					int Bytes = BMPWIDTH(pDC->X() * pDC->GetBits());
					for (int y=pDC->Y()-1; y>=0; y--)
					{
						uchar *d = (*pDC)[y];
						if (d)
						{
							memcpy(d, Source, Bytes);
						}
						Source += Bytes;
					}
				}
			}
		}
	}
	return pDC;
}

GdcBasePrimitives *GClipBoard::Bitmap()
{
	HGLOBAL hMem = GetClipboardData(CF_DIB);
	void *Ptr = GlobalLock(hMem);
	GdcBasePrimitives *pDC = 0;
	if (Ptr)
	{
		pDC = ConvertFromPtr(Ptr);
		GlobalUnlock(hMem);
	}
	return pDC;
}

bool GClipBoard::Binary(int Format, uchar *Ptr, int Len)
{
	bool Status = FALSE;

	Empty();
	if (Ptr AND Len > 0)
	{
		HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, Len);
		if (hMem)
		{
			char *Data = (char*) GlobalLock(hMem);
			if (Data)
			{
				memcpy(Data, Ptr, Len);
				GlobalUnlock(hMem);
				Status = SetClipboardData(Format, hMem) != 0;
			}
			else
			{
				GlobalFree(hMem);
			}
		}
	}

	return Status;
}

bool GClipBoard::Binary(int Format, uchar **Ptr, int *Len)
{
	bool Status = false;

	if (Ptr AND Len)
	{
		HGLOBAL hMem = GetClipboardData(Format);
		if (hMem)
		{
			*Len = GlobalSize(hMem);
			uchar *Data = (uchar*) GlobalLock(hMem);
			if (Data)
			{
				*Ptr = NEW(uchar[*Len]);
				if (*Ptr)
				{
					memcpy(*Ptr, Data, *Len);
					Status = true;
				}
			}

			GlobalUnlock(hMem);
		}
	}

	return Status;
}

