/*hdr
**      FILE:           Gui.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <time.h>
#include "Lgi.h"
#include "base64.h"

////////////////////////////////////////////////////////////////////////////
bool _lgi_check_file(char *Path)
{
	if (Path)
	{
		if (FileExists(Path))
		{
			// file is there
			return true;
		}
		else
		{
			// shortcut?
			char *e = Path + strlen(Path);
			strcpy(e, ".lnk");
			if (FileExists(Path))
			{
				// resolve shortcut
				char Link[256];
				if (ResolveShortcut(Path, Link, sizeof(Link)))
				{
					// check destination of link
					if (FileExists(Link))
					{
						strcpy(Path, Link);
						return true;
					}
				}
			}
			*e = 0;
		}
	}

	return false;
}

// Mouse wheel stuff
#include "zmouse.h"
int MouseRollMsg = 0;
#define SetKeyFlag(v, k, f)		if (GetKeyState(k)&0xFF00) { v |= f; }

int _lgi_mouse_wheel_lines()
{
	OSVERSIONINFO Info;
	ZeroObj(Info);
	Info.dwOSVersionInfoSize = sizeof(Info);
	if (GetVersionEx(&Info) AND
		Info.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS AND
		Info.dwMajorVersion == 4 AND
		Info.dwMinorVersion == 0)
	{
       HWND hdlMSHWheel=NULL;
       UINT msgMSHWheelGetScrollLines=NULL;
       UINT uiMsh_WheelScrollLines;

       msgMSHWheelGetScrollLines = 
               RegisterWindowMessage(MSH_SCROLL_LINES);
       hdlMSHWheel = FindWindow(MSH_WHEELMODULE_CLASS, 
                                MSH_WHEELMODULE_TITLE);
       if (hdlMSHWheel && msgMSHWheelGetScrollLines)
       {
			return SendMessage(hdlMSHWheel, msgMSHWheelGetScrollLines, 0, 0);
       }
	}
	else
	{
		UINT nScrollLines;
		if (SystemParametersInfo(	SPI_GETWHEELSCROLLLINES, 
									0, 
									(PVOID) &nScrollLines, 
									0))
		{
			return nScrollLines;
		}
	}

	return 3;
}

VMouse _lgi_adjust_click_for_window(VMouse &Info, GView *Wnd)
{
	VMouse Temp = Info;
	if (Wnd)
	{
		GdcPt2 Offset(0, 0);
		if (Wnd->WindowVirtualOffset(&Offset))
		{
			Temp.x -= Offset.x;
			Temp.y -= Offset.y;
		}
	}
	return Temp;
}

//////////////////////////////////////////////////////////////////////////
#ifdef WIN32
#define RecursiveFileSearch_Wildcard "*.*"
#else // unix'ish OS
#define RecursiveFileSearch_Wildcard "*"
#endif

bool LgiRecursiveFileSearch(char *Root,
							List<char> &Ext,
							List<char> *Files,
							quad *Size,
							quad *Count,
							RecursiveFileSearch_Callback Callback)
{
	bool Status = false;
	char Buf[256];

	// validate args
	if (NOT Root) return 0;

	// setup search string
	strcpy(Buf, Root);
	if (Buf[strlen(Buf)-1] != DIR_CHAR)
	{
		strcat(Buf, DIR_STR);
	}
	char *Wildcard = Buf+strlen(Buf);
	strcpy(Wildcard, RecursiveFileSearch_Wildcard);

	// get directory enumerator
	VDirectory *Dir = FileDev->GetDir(0);
	if (Dir)
	{
		Status = true;

		// enumerate the directory contents
		for (bool Found = Dir->First(Buf, 0, -1); Found; Found = Dir->Next())
		{
			char *Name = Dir->GetName();
			if (Name AND
				stricmp(Name, ".") != 0 AND
				stricmp(Name, "..") != 0)
			{
				strcpy(Wildcard, Name);

				if (Dir->IsDir())
				{
					// dir
					LgiRecursiveFileSearch(	Buf,
											Ext,
											Files,
											Size,
											Count,
											Callback);

				}
				else
				{
					// process file
					bool Match = true; // if no Ext's then default to match
					for (char *e = Ext.First(); e; e = Ext.Next())
					{
						if (Match = MatchStr(e, Name))
						{
							break;
						}
					}

					if (Match)
					{
						// file matched... process:
						if (Files)
						{
							Files->Insert(NewStr(Buf));
						}

						if (Size)
						{
							*Size += Dir->GetSize();
						}

						if (Count)
						{
							*Count++;
						}

						if (Callback)
						{
							if (NOT Callback(Buf, Dir))
							{
								Status = false;
								break;
							}
						}
					}
				}

				strcpy(Wildcard, RecursiveFileSearch_Wildcard);
			}
		}

		DeleteObj(Dir);
	}

	return Status;
}

#define LGI_TRACE_TO_FILE

void LgiTrace(char *Msg, ...)
{
	if (Msg)
	{
		char Buffer[1025];
		LgiGetExeFile(Buffer);
		char *Dot = strrchr(Buffer, '.');
		if (Dot AND NOT strchr(Dot, DIR_CHAR))
		{
			strcpy(Dot+1, "txt");
		}
		else
		{
			strcat(Buffer, ".txt");
		}

		#ifdef LGI_TRACE_TO_FILE
		VFile f;
		bool FileOpen = f.Open(Buffer, O_WRITE);
		#endif

		va_list Arg;
		va_start(Arg, Msg);
		vsprintf(Buffer, Msg, Arg);
		va_end(Arg);

		#ifdef LGI_TRACE_TO_FILE
		f.Seek(0, SEEK_END);
		f.Write(Buffer, strlen(Buffer));
		f.Write("\n", 1);
		f.Close();
		#endif

		OutputDebugString(Buffer);
	}
}

bool LgiGetFileMimeType(char *File, char *Mime)
{
	bool Status = false;

	if (File AND Mime)
	{
		char *Dot = strrchr(File, '.');
		if (Dot)
		{
			HKEY hClass;
			if (RegOpenKey(HKEY_CLASSES_ROOT, Dot, &hClass) == ERROR_SUCCESS)
			{
				char Str[256];
				DWORD StrLen = sizeof(Str), Type = 0;
				if (RegQueryValueEx(hClass, "Content Type", 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
				{
					strcpy(Mime, Str);
					Status = true;
				}

				RegCloseKey(hClass);
			}
		}

		if (NOT Status)
		{
			// no extension?
			// no registry entry for file type?
			strcpy(Mime, "application/octet-stream");
			Status = true;
		}
	}

	return Status;
}

bool LgiGetAppForMimeType(char *Mime, char *AppPath)
{
	bool Status = false;

	if (AppPath AND Mime)
	{
		if (stricmp(Mime, "application/email") == 0)
		{
			// get email app
			HKEY hKey;
			if (RegOpenKey(HKEY_CLASSES_ROOT, "mailto\\shell\\open\\command", &hKey) == ERROR_SUCCESS)
			{
				char Str[256];
				DWORD StrLen = sizeof(Str), Type = 0;

				// get app path
				if (RegQueryValueEx(hKey, 0, 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
				{
					strcpy(AppPath, Str);
					Status = true;
				}

				RegCloseKey(hKey);
			}
		}
		else
		{
			// get generic app
			HKEY hKey;
			char KeyName[256];
			sprintf(KeyName, "MIME\\Database\\Content Type\\%s", Mime);
			if (RegOpenKey(HKEY_CLASSES_ROOT, KeyName, &hKey) == ERROR_SUCCESS)
			{
				char Str[256];
				DWORD StrLen = sizeof(Str), Type = 0;

				// get extension
				if (RegQueryValueEx(hKey, "Extension", 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
				{
					RegCloseKey(hKey);

					// get classes location
					if (RegOpenKey(HKEY_CLASSES_ROOT, Str, &hKey) == ERROR_SUCCESS)
					{
						StrLen = sizeof(Str);
						if (RegQueryValueEx(hKey, 0, 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
						{
							RegCloseKey(hKey);
							
							// get app path
							strcat(Str, "\\shell\\open\\command");
							if (RegOpenKey(HKEY_CLASSES_ROOT, Str, &hKey) == ERROR_SUCCESS)
							{
								StrLen = sizeof(Str);
								if (RegQueryValueEx(hKey, 0, 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
								{
									strcpy(AppPath, Str);
									Status = true;
								}

								RegCloseKey(hKey);
							}
						}
						else
						{
							RegCloseKey(hKey);
						}
					}
				}
				else
				{
					RegCloseKey(hKey);
				}
			}
		}

		char *RootVar = "%SystemRoot%";
		char *SysRoot = strstr(RootVar, AppPath);
		if (SysRoot)
		{
			// correct path for variables
			char Temp[256];
			GetWindowsDirectory(Temp, sizeof(Temp));
			if (Temp[strlen(Temp)-1] != DIR_CHAR) strcat(Temp, DIR_STR);
			char *End = AppPath + strlen(RootVar);
			strcat(Temp, (*End == DIR_CHAR) ? End + 1 : End);
			strcpy(AppPath, Temp);
		}
	}

	return Status;
}

int LgiRand(int i)
{
	return (rand() % i);
}

int LgiMsg(GView *Parent, char *Msg, char *Title, int Type, ...)
{
	int Result = 0;
	if (Msg)
	{
		char Buffer[1025];

		va_list Arg;
		va_start(Arg, Type);
		vsprintf(Buffer, Msg, Arg);
		va_end(Arg);

		Result = MessageBox((Parent)?Parent->Handle():0,
							Buffer,
							(Title)?Title:"Message",
							Type);
	}

	return Result;
}

/*
bool LgiGetCursorPos(GdcPt2 *Pt, int *Buttons)
{
	bool Status = false;
	POINT p;

	if (Pt AND GetCursorPos(&p))
	{
		Pt->x = p.x;
		Pt->y = p.y;
		Status = true;
	}
	if (Buttons)
	{
		*Buttons =	(GetAsyncKeyState(VK_LBUTTON)?MK_LEFT:0) |
					(GetAsyncKeyState(VK_MBUTTON)?MK_MIDDLE:0) |
					(GetAsyncKeyState(VK_RBUTTON)?MK_RIGHT:0);
	}

	return Status;
}
*/

bool LgiExecute(char *File, char *Arguments, char *Dir)
{
	int Status = (int) ShellExecute(NULL, "open", File, Arguments, Dir, 5);
	return Status > 32;
}

bool LgiTrimDir(char *Path)
{
	if (Path)
	{
		char *p = strrchr(Path, DIR_CHAR);
		if (p)
		{
			*p = 0;
			return true;
		}
	}
	return false;
}

bool LgiGetExeFile(char *Dst)
{
	if (Dst)
	{
		#ifdef WIN32
		char Exe[256];
		if (GetModuleFileName(NULL, Exe, sizeof(Exe)) > 0)
		{
			strlwr(Exe);
			strcpy(Dst, Exe);
			return TRUE;
		}
		#else if BEOS
		app_info Info;
		if (LgiApp->GetAppInfo(&Info) == B_OK)
		{
			BEntry e(&Info.ref);
			BPath p;
			if (e.GetPath(&p) == B_OK)
			{
				strcpy(Dst, p.Path());
				return TRUE;
			}
		}
		#endif
	}
	return FALSE;
}

bool LgiGetExePath(char *Dst)
{
	if (Dst)
	{
		if (LgiGetExeFile(Dst))
		{
			LgiTrimDir(Dst);
			return TRUE;
		}
	}
	return FALSE;
}

char *LgiFindFile(char *Name)
{
	char *Result = 0;

	if (Name)
	{
		char *JustFile = Name + (strlen(Name) - 1);
		for (; JustFile>Name AND *JustFile != DIR_CHAR; JustFile--);
		if (JustFile > Name) JustFile++;

		#ifdef WIN32
		char ExePath[256];
		char File[256];
		if (LgiGetExePath(ExePath))
		{
			sprintf(File, "%s\\%s", ExePath, JustFile);
			if (_lgi_check_file(File))
			{
				Result = NewStr(File);
			}
			else
			{
				// ..
				strcpy(File, ExePath);
				LgiTrimDir(File);
				sprintf(File+strlen(File), "\\%s", JustFile);
				if (_lgi_check_file(File))
				{
					Result = NewStr(File);
				}
				else
				{
					// ..\\code
					strcpy(File, ExePath);
					LgiTrimDir(File);
					sprintf(File+strlen(File), "\\code\\%s", JustFile);
					if (_lgi_check_file(File))
					{
						Result = NewStr(File);
					}
				}
			}
		}
		#else if BEOS
		char ExePath[256];
		char File[256];
		if (LgiGetExePath(ExePath))
		{
			sprintf(File, "%s/%s", ExePath, JustFile);
			if (FileExists(File))
			{
				Result = NewStr(File);
			}
			else
			{
				// ..
				strcpy(File, ExePath);
				LgiTrimDir(File);
				sprintf(File+strlen(File), "/%s", JustFile);
				if (FileExists(File))
				{
					Result = NewStr(File);
				}
				else
				{
					// code
					strcpy(File, ExePath);
					sprintf(File+strlen(File), "/Code/%s", JustFile);
					if (FileExists(File))
					{
						Result = NewStr(File);
					}
				}
			}
		}
		#endif
	}

	return Result;
}

bool LgiPlaySound(char *FileName, int Flags)
{
	bool Status = false;
	HMODULE hDll = LoadLibrary("winmm.dll");
	if (hDll)
	{
		typedef BOOL (__stdcall *Proc_sndPlaySound)(LPCSTR pszSound, UINT fuSound);
		Proc_sndPlaySound psndPlaySound = (Proc_sndPlaySound)GetProcAddress(hDll, "sndPlaySound");

		if (psndPlaySound)
		{
			Status = psndPlaySound(FileName, Flags);
		}

		FreeLibrary(hDll);
	}

	return Status;
}

/////////////////////////////////////////////////////////////////////////////////////////////
void DrawBox(GdcBasePrimitives *pDC, GdcRegion &r, bool Sunken, bool Fill)
{
	if (Fill)
	{
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle(r.x1+1, r.y1+1, r.x2-1, r.y2-1);
	}

	pDC->Colour((Sunken) ? LC_LIGHT : LC_LOW, 24);
	pDC->Line(r.x2, r.y2, r.x2, r.y1);
	pDC->Line(r.x2, r.y2, r.x1, r.y2);

	pDC->Colour((Sunken) ? LC_LOW : LC_LIGHT, 24);
	pDC->Line(r.x1, r.y1, r.x1, r.y2);
	pDC->Line(r.x1, r.y1, r.x2, r.y1);
}

void DrawMoveDeepEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	if (NOT pDC) return;
	COLOUR Old = pDC->Colour();
	COLOUR VLow = LC_SHADOW;
	COLOUR Low = LC_LOW;
	COLOUR High = LC_HIGH;
	COLOUR VHigh = LC_LIGHT;

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(Low, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(VLow, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(High, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VHigh, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case RAISED:
		{
			pDC->Colour(VHigh, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-1, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-1);

			pDC->Colour(Low, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VLow, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case CHISEL:
		{
			pDC->Colour(Low, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(VHigh, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(Low, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VHigh, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
	}

	r.Size(2, 2);
	pDC->Colour(Old);
}

void DrawMoveShallowEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	if (NOT pDC) return;
	COLOUR Old = pDC->Colour();

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(LC_LIGHT, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(LC_LOW, 24);
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
		case RAISED:
		{
			pDC->Colour(LC_LOW, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(LC_LIGHT, 24);
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
	}

	pDC->Colour(Old);
}

void DrawMoveFill(GdcBasePrimitives *pDC, GdcRegion &r, int Width)
{
	pDC->Colour(LC_MED, 24);
	if (Width < 1 OR r.X() < (2 * Width) OR r.Y() < (2 * Width))
	{
		pDC->Rectangle(&r);
		r.ZOff(-1, -1);
	}
	else
	{
		pDC->Rectangle(r.x1, r.y1, r.x2, r.y1+Width-1);
		pDC->Rectangle(r.x1, r.y2-Width+1, r.x2, r.y2);
		pDC->Rectangle(r.x1, r.y1+Width, r.x1+Width-1, r.y2-Width);
		pDC->Rectangle(r.x2-Width+1, r.y1+Width, r.x2, r.y2-Width);
		r.Size(Width, Width);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GFlags::GFlags()
{
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
}

bool GFlags::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GFlags::Enabled(bool i)
{
	if (i) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

bool GFlags::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GFlags::Visible(bool i)
{
	if (i) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);
}

bool GFlags::Focus()
{
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GFlags::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

bool GFlags::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

void GFlags::DropTarget(bool i)
{
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
}

#define GWF_ALL (GWF_SUNKEN | GWF_FLAT | GWF_RAISED)

bool GFlags::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GFlags::Sunken(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
}

bool GFlags::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GFlags::Flat(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_FLAT);
}

bool GFlags::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GFlags::Raised(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_RAISED);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GWin32Class GLayout::Class;

////////////////////////////////////////////////////////////////////////////////////////////////////
HHOOK	hMouseHook = 0;
HWND	hMouseOver = 0;

LRESULT CALLBACK MouseHook(int nCode, WPARAM wParam, LPARAM lParam)
{
	// check whether the mouse has moved off the last window
	MOUSEHOOKSTRUCT *Ms = (MOUSEHOOKSTRUCT*) lParam;
	if (Ms)
	{
		RECT rc;
		GetClientRect(Ms->hwnd, &rc);
		ScreenToClient(Ms->hwnd, &Ms->pt);

		// is the mouse inside the client area?
		bool Inside = NOT	(Ms->pt.x < 0 OR
							Ms->pt.y < 0 OR
							Ms->pt.x >= rc.right-rc.left OR
							Ms->pt.y >= rc.bottom-rc.top);

		HWND hWnd = (Inside) ? Ms->hwnd : 0;

		if (hWnd != hMouseOver)
		{
			// window has changed
			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is LOSING mouse
				SendMessage(hMouseOver, WM_MOUSEEXIT, 0, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}

			// set current window
			hMouseOver = hWnd;

			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is GETTING mouse
				SendMessage(hMouseOver, WM_MOUSEENTER, Inside, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}
		}
	}

	return CallNextHookEx(hMouseHook, nCode, wParam, lParam);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK GWin32Class::Redir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		
		GView *NewWnd = (GView*) Info->lpCreateParams;
		NewWnd->hWnd = hWnd;
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GView *Wnd = (GView*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	return DefWindowProc(hWnd, m, a, b);
}

LRESULT CALLBACK GWin32Class::SubClassRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		GView *NewWnd = 0;
		if (Info->lpCreateParams)
		{
			NewWnd = (GView*) Info->lpCreateParams;
			NewWnd->hWnd = hWnd;
		}
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GView *Wnd = (GView*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		int Status = Wnd->OnEvent(&Msg);
		return Status;
	}

	return DefWindowProc(hWnd, m, a, b);
}

GWin32Class::GWin32Class()
{
	Class.style = CS_DBLCLKS; //  | CS_HREDRAW | CS_VREDRAW
	Class.lpfnWndProc = (WNDPROC) Redir;
	Class.cbClsExtra = 0;
	Class.cbWndExtra = 0;
	Class.hInstance = 0;
	Class.hIcon = NULL;
	Class.hCursor = 0; // LoadCursor(hAppInst(), MAKEINTRESOURCE(IDC_ARROW));
	Class.hbrBackground = NULL;
	Class.lpszMenuName = NULL;
	Class.lpszClassName = NULL;

	ParentProc = 0;
}

GWin32Class::~GWin32Class()
{
	UnregisterClass(Name(), hAppInst());
	Class.lpszClassName = NULL;
}

bool GWin32Class::Create(char *ClassName)
{
	bool Status = FALSE;

	if (NOT Class.lpszClassName)
	{
		Name(ClassName);
		Class.hInstance = hAppInst();
		Class.lpszClassName = ClassName;
		Status = RegisterClass(&Class) != 0;
	}
	else
	{
		Status = TRUE;
	}

	return Status;
}

bool GWin32Class::SubClass(char *ClassName, char *Parent)
{
	bool Status = FALSE;

	if (NOT Class.lpszClassName)
	{
		HBRUSH hBr = Class.hbrBackground;

		if (GetClassInfo(hAppInst(), Parent, &Class))
		{
			Name(ClassName);
			ParentProc = Class.lpfnWndProc;
			if (hBr)
			{
				Class.hbrBackground = hBr;
			}
			Class.hInstance = hAppInst();
			Class.lpfnWndProc = (WNDPROC) SubClassRedir;
			Class.lpszClassName = ClassName;
			Status = RegisterClass(&Class) != 0;
		}
	}
	else
	{
		return TRUE;
	}

	return Status;
}

LRESULT CALLBACK GWin32Class::CallParent(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (NOT ParentProc) return 0;
	#if _MSC_VER == 1100
	return CallWindowProc((FARPROC) ParentProc, hWnd, m, a, b);
	#else
	return CallWindowProc(ParentProc, hWnd, m, a, b);
	#endif
}

//////////////////////////////////////////////////////////////////////////////
GView *GWindowFromHandle(HWND hWnd)
{
	return dynamic_cast<GView*>((GView*) GetWindowLong(hWnd, GWL_USERDATA));
}

//////////////////////////////////////////////////////////////////////////////
GView *GView::Capturing = 0;
GView *GView::Over = 0;
HWND GView::hPrevCapture = 0;

GView::GView()
{
	hWnd = 0;
	CtrlId = -1;

	WndFlags = 0;
	WndStyle = WS_CHILD | WS_VISIBLE;
	WndExStyle = 0;
	WndClass = 0;
	WndDlgCode = 0;
	_Win32_TimerId = 0;

	Pos.ZOff(0, 0);
	Parent = 0;
	Notify = 0;
	_MinimumSize.x = _MinimumSize.y = 0;
}

GView::~GView()
{
	// Delete myself out of my parent's list
	if (Parent)
	{
		Parent->Children.Delete(this);
	}

	// Delete all children
	GView *c;
	while (c = Children.First())
	{
		ASSERT(c->Parent == this);
		DeleteObj(c);
	}

	// Delete the OS representation of myself
	if (hWnd AND IsWindow(hWnd))
	{
		WndFlags |= GWF_DESTRUCTOR;
		DestroyWindow(hWnd);
	}
	
	// NULL my handles and flags
	hWnd = 0;
	WndFlags = 0;

	// Remove static references to myself
	if (Over == this) Over = 0;
	if (Capturing == this) Capturing = 0;
	if (LgiApp->AppWnd == this)
	{
		LgiApp->AppWnd = 0;
	}
}

void GView::OnPosChange()
{
}

void GView::_Paint(GdcBasePrimitives *pDC)
{
	// Paint me
	OnPaint(pDC);

	// Paint any virtual children
	int x, y;
	pDC->GetOrigin(x, y);
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		if (NOT w->Handle())
		{
			GdcRegion r = w->GetPos();
			GdcRegion c = r;

			c.Offset(-x, -y);
			c.Intersection(&pDC->ClipRgn());
			pDC->ClipRgn(&c);
			pDC->SetOrigin(x-r.x1, y-r.y1);

			w->_Paint(pDC);

			pDC->SetOrigin(x, y);
			pDC->ClipRgn(NULL);
		}
	}
}

#define ADJ_LEFT			1
#define ADJ_RIGHT			2
#define ADJ_UP				3
#define ADJ_DOWN			4

int IsAdjacent(GdcRegion &a, GdcRegion &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

GdcRegion JoinAdjacent(GdcRegion &a, GdcRegion &b, int Adj)
{
	GdcRegion t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

GdcRegion *GView::FindLargest(GdcComplexRegion &r)
{
	int Pixels = 0;
	GdcRegion *Best = 0;
	static GdcRegion Final;

	// do initial run through the list to find largest single region
	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GdcRegion LastRgn = Final;
		int ThisPixels = Pixels;
		GdcRegion ThisRgn = Final;

		GdcComplexRegion TempList;
		for (GdcRegion *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GdcRegion *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GdcRegion t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GdcRegion *GView::FindSmallestFit(GdcComplexRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GdcRegion *Best = 0;

	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		if (	(i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GdcRegion *GView::FindLargestEdge(GdcComplexRegion &r, int Edge)
{
	GdcRegion *Best = 0;

	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

void GView::Quit(bool DontDelete)
{
	if (hWnd)
	{
		if (NOT DontDelete)
		{
			WndFlags |= GWF_QUIT_WND;
		}

		DestroyWindow(hWnd);
	}
}

uint32 GView::Style()
{
	return WndStyle;
}

uint32 GView::ExStyle()
{
	return WndExStyle;
}

char *GView::Class()
{
	return WndClass;
}

GView *GView::FindReal(GdcPt2 *Offset)
{
	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
	}

	GView *p = Parent;
	while (p AND NOT p->Handle())
	{
		if (Offset)
		{
			Offset->x += Pos.x1;
			Offset->y += Pos.y1;
		}

		p = p->Parent;
	}
	
	if (p AND p->Handle())
	{
		return p;
	}

	return NULL;
}

bool GView::Capture(bool c)
{
	return HandleCapture(this, c);
}

bool GView::IsCapturing()
{
	return Capturing != 0;
}

bool GView::HandleCapture(GView *Wnd, bool c)
{
	bool Status = FALSE;

	if (hWnd)
	{
		if (c)
		{
			if (Capturing)
			{
				Capturing->HandleCapture(NULL, FALSE);
			}

			hPrevCapture = SetCapture(hWnd);
			Status = TRUE;
			Capturing = Wnd;
		}
		else
		{
			if (Capturing)
			{
				ReleaseCapture();
				Capturing = 0;
			}
		}
	}
	else
	{
		if (Parent)
		{
			Status = Parent->HandleCapture(Wnd, c);
		}
	}

	return Status;
}

bool GView::Enabled()
{
	if (hWnd)
	{
		return IsWindowEnabled(hWnd);
	}
	return NOT TestFlag(WndStyle, WS_DISABLED);
}

void GView::Enabled(bool i)
{
	if (NOT i) SetFlag(WndStyle, WS_DISABLED);
	else ClearFlag(WndStyle, WS_DISABLED);

	if (hWnd)
	{
		EnableWindow(hWnd, i);
	}
	else
	{
		Invalidate();
	}
}

bool GView::Visible()
{
	if (hWnd)
	{
		return IsWindowVisible(hWnd);
	}

	return TestFlag(WndStyle, WS_VISIBLE);
}

void GView::Visible(bool i)
{
	if (i) SetFlag(WndStyle, WS_VISIBLE);
	else ClearFlag(WndStyle, WS_VISIBLE);

	if (hWnd)
	{
		ShowWindow(hWnd, (i) ? SW_SHOWNORMAL : SW_HIDE);
	}
	Invalidate();
}

bool GView::Focus()
{
	if (hWnd)
	{
		return GetFocus() == hWnd;
	}
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GView::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);

	if (hWnd)
	{
		SetFocus((i) ? hWnd : GetDesktopWindow());
	}
	else
	{
		Invalidate();
	}
}

bool GView::Sunken()
{
	return TestFlag(WndExStyle, WS_EX_CLIENTEDGE);
}

void GView::Sunken(bool i)
{
	ClearFlag(WndExStyle, WS_EX_WINDOWEDGE);
	
	if (i) SetFlag(WndExStyle, WS_EX_CLIENTEDGE);
	else ClearFlag(WndExStyle, WS_EX_CLIENTEDGE);
}

bool GView::Flat()
{
	return	NOT TestFlag(WndExStyle, WS_EX_CLIENTEDGE) AND
			NOT TestFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

void GView::Flat(bool i)
{
	ClearFlag(WndExStyle, WS_EX_CLIENTEDGE);
	ClearFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

bool GView::Raised()
{
	return TestFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

void GView::Raised(bool i)
{
	ClearFlag(WndExStyle, WS_EX_CLIENTEDGE);
	
	if (i) SetFlag(WndExStyle, WS_EX_WINDOWEDGE);
	else ClearFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

int GView::GetCtrlValue(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Value() : -1;
}

void GView::SetCtrlValue(int Id, int i)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

void GView::SetId(int i)
{
	CtrlId = i;
	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_ID, CtrlId);
	}
}

bool GView::Create(char *Class, GWndRef Parent)
{
	CreateWindowEx(	ExStyle(),
					Class,
					Name(),
					Style(),
					Pos.x1, Pos.y1,
					Pos.X(), Pos.Y(),
					Parent,
					NULL,
					hAppInst(),
					this);

	return hWnd != NULL;
}

bool GView::IsAttached()
{
	return hWnd AND IsWindow(hWnd);
}

bool GView::Attach(GView *p)
{
	bool Status = false;
	if (p AND NOT p->Children.HasItem(this))
	{
		Parent = p;

		if (Class())
		{
			// Real window with HWND
			if (Create(Class(), p->Handle()))
			{
				Status = (hWnd != 0);
			}
		}
		else
		{
			// Virtual window (no HWND)
			Status = true;
		}

		if (Status)
		{
			Parent->Children.Insert(this);
			Parent->OnChildrenChanged(this, true);
		}
		else
		{
			Parent = 0;
		}
	}
	return Status;
}

bool GView::AttachChildren()
{
	int Childs = Children.GetItems();

	List<GView> Temp;
	for (GView *c=Children.First(); c; c=Children.Next())
	{
		Temp.Insert(c);
	}
	Children.Empty();

	for (c=Temp.First(); c; c=Temp.Next())
	{
		LgiAssert(c->Handle() == 0);

		if (c->Attach(this))
		{
			Childs--;
		}
	}

	return Childs == 0;
}

bool GView::Detach()
{
	bool Status = false;
	if (Parent)
	{
		Visible(false);
		Parent->Children.Delete(this);
		Parent->OnChildrenChanged(this, false);
		Parent = 0;
		Status = true;

		if (hWnd)
		{
			WndFlags &= ~GWF_QUIT_WND;
			DestroyWindow(hWnd);
		}
	}
	return Status;
}

GdcRegion &GView::GetClient()
{
	RECT rc;

	if (hWnd)
	{
		GetClientRect(hWnd, &rc);
	}
	else
	{
		rc = Pos;
	}

	static GdcRegion Client;
	Client = rc;
	return Client;
}

GFont *GView::GetFont()
{
	/*
	GFontRef Fnt = 0; 
	if (hWnd)
	{
		Fnt = (GFontRef) SendMessage(hWnd, WM_GETFONT, 0, 0);
	}

	return Fnt;
	*/

	return SysFont;
}

void GView::SetFont(GFont *Font)
{
	if (hWnd)
	{
		SendMessage(hWnd, WM_SETFONT, (WPARAM) Font->Handle(), 0);
	}
}

bool GView::IsOver(VMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

bool GView::GetMouse(VMouse &m, bool ScreenCoords)
{
	// position
	POINT p;
	GetCursorPos(&p);
	if (NOT ScreenCoords)
	{
		ScreenToClient(hWnd, &p);
	}
	m.x = p.x;
	m.y = p.y;

	// buttons
	m.Flags =	((GetAsyncKeyState(VK_LBUTTON)&0x8000) ? MK_LEFT : 0) |
				((GetAsyncKeyState(VK_MBUTTON)&0x8000) ? MK_MIDDLE : 0) |
				((GetAsyncKeyState(VK_RBUTTON)&0x8000) ? MK_RIGHT : 0) |
				((GetAsyncKeyState(VK_CONTROL)&0x8000) ? MK_CTRL : 0) |
				((GetAsyncKeyState(VK_MENU)&0x8000) ? MK_ALT : 0) |
				((GetAsyncKeyState(VK_SHIFT)&0x8000) ? MK_SHIFT : 0);

	return true;
}

bool GView::WindowVirtualOffset(GdcPt2 *Offset)
{
	bool Status = false;

	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
		GView *Wnd = this;
		while (Wnd)
		{
			GdcRegion r = Wnd->GetPos();
			GView *Parent = Wnd->GetParent();
			if (NOT Wnd->Handle())
			{
				Offset->x += r.x1;
				Offset->y += r.y1;
				Wnd = Parent;
				Status = true;
			}
			else
			{
				Wnd = 0;
			}
		}
	}

	return Status;
}

GView *GView::WindowFromPoint(int x, int y)
{
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		GdcRegion r = c->GetPos();
		if (r.Overlap(x, y) AND
			c->Visible())
		{
			GView *Child = c->WindowFromPoint(x - r.x1, y - r.y1);
			if (Child)
			{
				return Child;
			}
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

bool GView::SetPos(GdcRegion &p, bool Repaint)
{
	bool Status = TRUE;
	Pos = p;
	if (hWnd)
	{
		Status = SetWindowPos(	hWnd,
								NULL,
								Pos.x1,
								Pos.y1,
								Pos.X(),
								Pos.Y(),
								((Repaint) ? 0 : SWP_NOREDRAW) |
								SWP_NOZORDER);
	}
	else if (Repaint)
	{
		Invalidate();
	}
	return Status;
}

bool GView::InThread()
{
	DWORD CurThread = GetCurrentThreadId();
	DWORD WndThread = GetWindowThreadProcessId(hWnd, NULL);
	
	return CurThread == WndThread;
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	return PostMessage(hWnd, Cmd, a, b);
}

bool GView::Invalidate(GdcRegion *r, bool Repaint)
{
	if (hWnd)
	{
		return InvalidateRect(hWnd, (r) ? &((RECT)*r) : NULL, Repaint);

		/*
		else
		{
			int a = (r) ? (r->x1 << 16) | (r->y1 & 0xffff) : 0xffffffff;
			int b = (r) ? (r->x2 << 16) | (r->y2 & 0xffff) : 0xffffffff;
			PostEvent(WM_INVALIDATE, a, b);
			if (r)
			{
				LgiTrace("sending WM_INVALIDATE from thread %X, hWnd %i, region(%i, %i, %i, %i)", LgiGetCurrentThread(), hWnd, r->x1, r->y1, r->x2, r->y2);
			}
			else
			{
				LgiTrace("sending WM_INVALIDATE from thread %X, hWnd %i, region(0)", LgiGetCurrentThread(), hWnd);
			}
		}
		*/
	}
	else
	{
		GdcRegion Up;
		GView *p = this;

		if (r)
		{
			Up = *r;
		}
		else
		{
			Up.Set(0, 0, Pos.X()-1, Pos.Y()-1);
		}

		while (p AND NOT p->Handle())
		{
			Up.Offset(p->Pos.x1, p->Pos.y1);
			p = p->Parent;
		}

		if (p AND p->Handle())
		{
			return p->Invalidate(&Up);
		}
	}
	return FALSE;
}

bool GView::InvalidateFrame(GdcRegion *r, bool Repaint)
{
	if (hWnd)
	{
		RedrawWindow(	hWnd,
						NULL,
						NULL,
						RDW_FRAME |
						RDW_INVALIDATE |
						RDW_ALLCHILDREN |
						((Repaint) ? RDW_UPDATENOW : 0));

		return TRUE;
	}

	return FALSE;
}

void
CALLBACK
GView::TimerProc(HWND hwnd, UINT uMsg, UINT idEvent, uint32 dwTime)
{
	GView *View = (GView*) idEvent;
	if (View)
	{
		View->OnPulse();
	}
}

void GView::SetPulse(int Length)
{
	if (hWnd)
	{
		if (Length > 0)
		{
			_Win32_TimerId = SetTimer(hWnd, (UINT) this, Length, (TIMERPROC) TimerProc);
		}
		else
		{
			KillTimer(hWnd, _Win32_TimerId);
			_Win32_TimerId = 0;
		}
	}
}

int GView::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl AND Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

GButton *FindDefault(GView *w)
{
	for (GView *c = w->Children.First(); c; c = w->Children.Next())
	{
		GButton *But = dynamic_cast<GButton*>(c);
		if (But AND But->Default())
		{
			return But;
		}
		
		But = FindDefault(c);
		if (But)
		{
			return But;
		}
	}

	return NULL;
}

List<GView> *GetChildrenList(GView *w)
{
	if (w)
	{
		GTabPage *Page = dynamic_cast<GTabPage*>(w);
		GTabControl *TabCtrl = dynamic_cast<GTabControl*>(Page ? Page->GetParent() : w);
		if (TabCtrl)
		{
			GTabPage *CurTab = TabCtrl->TabAt(TabCtrl->Value());
			if (CurTab)
			{
				return &CurTab->Children;
			}
		}
		else
		{
			return &w->Children;
		}
	}

	return 0;
}

GView *GetNextTabStop(GView *Wnd, bool Back)
{
	GView *w = Wnd;
	do
	{
		// move
		if (w AND w->GetParent())
		{
			List<GView> *CList = GetChildrenList(w);
			List<GView> *PList = GetChildrenList(w->GetParent());
			if (CList AND PList)
			{
				int Me = PList->IndexOf(w);
				if (Back)
				{
					GView *Prev = PList->ItemAt(Me-1);
					while (Prev)
					{
						GView *w = Prev->Children.ItemAt(Prev->Children.GetItems()-1);
						if (w)
						{
							Prev = w;
						}
						else break;
					}

					w = Prev;
				}
				else
				{
					GView *Next = CList->First();
					if (Next)
					{
						// Make child the next focus
						w = Next;
					}
					else
					{
						// No children
						Next = PList->ItemAt(Me+1);
						if (Next)
						{
							// Make next sibling the focus...
							w = Next;
						}
						else
						{
							// No sibling
							while (w)
							{
								GView *Parent = w->GetParent();
								if (dynamic_cast<GWindow*>(Parent) OR
									dynamic_cast<GDialog*>(Parent))
								{
									int Index = Parent->Children.IndexOf(w);
									Next = Parent->Children.ItemAt(Index+1);
									if (NOT Next)
									{
										Next = Parent->Children.ItemAt(0);
									}
									if (Next)
									{
										w = Next;
									}
									break;
								}
								else if (Parent)
								{
									List<GView> *Lst = GetChildrenList(Parent);
									int Index = (Lst) ? Lst->IndexOf(w) : -1;
									Next = (Lst AND Index >= 0) ? Lst->ItemAt(Index+1) : 0;
									if (Next AND Lst)
									{
										w = Next;
										break;
									}
									else
									{
										w = Parent;
									}
								}
								else break;
							}
						}
					}
				}
			}
			else
			{
				w = 0;
			}
		}
		else
		{
			break;
		}

		// test
		if (w AND
			w->Enabled() AND
			TestFlag(w->Style(), WS_TABSTOP))
		{
			return w;
		}
	}
	while (w AND w != Wnd);

	return 0;
}

bool SysOnKey(GView *w, GMessage *m)
{
	if (m->a == VK_TAB AND
		(m->Msg == WM_KEYDOWN OR
		m->Msg == WM_SYSKEYDOWN) )
	{
		if (NOT TestFlag(w->WndDlgCode, DLGC_WANTTAB) AND
			NOT TestFlag(w->WndDlgCode, DLGC_WANTALLKEYS))
		{
			// push the focus to the next control
			bool Shifted = (GetAsyncKeyState(VK_SHIFT) & 0x8000) != 0;
			GView *Wnd = GetNextTabStop(w, Shifted);
			if (Wnd)
			{
				::SetFocus(Wnd->Handle());
				return true;
			}
		}
	}

	return false;
}


#define KEY_FLAGS		(~(MK_LBUTTON | MK_MBUTTON | MK_RBUTTON))

int GView::OnEvent(GMessage *Msg)
{
	int Status = 0;

	if (Message(Msg) == MouseRollMsg)
	{
		HWND hFocus = GetFocus();
		if (hWnd)
		{
			int Flags = ((GetKeyState(VK_SHIFT)&0xF000) ? VK_SHIFT : 0) | 
						((GetKeyState(VK_CONTROL)&0xF000) ? VK_CONTROL : 0);
			
			PostMessage(hFocus, WM_MOUSEWHEEL, MAKELONG(Flags, (short)Msg->a), Msg->b);
		}
		return 0;
	}

	if (hWnd)
	{
		switch (Msg->Msg)
		{
			/*
			case WM_INVALIDATE:
			{
				LgiTrace("receiving WM_INVALIDATE in thread %X, hWnd %i", LgiGetCurrentThread(), hWnd);
				if (Msg->a != 0xffffffff)
				{
					GdcRegion r;
					r.x1 = Msg->a >> 16;
					r.y1 = Msg->a & 0xffff;
					r.x2 = Msg->b >> 16;
					r.y2 = Msg->b & 0xffff;

					RECT rc = r;
					InvalidateRect(hWnd, &rc, true);
				}
				else
				{
					InvalidateRect(hWnd, 0, true);
				}
				break;
			}
			*/
			case WM_MENUCHAR:
			case WM_MEASUREITEM:
			case WM_DRAWITEM:
			{
				return GMenu::_OnEvent(Msg);
				break;
			}
			case WM_ENABLE:
			{
				Invalidate();
				break;
			}
			case WM_HSCROLL:
			case WM_VSCROLL:
			{
				GView *Wnd = dynamic_cast<GView*>(FindControl((HWND) Msg->b));
				if (Wnd)
				{
					Wnd->OnEvent(Msg);
				}
				break;
			}
			case WM_GETDLGCODE:
			{
				// we handle all tab control stuff
				return WndDlgCode | DLGC_WANTTAB;
			}
			case WM_MOUSEWHEEL:
			{
				short fwKeys = LOWORD(Msg->a);			// key flags
				short zDelta = (short) HIWORD(Msg->a);	// wheel rotation
				short xPos = (short) LOWORD(Msg->b);	// horizontal position of pointer
				short yPos = (short) HIWORD(Msg->b);	// vertical position of pointer
				int nScrollLines = - _lgi_mouse_wheel_lines();
				OnMouseWheel( ((int) zDelta / WHEEL_DELTA) * (int) nScrollLines );
				return 0;
			}
			case WM_CHANGE:
			{
				GView *Ctrl = dynamic_cast<GView*>((GView*) Msg->a);
				if (Ctrl)
				{
					OnNotify(Ctrl, Msg->b);
				}
				break;
			}
			case WM_NOTIFY:
			{
				LPNMHDR Hdr = (LPNMHDR) Msg->b;
				GView *Ctrl = FindControl(Hdr->hwndFrom);

				int i1 = LVN_GETDISPINFO;
				if (Ctrl) 
				{
					switch (Hdr->code)
					{
						#ifdef __GLIST_H
						// List View Stuff
						//
						// this is only required when using the Win32 ListView control
						// in GList.h/GList.cpp
						// for xp compatibility use the owner draw list control in
						// GList2.h/GList2.cpp
						case LVN_GETDISPINFOW:
						case LVN_GETDISPINFOA:
						{
							bool Unicode = Hdr->code == LVN_GETDISPINFOW;
							LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
							GListItem *Item = (GListItem*) Info->item.lParam;

							if (Item)
							{
								if (Info->item.mask & LVIF_TEXT)
								{
									if (Info->item.pszText)
									{
										char *Text = Item->GetText(Info->item.iSubItem);
										if (Text)
										{
											if (Unicode)
											{
												short *d = (short*) Info->item.pszText;
												char *s = Text;
												int Size = Info->item.cchTextMax - 2;
												while (*s AND Size > 1)
												{
													*d++ = *s++;
													Size -= 2;
												}
												*d++ = 0;
											}
											else
											{
												strcpy(Info->item.pszText, Text);
											}
										}
										else
										{
											Info->item.pszText[0] = 0;
										}
									}
									else
									{
										if (Unicode)
										{
											static short Error[] = { 'E', 'r', 'r', 'o', 'r', 0 };
											Info->item.pszText = (char*) Error;
										}
										else
										{
											// what the?
											// don't know what causes windows to call using a NULL
											// pointer. Any ideas ppl?
											Info->item.pszText = Item->GetText(Info->item.iSubItem);
										}
									}
								}

								if (Info->item.mask & LVIF_IMAGE)
								{
									Info->item.iImage = Item->GetImage();
								}
							}
							break;
						}
						case LVN_ENDLABELEDIT:
						{
							LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
							GListItem *Item = (GListItem*) Info->item.lParam;
							if (Info AND Item)
							{
								if (Info->item.mask & LVIF_TEXT)
								{
									return Item->SetText(Info->item.pszText);
								}
							}
							break;
						}
						case LVN_BEGINDRAG:
						case LVN_BEGINRDRAG:
						{
							NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
							LV_ITEM Item;

							ZeroObj(Item);
							Item.mask = LVIF_PARAM;
							Item.iItem = Info->iItem;
							if (ListView_GetItem(Ctrl->Handle(), &Item))
							{
								VMouse m;
								m.Flags = (Hdr->code == LVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON;
								((GList*) Ctrl)->OnItemBeginDrag((GListItem*) Item.lParam, m);
							}
							break;
						}
						case LVN_COLUMNCLICK:
						{
							NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
							((GList*) Ctrl)->OnColumnClick(Info->iSubItem);
							break;
						}
						case LVN_ITEMCHANGED:
						{
							LPNMLISTVIEW Info = (LPNMLISTVIEW) Msg->b;
							if ((Info->uNewState | Info->uOldState) & LVIS_SELECTED)
							{
								GList *Lst = (GList*) Ctrl;
								if (Info->uNewState & LVIS_SELECTED)
								{
									Lst->OnItemSelect((GListItem*) Info->lParam);
								}
								else
								{
									Lst->OnItemSelect(0);
								}
							}
							break;
						}
						#endif

						// Tab Control Stuff
						case TCN_SELCHANGING:
						{
							Status = Ctrl->SysOnNotify(Hdr->code);
							break;
						}
						case TCN_SELCHANGE:
						{
							Ctrl->SysOnNotify(Hdr->code);
							OnNotify(Ctrl, 0);
							break;
						}

						#ifdef __GTREE_H
						// Tree View Control
						//
						// this is only relevent when using the Win32 Tree control
						// which resides in GTree.h and GTree.cpp
						// for xp compatibility use the owner draw tree control
						// in GTree2.h and GTree2.cpp
	 					case TVN_BEGINRDRAG:
						case TVN_BEGINDRAG:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemBeginDrag((GTreeItem*) Item->itemNew.lParam, (Hdr->code == TVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON);
							break;
						}
						case TVN_SELCHANGED:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemSelect((GTreeItem*) Item->itemNew.lParam);
							break;
						}
						case TVN_ITEMEXPANDED:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemExpand((GTreeItem*) Item->itemNew.lParam,
															(Item->itemNew.state & TVIS_EXPANDED) != 0);
							break;
						}
						case TVN_GETDISPINFO:
						{
							TV_DISPINFO *Info = (TV_DISPINFO*) Msg->b;
							GTreeItem *Item = (GTreeItem*) Info->item.lParam;
							if (Item)
							{
								if (Info->item.mask & TVIF_TEXT)
								{
									char *Text = Item->GetText();
									if (NOT Text) Text = "";
									if (Info->item.pszText)
									{
										strcpy(Info->item.pszText, Text);
									}
									else
									{
										Info->item.pszText = Text;
									}								
								}

								if (Info->item.mask & TVIF_IMAGE)
								{
									Info->item.iImage = Item->GetImage();
								}
								
								if (Info->item.mask & TVIF_SELECTEDIMAGE)
								{
									Info->item.iSelectedImage = Item->GetImage(1);
								}							
							}
							break;
						}
						case TVN_DELETEITEM:
						{
							NM_TREEVIEW *pnmtv = (NM_TREEVIEW*) Msg->b;
							if (pnmtv)
							{
								GTreeItem *Item = (GTreeItem*) pnmtv->itemOld.lParam;
								if (Item)
								{
									Item->hTreeItem = 0;
								}
							}
							break;
						}
						#endif
					}
				}
				break;
			}
			case WM_COMMAND:
			{
				GView *Ctrl = FindControl((HWND) Msg->b);
				if (Ctrl)
				{
					short Code = HIWORD(Msg->a);
					switch (Code)
					{
						case BN_CLICKED: // BUTTON
						{
							OnNotify(Ctrl, 0);
							break;
						}
						case EN_CHANGE: // EDIT
						{
							int n = Ctrl->SysOnNotify(Code);
							OnNotify(Ctrl, n);
							break;
						}
						case CBN_CLOSEUP:
						{
							PostMessage(hWnd, WM_COMMAND, MAKELONG(Ctrl->GetId(), CBN_EDITCHANGE), Msg->b);
							break;
						}
						case CBN_EDITCHANGE: // COMBO
						{
							Ctrl->SysOnNotify(Code);
							OnNotify(Ctrl, 0);
							break;
						}
					}
				}
				break;
			}
			case WM_NCDESTROY:
			{
				SetWindowLong(hWnd, GWL_USERDATA, 0);
				hWnd = 0;
				if (WndFlags & GWF_QUIT_WND)
				{
					delete this;
				}
				break;
			}
	 		case WM_CLOSE:
			{
				if (OnRequestClose())
				{
					Quit();
				}
				break;
			}
			case WM_DESTROY:
			{
				OnDestroy();
				break;
			}
			case WM_CREATE:
			{
				SetId(CtrlId);
				OnCreate();
				break;
			}
			case WM_SETFOCUS:
			{
				OnFocus(true);
				break;
			}
			case WM_KILLFOCUS:
			{
				OnFocus(false);
				break;
			}
			case WM_WINDOWPOSCHANGED:
			{
				WINDOWPOS *Info = (LPWINDOWPOS) Msg->b;
				if (Info)
				{
					Pos.ZOff(Info->cx-1, Info->cy-1);
					Pos.Offset(Info->x, Info->y);
				}
				OnPosChange();

				if (NOT (WndFlags & GWF_DIALOG))
				{
					return DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
			case WM_CAPTURECHANGED:
			{
				Capturing = 0;
				break;
			}
			case WM_MOUSEENTER:
			{
				if (Msg->a AND NOT TestFlag(WndFlags, GWF_OVER))
				{
					VMouse Ms;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;

					GView *MouseOver = WindowFromPoint(Ms.x, Ms.y);
					if (Capturing)
					{
						if (MouseOver == Capturing)
						{
							Capturing->OnMouseEnter(_lgi_adjust_click_for_window(Ms, Capturing));
						}
					}
					else
					{
						if (Over != MouseOver)
						{
							if (Over)
							{
								Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, Over));
							}

							Over = MouseOver;

							if (Over)
							{
								Over->OnMouseEnter(_lgi_adjust_click_for_window(Ms, Over));
							}
						}
					}

					SetFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEEXIT:
			{
				if (NOT Msg->a AND TestFlag(WndFlags, GWF_OVER))
				{
					VMouse Ms;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;

					if (Capturing)
					{
						Capturing->OnMouseExit(_lgi_adjust_click_for_window(Ms, Capturing));
					}
					else
					{
						if (Over)
						{
							Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, Over));
							Over = 0;
						}
						else
						{
							OnMouseExit(Ms);
						}
					}

					ClearFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEMOVE:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = Msg->a;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down((Msg->a & (MK_LBUTTON|MK_MBUTTON|MK_RBUTTON)) != 0);

				int Hit = OnHitTest(Ms.x, Ms.y);
				if (Hit < 0)
				{
					SetCursor(LgiApp->hNormalCursor);
				}

				GView *MouseOver = WindowFromPoint(Ms.x, Ms.y);
				if (Over != MouseOver)
				{
					if (Over)
					{
						Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, Over));
					}

					Over = MouseOver;

					if (Over)
					{
						Over->OnMouseEnter(_lgi_adjust_click_for_window(Ms, Over));
					}
				}

				if (Capturing)
				{
					Capturing->OnMouseMove(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseMove(_lgi_adjust_click_for_window(Ms, Over));
				}
				break;
			}
			/*
			case WM_NCHITTEST:
			{
				POINT Pt = { LOWORD(Msg->b), HIWORD(Msg->b) };
				ScreenToClient(hWnd, &Pt);
				int Hit = OnHitTest(Pt.x, Pt.y);
				if (Hit >= 0)
				{
					return Hit;
				}
				if (NOT (WndFlags & GWF_DIALOG))
				{
					return DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
			*/
			case WM_LBUTTONDBLCLK:
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = (Msg->a & KEY_FLAGS) | MK_LEFT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down(Msg->Msg != WM_LBUTTONUP);
				Ms.Double(Msg->Msg == WM_LBUTTONDBLCLK);

				if (Capturing)
				{
					Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_RBUTTONDBLCLK:
			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = (Msg->a & KEY_FLAGS) | MK_RIGHT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down(Msg->Msg != WM_RBUTTONUP);
				Ms.Double(Msg->Msg == WM_RBUTTONDBLCLK);

				if (Capturing)
				{
					Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_MBUTTONDBLCLK:
			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = (Msg->a & KEY_FLAGS) | MK_MIDDLE;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
 				Ms.Down(Msg->Msg != WM_MBUTTONUP);
				Ms.Double(Msg->Msg == WM_MBUTTONDBLCLK);

				if (Capturing)
				{
					Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_SYSKEYUP:
			case WM_SYSKEYDOWN:
			case WM_KEYDOWN:
			case WM_KEYUP:
			{
				bool IsDialog = TestFlag(WndFlags, GWF_DIALOG);
				HWND hwnd = hWnd;

				if (NOT SysOnKey(this, Msg))
				{
					VKey Key;
					Key.c = Msg->a;		// virtual-key code 
					Key.Flags = 0;
					SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
					SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
					SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
					Key.Data = Msg->b;
					Key.Down =	Msg->Msg == WM_KEYDOWN OR
								Msg->Msg == WM_SYSKEYDOWN;
					Key.IsChar = FALSE;

					OnKey(Key);
				}

				if (NOT IsDialog)
				{
					// required for Alt-Key function (eg Alt-F4 closes window)
					return DefWindowProc(hwnd, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
			case WM_CHAR:
			{
				VKey Key;
				Key.c = Msg->a;		// virtual-key code 
				Key.Flags = 0;
				SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
				SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
				SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
				Key.Data = Msg->b;
				Key.Down = TRUE;
				Key.IsChar = TRUE;

				OnKey(Key);
				break;
			}
			case WM_PAINT:
			{
				GdcGdiScreen DeviceContext(hWnd);
				_Paint(&DeviceContext);
				break;
			}
			default:
			{
				if (NOT (WndFlags & GWF_DIALOG))
				{
					return DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
		}
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);
	if (hWnd) SetWindowText(hWnd, GObject::Name());
	return TRUE;
}

char *GView::Name()
{
	if (hWnd)
	{
		int Length = GetWindowTextLength(hWnd) + 1;
		char *Buf = NEW(char[Length]);
		if (Buf AND GetWindowText(hWnd, Buf, Length) > 0)
		{
			if (NOT GObject::Name() OR
				strcmp(GObject::Name(), Buf) != 0)
			{
				GObject::Name(Buf);
			}
		}
		DeleteArray(Buf);
	}
	return GObject::Name();
}

GView *GView::FindControl(GWndRef hCtrl)
{
	if (hWnd == hCtrl)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GView *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

void GView::MoveOnScreen()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);

	if (Pos.x2 >= Screen.X())
		Pos.Offset(Screen.X() - Pos.x2, 0);
	if (Pos.y2 >= Screen.Y())
		Pos.Offset(0, Screen.Y() - Pos.y2);
	if (Pos.x1 < 0)
		Pos.Offset(-Pos.x1, 0);
	if (Pos.y1 < 0)
		Pos.Offset(0, -Pos.y1);
}

void GView::MoveToCenter()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	Pos.Offset(-Pos.x1, -Pos.y1);
	Pos.Offset((Screen.X() - Pos.X()) / 2, (Screen.Y() - Pos.Y()) / 2);

	if (Handle())
	{
		SetPos(Pos, true);
	}
}

void GView::MoveToMouse()
{
	POINT Pt;
	if (GetCursorPos(&Pt))
	{
		Pos.Offset(-Pos.x1, -Pos.y1);
		Pos.Offset(Pt.x-(Pos.X()/2), Pt.y-(Pos.Y()/2));
		MoveOnScreen();
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
GWindow::GWindow()
{
	Menu = 0;
	WndStyle |= WS_TILEDWINDOW | WS_CLIPCHILDREN; // | WS_CLIPSIBLINGS;
	WndStyle &= ~WS_CHILD;
	WndExStyle |= WS_EX_CONTROLPARENT | WS_EX_ACCEPTFILES;
	LastFocus = 0;
	_QuitOnClose = true;
}

GWindow::~GWindow()
{
	if (Menu)
	{
		Menu->Detach();
		DeleteObj(Menu);
	}
}

void GWindow::OnPaint(GdcBasePrimitives *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

bool GWindow::Visible()
{
	return GView::Visible();
}

void GWindow::Visible(bool v)
{
	if (v) Pour();
	GView::Visible(v);
}

void GWindow::Pour()
{
	GdcComplexRegion Client(GetClient());
	GdcComplexRegion Update;

	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GdcRegion OldPos = v->Pos;
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(true);
			}

			if (OldPos != v->Pos)
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->Pos);
			Update.Subtract(&v->Pos);
		}
		else
		{
			// make the view not visible
			// v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.GetSize(); i++)
	{
		Invalidate(Update[i]);
	}
}

int GWindow::OnEvent(GMessage *Msg)
{
	int Status = 0;

	switch (Message(Msg))
	{
		case WM_TRANSLATE_ACCELERATOR:
		{
			if (Menu)
			{
				Menu->OnEvent(Msg);
			}
			break;
		}
		case WM_SYSCOLORCHANGE:
		{
			LgiInitColours();
			break;
		}
		case WM_WINDOWPOSCHANGING:
		{
			WINDOWPOS *Info = (LPWINDOWPOS) Msg->b;
			if (Info)
			{
				if (_MinimumSize.x AND
					_MinimumSize.x > Info->cx)
				{
					Info->cx = _MinimumSize.x;
				}

				if (_MinimumSize.y AND
					_MinimumSize.y > Info->cy)
				{
					Info->cy = _MinimumSize.y;
				}
			}
			break;
		}
		case WM_ACTIVATE:
		{
			// This is a hack to make Windows set the focus of a child
			// control when you Alt-Tab in and out of a top level window
			if (LOWORD(Msg->a) != WA_INACTIVE)
			{
				// gaining focus
				if (LastFocus)
				{
					SetFocus(LastFocus);
					LastFocus = 0;
				}
			}
			else
			{
				// losing focus
				LastFocus = 0;
				HWND f = GetFocus();
				for (HWND p = ::GetParent(f); p; p = ::GetParent(p))
				{
					if (p == Handle())
					{
						LastFocus = f;
					}
				}
			}
			break;
		}
		case WM_CREATE:
		{
			Pour();
			OnCreate();
			break;
		}
		case WM_QUERYENDSESSION:
	 	case WM_CLOSE:
		{
			bool QuitApp = false;
			if (OnRequestClose())
			{
				QuitApp = true;

				if (GetQuitOnClose())
				{
					LgiCloseApp();
				}
				else
				{
					Quit();
				}
			}

			if (Message(Msg) == WM_CLOSE)
			{
				return 0;
			}
			else
			{
				return QuitApp;
			}
			break;
		}
		case WM_DROPFILES:
		{
			HDROP hDrop = (HDROP) Msg->a;
			if (hDrop)
			{
				List<char> FileNames;
				int Count = DragQueryFile(hDrop, -1, NULL, 0);;
				char FileName[256];

				for (int i=0; i<Count; i++)
				{
					if (DragQueryFile(hDrop, i, FileName, sizeof(FileName)-1) > 0)
					{
						FileNames.Insert(NewStr(FileName));
					}
				}

				OnReceiveFiles(FileNames);
				DeleteObjList(&FileNames, char);
			}
			break;
		}
		case WM_COMMAND:
		{
			HWND OurWnd = hWnd; // copy onto the stack, because
								// we might lose the 'this' object in the
								// OnCommand handler which would delete
								// the memory containing the handle.

			Status = OnCommand(LOWORD(Msg->a), HIWORD(Msg->a), (HWND) Msg->b);
			if (NOT IsWindow(OurWnd))
			{
				// The window was deleted so break out now
				break;
			}
			// otherwise fall thru to the GView handler
		}
		default:
		{
			Status = GView::OnEvent(Msg);
			break;
		}
	}

	return Status;
}

void GWindow::OnPosChange()
{
	Pour();
}

/////////////////////////////////////////////////////////////////////////////
GApp *GApp::TheApp = 0;

GApp::GApp(HINSTANCE hInst, char *lpCmdLine, int nShow)
{
	// Vars...
	Config = 0;
	AppWnd = 0;
	CmdShow = nShow;
	hAppInst = hInst;
	CmdLine = NewStr(lpCmdLine);

	// GApp instance
	LgiAssert(TheApp == 0);
	TheApp = this;

	// System font
	SystemFont = 0;
	GdcFontType SysFontType;
	if (SysFontType.GetSystemFont("System"))
	{
		SystemFont = SysFontType.CreateFont();
		if (SystemFont)
		{
			SystemFont->Transparent(true);
		}
	}

	if (NOT SystemFont)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}

	// Other vars and init
	hNormalCursor = LoadCursor(NULL, IDC_ARROW);
	srand(((unsigned)time(NULL))*((uint)GetCurrentThread()));
	MouseRollMsg = RegisterWindowMessage("MSWHEEL_ROLLMSG");
	hMouseHook = SetWindowsHookEx(WH_MOUSE, (HOOKPROC) MouseHook, NULL, (DWORD) GetCurrentThreadId());
	LgiInitColours();
}

GApp::~GApp()
{
	DeleteArray(CmdLine);
	DeleteObj(AppWnd);	
	DeleteObj(SystemFont);
	if (hMouseHook)
	{
		UnhookWindowsHookEx(hMouseHook);
	}
	TheApp = 0;
}

XmlTag *GApp::GetConfig(char *Tag)
{
	if (NOT Config)
	{
		char Path[256];
		if (LgiGetExePath(Path))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, "lgi.conf");
			if (FileExists(Path))
			{
				Config = NEW(Xml);
				if (Config)
				{
					if (Config->ParseXmlFile(Path) <= 0)
					{
						DeleteObj(Config);
					}
				}
			}
		}
	}

	if (Tag AND Config)
	{
		for (XmlTag *t=Config->Values.First(); t; t=Config->Values.Next())
		{
			if (t->Name AND
				stricmp(t->Name, Tag) == 0)
			{
				return t;
			}
		}
	}

	return 0;
}

bool GApp::GetOption(char *Dest, char Option)
{
	char WhiteSpace[] = " \t\r\n";
	char *c = CmdLine;

	while (c AND *c)
	{
		if (*c == '/' OR *c == '-')
		{
			c++;
			if (*c == Option)
			{
				c++;
				if (*c)
				{
					// skip leading whitespace
					while (*c AND strchr(WhiteSpace, *c))
					{
						c++;
					}

					// write str out if they want it
					if (Dest)
					{
						char End = (*c == '\'') ? *c++ : ' ';

						while (*c AND *c != End)
						{
							*Dest++ = *c++;
						}

						*Dest++ = 0;
					}
				}

				// yeah we got the option
				return true;
			}
		}

		c = strchr(c, ' ');
		if (c) c++;
	}

	return false;
}

#include "ginput.h"

void GApp::OnCmdLine(char *Cmd)
{
	char WhiteSpace[] = " \r\n\t";

	DeleteArray(CmdLine);
	CmdLine = NewStr(Cmd);

	/*
	GInput dlg(0, CmdLine);
	dlg.DoModal();
	*/
	
	/*
	if (Cmd AND strlen(Cmd) > 0)
	{
		List<char> l;

		// search for args
		for (char *c = Cmd; *c; )
		{
			// skip white
			while (strchr(WhiteSpace, *c)) c++;
			if (*c)
			{
				char End = (*c == '\"') ? '\"' : ' ';
				if (End == '\"') c++;
				char *Start = c;
				while (*c AND *c != End) c++;
				char *Str = NewStr(Start, (int)c-(int)Start);
				if (Str) l.Insert(Str);
				if (End == '\"') c++;
			}
		}

		if (l.GetItems() > 0)
		{
			// read args into an array
			char **Arg = NEW(char*[l.GetItems()]);
			if (Arg)
			{
				int n=0;
				for (char *a = l.First(); a; a = l.Next())
				{
					Arg[n++] = a;
				}

				// call the client
				OnArgs(l.GetItems(), Arg);

				// clean up
				for (n=0; n<l.GetItems(); n++)
				{
					DeleteArray(Arg[n]);
				}
				DeleteArray(Arg);
			}
		}
	}
	*/

	if (CmdLine)
	{
		List<char> Files;
		
		/*
		char File[256] = "";

		for (int i=0; i<Args; i++)
		{
			if (strlen(File) == 0)
			{
				strcpy(File, Arg[i]);
				if (FileExists(File) OR
					DirExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
			else
			{
				strcat(File, " ");
				strcat(File, Arg[i]);
				if (FileExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
		}
		*/

		char *Delim = "\'\"";
		char *s;
		for (s = CmdLine; *s; )
		{
			// skip ws
			while (strchr(WhiteSpace, *s)) s++;

			// read to end of token
			char *e = s;
			if (strchr(Delim, *s))
			{
				for (e = ++s; *e AND NOT strchr(Delim, *e); e++);
			}
			else
			{
				for (; *e AND NOT strchr(WhiteSpace, *e); e++);
			}

			char *Arg = NewStr(s, (int)e-(int)s);
			if (Arg)
			{
				if (FileExists(Arg))
				{
					Files.Insert(Arg);
				}
				else
				{
					DeleteArray(Arg);
				}
			}

			// next
			s = (*e) ? e + 1 : e;
		}

		// call app
		if (Files.GetItems() > 0)
		{
			OnReceiveFiles(Files);
		}

		// clear up
		DeleteArrayList(&Files, char);
	}
}

void GApp::OnReceiveFiles(List<char> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

bool IsTabStop(HWND hWnd)
{
	int Style = GetWindowLong(hWnd, GWL_STYLE);
	return ((Style & WS_TABSTOP) != 0) AND ((Style & WS_VISIBLE) != 0);
}

HWND GetTabControl(HWND hWnd)
{
	// Hack for selecting TabControls first even though they
	// are last in the list
	HWND hTab = GetWindow(hWnd, GW_HWNDLAST);

	// Ok we're not the first
	// Get the class.
	char ClassName[256];
	if (GetClassName(hTab, ClassName, sizeof(ClassName)) > 0)
	{
		if (stricmp(ClassName, LGI_TABCONTROL) == 0)
		{
			return hTab;
		}
	}

	return NULL;
}


HWND GetTabstop(HWND hWnd, bool Forward = true)
{
	HWND h = GetWindow(hWnd, (Forward) ? GW_HWNDNEXT : GW_HWNDPREV);
	while (h)
	{
		if (h == hWnd)
		{
			return NULL;
		}

		if (IsTabStop(h))
		{
			// Is a tabstop...
			return h;
		}

		h = GetWindow(h, (Forward) ? GW_HWNDNEXT : GW_HWNDPREV);
	}

	return NULL;
}

/*
if (hTab == hWnd)
{
	if (Shifted)
	{
		hNext = 0;
	}
	else
	{
		hNext = GetWindow(hWnd, GW_CHILD);
		if (NOT IsTabStop(hNext))
		{
			hNext = GetTabstop(hNext, NOT Shifted);
		}
	}
}
else

BOOL ProcessDialogMessage(HWND hDlg, MSG *Msg)
{
	if (Msg->message == WM_KEYDOWN AND
		Msg->wParam == VK_TAB AND
		NOT (Msg->lParam & 0x80000000))
	{
		bool Shifted = (GetAsyncKeyState(VK_SHIFT) & 0x8000) != 0;
		HWND hWnd = GetFocus();
		if (hWnd)
		{
			int DlgCode = SendMessage(hWnd, WM_GETDLGCODE, 0, 0);
			if (NOT (DlgCode & (DLGC_WANTTAB | DLGC_WANTALLKEYS)))
			{
				// Doesn't want tab's so we process the tab
				// HWND hTab = GetTabControl(hWnd);
				HWND hNext = 0;

				// step down to child level
				hNext = GetWindow(hWnd, GW_CHILD);
				if (hNext)
				{
					if (Shifted)
					{
						hNext = GetWindow(hNext, GW_HWNDLAST);
					}

					if (NOT IsTabStop(hNext))
					{
						hNext = GetTabstop(hNext, NOT Shifted);
					}
				}

				if (NOT hNext)
				{
					// no children... so along
					hNext = GetTabstop(hWnd, NOT Shifted);
				}

				if (hNext)
				{
					SetFocus(hNext);
				}
				else
				{
					// end of list
					
					// go up till we find a "Next"
					bool ParentExit = false;
					HWND hParent = GetParent(hWnd);
					while (	hParent AND
							NOT (hNext = GetWindow(hParent, (Shifted) ? GW_HWNDPREV : GW_HWNDNEXT)) AND
							GetParent(hParent))
					{
						HWND p = GetParent(hParent);
						if (GetParent(p))
						{
							hParent = p;
						}
						else
						{
							ParentExit = true;
							break;
						}
					}

					if (ParentExit AND NOT hNext)
					{
						hNext = GetWindow(hParent, (Shifted) ? GW_HWNDLAST : GW_HWNDFIRST);
					}

					while (hNext)
					{
						// search through all the children for tabstops
						HWND hChild = GetWindow(hNext, GW_CHILD);
						if (hChild)
						{
							hChild = GetWindow(hChild, (Shifted) ? GW_HWNDLAST : GW_HWNDFIRST);
							HWND hTab = GetTabControl(hChild);
							if (hTab)
							{
								hChild = hTab;
							}
							else if (NOT IsTabStop(hChild))
							{
								hChild = GetTabstop(hChild, NOT Shifted);
							}
						}
						else
						{
							if (IsTabStop(hNext))
							{
								hChild = hNext;
							}
							else
							{
								hChild = GetTabstop(hNext, NOT Shifted);
							}
						}

						if (hChild)
						{
							SetFocus(hChild);
							return 0;
						}

						HWND n = GetWindow(hNext, (Shifted) ? GW_HWNDPREV : GW_HWNDNEXT);
						if (NOT n)
						{
							hNext = GetWindow(hNext, (Shifted) ? GW_HWNDLAST : GW_HWNDFIRST);
						}
						else
						{
							hNext = n;
						}
					}
				}
				return true;
			}
		}
	}

	return IsDialogMessage(hDlg, Msg);
}
*/

/*
BOOL ProcessDialogMessage(HWND hDlg, MSG *Msg)
{
	switch (Msg->message)
	{
		case WM_KEYDOWN:
		{
			if (Msg->wParam == VK_RETURN)
			{
				int i = SendMessage(Msg->hwnd, WM_GETDLGCODE, 0, 0);
				if (i & DLGC_WANTALLKEYS == 0)
				{
					int DefId = SendMessage(hDlg, DM_GETDEFID, 0, 0);
					Msg->hwnd = GetDlgItem(hDlg, DefId);

					TranslateMessage(Msg);
					DispatchMessage(Msg);
					return true;
				}
			}
			break;
		}
	}

	return IsDialogMessage(hDlg, Msg);
}
*/

HWND GApp::ActiveDialog = NULL;
bool GApp::Run(bool Loop)
{
	MSG Msg;
	bool status = true;

	if (Loop)
	{
		while (GetMessage(&Msg, NULL, 0, 0) > 0)
		{
			/*
			if (ActiveDialog)
			{
				// if (ProcessDialogMessage(ActiveDialog, &Msg))
				if (IsDialogMessage(ActiveDialog, &Msg))
				{
					continue;
				}
			}
			*/

			if (Msg.message == WM_KEYDOWN)
			{
				HWND hTop = Msg.hwnd, hNext;
				while (hNext = GetParent(hTop))
				{
					hTop = hNext;
				}

				if (hTop)
				{
					SendMessage(hTop, WM_TRANSLATE_ACCELERATOR, 0, (LPARAM) &Msg);
				}
			}

			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}
	}
	else
	{
		while (	PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE) AND
				Msg.message != WM_QUIT)
		{
			/*
			if (ActiveDialog)
			{
				// if (ProcessDialogMessage(ActiveDialog, &Msg))
				if (IsDialogMessage(ActiveDialog, &Msg))
				{
					continue;
				}
			}
			*/

			if (Msg.message == WM_KEYDOWN)
			{
				HWND hTop = Msg.hwnd, hNext;
				while (hNext = GetParent(hTop))
				{
					hTop = hNext;
				}
				if (hTop)
				{
					SendMessage(hTop, WM_TRANSLATE_ACCELERATOR, 0, (LPARAM) &Msg);
				}
			}

			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}
	}

	return Msg.message != WM_QUIT;
}

void GApp::Exit(int Code)
{
	if (Code)
	{
		exit(0);
	}
	else
	{
		PostQuitMessage(Code);
	}
}

//////////////////////////////////////////////////////////////////////////////
#include <process.h>

uint WINAPI ThreadEntryPoint(int i)
{
#ifdef	_MT
	GThread *Thread = (GThread*) i;
	if (Thread)
	{
		Thread->OnBeforeMain();
		Thread->State = LGITHREAD_RUNNING;
		int Status = Thread->ReturnValue = Thread->Main();
		Thread->State = LGITHREAD_EXITED;
		Thread->OnAfterMain();

		if (Thread->DeleteOnExit)
		{
			DeleteObj(Thread);
		}

		_endthreadex(Status);
	}
#endif

	return 0;
}

GThread::GThread()
{
	State = LGITHREAD_INIT;
	hThread = 0;
	ThreadId = 0;
	ReturnValue = 0;
	DeleteOnExit = false;

#ifdef	_MT
	hThread = (HANDLE) _beginthreadex(NULL,
									16<<10,
									(unsigned int (__stdcall *)(void *)) ThreadEntryPoint,
									(LPVOID) this,
									CREATE_SUSPENDED,
									&ThreadId);
#else
	#error "You are tryed to compile GThread support without the WIN32 multithreaded libs.";
#endif
}

GThread::~GThread()
{
}

bool GThread::IsExited()
{
	return (ExitCode() != STILL_ACTIVE);
}

bool GThread::IsRunning()
{
	return State == LGITHREAD_RUNNING;
}

void GThread::Run(bool b)
{
	DWORD Status = 0;
	DWORD Error = 0;

	if (b)
	{
		Status = ResumeThread(hThread);
		State = LGITHREAD_RUNNING;
	}
	else
	{
		State = LGITHREAD_ASLEEP;	// set before we sleep otherwise it can't 
									// be called from the actual thread.
		Status = SuspendThread(hThread);
	}

	OnRun(b);
}

void GThread::Terminate()
{
	if (hThread)
	{
		TerminateThread(hThread, 0);
		while (NOT IsExited());
	}
}

int GThread::ExitCode()
{
	DWORD d;
	GetExitCodeThread(hThread, &d);
	ReturnValue = d;
	return ReturnValue;
}

int GThread::Main()
{
	return 0;
}

int GThread::Priority()
{
	int p = 0;
	if (hThread)
	{
		p = GetThreadPriority(hThread);
	}
	return p;
}

void GThread::Priority(int p)
{
	if (hThread)
	{
		SetThreadPriority(hThread, p);
	}
}

//////////////////////////////////////////////////////////////////////////////
int GTrayIcon::Id = 1;
int GTrayIcon::TrayCreateMsg = 0;

GTrayIcon::GTrayIcon(int icons)
{
	ZeroObj(TrayIcon);
	Val = 0;
	Msg = WM_TRAY_NOTIFY;
	Notify = 0;
	MyId = 0;
	Icons = max(icons, 1);
	hIcon = NEW(HICON[Icons]);
	TrayCreateMsg = RegisterWindowMessage(TEXT("TaskbarCreated"));
}

GTrayIcon::~GTrayIcon()
{
	Visible(FALSE);

	if (hIcon)
	{
		for (int i=0; i<Icons; i++)
		{
			// DeleteObject(hIcon[i]);
		}

		DeleteArray(hIcon);
	}
}

bool GTrayIcon::LoadIcon(char *Str, int Index)
{
	if (hIcon AND Index >= 0 AND Index < Icons)
	{
		hIcon[Index] = ::LoadIcon(hAppInst(), Str);
		return hIcon[Index] != NULL;
	}

	return FALSE;
}

bool GTrayIcon::Visible()
{
	return TrayIcon.cbSize != 0;
}

void GTrayIcon::Visible(bool v)
{
	if (Visible() != v)
	{
		if (v)
		{
			if (Notify AND hIcon)
			{
				ZeroObj(TrayIcon);
				TrayIcon.cbSize = sizeof(TrayIcon);
				TrayIcon.hWnd = Notify->Handle();
				TrayIcon.uID = MyId = Id++;
				TrayIcon.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
				TrayIcon.uCallbackMessage = Msg;
				TrayIcon.hIcon = hIcon[Val];
				strncpy(TrayIcon.szTip, Name(), sizeof(TrayIcon.szTip));

				Shell_NotifyIcon(NIM_ADD, &TrayIcon);
			}
		}
		else
		{
			Shell_NotifyIcon(NIM_DELETE, &TrayIcon);
			ZeroObj(TrayIcon);
		}
	}
}

int GTrayIcon::Value()
{
	return Val;
}

void GTrayIcon::Value(int v)
{
	Val = v;

	if (Visible() AND hIcon)
	{
		ZeroObj(TrayIcon);
		TrayIcon.cbSize = sizeof(TrayIcon);
		TrayIcon.hWnd = Notify->Handle();
		TrayIcon.uID = MyId;
		TrayIcon.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
		TrayIcon.uCallbackMessage = Msg;
		TrayIcon.hIcon = hIcon[Val];
		strncpy(TrayIcon.szTip, Name(), sizeof(TrayIcon.szTip));

		Shell_NotifyIcon(NIM_MODIFY, &TrayIcon);
	}
}

int GTrayIcon::OnEvent(GMessage *Message)
{
	if (Message(Message) == Msg AND
		MessageA(Message) == MyId)
	{
		// got a notification from the icon
		VMouse m;
		ZeroObj(m);
		switch (MessageB(Message))
		{
			case WM_LBUTTONDBLCLK:
			case WM_LBUTTONDOWN:
			case WM_RBUTTONDBLCLK:
			case WM_RBUTTONDOWN:
			case WM_MBUTTONDBLCLK:
			case WM_MBUTTONDOWN:
			{
				m.Down(true);
				break;
			}
		}

		switch (MessageB(Message))
		{
			case WM_LBUTTONDBLCLK:
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP:
			{
				m.Flags = MK_LEFT;
				break;
			}
			case WM_RBUTTONDBLCLK:
			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
			{
				m.Flags = MK_RIGHT;
				break;
			}
			case WM_MBUTTONDBLCLK:
			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP:
			{
				m.Flags = MK_MIDDLE;
				break;
			}
		}

		switch (MessageB(Message))
		{
			case WM_LBUTTONDBLCLK:
			case WM_RBUTTONDBLCLK:
			case WM_MBUTTONDBLCLK:
			{
				m.Double(true);
				break;
			}
		}

		OnMouseClick(m);
	}
	else if (Message(Message) == TrayCreateMsg)
	{
		// explorer crashed... reinit the icon
		TrayIcon.cbSize = 0; // we're not visible
		Visible(true);
	}

	return 0;
}

void GTrayIcon::OnMouseClick(VMouse &m)
{
	// override to do something useful
}

//////////////////////////////////////////////////////////////////////
GPrinter::GPrinter()
{
	ZeroObj(Info);
	Info.lStructSize = sizeof(Info);
}

GPrinter::~GPrinter()
{
}

HDC GPrinter::StartDC(GView *Parent)
{
	Info.hwndOwner = (Parent) ? Parent->Handle() : 0;
	Info.Flags = PD_RETURNDC;
	if (NOT Parent OR PrintDlg(&Info))
	{
		if (NOT Info.hDC)
		{
			GMem DevMode(Info.hDevMode);
			GMem DevNames(Info.hDevNames);
			DEVNAMES *Names = (DEVNAMES*) DevNames.Lock();
			if (Names)
			{
				Info.hDC = ::CreateDC(	NULL,
										((char*) Names) + Names->wDeviceOffset,
										NULL,
										(DEVMODE*) DevMode.Lock());
			}
		}
	}

	return Info.hDC;
}

void GPrinter::EndDC()
{
	if (Info.hDC)
	{
		DeleteDC(Info.hDC);
		Info.hDC = 0;
	}
}

bool GPrinter::Browse(GView *Parent)
{
	Info.hwndOwner = (Parent) ? Parent->Handle() : 0;
	Info.Flags = PD_PRINTSETUP;
	return PrintDlg(&Info);
}

#define MAGIC_PRINTDLG					0xAAFF0100
#define MAGIC_DEVMODE					0xAAFF0101
#define MAGIC_DEVNAMES					0xAAFF0102

bool GPrinter::Serialize(pchar &Str, bool Write)
{
	int Size = sizeof(Info);
	if (Write)
	{
		GMem DevMode(Info.hDevMode);
		GMem DevNames(Info.hDevNames);
		DEVMODE *Mode = (DEVMODE*) DevMode.Lock();
		DEVNAMES *Names = (DEVNAMES*) DevNames.Lock();
		BytePipe Temp;
		int m;

		// Dlg
		m = MAGIC_PRINTDLG;
		Temp.Push(m);
		m = sizeof(Info);
		Temp.Push(m);
		Temp.Push((uchar*)&Info, m);

		// Mode
		if (Mode)
		{
			m = MAGIC_DEVMODE;
			Temp.Push(m);
			m = DevMode.GetSize();
			Temp.Push(m);
			Temp.Push((uchar*) Mode, m);
		}

		// Names
		if (Names)
		{
			m = MAGIC_DEVNAMES;
			Temp.Push(m);
			m = DevNames.GetSize();
			Temp.Push(m);
			Temp.Push((uchar*) Names, m);
		}

		// Convert to base64
		int BinLen = Temp.Sizeof();
		uchar *Bin = NEW(uchar[BinLen]);
		if (Bin AND Temp.Pop(Bin, BinLen))
		{
			int Base64Len = BufferLen_BinTo64(BinLen);
			char *Base64 = NEW(char[Base64Len+1]);
			if (Base64)
			{
				memset(Base64, 0, Base64Len+1);
				ConvertBinaryToBase64(Base64, Base64Len, Bin, BinLen);
				
				DeleteArray(Str);
				Str = Base64;
			}
			return true;
		}
	}
	else
	{
		bool Status = false;

		ZeroObj(Info);
		Info.lStructSize = Size;
		if (Str)
		{
			// Convert to binary
			BytePipe Temp;

			int Base64Len = strlen(Str);
			int BinaryLen = BufferLen_64ToBin(Base64Len);
			uchar *Binary = NEW(uchar[BinaryLen]);
			if (Binary)
			{
				int Len = ConvertBase64ToBinary(Binary, BinaryLen, Str, Base64Len);
				Temp.Push(Binary, Len);

				bool Done = false;
				do
				{
					int Type = 0;
					int Size = 0;

					if (Temp.Pop(Type) AND
						Temp.Pop(Size))
					{
						switch (Type)
						{
							case MAGIC_PRINTDLG:
							{
								if (Size == sizeof(Info))
								{
									Temp.Pop((uchar*) &Info, Size);
									Status = true;
								}
								else
								{
									// Grrrr: Error
									Done = true;
								}
								break;
							}
							case MAGIC_DEVMODE:
							{
								GMem DevMode(Size);
								uchar *Mode = (uchar*) DevMode.Lock();
								if (Mode)
								{
									Temp.Pop(Mode, Size);
									Info.hDevMode = DevMode.Handle();
									DevMode.Detach();
								}
								break;
							}
							case MAGIC_DEVNAMES:
							{
								GMem DevNames(Size);
								uchar *Names = (uchar*) DevNames.Lock();
								if (Names)
								{
									Temp.Pop(Names, Size);
									Info.hDevNames = DevNames.Handle();
									DevNames.Detach();
								}
								break;
							}
							default:
							{
								// Eof
								Done = true;
								break;
							}
						}
					}
					else
					{
						Done = true;
					}
				}
				while (NOT Done);
			}
		}

		return Status;
	}

	return false;
}

