/*hdr
**      FILE:           GuiCommon.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Common windows and controls
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
GFileSelect::GFileSelect()
{
	Size = 4096;
	FileStr = NEW(char[Size]);
	if (FileStr)
	{
		*FileStr = 0;
	}
	else
	{
		Size = 0;
	}

	ZeroObj(Info);
	Info.lStructSize = sizeof(Info);
	Info.hInstance = hAppInst();
}

GFileSelect::~GFileSelect()
{
	DeleteArray(FileStr);
	ClearTypes();
}

char *GFileSelect::TypeStr()
{
	char *Str = NEW(char[4096]);
	if (Str)
	{
		char *pstr = Str;

		for (GFileType *Type = TypeList.First(); Type; Type = TypeList.Next())
		{
			strcpy(pstr, Type->Description());
			pstr += strlen(Type->Description()) + 1;
			strcpy(pstr, Type->Extension());
			pstr += strlen(Type->Extension()) + 1;
		}

		*pstr = 0;
		Info.lpstrFilter = Str;
	}

	return Str;
}

char *GFileSelect::Name()
{
	return FileStr;
}

bool GFileSelect::Name(char *n)
{
	bool Status = FALSE;

	if (n AND FileStr)
	{
		strncpy(FileStr, n, Size);
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;
	}

	return Status;
}

char *GFileSelect::operator [](int i)
{
	if (Info.nFileOffset > strlen(FileStr))
	{
		// multi-select
		char *File = FileStr + Info.nFileOffset;
		for (int n=0; n<i AND *File; n++)
		{
			File = File + strlen(File) + 1;
		}

		static char Buf[256];
		sprintf(Buf, "%s%s%s", FileStr, DIR_STR, File);
		return Buf;
	}
	else
	{
		// single select
		return FileStr;
	}
}

int GFileSelect::Names()
{
	if (Info.nFileOffset > strlen(FileStr))
	{
		// multi-select
		char *File = FileStr + Info.nFileOffset;
		int n;
		for (n = 0; *File; n++)
		{
			File = File + strlen(File) + 1;
		}

		return n;
	}
	else
	{
		// single select
		return 1;
	}
}

char *GFileSelect::File()
{
	return FileStr + Info.nFileOffset;
}

char *GFileSelect::Extension()
{
	return FileStr + Info.nFileExtension;
}

int GFileSelect::Types()
{
	return TypeList.GetItems();
}

int GFileSelect::SelectedType()
{
	return Info.nFilterIndex - 1;
}

GFileType *GFileSelect::TypeAt(int n)
{
	return TypeList.ItemAt(n);
}

void GFileSelect::ClearTypes()
{
	DeleteObjList(&TypeList, GFileType);
}

bool GFileSelect::Type(char *Description, char *Extension, int Data)
{
	GFileType *Type = NEW(GFileType);
	if (Type)
	{
		Type->Description(Description);
		Type->Extension(Extension);
		Type->Data(Data);
		TypeList.Insert(Type);
	}

	return Type != 0;
}

GView *GFileSelect::Parent()
{
	return ParentWnd;
}

void GFileSelect::Parent(GView *Window)
{
	ParentWnd = Window;
	if (ParentWnd)
	{
		Info.hwndOwner = Window->Handle();
	}
}

bool GFileSelect::MultiSelect()
{
	return (Info.Flags & OFN_ALLOWMULTISELECT) != 0;
}

void GFileSelect::MultiSelect(bool Multi)
{
	if (Multi)
	{
		Info.Flags |= OFN_ALLOWMULTISELECT;
	}
	else
	{
		Info.Flags &= ~OFN_ALLOWMULTISELECT;
	}

	Info.Flags |= OFN_EXPLORER;
}

char *GFileSelect::InitialDir()
{
	return (char*) Info.lpstrInitialDir;
}

void GFileSelect::InitialDir(char *InitDir)
{
	Info.lpstrInitialDir = InitDir;
}

char *GFileSelect::Title()
{
	return (char*) Info.lpstrTitle;
}

void GFileSelect::Title(char *Title)
{
	Info.lpstrTitle = Title;
}

char *GFileSelect::DefaultExtension()
{
	return (char*) Info.lpstrDefExt;
}

void GFileSelect::DefaultExtension(char *DefExt)
{
	Info.lpstrDefExt = DefExt;
}

bool GFileSelect::Open()
{
	bool Status = FALSE;

	char *Str = TypeStr();
	if (Str)
	{
		Info.lpstrFilter = Str;		
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;

		Status = GetOpenFileName(&Info);

		DeleteObj(Str);
	}

	return Status AND strlen(FileStr) > 0;
}

#include "shlobj.h"

char *_GFileSelect_DefaultFolder = 0;
int CALLBACK GFileSelectBrowseCallback(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
{
	if (uMsg == BFFM_INITIALIZED AND
		_GFileSelect_DefaultFolder)
	{
		PostMessage(hwnd, BFFM_SETSELECTION, true, (uint)_GFileSelect_DefaultFolder);
	}

	return 0;
}

bool GFileSelect::OpenFolder()
{
	LPMALLOC pMalloc; // Gets the Shell's default allocator
	bool Status = false;
	_GFileSelect_DefaultFolder = Name();

	if  (FileStr AND ::SHGetMalloc(&pMalloc) == NOERROR)
	{
		BROWSEINFO      bi;
		char            pszBuffer[MAX_PATH];
		LPITEMIDLIST    pidl;

		ZeroObj(bi);

		/* Set the root folder?
		if (Name() AND strlen(Name()))
		{
			OLECHAR			olepath[256];
			ulong			cheaten;
			ulong			dwattributes;
			HRESULT			hr;
			LPSHELLFOLDER	pdesktopfolder;

			// // get a pointer to the desktop's ishellfolder interface. //
			if (SUCCEEDED(SHGetDesktopFolder(&pdesktopfolder)))
			{
				// ishellfolder::parsedisplayname requires the file name be in unicode.
				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, Name(), -1, olepath, CountOf(olepath));

				// convert the path to an itemidlist.
				hr = pdesktopfolder->ParseDisplayName(	0,
														0,
														olepath,
														&cheaten,
														&pidl,
														&dwattributes);
				bi.pidlRoot = pidl;
			}
		}
		*/

		// Get help on BROWSEINFO struct - it's got all the bit settings.
		bi.hwndOwner        = (ParentWnd) ? ParentWnd->Handle() : 0;
		bi.pszDisplayName   = pszBuffer;
		bi.lpszTitle        = "Select a target Directory";
		bi.ulFlags          = BIF_RETURNFSANCESTORS | BIF_RETURNONLYFSDIRS;
		bi.lpfn				= GFileSelectBrowseCallback;

		// This next call issues the dialog box.
		if  (pidl = ::SHBrowseForFolder(&bi))
		{
			if  (::SHGetPathFromIDList(pidl, pszBuffer))
			{
				// At this point pszBuffer contains the selected path
				strcpy(FileStr, pszBuffer);
				Status = true;
			}

			// Free the PIDL allocated by SHBrowseForFolder.
			pMalloc->Free(pidl);
		}

		// Release the shell's allocator
		pMalloc->Release();
	}

	return Status;
}

bool GFileSelect::Save()
{
	bool Status = FALSE;

	char *Str = TypeStr();
	if (Str)
	{
		Info.lpstrFilter = Str;		
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;

		Status = GetSaveFileName(&Info);

		DeleteObj(Str);
	}

	return Status AND strlen(FileStr) > 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GFindReplace::GFindReplace(GView *parent)
{
	Parent = parent;
	FindOp = true;
	FindMsgID = RegisterWindowMessage(FINDMSGSTRING);
	Find[0] = 0;
	Replace[0] = 0;
	hDlg = 0;

	ZeroObj(FindInfo);
	FindInfo.lStructSize = sizeof(FindInfo);
	FindInfo.hInstance = hAppInst();
	FindInfo.Flags = FR_HIDEUPDOWN;
	FindInfo.hwndOwner = Parent ? Parent->Handle() : 0;
	FindInfo.lpstrFindWhat = Find;
	FindInfo.lpstrReplaceWith = Replace;
	FindInfo.wFindWhatLen = sizeof(Find);
	FindInfo.wReplaceWithLen = sizeof(Replace);
}

GFindReplace::~GFindReplace()
{
	if (hDlg)
	{
		DestroyWindow(hDlg);
		hDlg = 0;
	}
}

void GFindReplace::SetParent(GView *parent)
{
	Parent = parent;
}

bool GFindReplace::OpenFind()
{
	FindOp = true;
	FindInfo.hwndOwner = Parent ? Parent->Handle() : 0;
	hDlg = FindText(&FindInfo);
	if (hDlg)
	{
		MSG Msg;
		while (GetMessage(&Msg, NULL, 0, 0) > 0 AND IsWindow(hDlg))
		{
			if (NOT hDlg OR NOT IsDialogMessage(hDlg, &Msg))
			{
				TranslateMessage(&Msg);
				DispatchMessage(&Msg);
			}
		}

		return TRUE;
	}
	else
	{
		DWORD e = CommDlgExtendedError();
		int k=0;
	}

	return FALSE;
}

bool GFindReplace::OpenReplace()
{
	FindOp = FALSE;
	hDlg = ReplaceText(&FindInfo);
	if (hDlg)
	{
		MSG Msg;
		while (GetMessage(&Msg, NULL, 0, 0) > 0 AND IsWindow(hDlg))
		{
			if (NOT hDlg OR NOT IsDialogMessage(hDlg, &Msg))
			{
				TranslateMessage(&Msg);
				DispatchMessage(&Msg);
			}
		}

		return TRUE;
	}

	return FALSE;
}

int GFindReplace::OnEvent(GMessage *Msg)
{
	int Status = FALSE;

	if (Msg->Msg == FindMsgID)
	{
		if (FindOp)
		{
			Status = OnFind();
		}
		else
		{
			Status = OnReplace();
		}
	}

	return Status;
}

bool GFindReplace::OnFind()
{
	DestroyWindow(hDlg);
	hDlg = 0;
	return TRUE;
}

bool GFindReplace::OnReplace()
{
	DestroyWindow(hDlg);
	hDlg = 0;
	return TRUE;
}

bool GFindReplace::CaseSensitive()
{
	return (FindInfo.Flags & FR_MATCHCASE) != 0;
}

void GFindReplace::CaseSensitive(bool i)
{
	if (i)
	{
		FindInfo.Flags |= FR_MATCHCASE;
	}
	else
	{
		FindInfo.Flags &= ~FR_MATCHCASE;
	}
}

bool GFindReplace::WholeWord()
{
	return (FindInfo.Flags & FR_WHOLEWORD) != 0;
}

void GFindReplace::WholeWord(bool i)
{
	if (i)
	{
		FindInfo.Flags |= FR_WHOLEWORD;
	}
	else
	{
		FindInfo.Flags &= ~FR_WHOLEWORD;
	}
}

bool GFindReplace::BtnFindNext()
{
	return (FindInfo.Flags & FR_FINDNEXT) != 0;
}

bool GFindReplace::BtnReplace()
{
	return (FindInfo.Flags & FR_REPLACE) != 0;
}

bool GFindReplace::BtnReplaceAll()
{
	return (FindInfo.Flags & FR_REPLACEALL) != 0;
}

//////////////////////////////////////////////////////////////////////////////
GInput::GInput(GView *parent, char *InitStr, char *Msg, char *Title, bool Password)
{
	SetParent(parent);
	Str = 0;
	Pos.ZOff(300, 110);
	GView::Name(Title);
	MoveToCenter();
	Children.Insert(NEW(GText(-1, 5, 5, 280, 20, Msg)));
	Children.Insert(Edit = NEW(GEdit(200, 5, 30, 280, 20, InitStr)));
	if (Edit)
	{
		Edit->Password(Password);
	}
	GButton *Ok = 0;
	Children.Insert(Ok = NEW(GButton(IDOK, 155, 60, 60, 20, "Ok")));
	Children.Insert(NEW(GButton(IDCANCEL, 225, 60, 60, 20, "Cancel")));

	if (Ok)
	{
		Ok->Default(true);
	}
}

GInput::~GInput()
{
	DeleteArray(Str);
}

int GInput::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDOK:
		{
			Str = NewStr(Edit->Name());
			// fall thru
		}
		case IDCANCEL:
		{
			EndModal(Ctrl->GetId() == IDOK);
			break;
		}
	}

	return 0;
}

