/*
**	FILE:		GuiViews.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Standard Views
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
/*
bool IsWindowThread(GView *w)
{
	return	w AND
			GetCurrentThreadId() == GetWindowThreadProcessId(w->Handle(), NULL);
}
*/

void HashRectangle(GView *Wnd, GdcRegion *r)
{
	if (r AND Wnd AND Wnd->Handle())
	{
		HDC hdc = GetDC(Wnd->Handle());
		if (hdc)
		{
			BITMAPINFO *bmp = (BITMAPINFO*) NEW(char[sizeof(BITMAPINFO)+sizeof(RGBQUAD)+32]);
			int Pat[8] = {0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555};
			if (bmp)
			{
				bmp->bmiHeader.biSize = sizeof(bmp->bmiHeader);
				bmp->bmiHeader.biWidth = 8;
				bmp->bmiHeader.biHeight = 8;
				bmp->bmiHeader.biPlanes = 1;
				bmp->bmiHeader.biBitCount = 1;
				bmp->bmiHeader.biCompression = BI_RGB;
				bmp->bmiHeader.biSizeImage = 32;
				bmp->bmiHeader.biXPelsPerMeter = 3000;
				bmp->bmiHeader.biYPelsPerMeter = 3000;
				bmp->bmiHeader.biClrUsed = 0;
				bmp->bmiHeader.biClrImportant = 0;

				bmp->bmiColors[0].rgbRed = 255;
				bmp->bmiColors[0].rgbGreen = 255;
				bmp->bmiColors[0].rgbBlue = 255;
				bmp->bmiColors[1].rgbRed = 0;
				bmp->bmiColors[1].rgbGreen = 0;
				bmp->bmiColors[1].rgbBlue = 0;

				memcpy(bmp->bmiColors + 2, Pat, 32);
				
				LOGBRUSH brush;
				brush.lbStyle = BS_DIBPATTERNPT;
				brush.lbColor = DIB_RGB_COLORS;
				brush.lbHatch = (LONG) bmp;

				HBRUSH hbrush = CreateBrushIndirect(&brush);
				if (hbrush)
				{
					HPEN hpen = (HPEN) GetStockObject(NULL_PEN);
					hpen = (HPEN) SelectObject(hdc, hpen);
					hbrush = (HBRUSH) SelectObject(hdc, hbrush);

					SetTextColor(hdc, RGB(255, 255, 255));
					SetBkColor(hdc, RGB(0, 0, 0));
					SetROP2(hdc, R2_XORPEN);
					SetBkMode(hdc, TRANSPARENT);
					::Rectangle(hdc, r->x1, r->y1, r->x2+1, r->y2+1);

					hbrush = (HBRUSH) SelectObject(hdc, hbrush);
					hpen = (HPEN) SelectObject(hdc, hpen);
					DeleteObject(hbrush);
				}

				ReleaseDC(Wnd->Handle(), hdc);
				DeleteArray(bmp);
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
GLayout::GLayout()
{
	_SettingScrollBars = 0;
	_PourLargest = 0;
	VScroll = 0;
	HScroll = 0;

	// Class.Class.style |= CS_HREDRAW | CS_VREDRAW;
	Class.Create("GLayout");
	WndClass = Class.Name();
	WndStyle |= WS_CHILD | WS_CLIPSIBLINGS; // | WS_CLIPCHILDREN
	WndExStyle |= WS_EX_CONTROLPARENT;
}

GLayout::~GLayout()
{
	DeleteObj(VScroll);
	DeleteObj(HScroll);
}

void GLayout::GetScrollPos(int &x, int &y)
{
	if (HScroll)
	{
		x = HScroll->Pos();
	}
	else
	{
		x = 0;
	}

	if (VScroll)
	{
		y = VScroll->Pos();
	}
	else
	{
		y = 0;
	}
}

void GLayout::SetScrollPos(int x, int y)
{
	if (HScroll)
	{
		HScroll->SetPos(x);
	}

	if (VScroll)
	{
		VScroll->SetPos(y);
	}
}

bool GLayout::SetScrollBars(bool x, bool y)
{
	bool Status = FALSE;

	if (NOT _SettingScrollBars)
	{
		_SettingScrollBars = true;

		/*
		if (Handle())
		{
			WndStyle = GetWindowLong(Handle(), GWL_STYLE);
		}
		*/

		if (y)
		{
			if (NOT VScroll)
			{
				VScroll = NEW(GScrollBar);
				if (VScroll)
				{
					// WndStyle |= WS_VSCROLL;
					VScroll->SetVertical(TRUE);
					VScroll->SetParent(this);
					Status = TRUE;
				}
			}
		}
		else
		{
			if (VScroll)
			{
				// WndStyle &= ~WS_VSCROLL;
				DeleteObj(VScroll);
				Status = TRUE;
			}
		}
		
		if (x)
		{
			if (NOT HScroll)
			{
				HScroll = NEW(GScrollBar);
				if (HScroll)
				{
					// WndStyle |= WS_HSCROLL;
					HScroll->SetVertical(FALSE);
					HScroll->SetParent(this);
					Status = TRUE;
				}
			}
		}
		else
		{
			if (HScroll)
			{
				// WndStyle &= ~WS_HSCROLL;
				DeleteObj(HScroll);
				Status = TRUE;
			}
		}

		/*
		if (Handle())
		{
			SetWindowLong(Handle(), GWL_STYLE, WndStyle);
			SetWindowPos(	Handle(),
							0,
							Pos.x1, Pos.y1, Pos.X(), Pos.Y(),
							SWP_NOMOVE | SWP_NOZORDER | SWP_NOSIZE | SWP_FRAMECHANGED);
		}
		*/

		_SettingScrollBars = false;
	}

	return Status;
}

bool GLayout::GetPourLargest()
{
	return _PourLargest;
}

void GLayout::SetPourLargest(bool i)
{
	_PourLargest = i;
}

bool GLayout::Pour(GdcComplexRegion &r)
{
	if (_PourLargest)
	{
		GdcRegion *Best = FindLargest(r);
		if (Best)
		{
			SetPos(*Best);
			return true;
		}
	}
	return false;
}

int GLayout::OnEvent(GMessage *Msg)
{
	if (Message(Msg) == WM_CHANGE AND
		Parent)
	{
		Parent->OnEvent(Msg);
	}

	if (VScroll) VScroll->OnEvent(Msg);
	if (HScroll) HScroll->OnEvent(Msg);
	return GView::OnEvent(Msg);
}

//////////////////////////////////////////////////////////////////////////////////////
GWin32Class GScrollBar::WndClass;

GScrollBar::GScrollBar()
#ifdef LGI_RES
	: ResObject(Res_ScrollBar)
#endif
{
	Type = 0;
	Parent = 0;
	ZeroObj(Info);
	Info.cbSize = sizeof(Info);
	Info.fMask = SIF_ALL; // | SIF_DISABLENOSCROLL;
	Info.nMax = -1;
}

GScrollBar::GScrollBar(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_ScrollBar)
#endif
{
	SetId(id);
	GView::Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);
	WndClass.SubClass(Class(), "SCROLLBAR");

	Type = SB_HORZ;
	ZeroObj(Info);
	Info.cbSize = sizeof(Info);
	Info.fMask = SIF_ALL | SIF_DISABLENOSCROLL;
	Info.nMax = -1;
}

GScrollBar::~GScrollBar()
{
	if (NOT Handle())
	{
		SetParentFlag(false);
		Parent = 0;
	}
}

uint32 GScrollBar::Style()
{
	return	WS_CHILD |
			WS_TABSTOP |
			WS_VISIBLE |
			GView::Style() |
			((Type == SB_VERT) ? SBS_VERT : SBS_HORZ);
}

bool GScrollBar::Vertical()
{
	return Type == SB_VERT;
}

void GScrollBar::SetVertical(bool v)
{
	if (v)
	{
		Type = SB_VERT;
	}
	else
	{
		Type = SB_HORZ;
	}
}

void GScrollBar::SetParentFlag(bool Set)
{
	if (NOT Handle() AND
		Parent AND
		Parent->Handle())
	{
		DWORD Flags = GetWindowLong(Parent->Handle(), GWL_STYLE);
		int Flag = (Vertical()) ? WS_VSCROLL : WS_HSCROLL;

		if (Set)
		{
			Flags |= Flag;
		}
		else
		{
			Flags &= ~Flag;

			// clear limits.
			Info.nMin = 0;
			Info.nMax = -1;
			Update();
		}

		// check that we're in thread...
		if (Parent->InThread())
		{
			// you can't call 'SetWindowLong' from a thread other than
			// the one that created the window.
			SetWindowLong(Parent->Handle(), GWL_STYLE, Flags);
			SetWindowPos(	Parent->Handle(),
							0, 0, 0, 0, 0,
							SWP_NOMOVE | SWP_NOZORDER | SWP_NOSIZE | SWP_FRAMECHANGED);

			if (Set)
			{
				Update();
			}
		}
	}
}

void GScrollBar::Update()
{
	if (Parent AND Parent->Handle())
	{
		if (Handle())
		{
			if (InThread())
			{
				SetScrollInfo(Handle(), SB_CTL, &Info, true);
				EnableScrollBar(Handle(), SB_CTL, (Info.nMax < Info.nMin) ? ESB_DISABLE_BOTH : ESB_ENABLE_BOTH);
			}
		}
		else if (Parent)
		{
			if (Parent->InThread())
			{
				Info.fMask = SIF_ALL;
				SetScrollInfo(Parent->Handle(), Type, &Info, true);
			}
		}
	}
}

void GScrollBar::SetParent(GView *p)
{
	Parent = p;
	SetParentFlag(true);
}

int GScrollBar::Pos()
{
	return Info.nPos;
}

void GScrollBar::SetPos(int p)
{
	Info.nPos = limit(p, Info.nMin, Info.nMax);
	Update();
}

void GScrollBar::Limits(int &Low, int &High)
{
	Low = Info.nMin;
	High = Info.nMax;
}

void GScrollBar::SetLimits(int Low, int High)
{
	Info.nMin = Low;
	Info.nMax = High;
	Update();
}

int GScrollBar::Page()
{
	return Info.nPage;
}

void GScrollBar::SetPage(int p)
{
	Info.nPage = p;
	Update();
}

void GScrollBar::Invalidate()
{
	SetScrollInfo(Parent->Handle(), Type, &Info, true);
}

int GScrollBar::OnEvent(GMessage *Msg)
{
	int Status = WndClass.CallParent(hWnd, Msg->Msg, Msg->a, Msg->b);

	if ((Msg->Msg == WM_HSCROLL AND Type == SB_HORZ) OR
		(Msg->Msg == WM_VSCROLL AND Type == SB_VERT))
	{
		int OldPos = Info.nPos;
		int NewPos = Info.nPos;
		int ScrollCode = (int) LOWORD(Msg->a);
		int nPos = (short) HIWORD(Msg->a);

		switch (ScrollCode)
		{
			case SB_BOTTOM:
			case SB_LINEDOWN:
			{
				NewPos = limit(NewPos+1, Info.nMin, Info.nMax);
				break;
			}
			case SB_TOP:
			case SB_LINEUP:
			{
				NewPos = limit(NewPos-1, Info.nMin, Info.nMax);
				break;
			}
			case SB_PAGEDOWN:
			{
				NewPos = limit(NewPos+(int)Info.nPage, Info.nMin, Info.nMax);
				break;
			}
			case SB_PAGEUP:
			{
				NewPos = limit(NewPos-(int)Info.nPage, Info.nMin, Info.nMax);
				break;
			}
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK:
			{
				NewPos = limit(nPos, Info.nMin, Info.nMax);
				break;
			}
		}

		if (OldPos != NewPos)
		{
			Info.nPos = NewPos;
			Update();
			OnChange(NewPos);

			if (Parent)
			{
				GView *w = dynamic_cast<GView*>(this);
				GMessage m(WM_CHANGE, (int)w, NewPos);
				
				Parent->OnEvent(&m);
			}
		}
	}

	if (Message(Msg) == WM_CREATE)
	{
		Update();
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////
GStatusBar::GStatusBar()
{
	Name("LGI_StatusBar");

	Font = NEW(GFont);
	if (Font)
	{
		Font->Height(8);
		Font->Face("Ms Sans Serif");
		Font->Transparent(true);
		ASSERT(Font->Create());
	}
}

GStatusBar::~GStatusBar()
{
	DeleteObj(Font);
}

bool GStatusBar::Pour(GdcComplexRegion &r)
{
	GdcRegion *Best = FindLargestEdge(r, GV_EDGE_BOTTOM);
	if (Best)
	{
		GdcRegion Take = *Best;
		if (Take.Y() > 21)
		{
			Take.y1 = max((Take.y2 - 21), Take.y1);
			SetPos(Take);
			RePour();
			return TRUE;
		}
	}
	return FALSE;
}

void GStatusBar::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, X()-1, Y()-1);

	DrawMoveShallowEdge(pDC, r, RAISED);

	GdcComplexRegion Cr(r);
	#ifdef WIN32
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		GdcRegion n = c->GetPos();
		Cr.Subtract(&n);
	}
	#endif
	pDC->Colour(LC_MED, 24);
	for (GdcRegion *i = Cr.First(); i; i = Cr.Next())
	{
		pDC->Rectangle(i);
	}

	GView::OnPaint(pDC);
}

void GStatusBar::RePour()
{
	int X = Pos.X() - 5;
	int i = Children.GetItems()-1;

	for (	GView *w = Children.Last();
			w;
			w = Children.Prev(), i--)
	{
		GStatusPane *Pane = dynamic_cast<GStatusPane*>(w);
		if (Pane)
		{
			if (i)
			{
				// not first
				X -= Pane->Width; 
				Pane->Pos.ZOff(Pane->Width, Pos.Y()-7);
				Pane->Pos.Offset(X, 3);
				X -= STATUSBAR_SEPARATOR;
			}
			else
			{
				// first
				Pane->Pos.ZOff(X-2, Pos.Y()-4);
				Pane->Pos.Offset(2, 2);
				X = 0;
			}
		}
	}
}

GStatusPane *GStatusBar::AppendPane(char *Text, int Width)
{
	GStatusPane *Pane = 0;
	if (Text)
	{
		Pane = NEW(GStatusPane);
		if (Pane)
		{
			Pane->Parent = this;
			Pane->Name(Text);
			Pane->Width = Width;
			Children.Insert(Pane);
		}
	}
	return Pane;
}

bool GStatusBar::AppendPane(GStatusPane *Pane)
{
	if (Pane)
	{
		Pane->Parent = this;
		Children.Insert(Pane);
	}
	return TRUE;
}

// Pane
GStatusPane::GStatusPane()
{
	Parent = 0;
	Flags = 0;
	Width = 32;
	Pos.ZOff(Width-1, 20);
	pDC = 0;
}

GStatusPane::~GStatusPane()
{
	DeleteObj(pDC)
}

bool GStatusPane::Name(char *n)
{
	bool Status = GObject::Name(n);
	GdcRegion p(0, 0, Pos.X()-1, Pos.Y()-1);
	p.Size(1, 1);
	Invalidate(&p);
	return Status;
}

void GStatusPane::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, X()-1, Y()-1);
	int Off = 1;

	if (Sunken())
	{
		Off = 0;
		DrawMoveShallowEdge(pDC, r, SUNKEN);
	}

	GStatusBar *Par = (GStatusBar*) Parent;
	if (Par AND Par->Font)
	{
		char *t = Name();
		if (t AND strlen(t)>0)
		{
			int TabSize = Par->Font->TabSize();
			Par->Font->TabSize(0);
			Par->Font->Colour(Rgb24(0, 0, 0), LC_MED);
			Par->Font->Transparent(false);
			Par->Font->Text(pDC, 2+Off, 1+Off, Name(), -1, &r);
			Par->Font->TabSize(TabSize);
			return;
		}
	}

	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);
}

int GStatusPane::GetWidth()
{
	return Width;
}

void GStatusPane::SetWidth(int x)
{
	Width = x;
	if (Parent)
	{
		((GStatusBar*) Parent)->RePour();
	}
}

bool GStatusPane::Sunken()
{
	return TestFlag(Flags, GSP_SUNKEN);
}

void GStatusPane::Sunken(bool s)
{
	bool Old = Sunken();
	if (s) SetFlag(Flags, GSP_SUNKEN);
	else ClearFlag(Flags, GSP_SUNKEN);
	if (Old != Sunken() AND Parent)
	{
		Parent->Invalidate(&Pos);
	}
}

GdcBasePrimitives *GStatusPane::Bitmap()
{
	return pDC;
}

void GStatusPane::Bitmap(GdcBasePrimitives *pNewDC)
{
	DeleteObj(pDC);
	if (pNewDC)
	{
		pDC = NEW(GdcMemDC);
		if (pDC)
		{
			if (pDC->Create(pNewDC->X(), pNewDC->Y(), pNewDC->GetBits()))
			{
				pDC->Blt(0, 0, pNewDC);
			}
			else
			{
				DeleteObj(pDC);
			}
		}
	}

	if (Parent)
	{
		Parent->Invalidate(&Pos);
	}
}

//////////////////////////////////////////////////////////////////////////////
#define SPLITER_BORDER			4
#define SPLITER_BAR				5
#define SPLITER_INSET			2
#define SPLITER_MIN_X			(2*SPLITER_INSET)
#define SPLITER_MIN_Y			(2*SPLITER_INSET)

#define SPLITER_FULL_DRAG

GSplitter::GSplitter()
{
	Name("LGI_Spliter");
	WndFlags |= WS_CLIPCHILDREN;
	SplitSet = SplitPos = -1;
	Moving = FALSE;
	Cursor = 0;

	PosA.ZOff(0, 0);
	PosB.ZOff(0, 0);
	ViewA = 0;
	ViewB = 0;
	BorderA = true;
	BorderB = true;
	OverBar = false;
	SplitFollow = false;

	OldCursor = 0;
	IsVertical(true);
	Border(true);
	Raised(true);
	Split(0);

	SetPourLargest(true);
	// WndStyle |= WS_CLIPCHILDREN;
}

GSplitter::~GSplitter()
{
	DeleteObj(ViewA);
	DeleteObj(ViewB);
}

void GSplitter::IsVertical(bool v)
{
	Vertical = v;
	if (Cursor)
	{
		DeleteObject(Cursor);
		Cursor = 0;
	}
}

void GSplitter::SetViewA(GView *a, bool Border)
{
	BorderA = Border;
	ViewA = a;
	if (a)
	{
		a->Attach(this);
		a->Visible(true);

		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

void GSplitter::SetViewB(GView *b, bool Border)
{
	BorderB = Border;
	ViewB = b;
	if (b)
	{
		b->Attach(this);
		b->Visible(true);

		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

void GSplitter::Split(int s)
{
	SplitSet = s;
	int Limit = ((Vertical) ? Pos.X() : Pos.Y()) - 18;
	int NewPos = (s > 0) ? max(s, 4) : max(4, Limit+s);
	if (NewPos != SplitPos)
	{
		SplitPos = NewPos;
		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

GView *GSplitter::FindControl(HWND hCtrl)
{
	GView *c = 0;
	if (ViewA) c = ViewA->FindControl(hCtrl);
	if (c) return c;

	if (ViewB) c = ViewB->FindControl(hCtrl);
	return c;
}

void GSplitter::CalcRegions(GdcRegion Rect, bool Follow)
{
	PosA = Rect;
	PosA.Offset(-PosA.x1, -PosA.y1);
	if (Border())
	{
		PosA.Size(SPLITER_BORDER, SPLITER_BORDER);
	}

	PosB = PosA;

	if (SplitFollow AND Follow)
	{
		SplitPos = max(1, ((Vertical) ? Pos.X()-10 : Pos.Y()-10) + SplitSet);
	}

	if (Vertical)
	{
		PosA.x2 = PosA.x1 + SplitPos;
		PosB.x1 = PosA.x2 + SPLITER_BAR;
	}
	else
	{
		PosA.y2 = PosA.y1 + SplitPos;
		PosB.y1 = PosA.y2 + SPLITER_BAR;
	}

	if (ViewA)
	{
		GdcRegion r = PosA;
		if (BorderA)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		ViewA->SetPos(r);
		
		GdcComplexRegion Rgn(r);
		ViewA->Pour(Rgn);
	}
	if (ViewB)
	{
		GdcRegion r = PosB;
		if (BorderB)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		ViewB->SetPos(r);

		GdcComplexRegion Rgn(r);
		ViewB->Pour(Rgn);
	}
}

void GSplitter::OnPosChange()
{
	CalcRegions(Pos, true);
}

void ClipDC(HDC hDC, RECT rc)
{
	HRGN hRgn = CreateRectRgn(rc.left, rc.top, rc.right, rc.bottom);
	SelectClipRgn(hDC, hRgn);
}

void GSplitter::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r;
	r.ZOff(pDC->X()-1, pDC->Y()-1);

	if (Border())
	{
		if (Sunken())
		{
			DrawMoveShallowEdge(pDC, r, SUNKEN);
		}
		else if (Raised())
		{
			DrawMoveShallowEdge(pDC, r, RAISED);
		}

		pDC->Colour(LC_MED, 24);
		DrawMoveFill(pDC, r, SPLITER_BORDER);
	}
	else
	{
		pDC->Colour(LC_MED, 24);
	}

	// bar
	if (Vertical)
	{
		pDC->Rectangle(PosA.x2, PosA.y1, PosB.x1, PosA.y2);
	}
	else
	{
		pDC->Rectangle(PosA.x1, PosA.y2, PosA.x2, PosB.y1);
	}

	// PosA region
	r = PosA;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderA)
		{
			DrawMoveDeepEdge(pDC, r, SUNKEN);
		}

		if (ViewA)
		{
			if (NOT ViewA->Handle())
			{
				pDC->SetOrigin(PosA.x1, PosA.y1);
				ViewA->OnPaint(pDC);
				pDC->SetOrigin(0, 0);
			}
			else
			{
				// View is real and will paint itself
			}
		}
		else
		{
			pDC->Colour(LC_WORKSPACE, 24);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	r = PosB;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderB)
		{
			DrawMoveDeepEdge(pDC, r, SUNKEN);
		}

		if (ViewB)
		{
			if (NOT ViewB->Handle())
			{
				pDC->SetOrigin(PosB.x1, PosB.y1);
				ViewB->OnPaint(pDC);
				pDC->SetOrigin(0, 0);
			}
			else
			{
				// View is real and will paint itself
			}
		}
		else
		{
			pDC->Colour(LC_WORKSPACE, 24);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	pDC->EndDC();
}

void GSplitter::OnMouseClick(VMouse &m)
{
	if (m.Down())
	{
		if (OverSplit(m.x, m.y))
		{
			if (Vertical)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x2;
				Bar.y1 = PosA.y1;
				Bar.x2 = PosB.x1;
				Bar.y2 = PosA.y2;
				Offset = m.x - Bar.x1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
			else
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x1;
				Bar.y1 = PosA.y2;
				Bar.x2 = PosA.x2;
				Bar.y2 = PosB.y1;
				Offset = m.y - Bar.y1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
	}
	else
	{
		#ifdef SPLITER_FULL_DRAG
		Capture(FALSE);
		Moving = FALSE;
		#else
		if (Moving)
		{
			HashRectangle(this, &Bar);
			Capture(FALSE);
			Moving = FALSE;

			int NewPos;
			if (Vertical)
			{
				NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
			}
			else
			{
				NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
			}

			if (NewPos != SplitPos)
			{
				SplitPos = NewPos;
				CalcRegions(Pos);
				Invalidate();
			}
		}
		#endif
	}
}

bool GSplitter::OverSplit(int x, int y)
{
	if (Vertical)
	{
		return (x > PosA.x2 AND x < PosB.x1 AND y > PosA.y1 AND y < PosA.y2);
	}
	else
	{
		return (x > PosA.x1 AND x < PosA.x2 AND y > PosA.y2 AND y < PosB.y1);
	}
}

void GSplitter::OnMouseExit(VMouse &m)
{
	if (OldCursor)
	{
		SetCursor(OldCursor);
	}
}

void GSplitter::OnMouseMove(VMouse &m)
{
	if (Moving)
	{
		#ifdef SPLITER_FULL_DRAG
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}

		int NewPos;
		if (Vertical)
		{
			NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
		}
		else
		{
			NewPos = limit(Bar.y1, 4, Pos.Y()-18);
		}

		if (NewPos != SplitPos)
		{
			SplitPos = NewPos;
			CalcRegions(Pos);
			Invalidate();
			UpdateWindow(hWnd);
		}
		#else
		HashRectangle(this, &Bar);
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}
		HashRectangle(this, &Bar);
		#endif
	}
}

int GSplitter::OnHitTest(int x, int y)
{
	if (OverSplit(x, y))
	{
		if (NOT Cursor)
		{
			Cursor = LoadCursor(NULL, MAKEINTRESOURCE((Vertical) ? IDC_SIZEWE : IDC_SIZENS));
		}

		SetCursor(Cursor);
		return 1;
	}

	return -1;
}

bool GSplitter::Attach(GView *p)
{
	bool Status = GLayout::Attach(p);
	if (Status)
	{
		if (ViewA AND NOT ViewA->Handle())
		{
			ViewA->Attach(this);
		}

		if (ViewB AND NOT ViewB->Handle())
		{
			ViewB->Attach(this);
		}
	}

	return Status;
}
