/*hdr
**      FILE:           Win32Res.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           10/3/2000
**      DESCRIPTION:    Win32 resource interfaces
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <time.h>

#include "Lgi.h"
#include "Token.h"

/*
	Include this file in your project if you want to use the 
	Win32 resource system.

	Otherwise include the LgiRes.cpp for cross platform code
	or BeOSRes.cpp for beos specific code.
*/

/////////////////////////////////////////////////////////////////////////////////////////
// Classes
bool GDlgFunc::LoadFromResource(char *Resource, List<GView> &Children, GdcRegion *Pos, char *Name)
{
	bool Status = false;

	HRSRC hRsrc = FindResourceEx(NULL, RT_DIALOG, Resource, LgiGetLanguageId());

	if (NOT hRsrc)
	{
		hRsrc = FindResource(NULL, Resource, RT_DIALOG);
	}

	if (hRsrc)
	{
		int Size = SizeofResource(NULL, hRsrc);
		HGLOBAL hMem = LoadResource(NULL, hRsrc);
		if (hMem)
		{
			uchar *Ptr = (uchar*) LockResource(hMem);
			if (Ptr)
			{
				DLGTEMPLATE *Template = (DLGTEMPLATE*) Ptr;

				if (Pos)
				{
					Pos->ZOff((double) Template->cx * DIALOG_X, (double) Template->cy * DIALOG_Y);
					Pos->Offset((double) Template->x * DIALOG_X, (double) Template->y * DIALOG_Y);
				}

				char Temp[256];
				short *A = (short*) (Template+1);

				// menu
				A = StrExtract(A, Temp);
				// class
				A = StrExtract(A, Temp);
				// title
				A = StrExtract(A, Temp);
				if (strlen(Temp) > 0 AND Name)
				{
					strcpy(Name, Temp);
				}

				if (Template->style & DS_SETFONT)
				{
					// font
					int PointSize = *A++;
					// name
					A = StrExtract(A, Temp);
				}

				A = PadToDWord(A, TRUE);

				// loop thru controls
				for (int i=0; i<Template->cdit; i++)
				{
					GControlInfo *CtrlInfo = NEW(GControlInfo);
					if (CtrlInfo)
					{
						DLGITEMTEMPLATE *Item = (DLGITEMTEMPLATE*) A;
						memcpy(&CtrlInfo->Item, Item, sizeof(DLGITEMTEMPLATE));
						A = (short*) (Item+1);

						// class
						A = StrExtract(A, Temp, &CtrlInfo->Atom);
						if (strlen(Temp) > 0)
						{
							CtrlInfo->Class = NewStr(Temp);
						}

						// title
						A = StrExtract(A, Temp);
						if (strlen(Temp) > 0)
						{
							CtrlInfo->Text = NewStr(Temp);
						}

						// creation data
						int Size = *A++;
						if (Size > 0)
						{
							A = PadToDWord(A, TRUE);
							A = (short*) (((uchar*)A)+Size);
						}
						A = PadToDWord(A, TRUE);

						GView *c = CreateControl(CtrlInfo);
						if (c)
						{
							c->Visible(CtrlInfo->Item.style & WS_VISIBLE);
							Children.Insert(c);
						}

						DeleteObj(CtrlInfo);
					}
				}

				Status = TRUE;
				GlobalUnlock(hMem);
			}
		}
	}

	return Status;
}

////////////////////////////////////////////////////////////////////////
// Get language
int LgiGetLanguageId()
{
	return MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT);
	// return MAKELANGID(LANG_FRENCH, SUBLANG_FRENCH);
}

///////////////////////////////////////////////////////////////////////////////////////
// Load string
char *LgiLoadString(WORD Res)
{
	static char Str[512];

	Str[0] = 0;
	HRSRC hRes = FindResourceEx(NULL, RT_STRING, MAKEINTRESOURCE(Res), LgiGetLanguageId());
	if (hRes)
	{
			HGLOBAL hMem = LoadResource(hAppInst(), hRes);
			if (hMem)
			{
				char *Ptr = (char*) LockResource(hMem);
				if (Ptr)
				{
					char *d = Str;
					short *s = (short*) (Ptr + 2);
					int Len = *s++;

					while (Len--)
					{
						*d++ = *s++;
					}
					*d++ = 0;
				}
			}
	}

	if (strlen(Str) < 1)
	{
		// can't find the language dependent string
		LoadString(hAppInst(), Res, Str, sizeof(Str));
	}

	return Str;
}

//////////////////////////////////////////////////////////////////////
// Menus
bool GSubMenu::Load(HMENU hmenu)
{
	bool Status = FALSE;
	hMenu = hmenu;
	if (hMenu)
	{
		int Count = GetMenuItemCount(hMenu);
		for (int i=0; i<Count; i++)
		{
			MENUITEMINFO Info;
			char Name[256];

			ZeroObj(Info);
			Info.cbSize = sizeof(Info);
			Info.fMask = MIIM_CHECKMARKS | MIIM_DATA | MIIM_ID | MIIM_STATE | MIIM_SUBMENU | MIIM_TYPE;
			Info.dwTypeData = Name;
			Info.cch = sizeof(Name)-1;

			if (GetMenuItemInfo(hMenu,
								i,
								TRUE, 
								&Info))
			{
				GMenuItem *Item = NEW(GMenuItem);
				if (Item)
				{
					Info.fMask = 0;
					memcpy(&Item->Info, &Info, sizeof(Info));
					Item->Menu = Menu;
					Item->Parent = this;
					Item->Position = i;

					if (Info.fType == MFT_STRING)
					{
						Item->GObject::Name((char*) Info.dwTypeData);
						Item->ScanForAccel();
					}

					if (Info.hSubMenu)
					{
						Item->Child = NEW(GSubMenu("", true));
						if (Item->Child)
						{
							Item->Child->Parent = Item;
							Item->Child->Menu = Menu;
							Item->Child->Load(Info.hSubMenu);
						}
					}

					Items.Insert(Item);
				}
			}
		}

		Status = TRUE;
	}

	return Status;
}

bool GMenu::Load(GView *w, char *Res)
{
	bool Status = FALSE;

	if (Res)
	{
		HMENU hNewMenu = 0;

		// try looking for a language specific menu first
		HRSRC hRes = FindResourceEx(hAppInst(), RT_MENU, Res, LgiGetLanguageId());
		if (hRes)
		{
			HGLOBAL hMem = LoadResource(hAppInst(), hRes);
			if (hMem)
			{
				void *Ptr = LockResource(hMem);
				if (Ptr)
				{
					hNewMenu = LoadMenuIndirect((MENUTEMPLATE*) Ptr);
				}
			}
		}

		if (NOT hNewMenu)
		{
			// if the language specific load fails we do a normal load here
			hNewMenu = LoadMenu(hAppInst(), Res);
		}

		if (hNewMenu)
		{
			hMenu = hNewMenu;
			Menu = this;
			Status = GSubMenu::Load(hMenu);
		}
	}

	return Status;
}
